#ifndef FRAME_CPP__Common__Compression_hh
#define	FRAME_CPP__Common__Compression_hh

#include <limits.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

#include "general/types.hh"
#include "general/autoarray.hh"
#include "general/hash_map.hh"

namespace FrameCPP
{
  //=====================================================================
  // constants
  //=====================================================================

  static const INT_2S MIN_INT_2S(
#if SIZEOF_SHORT == 2
				 SHRT_MIN
#else
#error Cannot compute the value for MIN_INT_2S
#endif
				 );
  static const INT_4S MIN_INT_4S(
#if SIZEOF_INT == 4
				 INT_MIN
#else
#error Cannot compute the value for MIN_INT_4S
#endif
				 );

  //=====================================================================
  // Exceptions
  //=====================================================================

  class FrUncompressable
    : public std::runtime_error
  {
  public:
    inline
    FrUncompressable( const std::string& What )
      : std::runtime_error( What.c_str( ) )
    {
    }
  };

  class FrZCompRangeError
    : public FrUncompressable
  {
  public:
    inline FrZCompRangeError( )
      : FrUncompressable( format_msg( ) )
    {
    };

  private:
    inline std::string
    format_msg( )
    {
      std::ostringstream	oss;

      oss << "Buffer cannot be compressed using FrVectZCompShort because value of "
	  << MIN_INT_2S << " encountered";
      return oss.str( );
    }
  };

  class FrZCompIRangeError
    : public FrUncompressable
  {
  public:
    inline FrZCompIRangeError( )
      : FrUncompressable( format_msg( ) )
    {
    };
  private:
    inline std::string
    format_msg( )
    {
      std::ostringstream	oss;

      oss << "Buffer cannot be compressed using FrVectZCompInt because value of "
	  << MIN_INT_4S << " encountered";
      return oss.str( );
    }
  };

  //=====================================================================
  // Functions
  //=====================================================================

  void
  ZeroSuppressCompressShort( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData );

  void
  ZeroSuppressUncompressShort( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData );

  void
  ZeroSuppressCompressInt( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData );

  void
  ZeroSuppressUncompressInt( CHAR_U*& Data, INT_8U& NBytes, INT_4U NData );

  namespace Compression
  {
    //===================================================================
    // Enumerations
    //===================================================================
    enum data_types_type {
      FR_VECT_C = 0,
      FR_VECT_2S = 1,
      FR_VECT_8R = 2,
      FR_VECT_4R = 3,
      FR_VECT_4S = 4,
      FR_VECT_8S = 5,
      FR_VECT_8C = 6,
      FR_VECT_16C = 7,
      FR_VECT_STRING = 8,
      FR_VECT_2U = 9,
      FR_VECT_4U = 10,
      FR_VECT_8U = 11,
      FR_VECT_1U = 12
    };

    enum compress_type {
      MODE_RAW = 0,
      MODE_GZIP = 1,
      MODE_DIFF = 2,
      MODE_DIFF_GZIP = 3,
      MODE_ZERO_SUPPRESS_SHORT = 5,
      MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER = 6,
      MODE_ZERO_SUPPRESS_INT_FLOAT = 8,
      MODE_ZERO_SUPPRESS_OTHERWISE_GZIP = 10
    };

    //===================================================================
    // types
    //===================================================================
    typedef General::hash_map< INT_4U, data_types_type > data_type_mapping;
    typedef General::hash_map< INT_4U, compress_type > compress_type_mapping;

    //===================================================================
    // Compression/Decompression classes
    //===================================================================

    namespace Differential
    {
      template< class T > void Encode( CHAR_U* Data, INT_4U NData );
      template< class T > void Decode( CHAR_U* Data, INT_4U NData );
    } // namespace - Differential

    namespace GZip
    {
      //-----------------------------------------------------------------
      //: Compress the data with GZip.
      void Compress( CHAR_U*& Data, INT_8U& NBytes, size_t Level );
      //-----------------------------------------------------------------
      //: Expand the data with GZip.
      void Expand( CHAR_U*& Data,
		   INT_8U& DataLength,
		   const INT_8U ExpandedLength,
		   const compress_type_mapping& CompressionMapping );
    } // namespace - GZip

    void
    Compress( INT_4U& CompressionMode,
	      INT_4U Level,
	      const compress_type_mapping& CompressionMapping,
	      INT_4U DataType,
	      const data_type_mapping& DataTypeMapping,
	      CHAR_U*& Data,
	      INT_8U& DataLength,
	      INT_8U& DataByteLength );

    void
    Expand( bool NativeOrder,
	    INT_4U CompressionMode,
	    const compress_type_mapping& CompressionMapping,
	    INT_4U DataType,
	    const data_type_mapping& DataTypeMapping,
	    CHAR_U*& Data,
	    INT_8U& DataLength,
	    INT_8U& DataByteLength
	    );

  } // namespace -  Compression
  
} // namespace - FrameCPP

#endif /* FRAME_CPP__Common__Compression_hh */
