#include "framecpp/config.h"

#include "adcdata.hh"
#include "framereaderbase.hh"
#include <general/regexmatch.hh>
#include "reference.hh"
#include "output.hh"


using namespace FrameCPP::Version_3_4_5;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an AdcData object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> groupNumber = 0 </li>
//     <li> channelNumber = 0 </li>
//     <li> nBits = 8 </li>
//     <li> bias = 0 </li>
//     <li> slope = 1.0 </li>
//     <li> units = "" (dimensionless) </li>
//     <li> sampleRate = 0.0 </li>
//     <li> timeOffset = 0 seconds, 0 nanoseconds </li>
//     <li> fShift = 0.0 </li>
//     <li> phase = 0.0 </li>
//     <li> dataValid = 0 </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
AdcData::AdcData()
        : Base(), 
          mName( "none" ), mComment( "" ),
          mChannelGroup( 0 ), mChannelNumber( 0 ),
          mNBits( 8 ), mBias( 0 ), mSlope( 1.0 ), mUnits( "" ),
          mSampleRate( 0.0 ), mTimeOffset( 0, 0 ), mFShift( 0.0 ),
          mPhase( 0.0 ), mDataValid( 0 ), mData(), mAux(), mInstance(0xbad)
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates an AdcData object with the specified properties.
// Additionally, the comment is set to "" (no comment).
//
//!param: const std::string& name - Chanel name
//!param: INT_4U channelGroup - Channel groups
//!param: INT_4U channelNumber - Channel number
//!param: INT_4U nBits - Number of bits in A/D output
//!param: const REAL_8& sampleRate - Data acquisition rate, samples/sec.
//!param: REAL_4 bias - DC Bias on channel.  Default: 0.0.
//!param: REAL_4 slope - Input units/count.  Default: 1.0.
//!param: const std::string& units - Units for slope.  Default: "counts".
//!param: const REAL_8& fShift - Frequency shift if signal has been
//+     heterodyned before ADC.  Default: 0.0.
//!param: const REAL_4& phase - Heterodyning phase. Default: 0.0.
//!param: const Time& timeOffset - Time offset relative to frame start.
//+     Default: Time( 0, 0 ) (no offset).
//!param: bool dataValid - Data valid flag.  Default: 0 (data is valid)
//
//!exc: None.
//      
AdcData::AdcData( const std::string& name,
		  INT_4U group, INT_4U channel,
                  INT_4U nBits, const REAL_8& sampleRate, REAL_4 bias,
                  REAL_4 slope, const std::string& units,
                  const REAL_8& fShift,
                  const Time& timeOffset,
                  INT_2U dataValid,
		  const REAL_4& phase )
        : Base(), 
          mName( name ), mComment( "" ),
	  mChannelGroup( group ), mChannelNumber( channel ),
	  mNBits( nBits ), mBias( bias ),
          mSlope( slope ), mUnits( units ), mSampleRate( sampleRate ),
          mTimeOffset( timeOffset ), mFShift( fShift ), mPhase( phase ),
          mDataValid( dataValid ), mData(), mAux()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// A deep copy is performed for owned objects.
//
//!param: const AdcData& adcData - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
AdcData::AdcData( const AdcData& adcData )
        :
        Base( adcData ), 
        mName( adcData.mName ), mComment( adcData.mComment ),
	mChannelGroup( adcData.mChannelGroup ), mChannelNumber( adcData.mChannelNumber ),
	mNBits( adcData.mNBits ), mBias( adcData.mBias ), mSlope( adcData.mSlope ),
        mUnits( adcData.mUnits ), mSampleRate( adcData.mSampleRate ),
        mTimeOffset( adcData.mTimeOffset ), mFShift( adcData.mFShift ),
	mPhase( adcData.mPhase ),
        mDataValid( adcData.mDataValid ), mData( adcData.mData ),
        mAux( adcData.mAux ), mInstance( adcData.mInstance )
{
}


//-----------------------------------------------------------------------------
//
//: Channel name
//
// Gets the channel name.
//
//!return: const std::string& -- Channel name.
//
const std::string& AdcData::getName() const
{
    return mName;
}




//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
// A deep copy is performed for owned objects.
//
//!param: const AdcData& adcData - Obejct to be assigned.
//
//!return: const AdcData& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const AdcData& AdcData::operator=( const AdcData& adcData )
{
    if ( this != &adcData )
    {
	mInstance = adcData.mInstance;
        mName = adcData.mName;
        mComment = adcData.mComment;
        mChannelGroup = adcData.mChannelGroup;
        mChannelNumber = adcData.mChannelNumber;
        mNBits = adcData.mNBits;
        mBias = adcData.mBias;
        mSlope = adcData.mSlope;
        mUnits = adcData.mUnits;
        mSampleRate = adcData.mSampleRate;
        mTimeOffset = adcData.mTimeOffset;
        mFShift = adcData.mFShift;
        mPhase = adcData.mPhase;
        mDataValid = adcData.mDataValid;
        mData = adcData.mData;
        mAux = adcData.mAux;
    }

    return *this;
}

//-----------------------------------------------------------------------------
//
//: += Assignment Operator.
//
//!param: const AdcData& adcData -
//
//!return: const AdcData&
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: frame_mismatch - ADC mismatch
//
const AdcData& AdcData::operator+=( const AdcData& adcData )
{
  // Check the attributes that should match
  //
  // TO BE DONE
  //
  
  mData += adcData.mData;
  mAux += adcData.mAux;

  #ifdef not_def
    if ( this != &adcData )
    {
        mName = adcData.mName;
        mComment = adcData.mComment;
        mChannelGroup = adcData.mChannelGroup;
        mChannelNumkebr = adcData.mChannelNumber;
        mNBits = adcData.mNBits;
        mBias = adcData.mBias;
        mSlope = adcData.mSlope;
        mUnits = adcData.mUnits;
        mSampleRate = adcData.mSampleRate;
        mTimeOffset = adcData.mTimeOffset;
        mFShift = adcData.mFShift;
        mPhase = adcData.mPhase;
        mDataValid = adcData.mDataValid;
        mData = adcData.mData;
        mAux = adcData.mAux;
	mInstance = adcData.mInstance;
    }
#endif

    return *this;
}

   
//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//: Class ID.
//
// Gets the class ID.
//
// This is the internal ID used to identify this class type.  It is specified
// in "types.hh".
//
//!return: INT_2U -- Class ID.
//
//!exc: None.
//      
INT_2U AdcData::getClassId() const
{
   return ID_ADCDATA;
}
   

//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs the Structure Header for this class. 
//
//!return: SH - ADC data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH AdcData::getSH()
{
  SH sh( "FrAdcData", ID_ADCDATA, "ADC Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "channelGroup", "INT_4U", "" ) );
  sh.appendSE( SE( "channelNumber", "INT_4U", "" ) );
  sh.appendSE( SE( "nBits", "INT_4U", "" ) );
  sh.appendSE( SE( "bias", "REAL_4", "" ) );
  sh.appendSE( SE( "slope", "REAL_4", "" ) );
  sh.appendSE( SE( "units", "STRING", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_8", "" ) );
  sh.appendSE( SE( "timeOffsetS", "INT_4S", "" ) );
  sh.appendSE( SE( "timeOffsetN", "INT_4U", "" ) );
  sh.appendSE( SE( "fShift", "REAL_8", "" ) );
  sh.appendSE( SE( "phase", "REAL_4", "" ) );
  sh.appendSE( SE( "dataValid", "INT_2U", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "aux", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrAdcData *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs the data format version 4 Structure Header for this class. 
//
//!return: SH - ADC data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH AdcData::getSH4()
{
  SH sh( "FrAdcData", ID_ADCDATA, "ADC Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "channelGroup", "INT_4U", "" ) );
  sh.appendSE( SE( "channelNumber", "INT_4U", "" ) );
  sh.appendSE( SE( "nBits", "INT_4U", "" ) );
  sh.appendSE( SE( "bias", "REAL_4", "" ) );
  sh.appendSE( SE( "slope", "REAL_4", "" ) );
  sh.appendSE( SE( "units", "STRING", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_8", "" ) );
  sh.appendSE( SE( "timeOffsetS", "INT_4S", "" ) );
  sh.appendSE( SE( "timeOffsetN", "INT_4U", "" ) );
  sh.appendSE( SE( "fShift", "REAL_8", "" ) );
  sh.appendSE( SE( "dataValid", "INT_2U", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "aux", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrAdcData *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs the data format version 3 Structure Header for this class. 
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH AdcData::getSH3()
{
    SH sh( "FrAdcData", ID_ADCDATA, "ADC Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "crate", "INT_4U", "" ) );
    sh.appendSE( SE( "channel", "INT_4U", "" ) );
    sh.appendSE( SE( "nBits", "INT_4U", "" ) );
    sh.appendSE( SE( "bias", "REAL_4", "" ) );
    sh.appendSE( SE( "slope", "REAL_4", "" ) );
    sh.appendSE( SE( "units", "STRING", "" ) );
    sh.appendSE( SE( "sampleRate", "REAL_8", "" ) );
    sh.appendSE( SE( "timeOffsetS", "INT_4U", "" ) );
    sh.appendSE( SE( "timeOffsetN", "INT_4U", "" ) );
    sh.appendSE( SE( "fShift", "REAL_8", "" ) );
    sh.appendSE( SE( "overRange", "INT_2U", "" ) );
    sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "aux", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrAdcData *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an AdcData object from an FrameReaderBase object.
//
// This method
// creates a dynamically allocated object.  This method is called indirectly
// by the FrameReader class.  This is a private method, access is granted
// through the ObjectProxy< AdcData > class which is a friend.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: AdcData* - Newly allocated.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//
AdcData* AdcData::read( FrameReaderBase& in )
{
  AdcData& adc = *new AdcData();
  adc.mInstance = in.currentInstance();
    
  in >> adc.mName >> adc.mComment >> adc.mChannelGroup >> adc.mChannelNumber
     >> adc.mNBits >> adc.mBias
     >> adc.mSlope >> adc.mUnits >> adc.mSampleRate >> adc.mTimeOffset
     >> adc.mFShift >> adc.mPhase
     >> adc.mDataValid >> adc.mData >> adc.mAux;
  in.readNextNodeReference();
  return &adc;
}

//: Version 4 frame reader.
//   
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.   
//   
AdcData* AdcData::read4( FrameReaderBase& in )
{
  AdcData& adc = *new AdcData();
  adc.mInstance = in.currentInstance();
    
  in >> adc.mName >> adc.mComment >> adc.mChannelGroup >> adc.mChannelNumber
     >> adc.mNBits >> adc.mBias
     >> adc.mSlope >> adc.mUnits >> adc.mSampleRate >> adc.mTimeOffset >> adc.mFShift
     >> adc.mDataValid >> adc.mData >> adc.mAux;
  in.readNextNodeReference();
  return &adc;
}

//: Version 3 frame reader.
//   
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.   
//      
AdcData* AdcData::read3( FrameReaderBase& in )
{
  AdcData& adc = *new AdcData();
  adc.mInstance = in.currentInstance();

  INT_4U crate; // version 4 does not have crate number
    
  in >> adc.mName >> adc.mComment >> crate >> adc.mChannelNumber
     >> adc.mNBits >> adc.mBias
     >> adc.mSlope >> adc.mUnits >> adc.mSampleRate >> adc.mTimeOffset >> adc.mFShift
     >> adc.mDataValid >> adc.mData >> adc.mAux;
    
  adc.mChannelGroup = 0; // version 3 does not have channel group

  in.readNextNodeReference();
  return &adc;
}


//-----------------------------------------------------------------------------
//
//: Write an AdcData object
//
// Writes an AdcData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void AdcData::write( Output& out ) const
{
    out.startRecord( ID_ADCDATA, 75
		     + 4 // for mPhase
                     + mName.size() + mComment.size()
                     + mUnits.size(), this );

    out << mName << mComment << mChannelGroup << mChannelNumber << mNBits << mBias
        << mSlope << mUnits << mSampleRate << mTimeOffset
	<< mFShift << mPhase
        << mDataValid << mData << mAux;

    out.endRecord( this );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two AdcData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const AdcData& adcData - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//   
bool AdcData::operator==( const AdcData& adcData ) const
{
    return ( mName          == adcData.mName          &&
             mComment       == adcData.mComment       &&
             mChannelGroup  == adcData.mChannelGroup  &&
             mChannelNumber == adcData.mChannelNumber &&
             mNBits         == adcData.mNBits         &&
             mBias          == adcData.mBias          &&
             mSlope         == adcData.mSlope         &&
             mUnits         == adcData.mUnits         &&
             mSampleRate    == adcData.mSampleRate    &&
             mTimeOffset    == adcData.mTimeOffset    &&
             mFShift        == adcData.mFShift        &&
             mDataValid     == adcData.mDataValid     &&
             mData          == adcData.mData          &&
             mAux           == adcData.mAux );
}
