#ifndef FrameCppContainerHH
#define FrameCppContainerHH

#include "framecpp/config.h"
#include <vector>
#include <string>
#include "types.hh"
#include "errors.hh"
#include "containerbase.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

//-----------------------------------------------------------------------------
//
//: Generic Container.
//
// <p>This implements a generic container object.  It is a template class, and
// is used to store any type of frame object.  Its principal feature is that it
// keeps track of ownership.
//
// <p>One of the principle problems when creating or using a container class is
// how to store the elements.  If the user adds an element to the class,
// should it be copied, or should a reference to the added element be stored
// instead?  If a reference (or pointer) is stored, the user must be careful
// about deallocating objects.  If an element is destructed while a reference
// to it exists in a container, then a segmentation fault may occur.  On the
// other hand, if an object is added to the container and the user doesn't
// destruct it, then there is a memory leak.  If the programmer decides to
// implement a container which only copies elements, other trade-offs are made.
// If the container is storing large objects, then copying may take a
// significant amount of time and incur a large memory overhead.  Furthermore,
// nesting containers may become awkward as the order in which containers are
// added becomes important or other overhead is required to obtain a reference
// to the copied element in the container.
//
// <p>One way to solve these problems is to implement a reference counting
// system.  However, this can quickly become complicated.  The method I have
// chosen is an ownership system.
//
// <p>There are two things which need to be known when adding an element to a
// container:<p><ol>
//
//   <li> Do I copy the object or use a reference? </li>
//   <li> Who is responsible for deleting the object (i.e., who owns it)? </li>
//        </ol>
//
// <p>If the container copies the object upon insertion and subsequently owns
// it, then we have the first case above.  If the container just inserts a
// reference and doesn't own it then we have the second case.  Sometimes a
// mixture of the two is desired, for example if the user is creating objects
// (perhaps by reading them in from a stream) and inserting them into a
// container, they may not want to keep track of all of the pointers to delete
// them later (he wants the container to do that).  But he also doesn't want
// to do another wasteful copy of the data.  In this case, he would want a
// reference stored and the container to own the elements.
//
// <p>The system I have chosen solves the ownership issue by storing a flag
// along with each element which records the ownership state of that element.
// If the container owns the object, then it will get destructed whenever it is
// removed from the container (perhaps upon the destruction of the container
// itself).  However, if the object is not owned then it will not be destroyed.
// To summarize, we have the following two states:
//
// <ul>
//   <li>Owned - The container owns the element.  This means that the container
//       is responsible for destructing the object when it is no longer used.
//       Therefore, if the objects is removed from the container or if the
//       container itself is destructed, this object will also be destructed.
//       Furthermore, in a copy constructor and during an assignment, a deep
//       copy will be performed of an owned object (so that the new container
//       will own the copy).</li>
//   <li>Unowned - The container does not own the element.  The container will
//       not destruct this object under any circumstances.  When a copy
//       constructor or assignment operator is called, only the pointer to
//       this object will be copied (shallow copy).</li>
// </ul>
//
// <p>The ownership state of an element can be determined via the
// <code>isOwned</code> accessor.  There are two versions, one taking an index
// and the other taking a constant iterator.  It is possible to manipulate the
// ownership of an element via the <code>setOwnership</code> mutator.
//
// <p>When objects are inserted into the container, it is possible to specify
// whether they should be copied into the container or if only a reference
// should be stored.
//
// <p>By default, objects are copied into the container and the container will
// ownd the objects.
//
// <p><B>Implementation Notes:</B><p><ul type=disc>
//
// <li> The Container class stores its elements in a STL vector (vector< T* >).
//   The ownership information is stored in another STL vector
//   ( vector< bool > ).  Note that pointers to the elements are stored in the
//   element vector. </li>
//
// <li> Mutators which insert data (append & insert) have 3 forms which differ
//   in the parameter list:<ol>
//   <li> const T& - This type takes a constant reference to an object.  This
//      will always result in the element being copied and ownership belonging
//      to the container.  Passing an object as a const means that it must be
//      copied, since the container does not store constant pointers, just
//      pointers.  Furthermore, if one wants to insert an object without
//      copying, passing a pointer to the object seems imply that the object
//      is not being copied more than a reference would (one normally does not
//      take the address of the referenced object and keep it, however it would
//      be reasonable to suspect that one might keep the pointer if it was
//      passed.  I am trying to follow the principal of least astonishment).
//      </li>
//   <li> const T* - This also always results in the element being copiend and
//      ownership belonging to the container. </li>
//   <li> T*, bool copy, bool owns - If the object is being passed as a
//      non-const pointer, then the user can specify whether or not it should
//      be copied and whether or not the container will own it. </li></ol>
// </li></ul>
//
template< class T >
class Container : public ContainerBase
{
public:

    /* Typedef's */
    typedef typename std::vector< T* >::iterator iterator;
    typedef typename std::vector< T* >::const_iterator const_iterator;
    typedef typename std::vector< T* >::reverse_iterator reverse_iterator;
    typedef typename std::vector< T* >::const_reverse_iterator const_reverse_iterator;

    /* Constructors, Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Default constructor.
    //
    // This just creates an empty container.
    //
    Container();

    //-----------------------------------------------------------------------------
    //
    //: Container copy constructor.
    //
    // This is the copy constructor for this class.  It iterates through source
    // container, copying its objects into this container.  The manner in which it
    // copies the elements depends upon their ownership in the source container.
    //
    // If the source container owns the element, then a copy of that element is
    // made and inserted into the new container.  The new container now owns that
    // element.
    //
    // If the source container does not own the element, then the pointer to the
    // element is inserted into the new container.  A copy is not made, therefore
    // the new container does NOT own the element.
    //
    // In Summary, we have:
    //
    // <pre>
    //   source                                     dest
    //   ownership      copy method                 ownership
    //   ---------      -----------                 ---------
    //   owned     ---> allocate copy          ---> owned
    //   unowned   ---> assign current pointer ---> unowned
    // </pre>
    //
    //!param: const Container< T >& c - The container to copy from
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Container( const Container< T >& c );

    //-----------------------------------------------------------------------------
    //
    //: Destructor.
    //
    // This deletes all of the owned objects in the container.
    //
    //!exc: None.
    //
    virtual ~Container();

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment operator.
    //
    // This copies another container to this one.  Ownership is handled in the same
    // manner as in the copy constructor.
    //
    //!param: const Container< T >& c - The container to assign from.
    //
    //!return: const Container< T >& -- This container.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Container< T >& operator=( const Container< T >& c );
 
    //-----------------------------------------------------------------------------
    //
    //: += Assignment operator.
    //
    // This concatenates another container to this one.  
    //
    //!param: const Container< T >& c - The container to assign from.
    //
    //!return: const Container< T >& -- This container.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: frame_mismatch - `T' mismatch.
    //
    const Container< T >& operator+=( const Container< T >& c );
 
    //-----------------------------------------------------------------------------
    //
    //: Equality operator.
    //
    // This checks to see if two containers contain identical elements.  The
    // ownership of the elements does not matter.
    //
    //!param: const Container< T >& c - The container to compare with.
    //
    //!return: bool -- true if the containers are equal.
    //
    //!exc: None.
    //
    bool operator==( const Container< T >& c ) const;

    //-----------------------------------------------------------------------------
    //
    // Inequality operator.
    //
    // This chacks to see if two containers do not contain identical elements.  The
    // ownership of the elements does not matter.
    //
    //!param: const Container< T >& c - The container to compare with.
    //
    //!return: bool -- true if the containers are not equal.
    //
    //!exc: None.
    //
    bool operator!=( const Container< T >& c ) const;

    //-----------------------------------------------------------------------------
    //
    // Array access operator.
    //
    // This returns the element at the specified index or a null pointer if the
    // index is out-of-range.
    //
    //!param: size_t index - The index of the desired element.
    //
    //!return: T* -- A pointer to the object stored at the index.
    //!return: T* -- A constant pointer to the object stored at the index.
    //
    //!exc: None.
    //
    virtual T* operator[]( unsigned int index );

    //-----------------------------------------------------------------------------
    //
    // Array access operator.
    //
    // This returns the element at the specified index or a null pointer if the
    // index is out-of-range.
    //
    //!param: size_t index - The index of the desired element.
    //
    //!return: const T* -- A constant pointer to the object stored at the index.
    //
    //!exc: None.
    //
    virtual const T* operator[]( unsigned int index ) const;
    
    /* Accessors */
    //-----------------------------------------------------------------------------
    //
    //: Returns the ownership of an element.
    //
    // Currently, no range checking is done.
    //
    //!todo: Add range checking.
    //
    //!param: const_iterator iter - A constant iterator pointing to an element in
    //+     the container.
    //
    //!return: bool -- True if the object is owned by the container.
    //
    bool isOwned( const_iterator iter ) const;

    //-----------------------------------------------------------------------------
    //
    //: Returns the ownership of an element.
    //
    // Currently, no range checking is done.
    //
    //!todo: Add range checking.
    //
    //!param: size_t index - Index.
    //
    //!return: bool -- True if the object is owned by the container.
    //
    bool isOwned( size_t index ) const;

    //-----------------------------------------------------------------------------
    //
    //: Returns the size of the container.
    //
    //!return: size_t -- The number of objects in the container.
    //
    //!exc: None.
    //
    unsigned int size() const;

    //-----------------------------------------------------------------------------
    //
    //: Returns the first item in the container.
    //
    //!return: T* -- First item.
    //
    //!exc: None.
    //
    T* front();

    //-----------------------------------------------------------------------------
    //
    //: Returns the first item in the container.
    //
    //!return: const T* -- First item.
    //
    //!exc: None.
    //
    const T* front() const;

    //-----------------------------------------------------------------------------
    //
    //: Returns the last item in the container.
    //
    //!return: T* -- Last item.
    //
    //!exc: None.
    //
    T* back();

    //-----------------------------------------------------------------------------
    //
    //: Returns the last item in the container.
    //
    //!return: const T* -- Last item.
    //
    //!exc: None.
    //
    const T* back() const;

    //-----------------------------------------------------------------------------
    //
    //: Returns an iterator at the beginning of the container.
    //
    //!return: Container< T >::iterator -- Beginning iterator.
    //
    //!exc: None.
    //
    iterator begin();

    //-----------------------------------------------------------------------------
    //
    //: Returns an iterator at the beginning of the container.
    //
    //!return: Container< T >::const_iterator -- Beginning iterator.
    //
    //!exc: None.
    //
    const_iterator begin() const;

    //-----------------------------------------------------------------------------
    //
    //: Returns an iterator at the end of the container.
    //
    //!return: Container< T >::iterator -- End iterator.
    //
    //!exc: None.
    //
    iterator end();

    //-----------------------------------------------------------------------------
    //
    //: Returns an iterator at the end of the container.
    //
    //!return: Container< T >::const_iterator -- End iterator.
    //
    //!exc: None.
    //
    const_iterator end() const;

    /* Mutators */
    //-----------------------------------------------------------------------------
    //
    //: Set Ownership
    //
    // Sets the ownsership for the element pointed to by iter.
    //
    //!param: iterator iter - The element whose ownership will be modified.
    //!param: bool owns - true if the container should own the data, false otherwise.
    //
    //!todo: Boundary checking
    //
    void setOwnership( iterator iter, bool owns );

    //-----------------------------------------------------------------------------
    //
    //: Set Ownership.
    //
    // Sets the ownsership for the element at the given index.
    //
    //!param: size_t index - The element whose ownership will be modified.
    //!param: bool owns - True if the container should own the data, false otherwise.
    //
    //!todo: Boundary checking
    //
    void setOwnership( size_t index, bool owns );

    //-----------------------------------------------------------------------------
    //
    //: Append an element to the end of the container.
    //
    // Memory for the element will be allocated and the container will own the
    // new element.
    //
    //!param: const T& data - The item to append.
    //
    //!return: iterator -- The iterator correspinding to the appended object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    iterator append( const T& data );

    //-----------------------------------------------------------------------------
    //
    //: Append an element to the end of the container.
    //
    // Memory for the element will be allocated and the container will own the
    // new element.
    //
    //!param: const T* data - The item to append.
    //
    //!return: iterator -- The iterator corresponding to the appended object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    iterator append( const T* data );

    //-----------------------------------------------------------------------------
    //
    //: Append an element to the end of the container.
    //
    // The user may also specifiy whether or not memory is allocated for the
    // element (i.e., the element is copied) and if the container owns the element.
    //
    //!param: const T& data - The item to append.
    //!param: bool copy - True if memory should be allocated and the element copied
    //+     into the container (default: true).
    //!param: bool owns - True if the container will own the element (default: true).
    //
    //!return: iterator -- The iterator corresponding to the appended object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    iterator append( T* data, bool allocate = true, bool owns = true );

    //-----------------------------------------------------------------------------
    //
    //: Insert an element into the container.
    //
    // The element will be copied and the container will own the new element.
    //
    //!param: iterator iter - An iterator pointing to the location where the
    //+     element should be added.
    //!param: const T& data - The item to add.
    //
    //!return: iterator -- An iterator pointing to the added item.
    //
    //!exc: std::bad_alloc - Memory could not be allocated for the element.
    //
    iterator insert( iterator iter, const T& data );

    //-----------------------------------------------------------------------------
    //
    //: Insert an element into the container.
    //
    // The element will be copied and the container will own the new element.
    //
    //!param: iterator iter - An iterator pointing to the location where the
    //+     element should be added.
    //!param: const T* data - The item to add.
    //
    //!return: iterator -- An iterator pointing to the added item.
    //
    //!exc: std::bad_alloc - Memory could not be allocated for the element.
    //
    iterator insert( iterator iter, const T* data );

    //-----------------------------------------------------------------------------
    //
    //: Insert an element into the container.
    //
    //!param: iterator iter - An iterator pointing to the location where the
    //+     element should be added.
    //!param: T* data - The item to add.
    //!param: bool copy - Duplicate passed data or not.
    //!param: bool owns - Destroy the data when container is destroyed or not.
    //
    //!return: iterator -- An iterator pointing to the added item.
    //
    //!exc: std::bad_alloc - Memory could not be allocated for the element.
    //
    iterator insert( iterator iter, T* data, bool allocate = true,
                     bool owns = true );

    //-----------------------------------------------------------------------------
    //
    //: Insert an element into the container.
    //
    // The element will be copied and the container will own the new element.
    //
    //!param: size_t index - The index at which the element should be added.  If
    //+     the index is out of range then the element will be inserted at the end.
    //!param: const T& data - The item to add.
    //
    //!return: iterator -- An iterator pointing to the added item.
    //
    //!exc: std::bad_alloc - Memory could not be allocated for the element.
    //
   iterator insert( size_t index, const T& data );

    //-----------------------------------------------------------------------------
    //
    //: Insert an element into the container.
    //
    // The element will be copied and the container will own the new element.
    //
    //!param: size_t index - The index at which the element should be added.  If
    //+     the index is out of range then the element will be inserted at the end.
    //!param: const T* data - The item to add.
    //
    //!return: iterator -- An iterator pointing to the added item.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    iterator insert( size_t index, const T* data );

    //-----------------------------------------------------------------------------
    //
    //: Insert an element into the container.
    //
    //!param: size_t index - The index at which the element should be added.  If
    //+     the index is out of range then the element will be inserted at the end.
    //!param: T* data - The item to add.
    //!param: bool copy - Duplicate passed data or not.
    //!param: bool owns - Destroy the data when container is destroyed or not.
    //
    //!return: iterator -- An iterator pointing to the added item.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    iterator insert( size_t index, T* data, bool allocate = true,
                     bool owns = true );

    //-----------------------------------------------------------------------------
    //
    //: Erase an element.
    //
    // This erases an element at the specified index.  If the index is out of range
    // then nothing happens.
    //
    //!param: unsigned int index - Element index.
    //
    //!exc: None.
    //
    void erase( unsigned int index );

    //-----------------------------------------------------------------------------
    //
    //: Erases a range of elements.
    //
    // If end is less thatn start or any of the iterators are out of range
    // then nothing will happen.
    //
    //!param: iterator start - Range start.
    //!param: iterator finish - Range end.
    //
    //!exc: None.
    //   
    void erase( iterator start, iterator finish );
    
    //#if DEPRICATED
    inline unsigned int getSize( ) const
    {  
      return size( );
    }
    //#endif /* DEPRICATED */
private:

    typedef std::vector< bool >::iterator owns_iterator;

    //: Contained objects.
    std::vector< T* > mData;
    //: True if an object is owned by the container.
    std::vector< bool > mOwns;    
};


//-----------------------------------------------------------------------------
//
//: Default constructor.
//
// This just creates an empty container.
//
//!exc: None.
//   
template< class T >
inline Container< T >::Container()
        : mData(), mOwns()
{
}


//-----------------------------------------------------------------------------
//
//: Container copy constructor.
//
// This is the copy constructor for this class.  It iterates through source
// container, copying its objects into this container.  The manner in which it
// copies the elements depends upon their ownership in the source container.
//
// If the source container owns the element, then a copy of that element is
// made and inserted into the new container.  The new container now owns that
// element.
//
// If the source container does not own the element, then the pointer to the
// element is inserted into the new container.  A copy is not made, therefore
// the new container does NOT own the element.
//
// In Summary, we have:
//
// <pre>
//   source                                     dest
//   ownership      copy method                 ownership
//   ---------      -----------                 ---------
//   owned     ---> allocate copy          ---> owned
//   unowned   ---> assign current pointer ---> unowned
// </pre>
//
//!param: const Container< T >& c - The container to copy from
//
//!exc: std::bad_alloc - Memory allocation failed.
//
template< class T >
Container< T >::Container( const Container< T >& c )
        : mData(), mOwns()
{
    // Loop through all of the source elements
    for ( const_iterator iter = c.begin(); iter != c.end(); ++iter )
    {
        // check the ownership
        if ( c.isOwned( iter ) )
        {
            mData.push_back( new T( **iter ) );
            mOwns.push_back( true );
        }
        else
        {
            mData.push_back( *iter );
            mOwns.push_back( false );
        }
    }
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This deletes all of the owned objects in the container.
//
//!exc: None.
//      
template< class T >
Container< T >::~Container()
{
  for ( unsigned int i = 0; i < mData.size(); i++ )
    {
      if ( mOwns[i] )
	{
	  delete mData[i];
          mData[ i ] = 0;
	}
    }
}


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment operator.
//
// This copies another container to this one.  Ownership is handled in the same
// manner as in the copy constructor.
//
//!param: const Container< T >& c - The container to assign from.
//
//!return: const Container< T >& -- This container.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
template< class T >
const Container< T >& Container< T >::operator=(
    const Container< T >& c )
{
    if ( this != &c )
    {
        erase( begin(), end() );

        for ( const_iterator iter = c.begin(); iter != c.end(); ++iter )
        {
            // check the ownership
            if ( c.isOwned( iter ) )
            {
                mData.push_back( new T( **iter ) );
                mOwns.push_back( true );
            }
            else
            {
                mData.push_back( *iter );
                mOwns.push_back( false );
            }
        }
    }
    
    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Assignment operator.
//
// This concatenates another container to this one.  
//
//!param: const Container< T >& c - The container to assign from.
//
//!return: const Container< T >& -- This container.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: frame_mismatch - `T' mismatch.
//
template< class T >
const Container< T >& Container< T >::operator+=( const Container< T >& c )
{
  // Both containers must be of equal size
  //
  if ( size() != c.size() )
    {
      throw frame_mismatch( "Containers have different size" );
    }

  const_iterator iter1, iter2;
  for ( iter1 = begin(),  iter2 = c.begin();
	iter1 != end() && iter2 != c.end(); ++iter1, ++iter2 )
    {
      **iter1 += **iter2;
    }
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Equality operator.
//
// This checks to see if two containers containe identical elements.  The
// ownership of the elements does not matter.
//
//!param: const Container< T >& c - The container to compare with.
//
//!return: bool -- true if the containers are equal.
//
//!exc: None.
//      
template< class T >
bool Container< T >::operator==( const Container< T >& c ) const
{
    bool equal( mData.size() == c.mData.size() );

    if ( equal )
    {
        for ( const_iterator iter1 = begin(), iter2 = c.begin();
              ( iter1 != end() ) && equal; ++iter1, ++iter2 )
        {
            equal = equal && ( **iter1 == **iter2 );
        }
    }

    return equal;
}


//-----------------------------------------------------------------------------
//
// Inequality operator.
//
// This chacks to see if two containers do not contain identical elements.  The
// ownership of the elements does not matter.
//
//!param: const Container< T >& c - The container to compare with.
//
//!return: bool -- true if the containers are not equal.
//
//!exc: None.
//      
template< class T >
bool Container< T >::operator!=( const Container< T >& c ) const
{
    return !( *this == c );
}


//-----------------------------------------------------------------------------
//
// Array access operator.
//
// This returns the element at the specified index or a null pointer if the
// index is out-of-range.
//
//!param: size_t index - The index of the desired element.
//
//!return: T* -- A pointer to the object stored at the index.
//!return: T* -- A constant pointer to the object stored at the index.
//
//!exc: None.
//      
template< class T >
T* Container< T >::operator[]( unsigned int index )
{
    if ( index >= mData.size() )
    {
        return 0;
    }

    return mData[ index ];
}

//-----------------------------------------------------------------------------
//
// Array access operator.
//
// This returns the element at the specified index or a null pointer if the
// index is out-of-range.
//
//!param: size_t index - The index of the desired element.
//
//!return: const T* -- A constant pointer to the object stored at the index.
//
//!exc: None.
//
template< class T >
const T* Container< T >::operator[]( unsigned int index ) const
{
    if ( index >= mData.size() )
    {
        return 0;
    }

    return mData[ index ];
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Returns the ownership of an element.
//
// Currently, no range checking is done.
//
//!todo: Add range checking.
//
//!param: const_iterator iter - A constant iterator pointing to an element in
//+     the container.
//
//!return: bool -- True if the object is owned by the container.
//
template< class T >
bool Container< T >::isOwned( const_iterator iter ) const
{
    return *( mOwns.begin() + ( iter - begin() ) );
}


//-----------------------------------------------------------------------------
//
//: Returns the ownership of an element.
//
// Currently, no range checking is done.
//
//!todo: Add range checking.
//
//!param: size_t index - Index.
//
//!return: bool -- True if the object is owned by the container.
//
template< class T >
bool Container< T >::isOwned( size_t index ) const
{
    return *( mOwns.begin() + index );
}


//-----------------------------------------------------------------------------
//
//: Returns the size of the container.
//
//!return: size_t -- The number of objects in the container.
//
//!exc: None.
//      
template< class T >
unsigned int Container< T >::size() const
{
    return mData.size();
}


//-----------------------------------------------------------------------------
//
//: Returns the first item in the container.
//
//!return: T* -- First item.
//
//!exc: None.
//      
template< class T >
inline T* Container< T >::front()
{
    return mData.front();
}

//-----------------------------------------------------------------------------
//
//: Returns the first item in the container.
//
//!return: const T* -- First item.
//
//!exc: None.
//      
template< class T >
inline const T* Container< T >::front() const
{
    return mData.front();
}


//-----------------------------------------------------------------------------
//
//: Returns the last item in the container.
//
//!return: T* -- Last item.
//
//!exc: None.
//      
template< class T >
inline T* Container< T >::back()
{
    return mData.back();
}

//-----------------------------------------------------------------------------
//
//: Returns the last item in the container.
//
//!return: const T* -- Last item.
//
//!exc: None.
//      
template< class T >
inline const T* Container< T >::back() const
{
    return mData.back();
}


//-----------------------------------------------------------------------------
//
//: Returns an iterator at the beginning of the container.
//
//!return: Container< T >::iterator -- Beginning iterator.
//
//!exc: None.
//      
template< class T >
inline typename Container< T >::iterator Container< T >::
begin()
{
    return mData.begin();
}


//-----------------------------------------------------------------------------
//
//: Returns an iterator at the beginning of the container.
//
//!return: Container< T >::const_iterator -- Beginning iterator.
//
//!exc: None.
//      
template< class T >
inline typename Container< T >::const_iterator Container< T >::
begin() const
{
    return mData.begin();
}


//-----------------------------------------------------------------------------
//
//: Returns an iterator at the end of the container.
//
//!return: Container< T >::iterator -- End iterator.
//
//!exc: None.
//      
template< class T >
inline typename Container< T >::iterator Container< T >::
end()
{
    return mData.end();
}


//-----------------------------------------------------------------------------
//
//: Returns an iterator at the end of the container.
//
//!return: Container< T >::const_iterator -- End iterator.
//
//!exc: None.
//      
template< class T >
inline typename Container< T >::const_iterator Container< T >::
end() const
{
    return mData.end();
}


//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Set Ownership
//
// Sets the ownsership for the element pointed to by iter.
//
//!param: iterator iter - The element whose ownership will be modified.
//!param: bool owns - true if the container should own the data, false otherwise.
//
//!todo: Boundary checking
//
template< class T >
void Container< T >::setOwnership( iterator iter, bool owns )
{
    *( mOwns.begin() + ( iter - begin() ) ) = owns;
}

    
//-----------------------------------------------------------------------------
//
//: Set Ownership.
//
// Sets the ownsership for the element at the given index.
//
//!param: size_t index - The element whose ownership will be modified.
//!param: bool owns - True if the container should own the data, false otherwise.
//
//!todo: Boundary checking
//
template< class T >
void Container< T >::setOwnership( size_t index, bool owns )
{
    *( mOwns.begin() + index ) = owns;
}


//-----------------------------------------------------------------------------
//
//: Append an element to the end of the container.
//
// Memory for the element will be allocated and the container will own the
// new element.
//
//!param: const T& data - The item to append.
//
//!return: iterator -- The iterator correspinding to the appended object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//      
template< class T >
typename Container< T >::iterator Container< T >::
append( const T& data )
{
    mData.push_back( new T( data ) );
    mOwns.push_back( true );

    return end() - 1;
}


//-----------------------------------------------------------------------------
//
//: Append an element to the end of the container.
//
// Memory for the element will be allocated and the container will own the
// new element.
//
//!param: const T* data - The item to append.
//
//!return: iterator -- The iterator corresponding to the appended object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
template< class T >
typename Container< T >::iterator Container< T >::
append( const T* data )
{
    mData.push_back( new T( *data ) );
    mOwns.push_back( true );
    return end() - 1;
}


//-----------------------------------------------------------------------------
//
//: Append an element to the end of the container.
//
// The user may also specifiy whether or not memory is allocated for the
// element (i.e., the element is copied) and if the container owns the element.
//
//!param: const T& data - The item to append.
//!param: bool copy - True if memory should be allocated and the element copied
//+     into the container (default: true).
//!param: bool owns - True if the container will own the element (default: true).
//
//!return: iterator -- The iterator corresponding to the appended object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
template< class T >
typename Container< T >::iterator Container< T >::
append( T* data, bool copy, bool owns )
{
    if ( copy )
    {
        mData.push_back( new T( *data ) );
    }
    else
    {
        mData.push_back( data );
    }

    mOwns.push_back( owns );
    
    return end() - 1;
}

        

//-----------------------------------------------------------------------------
//
//: Insert an element into the container.
//
// The element will be copied and the container will own the new element.
//
//!param: iterator iter - An iterator pointing to the location where the
//+     element should be added.
//!param: const T& data - The item to add.
//
//!return: iterator -- An iterator pointing to the added item.
//
//!exc: std::bad_alloc - Memory could not be allocated for the element.
//
template< class T >
typename Container< T >::iterator Container< T >::
insert( iterator iter, const T& data )
{
    mOwns.insert( mOwns.begin() + ( iter - begin() ), true );
    return mData.insert( iter, new T( data ) );
}


//-----------------------------------------------------------------------------
//
//: Insert an element into the container.
//
// The element will be copied and the container will own the new element.
//
//!param: iterator iter - An iterator pointing to the location where the
//+     element should be added.
//!param: const T* data - The item to add.
//
//!return: iterator -- An iterator pointing to the added item.
//
//!exc: std::bad_alloc - Memory could not be allocated for the element.
//
template< class T >
typename Container< T >::iterator Container< T >::
insert( iterator iter, const T* data )
{
    mOwns.insert( mOwns.begin() + ( iter - begin() ), true );
    return mData.insert( iter, new T( *data ) );
}


//-----------------------------------------------------------------------------
//
//: Insert an element into the container.
//
//!param: iterator iter - An iterator pointing to the location where the
//+     element should be added.
//!param: T* data - The item to add.
//!param: bool copy - Duplicate passed data or not.
//!param: bool owns - Destroy the data when container is destroyed or not.
//
//!return: iterator -- An iterator pointing to the added item.
//
//!exc: std::bad_alloc - Memory could not be allocated for the element.
//
template< class T >
typename Container< T >::iterator Container< T >::
insert( iterator iter, T* data, bool copy, bool owns )
{
    mOwns.insert( mOwns.begin() + ( iter - begin() ), owns );

    if ( copy )
    {
        return mData.insert( iter, new T( *data ) );
    }
    else
    {
        return mData.insert( iter, data );
    }
}


//-----------------------------------------------------------------------------
//
//: Insert an element into the container.
//
// The element will be copied and the container will own the new element.
//
//!param: size_t index - The index at which the element should be added.  If
//+     the index is out of range then the element will be inserted at the end.
//!param: const T& data - The item to add.
//
//!return: iterator -- An iterator pointing to the added item.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
template< class T >
typename Container< T >::iterator Container< T >::
insert( size_t index, const T& data )
{
    if ( index >= mData.size() )
    {
        mOwns.insert( mOwns.end(), true );
        return mData.insert( mData.end(), new T( data ) );
    }

    mOwns.insert( mOwns.begin() + index, true );
    return mData.insert( mData.begin() + index, new T( data ) );
}


//-----------------------------------------------------------------------------
//
//: Insert an element into the container.
//
// The element will be copied and the container will own the new element.
//
//!param: size_t index - The index at which the element should be added.  If
//+     the index is out of range then the element will be inserted at the end.
//!param: const T* data - The item to add.
//
//!return: iterator -- An iterator pointing to the added item.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
template< class T >
typename Container< T >::iterator Container< T >::
insert( size_t index, const T* data )
{
    if ( index >= mData.size() )
    {
        mOwns.insert( mOwns.end(), true );
        return mData.insert( mData.end(), new T( *data ) );
    }

    mOwns.insert( mOwns.end(), true );
    return mData.insert( mData.begin() + index, new T( *data ) );
}


//-----------------------------------------------------------------------------
//
//: Insert an element into the container.
//
//!param: size_t index - The index at which the element should be added.  If
//+     the index is out of range then the element will be inserted at the end.
//!param: T* data - The item to add.
//!param: bool copy - Duplicate passed data or not.
//!param: bool owns - Destroy the data when container is destroyed or not.
//
//!return: iterator -- An iterator pointing to the added item.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
template< class T >
typename Container< T >::iterator Container< T >::
insert( size_t index, T* data, bool copy, bool owns )
{
    if ( index >= mData.size() )
    {
        mOwns.insert( mOwns.end(), owns );

        if ( copy )
        {
            return mData.insert( mData.end(), new T( *data ) );
        }
        else
        {
            return mData.insert( mData.end(), data );
        }
    }

    mOwns.insert( mOwns.begin() + index, owns );
    if ( copy )
    {
        return mData.insert( mData.begin() + index, new T( *data ) );
    }
    else
    {
        return mData.insert( mData.begin() + index, data );
    }
}


//-----------------------------------------------------------------------------
//
//: Erase an element.
//
// This erases an element at the specified index.  If the index is out of range
// then nothing happens.
//
//!param: unsigned int index - Element index.
//
//!exc: None.
//   
template< class T >
void Container< T >::
erase( unsigned int index )
{
    if ( index >= mData.size() )
    {
        return;
    }

    // Create the range of one element to erase   
    iterator start( begin() );
    std::advance( start, index );
   
    iterator end( start);
    ++end; // point to the next element or to the end()
   
    erase( start, end );
}


//-----------------------------------------------------------------------------
//
//: Erases a range of elements.
//
// If end is less thatn start or any of the iterators are out of range
// then nothing will happen.
//
//!param: iterator start - Range start.
//!param: iterator finish - Range end.
//
//!exc: None.
//      
template< class T >
void Container< T >::
erase( iterator start, iterator finish )
{
    if ( finish < start || start < begin() || finish > end() )
    {
        return;
    }
    
    owns_iterator o_start( mOwns.begin() + ( start - begin() ) );
    owns_iterator o_end( mOwns.end() + ( finish - begin() ) );

    iterator iter( start );
    for ( owns_iterator owns = o_start; owns != o_end; ++iter, ++start )
    {
        if ( *owns )
        {
            delete *iter;
            *iter = 0;
        }
    }

    mOwns.erase( o_start, o_end );
    mData.erase( start, finish );
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppContainerHH
