#include "framecpp/config.h"

#include <stdexcept>

#include "detector.hh"
#include "framereaderbase.hh"
#include "output.hh"

#include <math.h>

using std::runtime_error;

using FrameCPP::Version_3_4_5::Detector;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an Detector object with the following properties:
// <ul>
//     <li> name = "none" </li>
//     <li> location of 0 degrees longitude, 0 degrees latitude </li>
//     <li> elevation = 0 meters </li>
//     <li> armXazimuth = 0 </li>
//     <li> armYazimuth = Pi/2 </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
Detector::Detector()
        : Base(), mName( "none" ), mLongitude( 0.), mLatitude( 0.), mElevation( 0. ),
          mArmXAzimuth( 0. ), mArmYAzimuth( 90. ),
          mArmXAltitude( 0. ), mArmYAltitude( 0. ),
          mArmXMidpoint( 0. ), mArmYMidpoint( 0. ),
	  mLocalTime( 0 ), mDataQuality( 0 ), mQaBitList( "" ),
          mMore(), mMoreTable(), mStatData()
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Detector object with the specified attributes.
//
//!param: const std::string& name - Instrument name
//!param: REAL_8 longitude - Instrument Longitude
//!param: REAL_8 latitude - Instrument Latitude
//!param: REAL_4 elevation - Vertex elevation
//!param: REAL_4 armXazimuth - Orientation of X arm, radians CCW from East.
//!param: REAL_4 armYazimuth - Orientation of Y arm, radians CCW from East.
//!param: REAL_4 armXaltitude - Orientation of X arm, degrees above horizon.
//!param: REAL_4 armYaltutude - Orientation of Y arm, degrees above horizon.
//!param: REAL_4 armXmidpoint - Midpoint of X cavity.
//!param: REAL_4 armYmidpoint - Midpoint of Y cavity.
//!param: INT_4S localTime - Local seasonal time minus UTC (seconds)
//!param: INT_4U dataQuality - Data quality word.
//!param: std::string& qaBitList - Data quality bit definition list.
//
//!exc: None.
//      
Detector::Detector(
    const std::string& name,
    REAL_8 longitude, REAL_8 latitude, REAL_4 elevation,
    REAL_4 armXazimuth, REAL_4 armYazimuth,
    REAL_4 armXaltitude, REAL_4 armYaltitude,
    REAL_4 armXmidpoint, REAL_4 armYmidpoint,
    INT_4S localTime, INT_4U dataQuality, const std::string& qaBitList )
        : mName( name ),
	  mLongitude( longitude ), mLatitude( latitude ),
	  mElevation( elevation ),
          mArmXAzimuth( armXazimuth ), mArmYAzimuth( armYazimuth ),
          mArmXAltitude( armXaltitude ), mArmYAltitude( armYaltitude ),
          mArmXMidpoint( armXmidpoint ), mArmYMidpoint( armYmidpoint ),
	  mLocalTime( localTime ),
	  mDataQuality( dataQuality ), mQaBitList( qaBitList ),
          mMore(), mMoreTable(), mStatData()
{
}


//-----------------------------------------------------------------------------
//
// Copy Constructor.  This performs a deep copy of the object.
//
//!param: const Detector& detector - Detector to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Detector::Detector( const Detector& detector ) 
        : Base( detector ), mName( detector.mName ),
          mLongitude( detector.mLongitude ),
          mLatitude( detector.mLatitude ),
          mElevation( detector.mElevation ),
          mArmXAzimuth( detector.mArmXAzimuth ),
          mArmYAzimuth( detector.mArmYAzimuth ),
          mArmXAltitude( detector.mArmXAltitude ),
          mArmYAltitude( detector.mArmYAltitude ),
          mArmXMidpoint( detector.mArmXMidpoint ),
          mArmYMidpoint( detector.mArmYMidpoint ),
	  mLocalTime( detector.mLocalTime ),
	  mDataQuality( detector.mDataQuality ),
	  mQaBitList( detector.mQaBitList ),
	  mMore( detector.mMore ),
	  mMoreTable( detector.mMoreTable ),
          mStatData( detector.mStatData )
{
}

   
//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!kw: virtual inline
//
//!exc: None.
//   
Detector::~Detector()
{
}


//-----------------------------------------------------------------------------
//
//: Gets the instrument name.
//
//!return: const std::string& -- Detector name.
//
//!kw: inline const
//
const std::string& Detector::getName() const
{
    return mName;
}

   
//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Detector& detector - Detector to be assigned.
//
//!return: const Detector& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Detector& Detector::operator=( const Detector& detector )
{
    if ( this != &detector )
    {
        mName = detector.mName;
        mLongitude = detector.mLongitude;
        mLatitude = detector.mLatitude;
        mElevation = detector.mElevation;
        mArmXAzimuth = detector.mArmXAzimuth;
        mArmYAzimuth = detector.mArmYAzimuth;
	mArmXAltitude = detector.mArmXAltitude;
	mArmYAltitude = detector.mArmYAltitude;
	mArmXMidpoint = detector.mArmXMidpoint;
	mArmYMidpoint = detector.mArmYMidpoint;
	mLocalTime = detector.mLocalTime;
	mDataQuality = detector.mDataQuality;
	mQaBitList = detector.mQaBitList;
        mMore = detector.mMore;
        mMoreTable = detector.mMoreTable;
        mStatData = detector.mStatData;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const Detector& detector - Data to be added.
//
//!return: const Detector& -- This object.
//
const Detector& Detector::operator+=( const Detector& detector )
{
   throw runtime_error( "Unimplemented Detector::operator+=() is called." );
  
   return *this;
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.  
//
//!return: SH - Detector structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Detector::getSH()
{
  SH sh( "FrDetector", ID_DETECTOR, "Detector Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "longitude", "REAL_8", "" ) );
  sh.appendSE( SE( "latitude", "REAL_8", "" ) );
  sh.appendSE( SE( "elevation", "REAL_4", "" ) );
  sh.appendSE( SE( "armXazimuth", "REAL_4", "" ) );
  sh.appendSE( SE( "armYazimuth", "REAL_4", "" ) );

  sh.appendSE( SE( "armXaltitude", "REAL_4", "" ) );
  sh.appendSE( SE( "armYaltitude", "REAL_4", "" ) );
  sh.appendSE( SE( "armXmidpoint", "REAL_4", "" ) );
  sh.appendSE( SE( "armYmidpoint", "REAL_4", "" ) );
  sh.appendSE( SE( "localTime", "INT_4S", "" ) );
  sh.appendSE( SE( "dataQuality", "INT_4U", "" ) );
  sh.appendSE( SE( "qaBitList", "STRING", "" ) );

  sh.appendSE( SE( "more", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "moreTable", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrDetector *)", "" ) );

  return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 4 Structure Header for this class.  
//
//!return: SH - Detector structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Detector::getSH4()
{
  SH sh( "FrDetector", ID_DETECTOR, "Detector Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "longitudeD", "INT_2S", "" ) );
  sh.appendSE( SE( "longitudeM", "INT_2S", "" ) );
  sh.appendSE( SE( "longitudeS", "REAL_4", "" ) );
  sh.appendSE( SE( "latitudeD", "INT_2S", "" ) );
  sh.appendSE( SE( "latitudeM", "INT_2S", "" ) );
  sh.appendSE( SE( "latitudeS", "REAL_4", "" ) );
  sh.appendSE( SE( "elevation", "REAL_4", "" ) );
  sh.appendSE( SE( "armXazimuth", "REAL_4", "" ) );
  sh.appendSE( SE( "armYazimuth", "REAL_4", "" ) );
  sh.appendSE( SE( "more", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "moreTable", "PTR_STRUCT(FrTable *)", "" ) );

  return sh;
}

//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 3 Structure Header for this class.  
//
//!return: SH - Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Detector::getSH3()
{
    SH sh( "FrDetector", ID_DETECTOR, "Detector Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "longitudeD", "INT_2S", "" ) );
    sh.appendSE( SE( "longitudeM", "INT_2S", "" ) );
    sh.appendSE( SE( "longitudeS", "REAL_4", "" ) );
    sh.appendSE( SE( "latitudeD", "INT_2S", "" ) );
    sh.appendSE( SE( "latitudeM", "INT_2S", "" ) );
    sh.appendSE( SE( "latitudeS", "REAL_4", "" ) );
    sh.appendSE( SE( "elevation", "REAL_4", "" ) );
    sh.appendSE( SE( "armXazimuth", "REAL_4", "" ) );
    sh.appendSE( SE( "armYazimuth", "REAL_4", "" ) );
    sh.appendSE( SE( "armLength", "REAL_4", "" ) );
    sh.appendSE( SE( "more", "PTR_STRUCT(FrVect *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an Detector object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: Detector* - Newly allocated.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure.
//
Detector* Detector::read( FrameReaderBase& in )
{
  Detector &d =  *new Detector();
  in >> d.mName
     >> d.mLongitude >> d.mLatitude
     >> d.mElevation
     >> d.mArmXAzimuth >> d.mArmYAzimuth
     >> d.mArmXAltitude >> d.mArmYAltitude
     >> d.mArmXMidpoint >> d.mArmYMidpoint
     >> d.mLocalTime >> d.mDataQuality >> d.mQaBitList
     >> d.mMore >> d.mMoreTable;

  in.readNextNodeReference();

  return &d;
}


//: Version 4 frame input method
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure.
//   
Detector* Detector::read4( FrameReaderBase& in )
{
  Detector &d =  *new Detector();
  INT_2S mLatD;
  INT_2S mLatM;
  REAL_4 mLatS;
  INT_2S mLonD;
  INT_2S mLonM;
  REAL_4 mLonS;

  in >> d.mName
     >> mLonD >> mLonM >> mLonS
     >> mLatD >> mLatM >> mLatS
     >> d.mElevation >> d.mArmXAzimuth
     >> d.mArmYAzimuth
     >> d.mMore >> d.mMoreTable;

  // Convert radian to degrees
  // d.mArmXAzimuth *= 180.0/M_PI;
  // d.mArmYAzimuth *= 180.0/M_PI;

  // Convert from North of East to East of North
  d.mArmXAzimuth = M_PI/2 - d.mArmXAzimuth;
  d.mArmYAzimuth = M_PI/2 - d.mArmYAzimuth;

  // Convert coordinates
  if ( mLonD >= 0 )
  {
    d.mLongitude = (REAL_8)mLonD + (REAL_8)mLonM / 60. + (REAL_8)mLonS / 3600.;
  }
  else
  {
    d.mLongitude = (REAL_8)mLonD - (REAL_8)mLonM / 60. - (REAL_8)mLonS / 3600.;
  }
  if ( mLatD >= 0 )
  {
    d.mLatitude = (REAL_8)mLatD + (REAL_8)mLatM / 60. + (REAL_8)mLatS / 3600.;
  }
  else
  {
    d.mLatitude = (REAL_8)mLatD - (REAL_8)mLatM / 60. - (REAL_8)mLatS / 3600.;
  }

  // Correct the position of Hanford and Livingston
  if ( ( mLatD == -119 ) && ( mLonD == 46 ) )
  {
    // Hanford
    d.mLongitude = -119.40765714;
    d.mLatitude = 46.4551467;
  }
  else if ( ( mLatD == -90 ) && ( mLonD == 30 ) )
  {
    // Livingston
    d.mLongitude = -90.77424039;
    d.mLatitude = 30.56289433;
  }
 
  d.mLongitude = (d.mLongitude * M_PI ) / 180.;
  d.mLatitude = (d.mLatitude * M_PI ) / 180.;

  return &d;
}

//: Version 3 frame input method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure.
//      
Detector* Detector::read3( FrameReaderBase& in )
{
  REAL_4 armLength; // dropped for version 4
  INT_2S mLatD;
  INT_2S mLatM;
  REAL_4 mLatS;
  INT_2S mLonD;
  INT_2S mLonM;
  REAL_4 mLonS;

  Detector &d =  *new Detector();
  in >> d.mName
     >> mLonD >> mLonM >> mLonS
     >> mLatD >> mLatM >> mLatS
     >> d.mElevation >> d.mArmXAzimuth
     >> d.mArmYAzimuth >> armLength >> d.mMore;

  // Convert radian to degrees
  // d.mArmXAzimuth *= 180.0/M_PI;
  // d.mArmYAzimuth *= 180.0/M_PI;

  // Convert from North of East to East of North
  d.mArmXAzimuth = M_PI/2 - d.mArmXAzimuth;
  d.mArmYAzimuth = M_PI/2 - d.mArmYAzimuth;

  // Convert coordinates
  if ( mLonD >= 0 )
  {
    d.mLongitude = (REAL_8)mLonD + (REAL_8)mLonM / 60. + (REAL_8)mLonS / 3600.;
  }
  else
  {
    d.mLongitude = (REAL_8)mLonD - (REAL_8)mLonM / 60. - (REAL_8)mLonS / 3600.;
  }
  if ( mLatD >= 0 )
  {
    d.mLatitude = (REAL_8)mLatD + (REAL_8)mLatM / 60. + (REAL_8)mLatS / 3600.;
  }
  else
  {
    d.mLatitude = (REAL_8)mLatD - (REAL_8)mLatM / 60. - (REAL_8)mLatS / 3600.;
  }

  // Correct the position of Hanford and Livingston
  if ( ( mLatD == -119 ) && ( mLonD == 46 ) )
  {
    // Hanford
    d.mLongitude = -119.40765714;
    d.mLatitude = 46.4551467;
  }
  else if ( ( mLatD == -90 ) && ( mLonD == 30 ) )
  {
    // Livingston
    d.mLongitude = -90.77424039;
    d.mLatitude = 30.56289433;
  }
 
  d.mLongitude = (d.mLongitude * M_PI ) / 180.;
  d.mLatitude = (d.mLatitude * M_PI ) / 180.;

  return &d;
}


//-----------------------------------------------------------------------------
//
//: Writes an Detector object to an Output object.
//
//!param: Output& out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void Detector::write( Output& out ) const
{
  out.startRecord( ID_DETECTOR,
		   78
		   + mName.size()
		   + mQaBitList.size(),
		   this );
  out << mName << mLongitude << mLatitude
      << mElevation
      << mArmXAzimuth << mArmYAzimuth
      << mArmXAltitude << mArmYAltitude
      << mArmXMidpoint << mArmYMidpoint
      << mLocalTime << mDataQuality << mQaBitList
      << mMore << mMoreTable;

  for ( StatDataContainer::const_iterator i = mStatData.begin();
	i != mStatData.end(); ++i )
    {
      out.queueObject( **i );
    }
  
  out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Find StatData corresponding to a particular time.
//
//!param: INT_4U time - Time for which to look.
//
//!return: const StatData* - Found data.
//
//!exc: not_found_error - Data not found.
//
const FrameCPP::Version_3_4_5::StatData* FrameCPP::Version_3_4_5::Detector::findStatData( INT_4U time )
    const
{
    StatDataContainer::const_iterator iter = mStatData.begin();
    while(
        ( iter != mStatData.end() ) &&
        ( ( time >= (*iter)->getTimeStart() ) && ( time < (*iter)->getTimeEnd() ) )
        );
    {
        ++iter;
    }
    
    if ( iter == mStatData.end() )
    {
        throw not_found_error( "The data was not found" );
    }

    return *iter;
}

//-----------------------------------------------------------------------------
//
//: Find StatData corresponding to a particular time.
//
//!param: INT_4U time - Time for which to look.
//
//!return: StatData* - Found data.
//
//!exc: not_found_error - Data not found.
//
FrameCPP::Version_3_4_5::StatData* FrameCPP::Version_3_4_5::Detector::findStatData( INT_4U time )
{
    StatDataContainer::iterator iter = mStatData.begin();
    while(
        ( iter != mStatData.end() ) &&
        ( ( time >= (*iter)->getTimeStart() ) && ( time < (*iter)->getTimeEnd() ) )
        );
    {
        ++iter;
    }
    
    if ( iter == mStatData.end() )
    {
        throw not_found_error( "The data was not found" );
    }

    return *iter;
}

    
//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two Detector objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const Detector& detector - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//   
bool Detector::operator==( const Detector& detector ) const
{
    if ( this == &detector )
    {
        return true;
    }

    return ( ( mName == detector.mName ) &&
             ( mLongitude == detector.mLongitude ) &&
             ( mLatitude == detector.mLatitude ) &&
             ( mElevation == detector.mElevation ) &&
             ( mArmXAzimuth == detector.mArmXAzimuth ) &&
             ( mArmYAzimuth == detector.mArmYAzimuth ) &&
             ( mArmXAltitude == detector.mArmXAltitude ) &&
             ( mArmYAltitude == detector.mArmYAltitude ) &&
             ( mArmXMidpoint == detector.mArmXMidpoint ) &&
             ( mArmYMidpoint == detector.mArmYMidpoint ) &&
	     ( mLocalTime == detector.mLocalTime ) &&
	     ( mDataQuality == detector.mDataQuality ) &&
	     ( mQaBitList == detector.mQaBitList ) &&
             ( mMore == detector.mMore ) &&
             ( mMoreTable == detector.mMoreTable ) &&
             ( mStatData == detector.mStatData ) );
}
   
   
//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- Class ID.
//
//!exc: None.
//      
INT_2U Detector::getClassId() const
{
    return ID_DETECTOR;
}
