#include "framecpp/config.h"

#include "endofframe.hh"
#include "framereaderbase.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::EndOfFrame;



//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an EndOfFrame object with the following properties:
// <ul>
//     <li> run = 0 </li>
//     <li> frame = 0 </li>
//     <li> chkType = 0 </li>
//     <li> chkSum = 0 </li>
// </ul>
//
//!exc: None.
//      
EndOfFrame::EndOfFrame()
  : Base(), mRun( 0 ), mFrame( 0 ), mChkType( 0 ), mChkSum( 0 )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.
//
//!param: const EndOfFrame& eofr - The object to be copied.
//
//!exc: None.
//      
EndOfFrame::EndOfFrame( const EndOfFrame& eofr )
  : Base( eofr ), mRun( eofr.mRun ), mFrame( eofr.mFrame ),
    mChkType( eofr.mChkType ), mChkSum( eofr.mChkSum )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: INT_4S run - Frame run number.
//!param: INT_4U frame - Frame number.
//!param: INT_4U chktype - Frame checksum scheme.
//!param: INT_4U chksum - File checksum.
//
//!exc: None.
//      
EndOfFrame::EndOfFrame( INT_4S run, INT_4U frame, INT_4U chktype, INT_4U chksum )
  : Base(), mRun( run ), mFrame( frame ), mChkType( chktype ), mChkSum( chksum )
{
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.
//   
EndOfFrame::~EndOfFrame()
{
}
   

//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const EndOfFrame& eofr - Object to be assigned.
//
//!return: const EndOfFrame& - This object.
//
//!exc: None
//
const EndOfFrame& EndOfFrame::operator=( const EndOfFrame& eofr )
{
    if ( this != &eofr )
    {
        mRun = eofr.mRun;
        mFrame = eofr.mFrame;
	mChkType = eofr.mChkType;
	mChkSum = eofr.mChkSum;
    }

    return *this;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.  
//
//!return: SH - End of frame structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH EndOfFrame::getSH()
{
    SH sh( "FrEndOfFrame", ID_ENDOFFRAME, "End of Frame Data" );
    sh.appendSE( SE( "run", "INT_4S", "" ) );
    sh.appendSE( SE( "frame", "INT_4U", "" ) );
    sh.appendSE( SE( "chkType", "INT_4U", "" ) );
    sh.appendSE( SE( "chkSum", "INT_4U", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 4 Structure Header for this class.  
//
//!return: SH - End of frame structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH EndOfFrame::getSH4()
{
    SH sh( "FrEndOfFrame", ID_ENDOFFRAME, "End of Frame Data" );
    sh.appendSE( SE( "run", "INT_4S", "" ) );
    sh.appendSE( SE( "frame", "INT_4U", "" ) );

    return sh;
}

//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 3 Structure Header for this class.  
//
//!return: SH - End of frame structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH EndOfFrame::getSH3()
{
    SH sh( "FrEndOfFrame", ID_ENDOFFRAME, "End of Frame Data" );
    sh.appendSE( SE( "run", "INT_4U", "" ) );
    sh.appendSE( SE( "frame", "INT_4U", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an EndOfFrame object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: EndOfFrame* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
EndOfFrame* EndOfFrame::read( FrameReaderBase& in )
{
  EndOfFrame& eof = *new EndOfFrame();
  in >> eof.mRun >> eof.mFrame >> eof.mChkType >> eof.mChkSum;
  return &eof;
}

//: Version 4 Frame read method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
EndOfFrame* EndOfFrame::read4( FrameReaderBase& in )
{
  EndOfFrame& eof = *new EndOfFrame();
  in >> eof.mRun >> eof.mFrame;
  return &eof;
}

//: Version 3 Frame read method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
EndOfFrame* EndOfFrame::read3( FrameReaderBase& in )
{
  EndOfFrame& eof = *new EndOfFrame();
  INT_4U run;
  in >> run >> eof.mFrame;
  eof.mRun = (INT_4S) run;
  return &eof;
}


//-----------------------------------------------------------------------------
//
//: Writes an EndOfFrame object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void EndOfFrame::write( Output& out ) const 
{
    out.startRecord( ID_ENDOFFRAME, 24, this );
    out << mRun << mFrame << mChkType << mChkSum;
    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- Class ID.
//
//!exc: None.
//      
INT_2U EndOfFrame::getClassId() const
{
    return ID_ENDOFFRAME;
}


    
