#ifndef FrameCppEventHH
#define FrameCppEventHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"
#include "framecpp/Version3_4_5/time.hh"

#include "framecpp/SearchContainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrEvent;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: Event Data
//
// This class represents event data.  It is modeled after the FrEvent
// structure in the Frame Specification, LIGO-T970130-B-E.
//
// <P>The Event class stores the following attributes:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B>                        </td></tr>
//   <tr><td> name       </td><td>Name of event                                </td></tr>
//   <tr><td> comment    </td><td>Descriptor of event                          </td></tr>
//   <tr><td> inputs     </td><td>Input channels and filter parameters to event
//                        process                                        </td></tr>
//   <tr><td> GTime      </td><td>Event occurrent time (integer seconds & nanoseconds)
//   <tr><td> timeBefore </td><td>Signal duration before event occurence time    </td></tr>
//   <tr><td> timeAfter  </td><td>Signal duration after event occurence time     </td></tr>
//   <tr><td> eventStatus</td><td>Defined by event                           </td></tr>
//   <tr><td> amplitude  </td><td>Continuous output amplitude returned by event</td></tr>
//   <tr><td> probability</td><td>Likelihood estimate of event.                  </td></tr>
//   <tr><td> statistics </td><td>Statistical description of event               </td></tr>
//   <tr><td> parameters </td><td>Additional event parameters                    </td></tr>
// </table>
//
// <p>The Event object also contains the following data:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Data</B></td><td><B>Description</B> </td></tr>
//   <tr><td> data</td><td>Vect Container for additional event results. </td></tr>
//   <tr><td> table</td><td>Table Container for additional event data. </td></tr>
// </table>
//
// <p>The data is available through the container object inside Event.
// This are accessed via the <code>refData</code> and <code>refTable</code> method.
//
class Event : public Base
{
public:

    //!ignore_begin:
    DECLARE_DEFAULT_CONTAINER( Data, data, Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    //!ignore_end:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a Event object with the following properties:
    // <ul type=disc>
    //     <li> name = "none" </li>
    //     <li> comment = "" </li>
    //     <li> inputs = "" </li>
    //     <li> GTime = 0 seconds, 0 nanoseconds </li>
    //     <li> timeBefore = 0 </li>
    //     <li> timeAfter = 0 </li>
    //     <li> eventStatus = 0 </li>
    //     <li> amplitude = 0 </li>
    //     <li> probability = -1 </li>
    //     <li> statistics = "" </li>
    //     <li> parameters = empty list </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.
    //   
    Event();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.
    //
    //!param: const Event& eventData - Event data to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Event( const Event& eventData );

    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    // Creates a Event object with the specified attributes but no data.
    //
    //!param: const std::string& name - Event name.
    //!param: const std::string& comment - Event description.
    //!param: const std::string& inputs - Input channels and filter parameters to
    //+     evenet process.
    //!param: const Time& time - Event occurrence time.
    //!param: REAL_4 timeBefore -  Signal duration before event occurrence.
    //!param: REAL_4 timeAfter - Signar dutaion after event occurrence.
    //!param: INT_4U eventStatus - Defined by event.
    //!param: REAL_4 amplitude - Continuous output amplitude returned by event.
    //!param: REAL_4 prob - Likelihood estimate of event.
    //!param: const std::string& statistics - Statistical description.
    //!param: const std::vector<std::pair<std::string, REAL_4> >& parameters - Additional event parameters.
    //
    //!exc: None.
    //      
    Event( const std::string& name, const std::string& comment, const std::string& inputs,
	   const Time& time,
	   REAL_4 timeBefore, REAL_4 timeAfter,
	   INT_4U eventStatus, REAL_4 amplitude,
	   REAL_4 prob,
	   const std::string& statistics,
	   const std::vector<std::pair<std::string, REAL_4> >& parameters );

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    // A deep copy is performed for owned data.
    //
    //!param: const Event& eventData - Data to be copied.
    //
    //!return: const Event& -- This event.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Event& operator=( const Event& eventData );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const Event& event - Data to be added.
    //
    //!return: const Event& -- This object.
    //
    const Event& operator+=( const Event& event );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two Event objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const Event& eventData - The object to compare with.
    //
    //!return: bool -- True if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.
    //      
    bool operator==( const Event& eventData ) const;
    bool operator!=( const Event& eventData ) const;

    /* Required Methods */
    //!exc: None.
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Structure Header.
    //
    // Constructs the Structure Header for this class.
    //
    //!return: SH -- Structure header for Event Data.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    static SH getSH();

    //-----------------------------------------------------------------------------
    //
    //: Structure Header.
    //
    // Constructs frame data format version 4 Structure Header for this class.
    //
    //!return: SH -- Structure header for Event Data.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    static SH getSH4();

    //-----------------------------------------------------------------------------
    //
    //: Structure Header.
    //
    // Constructs frame data format version 3 Structure Header for this class.
    //
    //!return: SH -- Structure header for Event Data.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    static SH getSH3();

    /* Accessors */
    const std::string& getName() const;
   
    //!exc: None.
    const std::string& getComment() const;
   
    //!exc: None.   
    const std::string& getInputs() const;
   
    //!exc: None.   
    const Time& getGTime() const;
   
    //!exc: None.   
    REAL_4 getTimeBefore() const;
   
    //!exc: None.   
    REAL_4 getTimeAfter() const;
   
    //!exc: None.   
    INT_4U getEventStatus() const;
   
    //!exc: None.   
    REAL_4 getAmplitude() const;
   
    //!exc: None.   
    REAL_4 getProbability() const;
   
    //!exc: None.   
    const std::string& getStatistics() const;

    //!exc: None.   
    const std::vector<std::pair<std::string, REAL_4> >& getParameters() const;

    /* Data Accessors */
    //!exc: None.   
    Event::DataContainer& refData();
   
    //!exc: None.   
    const Event::DataContainer& refData() const;

    //!exc: None.   
    Event::TableContainer& refTable();
   
    //!exc: None.   
    const Event::TableContainer& refTable() const;
    
private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrEvent;

    /* This method will be accessed by the Dictionary class but is
       not available for public use. */

    //-----------------------------------------------------------------------------
    //
    //: Reads a Event object from an Input object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: Input& in - Frame reader.
    //
    //!return: Event* -- Newly allocated and read data.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //!exc: read_failure - Read failed
    //
    static Event* read( FrameReaderBase& in );
   
    //: Read FrameCPP version 4 Trigger Data structure.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //!exc: read_failure - Read failed
    //   
    static Event* read4( FrameReaderBase& in );
   
    //: Read FrameCPP version 3 Trigger Data structure.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //!exc: read_failure - Read failed
    //   
    static Event* read3( FrameReaderBase& in );


    //-----------------------------------------------------------------------------
    //
    //: Writes an Event object to an Output object.
    //
    //!param: Output& out out - Output data object.
    //
    //!exc: write_failure - Write failed
    //
    virtual void write( Output& out ) const;

private:

    //: Name of event.
    std::string mName;
    //: Comment.
    std::string mComment;
    //: Input channels and filter parameters to event process.
    std::string mInputs;
    //: GPS time.
    Time mGTime;
    //: Signal duration before mGTime.
    REAL_4 mTimeBefore;
    //: Signal duration after mGTime.
    REAL_4 mTimeAfter;
    //: Defined by event.
    INT_4U mEventStatus;
    //: Continuous output amplitude returned by event.
    REAL_4 mAmplitude;
    //: Likelihood estimate of event (eq. -1, if can not be estimated).
    REAL_4 mProbability;
    //: Statistical description of event.
    std::string mStatistics;
    //: Additional parameters
    std::vector<std::pair<std::string, REAL_4> > mParameters;
    //: Additional event results.
    DataContainer mData;
    //: Additional event information.
    TableContainer mTable;
};


/******************/
/* Inline Methods */
/******************/



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal Comparison.
//
// Determines whether two objects are not equal.  All contained data is
// compared.
//
//!param: const Event& eventData - The object to compare with.
//
//!return: bool -- True if the objects are not equal, false otherwise.
//
//!exc: None.
//      
inline bool Event::operator!=( const Event& eventData ) const
{
    return !( *this == eventData );
}


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Event description.
//
// Gets the description of the event.
//
//!return: const std::string& -- The event description.
//
//!exc: None.
//      
inline const std::string& Event::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Input channels.
//
// Gets the input channels and filter parameters to the event process.
//
//!return: const std::string& -- The input channels and filter parameters.
//
//!exc: None.
//      
inline const std::string& Event::getInputs() const
{
    return mInputs;
}


//-----------------------------------------------------------------------------
//
//: Event occurrence time.
//
// Gets the event occurrence time.  This is the integer seconds and nanoseconds
// since the beginning of the frame.
//
//!return: const Time& -- The event occurrence time.
//
//!exc: None.
//      
inline const Time& Event::getGTime() const
{
    return mGTime;
}


//-----------------------------------------------------------------------------
//
//: Signal duration before event occurence
//
// Gets signal duration before event occurence.
//
//!return: REAL_4 -- Signal duration before event occurence.
//
//!exc: None.
//      
inline REAL_4 Event::getTimeBefore() const
{
  return mTimeBefore;
}


//-----------------------------------------------------------------------------
//
//: Signal duration after event occurence
//
// Gets signal duration after event occurence.
//
//!return: REAL_4  -- signal duration after event occurence.
//
//!exc: None.
//      
inline REAL_4 Event::getTimeAfter() const
{
  return mTimeAfter;
}


//-----------------------------------------------------------------------------
//
//: Event status defined by the event.
//
// Gets event status.
//
//!return: INT_4U -- Event status.
//
//!exc: None.
//      
inline INT_4U Event::getEventStatus() const
{
  return mEventStatus;
}


//-----------------------------------------------------------------------------
//
//: Continuous output amplitude returned by event.
//
// Gets continuous output amplitude returned by event.
//
//!return: REAL_4 -- Event amplitude.
//
//!exc: None.
//      
inline REAL_4 Event::getAmplitude() const
{
  return mAmplitude;
}


//-----------------------------------------------------------------------------
//
//: Probability.
//
// Returns the likelihood estimate of the event, if available.  This will be
// -1 if it cannot be estimated.
//
//!return: REAL_4 -- Probability.
//
//!exc: None.
//      
inline REAL_4 Event::getProbability() const
{
    return mProbability;
}


//-----------------------------------------------------------------------------
//
//: Statistical description.
//
// Gets the statistical description of the event (if relevant or applicable).
//
//!return: const std::string& -- Statistics.
//
//!exc: None.
//      
inline const std::string& Event::getStatistics() const
{
    return mStatistics;
}


//-----------------------------------------------------------------------------
//
//: Additional parameters
//
// Gets additional parameter vector
//
//!return: const std::vector<std::pair<std::string, REAL_4> > -- parameters.
//
//!exc: None.
//      
inline const std::vector<std::pair<std::string, REAL_4> >&
Event::getParameters() const
{
    return mParameters;
}


//-----------------------------------------------------------------------------
//
//: Data Container.
//
// Get the container which contains the data.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Data container.
//
//!kw: inline
//
//!exc: None.
//      
inline Event::DataContainer& Event::refData()
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Data Container
//
// Get the container which contains the data.
//
//!return: const SearchContainer< Vect, &Vect::getName >&
//
//!kw: inline const
//
//!exc: None.
//      
inline const
Event::DataContainer& Event::refData() const
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Table Container.
//
// Get table container.
//
//!return: SearchContainer< Table, &Table::getName >& -- Table.
//
//!kw: inline
//
//!exc: None.
//      
inline Event::TableContainer& Event::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Table Container
//
// Get table container.
//
//!return: const SearchContainer< Table, &Table::getName >&  -- Table.
//
//!kw: inline const 
//
//!exc: None.
//      
inline const
Event::TableContainer& Event::refTable() const
{
  return mTable;
}


} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppEventHH
