#include "framecpp/config.h"

#include "framecpp/FrameCPP.hh"

#include "fileheader.hh"
#include "input.hh"
#include "output.hh"

using FrameCPP::Version_3_4_5::FileHeader;

//-----------------------------------------------------------------------------
//
//: Copy constructor.
//
//!param: const FileHeader& header - Object to be copied.
//
//!except: std::bad_alloc - Memory allocation failed.
//
FileHeader::FileHeader( const FileHeader& header )
        : mInt2UReader( 0 ), mInt2SReader( 0 ), mInt4UReader( 0 ),
          mInt4SReader( 0 ), mInt8UReader( 0 ), mInt8SReader( 0 ),
          mReal4Reader( 0 ), mReal8Reader( 0 ), mData( 0 )
{
    if ( header.mData != 0 )
    {
        mData = new unsigned char[ 41 ];
        memcpy( mData, header.mData, 41 );
        update();
    }
}


//-----------------------------------------------------------------------------
//
//: Input constructor.
//
// This constructs a FileHeader by reading it from an Input object.
//
//!param: Input& in - Input data stream object.
//
//!except: std::bad_alloc - Memory allocation failed
//!exc: read_failure - Read failed.   
//
FileHeader::FileHeader( Input& in )
try
        : mInt2UReader( 0 ), mInt2SReader( 0 ), mInt4UReader( 0 ),
          mInt4SReader( 0 ), mInt8UReader( 0 ), mInt8SReader( 0 ),
          mReal4Reader( 0 ), mReal8Reader( 0 ), mData( 0 )
{
    mData = new unsigned char[ 41 ];
    
    in.read( mData, 40 );
    mData[ 40 ] = 0;
/*
    Not sure if I should check the magic word...

    if (mData[0] != 'I' || mData[1] != 'G' 
	|| mData[2] != 'W' || mData[3] != 'D')
	throw read_failure("bad frame file header", 0);   
*/

    update();
}
catch( ... )
{
    delete[] mData;
    mData = 0;
    throw;
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
//!exc: None
//         
FileHeader::~FileHeader()
{
    delete[] mData;
    mData = 0;
}


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const FileHeader& header - Object to be assigned.
//
//!return: const FileHeader& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
const FileHeader& FileHeader::operator=( const FileHeader& header )
{
    if ( this != &header )
    {
        if ( header.mData == 0 )
        {
            mData = 0;
            mInt2UReader = 0;
            mInt2SReader = 0;
            mInt4UReader = 0;
            mInt4SReader = 0;
            mInt8UReader = 0;
            mInt8SReader = 0;
            mReal4Reader = 0;
            mReal8Reader = 0;
        }
        else
        {
            delete[] mData;
            mData = 0;
   
            mData = new unsigned char[ 41 ];
            memcpy( mData, header.mData, 41 );
            update();
        }        
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Writes specified frame data format version file header to output.
//
//!param: out (Output&) -- output frame file
//!param: out (version) -- data format version
//
//!except: write_failure - Data write failed.
//
void
FileHeader::writeLocal( Output& out, int version )
{
  static unsigned char mLocal[41];
  static int mLocalInited = 0;

  // static flag makes initialization block run only first time
  // and be reused later
  if (! mLocalInited )
    {
      memset( mLocal, 0x00, 41 );
      INT_2U i2( 0x1234 );
      INT_4U i4( 0x12345678 );
      INT_8U i8( 0x01234567 );
      i8 <<= 32;
      i8 += 0x89ABCDEF;
      REAL_4 r4( 3.1415926535897932384 );
      REAL_8 r8( 3.1415926535897932384 );
      
      strcpy( (char*)mLocal, "IGWD" );
      mLocal[ 5 ] = Library::FRAMECPP_VERSION;
      mLocal[ 6 ] = Library::FRAMECPP_REVISION;
      mLocal[ 7 ] = sizeof( INT_2U );
      mLocal[ 8 ] = sizeof( INT_4U );
      mLocal[ 9 ] = sizeof( INT_8U );
      mLocal[ 10 ] = sizeof( REAL_4 );
      mLocal[ 11 ] = sizeof( REAL_8 );
      
      // What if these data types are larger than expected?
      // What is they are smaller than expected?
      memcpy( mLocal + 12, &i2, sizeof( INT_2U ) );
      memcpy( mLocal + 14, &i4, sizeof( INT_4U ) );
      memcpy( mLocal + 18, &i8, sizeof( INT_8U ) );
      memcpy( mLocal + 26, &r4, sizeof( REAL_4 ) );
      memcpy( mLocal + 30, &r8, sizeof( REAL_8 ) );    
      strcpy( (char*)( mLocal + 38 ), "AZ" );
      mLocalInited = 1;
    }

  // need to do three writes here to be thread safe (can't touch mLocal)
  //
  out.write( mLocal, 5 );
  unsigned char cver = (unsigned char) version;
  out.write( &cver, 1 );
  out.write( mLocal + 6, 40 - 6);
}


//-----------------------------------------------------------------------------
//
//: Read new file header from the input file
//
//!exc: std::bad_alloc - Memory allocation failed
//!exc: read_failure - Read failed.   
//   
FileHeader* FileHeader::read( Input& in )
{
    return new FileHeader( in );
}


//-----------------------------------------------------------------------------
//
//: Default constructor.
//
// This constructs a FileHeader and puts nothing into it
//
//
FileHeader::FileHeader( const char* Header )
{
  mData = new unsigned char[ 41 ];
  memcpy( mData, Header, 40 );
  mData[ 40 ] = '\0';
  update( );
#if 0
  if ( ByteOrder == FrameCPP::BYTE_ORDER_BIG_ENDIAN )
  {
    mInt2UReader = &FileHeader::readInt2U_B;
    mInt2SReader = &FileHeader::readInt2S_B;
    mInt4UReader = &FileHeader::readInt4U_4B;
    mInt4SReader = &FileHeader::readInt4S_4B;
    mInt8UReader = &FileHeader::readInt8U_8B;
    mInt8SReader = &FileHeader::readInt8S_8B;
    mReal4Reader = &FileHeader::readReal4_B;
    mReal8Reader = &FileHeader::readReal8_B;
  }
  else
  {
    mInt2UReader = &FileHeader::readInt2U_L;
    mInt2SReader = &FileHeader::readInt2S_L;
    mInt4UReader = &FileHeader::readInt4U_4L;
    mInt4SReader = &FileHeader::readInt4S_4L;
    mInt8UReader = &FileHeader::readInt8U_8L;
    mInt8SReader = &FileHeader::readInt8S_8L;
    mReal4Reader = &FileHeader::readReal4_L;
    mReal8Reader = &FileHeader::readReal8_L;
  }
#endif /* 0 */
}

//-----------------------------------------------------------------------------
//
//!exc: None.
//   
void FileHeader::update()
{
    checkInt2();
    checkInt4();
    checkInt8();
}

    
//-----------------------------------------------------------------------------
//
//!exc: None.
//      
void FileHeader::checkInt2()
{
    const unsigned char d1[ 2 ] = { 0x12, 0x34 };
    const unsigned char d2[ 2 ] = { 0x34, 0x12 };
    
    if ( getInt2Size() >= 2 )
    {
        if ( memcmp( d1, mData + 12, 2 ) == 0 )
        {
            mInt2UReader = &FileHeader::readInt2U_B;
            mInt2SReader = &FileHeader::readInt2S_B;
            mReal4Reader = &FileHeader::readReal4_B;
            mReal8Reader = &FileHeader::readReal8_B;
        }
        else if ( memcmp( d2, mData + 12, 2 ) == 0 )
        {
            mInt2UReader = &FileHeader::readInt2U_L;
            mInt2SReader = &FileHeader::readInt2S_L;
            mReal4Reader = &FileHeader::readReal4_L;
            mReal8Reader = &FileHeader::readReal8_L;
        }
        else
        {
            mInt2UReader = 0;
            mInt2SReader = 0;
            mReal4Reader = 0;
            mReal8Reader = 0;
        }        
    }
    else
    {
        mInt2UReader = 0;
        mInt2SReader = 0;
        mReal4Reader = 0;
        mReal8Reader = 0;
    }
}


//-----------------------------------------------------------------------------
//
//!exc: None.
//      
void FileHeader::checkInt4()
{
    const unsigned char d1[ 4 ] = { 0x12, 0x34, 0x56, 0x78 };
    const unsigned char d2[ 4 ] = { 0x78, 0x56, 0x34, 0x12 };
    const unsigned char d5[ 2 ] = { 0x56, 0x78 };
    const unsigned char d6[ 2 ] = { 0x78, 0x56 };
    
    if ( getInt4Size() >= 4 )
    {
        if ( memcmp( d1, mData + 14, 4 ) == 0 )
        {
            mInt4UReader = &FileHeader::readInt4U_4B;
            mInt4SReader = &FileHeader::readInt4S_4B;
        }
        else if ( memcmp( d2, mData + 14, 4 ) == 0 )
        {
            mInt4UReader = &FileHeader::readInt4U_4L;
            mInt4SReader = &FileHeader::readInt4S_4L;
        }
        else
        {
            mInt4UReader = 0;
            mInt4SReader = 0;
        }        
    }
    else if ( getInt4Size() == 2 )
    {
        if ( memcmp( d5, mData + 14, 2 ) == 0 )
        {
            mInt4UReader = &FileHeader::readInt4U_2B;
            mInt4SReader = &FileHeader::readInt4S_2B;
        }
        else if ( memcmp( d6, mData + 14, 2 ) == 0 )
        {
            mInt4UReader = &FileHeader::readInt4U_2L;
            mInt4SReader = &FileHeader::readInt4S_2L;
        }
        else
        {
            mInt4UReader = 0;
            mInt4SReader = 0;
        }
    }
    else
    {
        mInt4UReader = 0;
        mInt4SReader = 0;
    }
}


//-----------------------------------------------------------------------------
//
//!exc: None.
//     
void FileHeader::checkInt8()
{
    const unsigned char d1[ 8 ] = { 0x01, 0x23, 0x45, 0x67,
                                           0x89, 0xAB, 0xCD, 0xEF };
    const unsigned char d2[ 8 ] = { 0xEF, 0xCD, 0xAB, 0x89,
                                           0x67, 0x45, 0x23, 0x01 };
    const unsigned char d3[ 4 ] = { 0x01, 0x23, 0x45, 0x67 };
    const unsigned char d4[ 4 ] = { 0x67, 0x45, 0x23, 0x01 };
    const unsigned char d5[ 4 ] = { 0x89, 0xAB, 0xCD, 0xEF };
    const unsigned char d6[ 4 ] = { 0xEF, 0xCD, 0xAB, 0x89 };
    
    if ( getInt8Size() >= 8 )
    {
        if ( ( memcmp( d1, mData + 18, 8 ) == 0 ) ||
	     ( memcmp( d3, mData + 18, 4 ) == 0 ) )
        {
            mInt8UReader = &FileHeader::readInt8U_8B;
            mInt8SReader = &FileHeader::readInt8S_8B;
        }
        else if ( ( memcmp( d2, mData + 18, 8 ) == 0 ) ||
		  ( memcmp( d4, mData + 18, 4 ) == 0 ) )
        {
            mInt8UReader = &FileHeader::readInt8U_8L;
            mInt8SReader = &FileHeader::readInt8S_8L;
        }
        else
        {
            mInt8UReader = 0;
            mInt8SReader = 0;
        }        
    }
    else if ( getInt8Size() == 4 )
    {
        if ( ( memcmp( d5, mData + 18, 4 ) == 0 ) ||
             ( memcmp( d3, mData + 18, 4 ) == 0 ) )
        {
            mInt8UReader = &FileHeader::readInt8U_4B;
            mInt8SReader = &FileHeader::readInt8S_4B;
        }
        else if ( ( memcmp( d6, mData + 18, 4 ) == 0 ) ||
                  ( memcmp( d4, mData + 18, 4 ) == 0 ) )
        {
            mInt8UReader = &FileHeader::readInt8U_4L;
            mInt8SReader = &FileHeader::readInt8S_4L;
        }
        else
        {
            mInt8UReader = 0;
            mInt8SReader = 0;
        }
    }
    else
    {
        mInt8UReader = 0;
        mInt8SReader = 0;
    }
}



//-----------------------------------------------------------------------------
/****************/
/* Read Methods */
/****************/


/**********/
/* INT_2U */
/**********/


/**
 * File: 12 34
 * 2B:   12 34
 * 2L:   34 12
 */
//
//!exc: read_failure - Read failed.
//      
void FileHeader::readInt2U_B( Input& in, INT_2U* v, unsigned int n )
{
    in.read( v, 2 * n );
#   ifndef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 2 >( v, 1 );
            v++;
        }
    }
#   endif
}


//-----------------------------------------------------------------------------
/**
 * File: 34 12
 * 2B:   12 34
 * 2L:   34 12
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt2U_L( Input& in, INT_2U* v, unsigned int n )
{
    in.read( v, 2 * n );
#   ifdef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 2 >( v, 1 );
            v++;
        }
    }
#   endif
}


/**********/
/* INT_2S */
/**********/


/**
 * File: s2 34
 * 2B:   s2 34
 * 2L:   34 s2
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt2S_B( Input& in, INT_2S* v, unsigned int n )
{
    in.read( v, 2 * n );
#   ifndef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 2 >( v, 1 );
            v++;
        }
    }
#   endif
}


//-----------------------------------------------------------------------------
/**
 * File: 34 s2
 * 2B:   s2 34
 * 2L:   34 s2
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt2S_L( Input& in, INT_2S* v, unsigned int n )
{
    in.read( v, 2 * n );
#   ifdef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 2 >( v, 1 );
            v++;
        }
    }
#   endif
}


//-----------------------------------------------------------------------------
/**********/
/* INT_4U */
/**********/


/**
 * File: 01 23 45 67
 * 4B:   01 23 45 67
 * 4L:   67 45 23 01
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4U_4B( Input& in, INT_4U* v, unsigned int n )
{
    in.read( v, 4 * n );
#   ifndef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 4 >( v, 1 );
            v++;
        }
    }
#   endif
}


    
//-----------------------------------------------------------------------------
/**
 * File: 67 45 23 01
 * 4B:   01 23 45 67
 * 4L:   67 45 23 01
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4U_4L( Input& in, INT_4U* v, unsigned int n )
{
    in.read( v, 4 * n );
#   ifdef WORDS_BIGENDIAN
    {
        for( int i = n; i != 0; --i )
        {
            reverse< 4 >( v, 1 );
        }
    }
#   endif
}



//-----------------------------------------------------------------------------
/**
 * File: 45 67 xx xx
 * 4B:   00 00 45 67
 * 4L:   67 45 00 00
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4U_2B( Input& in, INT_4U* v, unsigned int n )
{
    memset( v, 0, 4 * n );
    for ( int i = n; i != 0; --i )
    {
#       ifdef WORDS_BIGENDIAN
        {
            in.read( ((char*)v) + 2, 2 );
        }
#       else
        {
            in.read( v, 2 );
            reverse< 2 >( v, 1 );
        }
#       endif        
        v++;
    }
}
    
    
//-----------------------------------------------------------------------------
/**
 * File: 67 45 xx xx
 * 4B:   00 00 45 67
 * 4L:   67 45 00 00
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4U_2L( Input& in, INT_4U* v, unsigned int n )
{
    memset( v, 0, 4 * n );
    for ( int i = n; i != 0; --i )
    {
#       ifdef WORDS_BIGENDIAN
        {
            in.read( ((char*)v) + 2, 2 );
            reverse< 2 >( ((char*)v) + 2, 1 );
        }
#       else
        {
            in.read( v, 2 );
        }
#       endif
        v++;
    }
}
    
    
//-----------------------------------------------------------------------------
/**********/
/* INT_4S */
/**********/


/**
 * File: s1 23 45 67
 * 4B:   s1 23 45 67
 * 4L:   67 45 23 s1
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4S_4B( Input& in, INT_4S* v, unsigned int n )
{
    in.read( v, 4 * n );
#   ifndef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 4 >( v, 1 );
            v++;
        }
    }
#   endif
}


    
//-----------------------------------------------------------------------------
/**
 * File: 67 45 23 s1
 * 4B:   s1 23 45 67
 * 4L:   67 45 23 s1
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4S_4L( Input& in, INT_4S* v, unsigned int n )
{
    in.read( v, 4 * n );
#   ifdef WORDS_BIGENDIAN
    {
        for( int i = n; i != 0; --i )
        {
            reverse< 4 >( v, 1 );
        }
    }
#   endif
}



//-----------------------------------------------------------------------------
/**
 * File: s5 67 xx xx
 * 4B:   ss ss s5 67
 * 4L:   67 s5 ss ss
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4S_2B( Input& in, INT_4S* v, unsigned int n )
{
    memset( v, 0, 4 * n );
    for ( int i = n; i != 0; --i )
    {
#       ifdef WORDS_BIGENDIAN
        {
            in.read( ((char*)v) + 2, 2 );
            if ( *( ((unsigned char*)v) + 2 ) & 0x80 != 0 )
            {
                memset( v, 0xFF, 2 );
            }
        }
#       else
        {
            in.read( v, 2 );
            if ( *(unsigned char*)v & 0x80 != 0 )
            {
                memset( (unsigned char*)v + 2, 0xFF, 2 );
            }
            reverse< 2 >( v, 1 );
        }
#       endif
        v++;
    }
}
    
    
//-----------------------------------------------------------------------------
/**
 * File: 67 s5 xx xx
 * 4B:   ss ss s5 67
 * 4L:   67 s5 ss ss
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt4S_2L( Input& in, INT_4S* v, unsigned int n )
{
    memset( v, 0, 4 * n );
    for ( int i = n; i != 0; --i )
    {
#       ifdef WORDS_BIGENDIAN
        {
            in.read( ((char*)v) + 2, 2 );
            reverse< 2 >( ((char*)v) + 2, 1 );
            if ( *( ((unsigned char*)v) + 2 ) & 0x80 != 0 )
            {
                memset( v, 0xFF, 2 );
            }
        }
#       else
        {
            in.read( v, 2 );
            if ( *( ((unsigned char*)v) + 1 ) & 0x80 != 0 )
            {
                memset( (unsigned char*)v + 2, 0xFF, 2 );
            }
        }
#       endif
        v++;
    }
}
    
    

//-----------------------------------------------------------------------------
/**********/
/* INT_8U */
/**********/


/**
 * Reads an 8-byte INT_8U in Big-Endian order.  If the reading machine has
 * only a 4-byte integer, then only the low-order bytes are stored.
 *
 * File: 01 23 45 67 89 AB CD EF
 * 8B:   01 23 45 67 89 AB CD EF
 * 8L:   EF CD AB 89 67 45 23 01
 * 4B:   89 AB CD EF 
 * 4L:   EF CD AB 89
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8U_8B( Input& in, INT_8U* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        in.read( v, 8 * n );
#       ifndef WORDS_BIGENDIAN
        {
            for ( int i = n; i != 0; --i )
            {
                reverse< 8 >( v, 1 );
                v++;
            }
        }
#       endif
    }
    else
    {
        for ( int i = n; i != 0; --i )
        {
            // Skip the high-order bytes
            in.skip( 4 );
            in.read( v, 4 );
#           ifndef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}

    
//-----------------------------------------------------------------------------
/**
 * Reads an 8-byte INT_8U in Little-Endian order.  If the reading machine has
 * only a 4-byte integer, then only the low-order bytes are stored.
 *
 * File: EF CD AB 89 67 45 23 01
 * 8B:   01 23 45 67 89 AB CD EF
 * 8L:   EF CD AB 89 67 45 23 01
 * 4B:   89 AB CD EF 
 * 4L:   EF CD AB 89
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8U_8L( Input& in, INT_8U* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        in.read( v, 8 * n );
#       ifdef WORDS_BIGENDIAN
        {
            for( int i = n; i != 0; --i )
            {
                reverse< 8 >( v, 1 );
            }
        }
#       endif
    }
    else
    {
        for ( int i = n; i != 0; --i )
        {
            in.read( v, 4 );
#           ifdef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            // Skip the high order bytes
            in.skip( 4 );
            v++;
        }
    }
}


//-----------------------------------------------------------------------------
/**
 * Reads a big-endian 4 byte INT_8U.
 *
 * File: 89 AB CD EF xx xx xx xx
 * 8B:   00 00 00 00 89 AB CD EF
 * 8L:   EF CD AB 89 00 00 00 00
 * 4B:   89 AB CD EF 
 * 4L:   EF CD AB 89 
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8U_4B( Input& in, INT_8U* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        memset( v, 0, 8 * n );
        for ( int i = n; i != 0; --i )
        {
#           ifdef WORDS_BIGENDIAN
            {
                in.read( ((char*)v) + 4, 4 );
            }
#           else
            {
                in.read( v, 4 );
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
    else
    {
        for ( int i = n; i != 0; --i )
        {
            in.read( v, 4 );
#           ifndef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}
    
    
//-----------------------------------------------------------------------------
/**
 * Reads a little-endian 4 byte INT_8U.
 *
 * File: EF CD AB 89 xx xx xx xx
 * 8B:   00 00 00 00 89 AB CD EF
 * 8L:   EF CD AB 89 00 00 00 00
 * 4B:   89 AB CD EF 
 * 4L:   EF CD AB 89 
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8U_4L( Input& in, INT_8U* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        memset( v, 0, 8 * n );
        for ( int i = n; i != 0; --i )
        {
#           ifdef WORDS_BIGENDIAN
            {
                in.read( ((char*)v) + 4, 4 );
                reverse< 4 >( ((char*)v) + 4, 1 );
            }
#           else
            {
                in.read( v, 4 );
            }
#           endif
            v++;
        }
    }
    else
    {
        for ( int i = n; i != 0; --i )
        {
            in.read( v, 4 );
#           ifdef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}
    
    
//-----------------------------------------------------------------------------
/**********/
/* INT_8S */
/**********/


/**
 * Reads an 8-byte INT_8S in Big-Endian order.  If the reading machine has
 * only a 4-byte integer, then only the low-order bytes are stored (with the
 * sign bit moved also).
 *
 * File: s1 23 45 67 89 AB CD EF
 * 8B:   s1 23 45 67 89 AB CD EF
 * 8L:   EF CD AB 89 67 45 23 s1
 * 4B:   s9 AB CD EF 
 * 4L:   EF CD AB s9
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8S_8B( Input& in, INT_8S* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        in.read( v, 8 * n );
#       ifndef WORDS_BIGENDIAN
        {
            for ( int i = n; i != 0; --i )
            {
                reverse< 8 >( v, 1 );
                v++;
            }
        }
#       endif
    }
    else
    {
        char dump[ 4 ];
        for ( int i = n; i != 0; --i )
        {
            in.read( dump, 4 );
            in.read( v, 4 );
            if ( *dump & 128 == 0 )
            {
                *v &= 0x7F;
            }
            else
            {
                *v |= 0x80;
            }
#           ifndef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}

    
//-----------------------------------------------------------------------------
/**
 * Reads an 8-byte INT_8S in Little-Endian order.  If the reading machine has
 * only a 4-byte integer, then only the low-order bytes are stored (with the
 * sign bit moved also).
 *
 * File: EF CD AB 89 67 45 23 01
 * 8B:   01 23 45 67 89 AB CD EF
 * 8L:   EF CD AB 89 67 45 23 01
 * 4B:   s9 AB CD EF 
 * 4L:   EF CD AB s9
 **/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8S_8L( Input& in, INT_8S* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        in.read( v, 8 * n );
#       ifdef WORDS_BIGENDIAN
        {
            for ( int i = n; i != 0; --i )
            {
                reverse< 8 >( v, 1 );
                v++;
            }
        }
#       endif
    }
    else
    {
        char dump[ 4 ];
        for ( int i = n; i != 0; --i )
        {
            in.read( v, 4 );
            in.read( dump, 4 );
            if ( *(dump+3) & 128 == 0 )
            {
                *(v+3) &= 0x7F;
            }
            else
            {
                *(v+3) |= 0x80;
            }
#           ifdef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}

    
//-----------------------------------------------------------------------------
/**
 * Reads a big-endian 4 byte INT_8S.
 *
 * File: s9 AB CD EF xx xx xx xx
 * 8B:   ss ss ss ss s9 AB CD EF
 * 8L:   EF CD AB s9 ss ss ss ss
 * 4B:   s9 AB CD EF 
 * 4L:   EF CD AB s9 
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8S_4B( Input& in, INT_8S* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        for ( int i = n; i != 0; --i )
        {
#           ifdef WORDS_BIGENDIAN
            {
                in.read( ((unsigned char*)v) + 4, 4 );
                if ( *( ((unsigned char*)v) + 4 ) & 0x80 == 0 )
                {
                    memset( v, 0x00, 4 );
                }
                else
                {
                    memset( v, 0xFF, 4 );
                }
            }
#           else
            {
                in.read( v, 4 );
                if ( *(unsigned char*)v & 0x80 == 0 )
                {
                    memset( ((unsigned char*)v) + 4, 0x00, 0 );
                }
                else
                {
                    memset( ((unsigned char*)v) + 4, 0xFF, 0 );
                }
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
    else
    {
        for ( int i = n; i != 0; --i )
        {
            in.read( v, 4 );
#           ifndef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}


//-----------------------------------------------------------------------------
/**
 * Reads a little-endian 4 byte INT_8S.
 *
 * File: EF CD AB s9 xx xx xx xx
 * 8B:   ss ss ss ss s9 AB CD EF
 * 8L:   EF CD AB s9 ss ss ss ss
 * 4B:   s9 AB CD EF 
 * 4L:   EF CD AB s9 
 */
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readInt8S_4L( Input& in, INT_8S* v, unsigned int n )
{
    if ( sizeof( INT_8U ) == 8 )
    {
        for ( int i = n; i != 0; --i )
        {
#           ifdef WORDS_BIGENDIAN
            {
                in.read( ((unsigned char*)v) + 4, 4 );
                reverse< 4 >( ((unsigned char*)v) + 4, 1 );
                if ( *( ((unsigned char*)v) + 4 ) & 0x80 == 0 )
                {
                    memset( v, 0x00, 4 );
                }
                else
                {
                    memset( v, 0xFF, 4 );
                }
            }
#           else
            {
                in.read( v, 4 );
                if ( *((unsigned char*)v + 3) & 0x80 == 0 )
                {
                    memset( ((unsigned char*)v) + 4, 0x00, 0 );
                }
                else
                {
                    memset( ((unsigned char*)v) + 4, 0xFF, 0 );
                }
            }
#           endif
            v++;
        }
    }
    else
    {
        for ( int i = n; i != 0; --i )
        {
            in.read( v, 4 );
#           ifdef WORDS_BIGENDIAN
            {
                reverse< 4 >( v, 1 );
            }
#           endif
            v++;
        }
    }
}
    

//-----------------------------------------------------------------------------
/**********/
/* REAL_4 */
/**********/
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readReal4_B( Input& in, REAL_4* v, unsigned int n )
{
    in.read( v, 4 * n );

#   ifndef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 4 >( v, 1 );
            v++;
        }
    }
#   endif
}


//-----------------------------------------------------------------------------
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readReal4_L( Input& in, REAL_4* v, unsigned int n )
{
    in.read( v, 4 * n );

#   ifdef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 4 >( v, 1 );
            v++;
        }
    }
#   endif    
}

            
//-----------------------------------------------------------------------------
/**********/
/* REAL_8 */
/**********/

//
//!exc: read_failure - Read failed.
//      
void FileHeader::readReal8_B( Input& in, REAL_8* v, unsigned int n )
{
    in.read( v, 8 * n );

#   ifndef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 8 >( v, 1 );
            v++;
        }
    }
#   endif
}


//-----------------------------------------------------------------------------
//
//!exc: read_failure - Read failed.
//         
void FileHeader::readReal8_L( Input& in, REAL_8* v, unsigned int n )
{
    in.read( v, 8 * n );

#   ifdef WORDS_BIGENDIAN
    {
        for ( int i = n; i != 0; --i )
        {
            reverse< 8 >( v, 1 );
            v++;
        }
    }
#   endif
}

