#ifndef FrameCppFrameHH
#define FrameCppFrameHH

#include "framecpp/config.h"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrameH;
  } // namespace - Version_6
  namespace Version_3_4_5
  {
    class Frame;
  } // namespace - Version_3_4_5
} // namespace - Frame



#include <string>

#include "framecpp/FrameHBase.hh"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/time.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"
#include "framecpp/Version3_4_5/container.hh"
#include "framecpp/Version3_4_5/detector.hh"
#include "framecpp/Version3_4_5/rawdata.hh"
#include "framecpp/Version3_4_5/procdata.hh"
#include "framecpp/Version3_4_5/simdata.hh"
#include "framecpp/Version3_4_5/simevent.hh"
#include "framecpp/Version3_4_5/history.hh"
#include "framecpp/Version3_4_5/event.hh"
#include "framecpp/Version3_4_5/summary.hh"
#include "framecpp/Version3_4_5/time.hh"

#include "framecpp/SearchContainer.hh"

#define FRAMECPP_DATAFORMAT_VERSION 5


namespace FrameCPP {

namespace Version_3_4_5 {


class FrameReaderBase;
class FrameReadPlan;


//-----------------------------------------------------------------------------
//
//: Frame
//
// This class represents a frame object.  It is modeled after the FrameH
// structure in the Frame Specification, LIGO-T970130.
//
class Frame : public Base, public FrameHBase
{
public:
    //!ignore_begin:
    /* Container Macros */
    DECLARE_CONTAINER( Type,     type,     Vect,     getName );
    DECLARE_CONTAINER( User,     user,     Vect,     getName );
    DECLARE_CONTAINER( DetectorSim,  detectorSim,  Detector,  getName );
    DECLARE_CONTAINER( DetectorProc,  detectorProc,  Detector,  getName );
    DECLARE_CONTAINER( History,  history,  History,  getName );
    DECLARE_CONTAINER( ProcData, procData, ProcData, getName );
    DECLARE_CONTAINER( SimData,  simData,  SimData,  getName );
    DECLARE_CONTAINER( Event, event, Event, getName );
    DECLARE_CONTAINER( SimEvent, simEvent, SimEvent, getName );
    DECLARE_CONTAINER( Summary,  summary,  Summary,  getName );
    DECLARE_CONTAINER( Aux,      aux,      Vect,     getName );
    DECLARE_CONTAINER( Table,    table,    Table,     getName );
    //!ignore_end:

    /* Constructors */
    //
    //: Default Constructor.
    //
    // This creates a Frame with the following characteristics:
    // <ul>
    //     <li>name = "none"</li>
    //     <li>run = 0</li>
    //     <li>frame = 0</li>
    //     <li>dataQuality = 0</li>
    //     <li>GTime = 0 seconds, 0 nanoseconds</li>
    //     <li>ULeapS = 0</li>
    //     <li>dt = 0</li>
    //     <li>No other data.</li>
    // </ul>
    //
    //!exc: None.
    //
    Frame();


    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of owned elements.
    //
    //!param: const Frame& frame - A frame to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Frame( const Frame& frame );


    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    // Constructs a frame with the specified parameters but no data.
    //
    //!param: const std::string& name - Project or experiment name.
    //!param: INT_4S run - Run number.
    //!param: INT_4U frame - Frame number.
    //!param: INT_4U dataQuality - data quality word.
    //!param: const Time& time - GPS start time.
    //!param: INT_2U uLeapS - Leap seconds in the epoch in which the frame was
    //+     written.
    //!param: const REAL_8& dt - Frame length in seconds.
    //
    Frame( const std::string& name, INT_4S run, INT_4U frame, const Time& time,
           INT_2U uLeapS, const REAL_8& dt, INT_4U dqual = 0 );
    
    /* Destructor */
    //-----------------------------------------------------------------------------
    //
    //: Destructor.
    //
    // This deletes the rawdata.
    //
    //!exc: None.
    //
    virtual ~Frame();

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const Frame& frame - Object to be assigned.
    //
    //!return: const Frame& -- This frame.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Frame& operator=( const Frame& frame );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    //!param: const Frame& frame - Frame which which to compare.
    //
    //!return: bool -- True if frames are equal.
    //
    //!exc: None.
    //   
    bool operator==( const Frame& frame ) const;
    bool operator!=( const Frame& frame ) const;

    //-----------------------------------------------------------------------------
    //
    //: += Assignment Operator.
    //
    //!param: const Frame& frame -
    //
    //!return: const Frame&
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: frame_mismatch - Couldn't add `frame' (time or ADC mismatch)
    //
    const Frame& operator+=( const Frame& frame );

    /* Required Methods */
    //
    //: Get Class ID
    //
    // Gets the class ID.
    //
    //!return: INT_2U
    //
    //!exc: None.
    //   
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Get Structure Header
    //
    // Constructs the Structure Header for this class.  
    //
    //!return: SH -- Frame structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();


    //-----------------------------------------------------------------------------
    //
    //: Get Structure Header
    //
    // Constructs the Structure Header for this class.
    // Frame Data format version 4.
    //
    //!return: SH -- Frame structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH4();

    //-----------------------------------------------------------------------------
    //
    //: Get Structure Header
    //
    // Constructs the Structure Header for this class.
    // Frame Data format version 3.
    //
    //!return: SH -- Frame structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH3();

    /* Accessors */
    const std::string& getName() const;
    INT_4S getRun() const;
    INT_4U getFrame() const;
    INT_4U getDataQuality() const;
    const Time& getGTime() const;
    INT_2U getULeapS() const;
    REAL_8 getDt() const;
    bool isRawDataOwned() const;
    
    //!exc: None.
    TypeContainer& refType();
    //!exc: None.   
    const TypeContainer& refType() const;

    //!exc: None.   
    UserContainer& refUser();
    //!exc: None.   
    const UserContainer& refUser() const;

    //!exc: None.   
    DetectorSimContainer& refDetectorSim();
    //!exc: None.   
    const DetectorSimContainer& refDetectorSim() const;

    //!exc: None.   
    DetectorProcContainer& refDetectorProc();
    //!exc: None.   
    const DetectorProcContainer& refDetectorProc() const;

    //!exc: None.   
    HistoryContainer& refHistory();
    //!exc: None.   
    const HistoryContainer& refHistory() const;

    //!exc: None.   
    RawData* getRawData();
    //!exc: None.   
    const RawData* getRawData() const;

    //!exc: None.   
    ProcDataContainer& refProcData();
    //!exc: None.   
    const ProcDataContainer& refProcData() const;
 
    //!exc: None.
    SimDataContainer& refSimData();
    //!exc: None.   
    const SimDataContainer& refSimData() const;

    //!exc: None.   
    EventContainer& refEvent();
    //!exc: None.   
    const EventContainer& refEvent() const;

    //!exc: None.   
    SimEventContainer& refSimEvent();
    //!exc: None.   
    const SimEventContainer& refSimEvent() const;

    //!exc: None.   
    SummaryContainer& refSummary();
    //!exc: None.   
    const SummaryContainer& refSummary() const;

    //!exc: None.   
    AuxContainer& refAux();
    //!exc: None.   
    const AuxContainer& refAux() const;

    //!exc: None.   
    TableContainer& refTable();
    //!exc: None.   
    const TableContainer& refTable() const;

    /* Mutators */

    //-----------------------------------------------------------------------------
    //
    //: Set frame length
    //
    // Sets the frame length in seconds.
    //
    //!return: Nothing
    //   
    void setDt( const REAL_8 dt );
    
    //-----------------------------------------------------------------------------
    //
    //: Set raw data
    //
    //!param: const RawData& r - RawData to set.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    void setRawData( const RawData& r );


    //-----------------------------------------------------------------------------
    //
    //: Set raw data
    //
    //!param: const RawData* r - RawData to set.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    void setRawData( const RawData* r );

    //-----------------------------------------------------------------------------
    //
    //: Set raw data
    //
    //!param: RawData* r - RawData to ste.
    //!param: bool allocate - Will allocate a copy of `r'.
    //!param: bool owns -  Frame will own its RawData.
    //
    //!exc: std::bad_alloc
    //
    void setRawData( RawData* r, bool allocate = true, bool owns = true );
    void setRawDataOwnership( bool owns );
    
private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrameH;

    //-----------------------------------------------------------------------------
    //
    //: Read a frame
    //
    // Reads a Frame object from an Input object.  This method creates a
    // dynamically allocated Frame object.  This frame is initialized from the
    // given Input object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: Input& in - Input data stream object.
    //
    //!return: Frame* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failed.
    //
    static Frame* read( FrameReaderBase& in );

    //: Version 4 frame read.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failed.
    //   
    static Frame* read4( FrameReaderBase& in );

    //: Version 3 frame read.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failed.
    //   
    static Frame* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Write a frame.
    //
    // Writes a frame to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failed.
    //
    virtual void write( Output& out ) const;

private: friend class DaqFrame; friend class FrameReadPlan;

    //: Name of project or other experiment descritpion.
    std::string mName;
    //: Run number.
    INT_4S mRun;
    //: Frame number
    INT_4U mFrame;
    //: Top level quality of data.
    INT_4U mDataQuality;
    //: Frame start time.
    Time mGTime;
    //: Leap seconds between GPS/TAI and UTC valid fo `mGTime'.
    INT_2U mULeapS;
    //: Frame length in seconds.
    REAL_8 mDt;
    //: General info.
    TypeContainer mType;
    //: Generic info.
    UserContainer mUser;
    //: Model or simulation parameter data definition.
    DetectorSimContainer mDetectorSim;
    //: Data used in generating post-processing outputs.
    DetectorProcContainer mDetectorProc;
    //: History of post-processing.
    HistoryContainer mHistory;
    //: Raw data.
    RawData* mRawData;
    //: Post-processed data.
    ProcDataContainer mProcData;
    //: Simulated data.
    SimDataContainer mSimData;
    //: Event data.
    EventContainer mEvent;
    //: Simulated event data.
    SimEventContainer mSimEvent;
    //: Statistical summary data.
    SummaryContainer mSummary;
    //: Auxiliary data.
    AuxContainer mAux;
    //: Auxiliary table data.
    TableContainer mTable;

    //: Frame owns its RawData.
    bool mOwnsRawData;
};


/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
// Determines whether two frame objects are not equal.  All contained data is
// compared.
//
//!param: const Frame& frame - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!exc: None.
//      
inline bool Frame::operator!=( const Frame& frame ) const
{
    return !( *this == frame );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get run number
//
// Gets the run number.
//
//!return: INT_4S -- Frame run number.
//
inline INT_4S Frame::getRun() const
{
    return mRun;
}


//-----------------------------------------------------------------------------
//
//: Get frame number
//
// Gets the frame number.
//
//!return: INT_4U -- Frame number.
//
inline INT_4U Frame::getFrame() const
{
    return mFrame;
}

//-----------------------------------------------------------------------------
//
//: Get data quality word
//
//!return: INT_4U -- Data quality.
//
inline INT_4U Frame::getDataQuality() const
{
    return mDataQuality;
}


//-----------------------------------------------------------------------------
//
//: Get start time
//
// Gets the frame start time.  This is the GPS time in integer seconds
// and the integer nanosecond residual since the GPS standard epoch.
// This is calid for approximately 143 years from time origin.
//
//!return: const Time& -- Frame start time.
//
inline const Time& Frame::getGTime() const
{
    return mGTime;
}


//-----------------------------------------------------------------------------
//
//: Get leap seconds
//
// Gets the number of leap seconds.  This is the integer number of leap seconds
// between GPS/TAI and UTC in the epoch when the frame is written:
// ULeapS = Int[TAI - UTC].
//
//!return: INT_2U -- Leap seconds.
//
inline INT_2U Frame::getULeapS() const
{
    return mULeapS;
}

//-----------------------------------------------------------------------------
//
//: Get frame length
//
// Gets the frame length in seconds.
//
//!return: REAL_8 -- Frame length in seconds.
//
inline REAL_8 Frame::getDt() const
{
    return mDt;
}

   
//-----------------------------------------------------------------------------
//
//: Set frame length
//
// Sets the frame length in seconds.
//
//!return: Nothing.
//
inline void Frame::setDt( const REAL_8 dt )
{
   mDt = dt;
   return;
}
   

//-----------------------------------------------------------------------------
//
//: Get 'type' container
//
// Gets the 'type' container.  This returns a reference to the 'type' container
// used to store general information.
//
//!return: TypeContainer& -- Data.
//
//!exc: None.
//      
inline
Frame::TypeContainer&
Frame::refType()
{
    return mType;
}


//-----------------------------------------------------------------------------
//
//: Get const 'type' container
//
// Gets the 'type' container.  This returns a constant reference to the 'type'
// container used to store general information.
//
//!return: const TypeContainer&  -- Data.
//
//!exc: None.
//      
inline
const Frame::TypeContainer&
Frame::refType() const
{
    return mType;
}


//-----------------------------------------------------------------------------
//
//: Get 'user' container
//
// Gets the 'user' container.  This returns a reference to the 'user' container
// used to store user-provided information.
//
//!return: UserContainer& -- User data.
//
//!exc: None.
//         
inline
Frame::UserContainer&
Frame::refUser()
{
    return mUser;
}


//-----------------------------------------------------------------------------
//
//: Get const 'user' container
//
// Gets the 'user' container.  This returns a constant reference to the 'user'
// container used to store user-provided information.
//
//!return: UserContainer& -- User data.
//
//!exc: None.
//         
inline
const Frame::UserContainer&
Frame::refUser() const
{
    return mUser;
}


//-----------------------------------------------------------------------------
//
//: Get sim detector container
//
//!return: DetectorSimContainer& -- Sim Detector.
//
//!exc: None.
//         
inline
Frame::DetectorSimContainer&
Frame::refDetectorSim()
{
    return mDetectorSim;
}


//-----------------------------------------------------------------------------
//
//: Get const sim detector container
//
//!return: const DetectorSimContainer& -- Sim Detector.
//
//!exc: None.
//         
inline const
Frame::DetectorSimContainer&
Frame::refDetectorSim() const
{
    return mDetectorSim;
}


//-----------------------------------------------------------------------------
//
//: Get proc detector container
//
//!return: SearchContainer& -- Proc Detector.
//
//!exc: None.
//         
inline
Frame::DetectorProcContainer&
Frame::refDetectorProc()
{
    return mDetectorProc;
}


//-----------------------------------------------------------------------------
//
//: Get const proc detector container
//
//!return: const DetectorProcContainer& -- Proc Detector.
//
//!exc: None.
//         
inline const
Frame::DetectorProcContainer&
Frame::refDetectorProc() const
{
    return mDetectorProc;
}


//-----------------------------------------------------------------------------
//
//: Get history container
//
// Gets the history container.  This contains the history of post-processing
// with which frame may have been generated.
//
//!return: HistoryContainer& -- History.
//
//!exc: None.
//         
inline
Frame::HistoryContainer&
Frame::refHistory()
{
    return mHistory;
}


//-----------------------------------------------------------------------------
//
//: Get const history container
//
// Gets the history container.  This contains the history of post-processing
// with which frame may have been generated.
//
//!return: const HistoryContainer& -- History.
//
//!exc: None.
//         
inline const
Frame::HistoryContainer&
Frame::refHistory() const
{
    return mHistory;
}


//-----------------------------------------------------------------------------
//
//: Get raw data
//
// Gets the raw data object.
//
//!return: RawData& -- Raw data.
//
//!exc: None.
//         
inline RawData* Frame::getRawData()
{
    return mRawData;
}


//-----------------------------------------------------------------------------
//
//: Get const raw data
//
// Gets the raw data object.
//
//!return: const RawData& -- Raw data.
//
//!exc: None.
//         
inline const RawData* Frame::getRawData() const
{
    return mRawData;
}


//-----------------------------------------------------------------------------
//
//: Get proc data container
//
// Gets the container used to store procData.
//
//!return: ProcDataContainer& -- Processed data.
//
//!exc: None.
//         
inline
Frame::ProcDataContainer&
Frame::refProcData()
{
    return mProcData;
}


//-----------------------------------------------------------------------------
//
//: Get const proc data container
//
// Gets the container used to store procData.
//
//!return: const ProcDataContainer& -- Processed data.
//
//!exc: None.
//         
inline const
Frame::ProcDataContainer&
Frame::refProcData() const
{
    return mProcData;
}


//-----------------------------------------------------------------------------
//
//: Get sim data container
//
// Gets the container used to store simulated data.
//
//!return: SimDataContainer& -- Simulated data.
//
//!exc: None.
//         
inline
Frame::SimDataContainer&
Frame::refSimData()
{
    return mSimData;
}


//-----------------------------------------------------------------------------
//
//: Get const sim data container
//
// Gets the container used to store simulated data.
//
//!return: const SimDataContainer& -- Simulated data.
//
//!exc: None.
//         
inline const
Frame::SimDataContainer&
Frame::refSimData() const
{
    return mSimData;
}


//-----------------------------------------------------------------------------
//
//: Get event data container
//
// Gets the container used to store event data. 
//
//!return: EventContainer& -- Event data.
//
//!exc: None.
//         
inline
Frame::EventContainer&
Frame::refEvent()
{
    return mEvent;
}


//-----------------------------------------------------------------------------
//
//: Get const event data container
//
// Gets the container used to store event data.
//
//!return: const EventContainer& -- Event data.
//
//!exc: None.
//         
inline const
Frame::EventContainer&
Frame::refEvent() const
{
    return mEvent;
}


//-----------------------------------------------------------------------------
//
//: Get simulated event data container
//
// Gets simulated event data container
//
//!return: SimEventContainer& -- Simulated event data.
//
//!exc: None.
//         
inline
Frame::SimEventContainer&
Frame::refSimEvent()
{
  return mSimEvent;
}


//-----------------------------------------------------------------------------
//
//: Get simulated event data container
//
// Gets simulated event data container
//
//!return: SimEventContainer& -- Simulated event data.
//
//!exc: None.
//         
inline const
Frame::SimEventContainer&
Frame::refSimEvent()  const 
{
  return mSimEvent;
}


//-----------------------------------------------------------------------------
//
//: Get summary data container
//
// Gets the container used to store statistical summary data.
//
//!return: SummaryContainer& -- Summary data.
//
//!exc: None.
//         
inline
Frame::SummaryContainer&
Frame::refSummary()
{
    return mSummary;
}


//-----------------------------------------------------------------------------
//
//: Get const summary data container
//
// Gets the container used to store statistical summary data.
//
//!return: const SummaryContainer& -- Summary data.
//
//!exc: None.
//         
inline const
Frame::SummaryContainer&
Frame::refSummary() const
{
    return mSummary;
}


//-----------------------------------------------------------------------------
//
//: Get aux data container
//
// Gets the container used to store auxiliary data.
//
//!return: AuxContainer& -- Auxiliary data.
//
//!exc: None.
//         
inline
Frame::AuxContainer&
Frame::refAux()
{
    return mAux;
}


//-----------------------------------------------------------------------------
//
//: Get const aux data container
//
// Gets the container used to store auxiliary data.
//
//!return: const AuxContainer& -- Auxiliary data.
//
//!exc: None.
//         
inline const
Frame::AuxContainer&
Frame::refAux() const
{
    return mAux;
}

//-----------------------------------------------------------------------------
//
//: Get table container
//
// Gets the container used to store table(s).
//
//!return: TableContainer& -- Auxiliary table data.
//
//!exc: None.
//         
inline Frame::TableContainer&
Frame::refTable()
{
    return mTable;
}

//-----------------------------------------------------------------------------
//
//: Get table container
//
// Gets the container used to store table(s).
//
//!return: const TableContainer& -- Auxiliary table data.
//
//!exc: None.
//         
inline 
const Frame::TableContainer&
Frame::refTable() const
{
    return mTable;
}

//: Get raw data ownership flag
inline bool Frame::isRawDataOwned() const
{
    return mOwnsRawData;
}

//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------

//: Set raw data ownership flag
inline void Frame::setRawDataOwnership( bool owns )
{
    mOwnsRawData = owns;
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppFrameHH
