#include "framecpp/config.h"

// System Header Files
#include <sstream>
   
// Local Header Files   
#include "framereader.hh"
#include "frame.hh"
#include "endofframe.hh"
#include "endoffile.hh"
#include "dictionary.hh"

using FrameCPP::Version_3_4_5::FrameReaderBase;
using namespace std;   


//-----------------------------------------------------------------------------
//
//: Constructor
//
//!param: std::istream& in - The stream to read from.
//!param: bool autoReset - True if the reader should automatically reset itself
//+     when an end-of-file structure is read.
//
//!exc: None.   
//      
FrameReaderBase::FrameReaderBase( std::istream& in, bool autoReset )
        : Input( &in ), mNFrames( 0 ), mAutoReset( autoReset ),
          mEndOfFile( false ), mFrame( 0 ), mCurrentObjectHeader(),
          mLinkedListStart(), mLinkedListNodes(), mObjectStorage(), mPointers()
{
}


//-----------------------------------------------------------------------------
//
//: Destructor
//
FrameReaderBase::~FrameReaderBase()
{
    // There are two types of allocated data stored internally.  These are in
    // the LinkedListStart container (RefHeadBase*) and ObjectStorage
    // container (Base*).
    
    for ( LinkedListStart::iterator iter = mLinkedListStart.begin();
          iter != mLinkedListStart.end(); ++iter )
    {
        delete iter->first;
        iter->first = 0;
    }
    // Erase all elements
    mLinkedListStart.clear();

    for ( ObjectStorage::iterator iter = mObjectStorage.begin();
          iter != mObjectStorage.end(); ++iter )
    {
        delete iter->second;
        iter->second = 0;
    }
    // Erase all elements   
    mObjectStorage.erase( mObjectStorage.begin(), mObjectStorage.end() );
}

//-----------------------------------------------------------------------------
//
//: Read the number of frames variable.
//
//!exc: read_failure - Read failed.
//      
INT_4U FrameReaderBase::getFrameNumber()
{
  // Save the state of the frame header.
  // This will have to be restored later.
  bool mh = mIsHeaderLoaded;

  // Header has to be loaded first if it is not loaded.
  // Only then it is possible to seek to the end of file.
  if ( mFileHeader == 0 )
    {
      mFileHeader = new FileHeader( *this );
      dictionary = mFileHeader->getDictionary();
    }
  unsigned long cur_pos = mStream->tellg();
  mStream->seekg( -20, std::ios::end );
  INT_4U nFrames;
  *this >> nFrames;

  // Keep Input object unaware of our uninvited intrusion into frame reading
  mIsHeaderLoaded = mh;
  mNBytes -= 4;
  //-------

  mStream->seekg( cur_pos );
  return nFrames;
}


//-----------------------------------------------------------------------------
//
//: Reset the Reader
//
//!exc: None.   
//      
void FrameReaderBase::reset()
{
    mPointers.erase( mPointers.begin(), mPointers.end() );
    mStatDetectorPointers.erase( mStatDetectorPointers.begin(),
                                 mStatDetectorPointers.end() );
    mLinkedListNodes.erase( mLinkedListNodes.begin(), mLinkedListNodes.end() );
    delete mFrame;
    mFrame = 0;
    
    for( LinkedListStart::iterator iter = mLinkedListStart.begin();
         iter != mLinkedListStart.end(); ++iter )
    {
        delete iter->first;
        iter->first = 0;
    }
    // Erase all elements
    mLinkedListStart.clear();

    for( ObjectStorage::iterator iter = mObjectStorage.begin();
         iter != mObjectStorage.end(); ++iter )
    {
        delete iter->second;
        iter->second = 0;
    }
    // Erase all elements
    mObjectStorage.erase( mObjectStorage.begin(), mObjectStorage.end() );

    mEndOfFile = false;
    Input::reset();
}


//-----------------------------------------------------------------------------
//
//: Read a Frame
//
//!return: Frame* - Newly allocated.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure accured.
//
FrameCPP::Version_3_4_5::Frame* FrameReaderBase::readFrame()
{
    // Erase any errors which were left over from a previous read
    clearErrors();

    // If we are already at the end of the file, reset if mAutoReset is true.
    if ( mEndOfFile )
    {
        if ( mAutoReset )
        {
            mEndOfFile = false;
            reset();
        }
        else
        {
            throw read_failure( "End of file reached.  The EndOfFile object "
                                "has been read.", ENDOFFILE );
        }
    }
    
    mFrame = 0;                 // Storage for frame pointer
    bool endOfFrame( false );   // Flag indicating if this frame is finished.
    INT_4U bytes( 0 );          // the number of bytes read.
    
    // Loop until the frame is finished
    while( !endOfFrame )
    {
        // Remember the current number of bytes read so that we can double
        // check the lengths of the file structures.
        bytes = getNBytes();

        // Read the next header, this is the 8 byte header in front of each
        // frame structure.
        mCurrentObjectHeader = readNextObjectHeader();

        // Read the structure.
        endOfFrame = readBase();

#ifdef not_def

	This check appears to be broken...


        // Check the length
        if ( bytes + mCurrentObjectHeader.getLength() != getNBytes() )
        {
            // All of this should be moved into the Input class
            ostringstream ss;
            ss << "Incorrect object length(class "
               << mCurrentObjectHeader.getLocalClassId() << "): "
               << mCurrentObjectHeader.getLength() << ". Actual length was: "
               << getNBytes() - bytes << ".";
            addError( ss.str() );
        }
#endif

    }
    
    // If we read a frame, reconnect the references.
    if ( mFrame != 0 )
    {
        // Reconnect references
        reconnect();
    }

    // If we are at the end of file, read in the end of file structure.
    if ( mEndOfFile )
    {
        readEndOfFile();
    }

    // We now need to delete stuff

    // Delete the linked list start information
    for ( LinkedListStart::iterator iter = mLinkedListStart.begin();
          iter != mLinkedListStart.end(); ++iter )
    {
        delete iter->first;
        iter->first = 0;
    }
    mLinkedListStart.erase( mLinkedListStart.begin(),
                            mLinkedListStart.end() );

    // Delete the list of nodes
    mLinkedListNodes.erase( mLinkedListNodes.begin(),
                            mLinkedListNodes.end() );

    // Delete the list of pointers
    mPointers.erase( mPointers.begin(), mPointers.end() );
    mStatDetectorPointers.erase( mStatDetectorPointers.begin(),
                                 mStatDetectorPointers.end() );

    // Delete any orphaned objects.  These are objects which existed in the
    // frame but were not referenced by any other objects.
    for ( ObjectStorage::iterator iter = mObjectStorage.begin();
          iter != mObjectStorage.end(); ++iter )
    {
        if ( iter->second != 0 )
        {
            ostringstream ss;
            ss << "Orphan found: (" << iter->first.getClass()
               << "," << iter->first.getInstance() << ")";
            addError( ss.str() );
            
            delete iter->second;
            iter->second = 0;
        }
    }

    mObjectStorage.erase( mObjectStorage.begin(), mObjectStorage.end() );    
    
    Frame* tmp = mFrame;
    mFrame = 0;
    return tmp;
}


//-----------------------------------------------------------------------------
//
//: Read a Frame Object
//
//!return: bool - True if the end-of-frame object was read.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure accured.
//
bool FrameReaderBase::readBase() 
{
    bool endOfFrame( false ); // end-of-frame flag
    
    // Check to see if this object is a frame
    if ( mCurrentObjectHeader.getLocalClassId() == ID_FRAME )
    {
        // Make sure we are not already reading a frame
        if ( mFrame != 0 )
        {
            addError( "End of frame missing. Another frame "
                      "header was encountered before the EndOfFrame object "
                      "for the previous frame was found." );

            // We will assume that the FrEndOfFrame was forgotten, therefore
            // end the frame now.
            endOfFrame = true;
        }
        
        // Set the reading flag and read the frame header.
        mFrame = dynamic_cast< Frame* >(
            dictionary->readElement( *this, ID_FRAME ) );
        readBaseHook( mFrame, ID_FRAME );
    }
    // Check for the end-of-file
    else if ( mCurrentObjectHeader.getLocalClassId() == ID_ENDOFFILE )
    {
        // Were we reading a frame?
        if ( mFrame != 0 )
        {
            addError( "End of frame missing.  The end of "
                      "file structure was found instead." );
        }
        // OK, there just wasn't a frame to read.
        else
        {
            addError( "Frame header not found.  The "
                      "end of file structure was found before any frame "
                      "header." );
        }

        // The EndOfFile structure will be read later, back in the readFrame
        // method.  For now just set the appropriate flags.
        endOfFrame = true;
        mEndOfFile = true;
    }
    // OK, it's something else
    else
    {
        // Read in the object (using the dictionary)
        Base* object = dictionary->readElement(
            *this, mCurrentObjectHeader.getLocalClassId() );
        readBaseHook( object, mCurrentObjectHeader.getLocalClassId() );

        // Is it the end of the frame?
        if ( mCurrentObjectHeader.getLocalClassId() == ID_ENDOFFRAME )
        {
            EndOfFrame* eofr = dynamic_cast< EndOfFrame* >( object );

            // Make sure we were reading a frame
            if ( mFrame == 0 )
            {
                // We found an end of frame but no frame header,
                // throw an error.
                addError( "Frame Header missing.  An FrEndOfFrame "
                          "object was encountered before a FrameH "
                          "object. " );
            }
            else
            {
                // Double check the frame and run numbers.
                if ( ( eofr->getRun() != mFrame->getRun() ) ||
                     ( eofr->getFrame() != mFrame->getFrame() ) )
                {
                    // The end of frame we found doesn't belong to this frame
                    // header
                    addError( "An FrEndOfFrame object "
                              "was found but did not correspond to this "
                              "frame object." );
                }
            }

            // End this frame, no need to keep the end-of-frame object
            delete eofr;
            eofr = 0;
            endOfFrame = true;
        }
        // OK, just store the object so we can dereference things later.
        else
        {
	  Reference myRef = mCurrentObjectHeader.getReference();

	  // Detector have to be put into the Object Storage
	  // because of the StatData dereferencing
	  //
	  if ( mCurrentObjectHeader.getLocalClassId() != ID_DETECTOR ) {

	    //  See if we can make a quick chain resolution
	    for(LinkedListStart::iterator llsIter = mLinkedListStart.begin();
		llsIter != mLinkedListStart.end(); ++llsIter )
	      {
		if (llsIter->second == myRef) {
		  llsIter->first->addItem(object);
		  LinkedListNodes::iterator 
		    riter = mLinkedListNodes.find(myRef);
		  
		  if ( riter == mLinkedListNodes.end() ) {
		    llsIter->second = Reference(0, 0);
		  } else {
		    llsIter->second = riter->second;
		    mLinkedListNodes.erase(riter);
		  }
		  if (llsIter->second == Reference(0, 0)) {
		    delete llsIter->first;
                    llsIter->first = 0;
   
		    mLinkedListStart.erase(llsIter);
		  }		  
		  return endOfFrame;
		}
	      }
	  }
	  mObjectStorage.insert( std::pair< Reference, Base* >(
              myRef, object ) );
        }
    }

    return endOfFrame;
}

   
//-----------------------------------------------------------------------------
//
//: Read Object Hook
//
//!exc: std::bad_alloc - Memory allocation failed.
//
void FrameReaderBase::readBaseHook( Base *base, INT_2U classId )
{}
   

//-----------------------------------------------------------------------------
//
//: Read next Linked-list Node
//
// This method reads in the next node of a linked list.  for example, this
// method is called when reading in the 'next' field in an FrAdcData struct.
//
//!exc: read_failure - Read failed.
//   
void FrameReaderBase::readNextNodeReference()
{
    // Read the reference
    Reference ref;
    *this >> ref;

    // Store the reference along with a reference to the current object.
    // Together, these will make a chain which can be traversed down when
    // references are resolved later.
    mLinkedListNodes.insert( std::pair< Reference, Reference >(
        mCurrentObjectHeader.getReference(), ref ) );
}


//-----------------------------------------------------------------------------
//
//: Resolve References
//
// This method puts the frame together, resolving the references which were
// read during the read process.
//
void FrameReaderBase::reconnect( bool ignoreErrors )
{
    // Connect StatData Detector pointers
    for(
        Pointers::iterator pIter = mStatDetectorPointers.begin();
        pIter != mStatDetectorPointers.end(); ++pIter )
    {
        ObjectStorage::iterator oIter =
            mObjectStorage.find( pIter->second );
        if ( oIter == mObjectStorage.end() )
        {
	  if (! ignoreErrors )
	    {
	      // It wasn't found.
	      addError( "StatData reference not found.  An object which was "
			"referenced by another class was not found." );
	    }
	  *(pIter->first) = 0;
        }
        else
        {
            *(pIter->first) = oIter->second;
        }
    }

    // Fill the containers
    for(
        LinkedListStart::iterator llsIter = mLinkedListStart.begin();
        llsIter != mLinkedListStart.end(); ++llsIter )
    {
        Reference ref = llsIter->second;
        
        while( ref != Reference( 0, 0 ) )
        {
            // Find the object being referenced
            ObjectStorage::iterator oiter =
                mObjectStorage.find( ref );
            if ( oiter == mObjectStorage.end() )
            {
	      if (! ignoreErrors )
		{
		  // It wasn't found.  Don't add anything, just end the loop.
		  ostringstream ss;
		  ss << "Reference not found: (" << ref.getClass()
		     << "," << ref.getInstance() << ")";
		  addError( ss.str() );
		}
                ref = Reference( 0, 0 );
            }
            else
            {
                // Ok, add the object
                llsIter->first->addItem( oiter->second );
		mObjectStorage.erase( oiter );

                // Now get the next item in the list
                LinkedListNodes::iterator riter = mLinkedListNodes.find(ref);
                if ( riter == mLinkedListNodes.end() )
                {
                    // What happened?  There should have been something.
                    ref = Reference( 0, 0 );
                }
                else
                {
                    ref = riter->second;
                }
            }
        }
    }

    // Connect Pointers
    for(
        Pointers::iterator pIter = mPointers.begin();
        pIter != mPointers.end(); ++pIter )
    {
        ObjectStorage::iterator oIter =
            mObjectStorage.find( pIter->second );
        if ( oIter == mObjectStorage.end() )
        {
	  if (! ignoreErrors )
	    {
	      // It wasn't found.
	      addError( "Reference not found.  An object which was "
			"referenced by another class was not found." );
	    }
            *(pIter->first) = 0;
        }
        else
        {
            *(pIter->first) = oIter->second;
            mObjectStorage.erase( oIter );
        }
    }

    // Find StatData
    for ( ObjectStorage::iterator oIter = mObjectStorage.begin();
          oIter != mObjectStorage.end(); ++oIter )
    {
        if ( oIter->second->getClassId() == ID_STATDATA )
        {
            StatData* stat = dynamic_cast< StatData* >( oIter->second );
            Detector* d = stat->getDetector();

            if ( d == 0 )
            {
	      if (! ignoreErrors )
		{ 
		  addError( "Orphaned StatData.  A FrStatData object was found "
			    "which did not possess a valid reference to a "
			    "FrDetector." );
		}
            }
            else
            {
                d->refStatData().append( stat, false );
                oIter->second = 0;
            }
        }
    }
}


//-----------------------------------------------------------------------------
//
//: Read a pointer to an object.
//
// Some frame structures have a pointer to another object which is not part
// of a linked-list. This method reads in those pointers.
//
//!param: Base** v - Target.
//
//!return: FrameReaderBase& - This object.
//
//!exc: read_failure - Read failure occured.
//
FrameCPP::Version_3_4_5::FrameReaderBase& FrameReaderBase::operator>>( Base** v )
{
    // Set the pointer to null for now.  It will be changed after all of the
    // objects have been read (during the reconnect method) if it was not a
    // null reference.
    *v = 0;

    // Read the reference
    Reference ref;
    *this >> ref;

    if ( ref != Reference( 0, 0 ) )
    {
        // If the reference wasn't null, add it to the pointers list to be
        // dereferenced during the reconnect phase.
        if ( mCurrentObjectHeader.getLocalClassId() != ID_STATDATA )
        {
            mPointers.push_back( std::pair< Base**, Reference >( v, ref ) );
        }
        else
        {
            mStatDetectorPointers.push_back( std::pair< Base**, Reference >(
                v, ref ) );
        }
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read End-Of-File
//
// This reads the end-of-file structure.
//
//!todo: double check the contents of the FrEndOfFile structure.
//
void FrameReaderBase::readEndOfFile()
{
    EndOfFile* eof = dynamic_cast< EndOfFile* >(
        dictionary->readElement( *this, ID_ENDOFFILE ) );

    mEndOfFile = true;
    delete eof;
    eof = 0;
}


