#ifndef FrameCppFrameReaderBaseHH
#define FrameCppFrameReaderBaseHH

#include "framecpp/VFrameReader.hh"

#include "input.hh"
#include "time.hh"
#include "location.hh"

namespace FrameCPP {

  namespace Version_6 {
    class FrameH;
  }

namespace Version_3_4_5 {


class Time;
class Location;


//-----------------------------------------------------------------------------
//
//: FrameReaderBase
//
// This class contains methods which are used by the Frame objects to read
// themselves in.
//
// <p>In a frame file, containers are implemented by using linked-lists.  Every
// object in the file which can be contained by another object has a pointer
// to the next element in the linked list.  This creates a class which is both
// a linked-list element and a linked-list container.  Internally, FrameCPP
// does not use linked-lists.  The "pointer to next" in the frame objects
// has been removed.
//
// <p>This is the general read procedure:
// <p><ol>
//   <li> readFrame is called. </li>
//   <li> FrameReaderBase calls Input::readNextObjectHeader to get information
//        about the next object. </li>
//   <li> Make sure that it is a frame header and call the read method for the
//        Frame class (via the object store). </li>
//   <li> Continue call readNextObjectHeader and reading objects (via the
//        store) until an EndOfFrame or EndOfFile occurs. </li>
//   <li> Reconnect references (see the next section). </li>
//   <li> Delete internal data. </li>
//   <li> Read EndOfFile if it is next. </li>
// </ol>
//
// <p>This is the object component read procedure:
// <p><ol>
//   <li> If the component is a simple data type, nothing special happens.
//        Just read it in.  For example, there is nothing special in reading a
//        REAL_4, FrameReaderBase just calls the Input method to read it.
//   <li> If the component is the beginning of a linked list, the extraction
//        operator for Container< T >& is called.  This will store a reference
//        to the Container (as a RefHeadBase) in the mLinkedListStart member,
//        along with the Reference to the starting member of the list.
//   <li> If the component is a next pointer, then the readNextNodeReference
//        method is called.  This reads a reference to the next object and
//        stores this reference, along with the reference to the current
//        object, in the mLinkedListNodes member. </li>
// </ol>
//
//!todo: It is now possible to insert elements without copying them, so much of
//!todo: this confusion can be replaced with a simpler read method.
//
class FrameReaderBase : protected Input
{
public:

    enum ErrorCode
    {
        BADSTREAM = Input::BADSTREAM,
        SKIPFAILED = Input::SKIPFAILED,
        ENDOFFILE = Input::ENDOFFILE
    };

    /* typedefs */
    typedef std::vector< std::pair< RefHeadBase*, Reference > > LinkedListStart;
    typedef std::map< Reference, Reference > LinkedListNodes;
    typedef std::map< Reference, Base* > ObjectStorage;
    typedef std::vector< std::pair< Base**, Reference > > Pointers;

    //-----------------------------------------------------------------------------
    //
    //: Constructor
    //
    //!param: std::istream& in - The stream to read from.
    //!param: bool autoReset - True if the reader should automatically reset itself
    //+     when an end-of-file structure is read.
    //
    //!exc: None.
    //      
    FrameReaderBase( std::istream& in, bool autoReset = false );

    //: Destructor
    virtual ~FrameReaderBase();

    /* Error Methods */
    Input::getError;
    Input::errorBegin;
    Input::errorEnd;
    Input::errorQuantity;
    Input::errorFirst;
    Input::errorLast;
    Input::clearErrors;
    Input::setErrorWatch;

    /* Accessors */
    Input::getNBytes;
    //!exc: None.
    INT_4U getFrames() const;
    //!exc: None.
    bool isAutoReset() const;

    /* Streaming */
    Input::read;
    //!exc: read_failure - Read failed.
    FrameReaderBase& operator>>( CHAR& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( CHAR_U& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( INT_2S& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( INT_2U& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( INT_4S& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( INT_4U& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( INT_8S& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( INT_8U& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( REAL_4& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( REAL_8& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( std::string& v );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( Time& t );
    //!exc: read_failure - Read failed.   
    FrameReaderBase& operator>>( Location& t );
    template< class T > void readType( T*, unsigned int n );


    //-----------------------------------------------------------------------------
    //
    //: Read a pointer to an object.
    //
    // Some frame structures have a pointer to another object which is not part
    // of a linked-list. This method reads in those pointers.
    //
    //!param: Base** v - Target.
    //
    //!return: FrameReaderBase& -- This object.
    //
    //!exc: read_failure - Read failure occured.
    //    
    FrameReaderBase& operator>>( Base** v );
    //!exc: read_failure - Read failure occured.   
    template< class T > FrameReaderBase& operator>>( Container< T >& c );

    //-----------------------------------------------------------------------------
    //
    //: Read next Linked-list Node
    //
    // This method reads in the next node of a linked list.  for example, this
    // method is called when reading in the 'next' field in an FrAdcData struct.
    //
    //!exc: read_failure - Read failure occured.   
    virtual void readNextNodeReference();

    //-----------------------------------------------------------------------------
    //
    //: Read a Frame
    //
    //!return: Frame* -- Newly allocated.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure accured.
    //
    Frame* readFrame();

    //-----------------------------------------------------------------------------
    //
    //: Read the `number of frames'.
    //
    //!return: INT_4U -- How many frames are there in the frame file.
    //
    //!exc: read_failure - Read failed.
    //
    INT_4U getFrameNumber();

    //: See if file input requires byte swapping
    bool byteSwapNeeded(){ return mFileHeader->byteSwapNeeded(); }
  
    /* Mutators */
    //!exc: None.
    void setAutoReset( bool autoReset );
    //: Reset the Reader
    //!exc: None.   
    void reset();
    
protected:

    /* Helper Methods */

    //-----------------------------------------------------------------------------
    //
    //: Read a Frame Object
    //
    //!return: bool -- True if the end-of-frame object was read.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure accured.
    //
    bool readBase();

    //-----------------------------------------------------------------------------
    //
    //: Read Object Hook
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    virtual void readBaseHook( Base *base, INT_2U classId );

    //-----------------------------------------------------------------------------
    //
    //: Resolve References
    //
    // This method puts the frame together, resolving the references which were
    // read during the read process.
    //
    void reconnect(bool ignoreErrors = false);

    //-----------------------------------------------------------------------------
    //
    //: Read End-Of-File
    //
    // This reads the end-of-file structure.
    //
    //!todo: double check the contents of the FrEndOfFile structure.
    //
    void readEndOfFile();

    //: The number of frames read.
    unsigned int mNFrames;

    /* Read Internal State */
    //: Input stream is reset at the end of frame file.
    bool mAutoReset;
    //: At the end of frame file.
    bool mEndOfFile;
    //: Current frame.
    Frame* mFrame;
    //: Current frame structure object header.
    ObjectHeader mCurrentObjectHeader;
    LinkedListStart mLinkedListStart;
    LinkedListNodes mLinkedListNodes;
    ObjectStorage mObjectStorage;
    Pointers mPointers;
    Pointers mStatDetectorPointers;

public:

    //: Get current input object's frame structure instance
    //!exc: None.   
    INT_2U currentInstance()
      {
        return mCurrentObjectHeader.getInstance();
      }
};


//-----------------------------------------------------------------------------
//
//: Check if AutoReset is on.
//
//!return: bool -- True is AutoReset is on.
//
//!exc: None.   
//   
inline bool FrameReaderBase::isAutoReset() const
{
    return mAutoReset;
}


//-----------------------------------------------------------------------------
//
//: Set the AutoReset flag.
//
// If AutoReset is on, then this class will reset itself whenever the
// end-of-file structure is read.
//
//!param: bool autoReset - True if want to enable autoreset.
//
//!exc: None.   
//      
inline void FrameReaderBase::setAutoReset( bool autoReset ) 
{
    mAutoReset = autoReset;
}


//-----------------------------------------------------------------------------
//
//: Get the number of frames read.
//
//!return: INT_4U -- The number of frames read.
//
//!exc: None.   
//      
inline INT_4U FrameReaderBase::getFrames() const 
{
    return mNFrames;
}


//-----------------------------------------------------------------------------
//
//: Read a container
//
//!param: Container< T >& c - The container to read into.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failed.
//      
template< class T >
inline FrameReaderBase& FrameReaderBase::operator>>(
    Container< T >& c )
{
    // Create a reference and read it
    Reference ref;
    *this >> ref;

    // If the reference is not null, then create a RefHead object which will
    // store the container.  Then add a pointer to this RefHead along with
    // the reference which belongs to it.  These references will be resolved
    // once the entire frame has been read.
    if ( ref != Reference( 0, 0 ) )
    {
        RefHead< T >* rh = new RefHead< T >( c );
        mLinkedListStart.push_back( std::pair< RefHeadBase*, Reference >(
            rh, ref ) );
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a CHAR.
//
//!param: CHAR& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( CHAR& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a CHAR_U
//
//!param: CHAR_U& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( CHAR_U& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_2S
//
//!param: INT_2S& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( INT_2S& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_2U
//
//!param: INT_2U& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( INT_2U& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_4S
//
//!param: INT_4S& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( INT_4S& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_4U
//
//!param: INT_4U& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( INT_4U& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_8S
//
//!param: INT_8S& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( INT_8S& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an INT_8U
//
//!param: INT_8U& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( INT_8U& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a REAL_4
//
//!param: REAL_4& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( REAL_4& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a REAL_8
//
//!param: REAL_8& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( REAL_8& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a std::string
//
//!param: std::string& v - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( std::string& v )
{
    Input::operator>>( v );
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a Time
//
//!param: Time& t - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( Time& t )
{
    ((Input&)(*this)) >> t;
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read a Location
//
//!param: Location& l - Target.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
inline FrameReaderBase& FrameReaderBase::operator>>( Location& l )
{
    (Input&)(*this) >> l;
    return *this;
}


//-----------------------------------------------------------------------------
//
//: Read an array
//
//!param: T* data - Target array.
//!param: unsigned int n - Number of array elements to read.
//
//!return: FrameReaderBase& -- This object.
//
//!exc: read_failure - Read failure occured.
//
template< class T >
void FrameReaderBase::readType( T* data, unsigned int n )
{
    Input::read( data, n * sizeof( T ) );
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppFrameReaderBaseHH
