#include "config.h"

#include <sstream>

#include "framereadplan.hh"
#include "adcdata.hh"

using namespace FrameCPP::Version_3_4_5;

// Constructor reads the table of contents from the
// template frame.
//   
//!exc: std::bad_alloc - Out of memory.
//!exc: read_failure - Read failed.
//      
FrameReadPlan::FrameReadPlan( std::istream& in )
  : TOCReader( in, true ), frame(),
    raw_data(""), adcDataOffsetVector()
{
  // Create required frame structures in the skeleton frame
  frame.setRawData (raw_data);
}

//!exc: None.   
FrameReadPlan::~FrameReadPlan() {}


// This is to build the TOC on the input frame file that has none
//   
//!exc: std::bad_alloc - Out of memory.
//         
void
FrameReadPlan::readBaseHook( Base *base, INT_2U classId )
{
  if (! mHasToc ) {

    // :TODO: fill in TOC structures. Use code from FrameCPP::Version_3_4_5::FrameWriterTOC 

    if ( classId == ID_ADCDATA ) {
      FrameCPP::Version_3_4_5::AdcData *adc = dynamic_cast<FrameCPP::Version_3_4_5::AdcData *>( base );
      std::cerr << adc->getName() << std::endl;
    }
  }
}

//   
//!exc: std::bad_alloc - Out of memory.
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.   
//         
void
FrameReadPlan::daqTriggerADC( const std::vector<std::string> &adcNames )
{
  // Goes through the adcs and adds them into the skeleton frame based on the TOC information.
  // Creates the update map, which allows fast frame updates from subsequent frame files.

  // :TODO: frames in one frame file could be different. Need to check and see if any of the requested ADCs
  // are different in different frames and create as many skeleton frames as needed.
  // Different skeleton frames will be used for different frames then.

  // Cleanup first
  FrameCPP::Version_3_4_5::RawData::AdcDataContainer& adc( frame.getRawData()->refAdc() );
  adc.erase( adc.begin(), adc.end() );
  adcDataOffsetVector.erase( adcDataOffsetVector.begin(), adcDataOffsetVector.end() );

  for ( std::vector<std::string>::const_iterator i = adcNames.begin(); i != adcNames.end(); i++) {
    // See if there exists named ADC
    //
    ANPI iter = TOC::adcNamePositionMap.find( *i );
    if ( iter == adcNamePositionMap.end() )
      {
	throw not_found_error( "bad ADC name" );    
      }
      
    // Position to the start of the ADC in frame zero
    //
    unsigned long pos = iter->second.first[ 0 ];
    if ( pos == 0 )
      {
	throw not_found_error( "no ADC in the first frame" );
      }
    mStream->seekg( pos , std::ios::beg );

    // Read ADC data structure and append to the skeletal frame
    //
    adc.append ( dynamic_cast< FrameCPP::Version_3_4_5::AdcData* >( TOCReader::readStructure() ), false );
    
    // Insert into internal the adc data state std::vector
    adcDataOffsetVector.push_back( iter->second.first );
  }
}


//: Get length of name at `offset'.
inline int
FrameReadPlan::daqNameLength ( INT_8U offset ) {
  mStream->seekg ( offset );
  INT_2U nameLength;
    //    mStream->read ( &nameLength, 2 );
  (*this) >> nameLength;
  return nameLength;
}

#if 0


//: Get frame run number.
inline INT_4S
FrameReadPlan::daqFrameRun ( INT_8U frame_offset ) {
  mStream->seekg ( frame_offset + daqNameLength ( frame_offset ) + 2 );
  INT_4S run; 
  (*this) >> run;
  return run;
}

//: Get frame number.
inline INT_4U
FrameReadPlan::daqFrameNumber ( INT_8U frame_offset ) {
  mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 4);
  INT_4U frame; 
  (*this) >> frame;
  return frame;
}


//: Get frame start time seconds.
inline INT_4U
FrameReadPlan::daqFrameGTimeS ( INT_8U frame_offset ) {
  if (mFileHeader->getDataFormatVersion() < 4)
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 8);
  else
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 12);
  INT_4U gps;
  (*this) >> gps;
  return gps;
}

//: Get frame start time nanoseconds.
inline INT_4U
FrameReadPlan::daqFrameGTimeN ( INT_8U frame_offset ) {
  if (mFileHeader->getDataFormatVersion() < 4)
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 12);
  else
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 16);
  INT_4U gps; 
  (*this) >> gps;
  return gps;
}

//: Get frame leap seconds.
inline INT_2U
FrameReadPlan::daqFrameULeapS ( INT_8U frame_offset ) {
  if (mFileHeader->getDataFormatVersion() < 4)
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 16);
  else
    mStream->seekg (frame_offset + daqNameLength (frame_offset) + 2 + 20);
  INT_2U s; 
  (*this) >> s;
  return s;
}

#endif //0 

//   
//!exc: cannot_update 
//!exc: read_failure - Read failed.
//         
Frame&
FrameReadPlan::readFrame( std::istream& in, INT_4U frameNumber )
{
  std::istream* mStreamSaved = mStream;
  mStream = &in;
  try {    
    if ( frameNumber >= TOC::frame.size() )
      throw read_failure( "Invalid frame number", BADSTREAM );

    INT_8U frame_offset = TOC::position[ frameNumber ] + 8;

    // update frame header attributes
    mStream->seekg ( frame_offset );
    std::string newFrameName;
    INT_4S newFrameRun;
    INT_4U newFrameNumber;

    (*this) >> newFrameName >> newFrameRun >> newFrameNumber;

    // check run number, it must be the same as in the template frame
    if ( TOC::runs[0] != newFrameRun)
      throw cannot_update ( "Run number changed" );

    frame.mName = newFrameName;
    frame.mFrame = newFrameNumber;
    frame.mRun = newFrameRun;
    if (mFileHeader->getDataFormatVersion() >= 4)
      (*this) >> frame.mDataQuality;
    (*this).operator>>(frame.mGTime) >> frame.mULeapS;
    if (mFileHeader->getDataFormatVersion() < 5) {	
      INT_4S localTime;
      (*this) >> localTime;
    }
    (*this) >> frame.mDt;
  
    // update active ADCs -- all the attributes that can change and the data
    //
    FrameCPP::Version_3_4_5::RawData::AdcDataContainer& adc( frame.getRawData () -> refAdc() );
    for ( unsigned i = 0; i < adcDataOffsetVector.size(); i++ ) {
      INT_4U adc_offset = adcDataOffsetVector[ i ][ frameNumber ];
      int NL = (adc_offset + 8) + daqNameLength (adc_offset + 8) + 2;
      mStream->seekg ((NL + daqNameLength (NL) + 2) + 12);
      (*this) >> adc[i]->mBias >> adc[i]->mSlope
	      >> adc[i]->mUnits >> adc[i]->mSampleRate
	      >> adc[i]->mTimeOffset >> adc[i]->mFShift
	      >> adc[i]->mDataValid;

      // Set data byteswap flag to false to avoid swapping bytes in getData()
      // on data that gets written over by next read().
      //
      FrameCPP::Version_3_4_5::Vect &vect = *adc[i]->refData()[0];
      vect.setDataNeedsByteSwap(false);

      // Skip to the start of std::vector
      mStream->seekg ( adc_offset ); // beginning of the ADC structure
      INT_4U length;   // length of structure in bytes
      (*this) >> length; // ADC structure length

      // Skip to the end of AdcData structure
      mStream->seekg ( adc_offset + length ); 

      // Need to skip all SH and SE structures
      for(;;) {
       INT_2U classId;
       INT_4U len;
         // Read next structure header and see if it is a vector
         (*this) >> len >> classId;
       if ( classId == ID_VECT ) {
          // skip two bytes of the instance number 
          // to the beginnig of the vector structure data
            mStream->seekg( 2, std::ios::cur );
          break;
       }
       // skip whole structure minus 6 bytes we just read
         mStream->seekg( len - 6, std::ios::cur );
      }

      (*this) >> vect.mName >> vect.mCompress >> vect.mType >> vect.mNData >> vect.mNBytes;

      // :FIXME: this fails for STRING data type
      mStream->read( (char*) vect.getData(),
		     vect.getNData() * Vect::getTypeSize( vect.getType() ));    

      // Set byteswap flag to whatever reference frame header dictates
      //
      vect.setDataNeedsByteSwap( mFileHeader->byteSwapNeeded() );
    }
  } catch ( ... ) {
    mStream = mStreamSaved;
    throw;
  }
  mStream = mStreamSaved;
  return frame;
}
