#ifndef FrameCppGeneralReaderHH
#define FrameCppGeneralReaderHH


#include "framecpp/Version3_4_5/daqframe.hh"
#include "framecpp/Version3_4_5/framereaderbase.hh"


namespace FrameCPP {

  namespace Version_3_4_5 {

    class ObjectHeader;

    //-------------------------------------------------------------------
    //
    //: GeneralReader
    //
    // This class reads either daq or frame frame files.
    //
    class GeneralReader : private FrameReaderBase, public VFrameReader
    {
    public:

      enum ErrorCode
      {
        BADSTREAM = FrameReaderBase::BADSTREAM,
        SKIPFAILED = FrameReaderBase::SKIPFAILED,
        ENDOFFILE = FrameReaderBase::ENDOFFILE
      };
    
      //: Constructor
      //!exc: read_failure - Read failed.
      GeneralReader( std::istream& in, bool autoReset = false );

      //: Destructor
      //!exc: None.   
      virtual ~GeneralReader();

      FrameReaderBase::getError;
      FrameReaderBase::errorBegin;
      FrameReaderBase::errorEnd;
      FrameReaderBase::errorQuantity;
      FrameReaderBase::errorFirst;
      FrameReaderBase::errorLast;
      FrameReaderBase::clearErrors;
      FrameReaderBase::setErrorWatch;
      FrameReaderBase::getNBytes;
      FrameReaderBase::getFrameNumber;
    
      //!exc: std::bad_alloc - Memory allocation failed.
      //!exc: read_failure - Read failed.
      Frame* readFrame();

      //: Overrides object header reader in the superclass.
      //   
      //!exc: read_failure
      //
      ObjectHeader readNextObjectHeader();
  
      // Accessors
      FrameReaderBase::getFrames;
      FrameReaderBase::isAutoReset;
      
      FrameReaderBase::setAutoReset;

      FrameReaderBase::reset;

      std::istream* GetStreamPtr( );

      //!exc: None.      
      virtual bool isEndOfFile() const;

      /* Maps (from frame builder's imageframewriter.hh) */

      //
      // Mapping works like this: (classId << 16 | instanceId) -> offset.
      //<br>
      // classId identifies a type of structure,
      // instanceId is ++ for every newly output structure of class classId.
      // `offset' is a number of bytes to the start of the structure.
      //<br>
      // (classId, instanceID) pairs are found in the frame image
      // and are specified in the documentation as PTR_STRUCT.
      // The pair serves as a pointer.
      //<br>
      // Map allows a quick lookup of the offset to the start of a structure
      // knowing the instance and class IDs.

      typedef std::pair<INT_4U, adc_map_info_struct> offset_pair_type;
      typedef std::map<INT_4U, adc_map_info_struct> offset_map_type;
      offset_map_type instanceOffsetMap;

      // Multimap is constructed to allow a lookup of
      // all the instances of a particular class.
      //<br>
      // The mapping works like this: classId -> instanceId.
      //<br>
      // Now, if you need to find all the instance of, say, ADC class,
      // you go and iterate over IMI for ID_ADCDATA,
      // then for every instance Id found yo can get the offset
      // from the instanceOffsetMap, using OMI.
      
      typedef std::pair< INT_4U, INT_4U > class_pair_type;
      typedef std::multimap<INT_2U, INT_2U> class_map_type;
      class_map_type classInstanceMultimap;

      //: Given the offset find the ADC data vector ptr.
      INT_4U adcDataVectorPtr ( int offset );
    };


    //: Constructor
    //
    //!exc: read_failure - Read failed.   
    //   
    inline GeneralReader::
    GeneralReader( std::istream& in, bool autoReset )
      : FrameReaderBase( in, autoReset )
    {
    }


    //: Destructor
    //
    //!exc: None.
    //      
    inline GeneralReader::
    ~GeneralReader()
    {
    }

    inline std::istream* GeneralReader::
    GetStreamPtr( )
    {
      return mStream;
    }

    //-------------------------------------------------------------------
    //
    //: See if FrEndOfFile structure was read, ie. the end of frame file
    //+ reached
    //
    //!return: bool - True if FrEndOfFile was read.
    //
    //!exc: None 
    //     
    inline bool GeneralReader::
    isEndOfFile() const
    {
      return mEndOfFile;
    }

  } // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppGeneralReaderHH
