#include "framecpp/config.h"

#include <stdexcept>

#include "history.hh"
#include "framereader.hh"
#include "output.hh"

using FrameCPP::Version_3_4_5::History;
using std::runtime_error;



//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a History object with the following properties:
// <ul>
//     <li> name = "none" </li>
//     <li> time = 0 </li>
//     <li> comment = "" </li>
// </ul>
//
//!exc: None.   
//      
History::History() 
        : mName( "none" ), mTime( 0 ), mComment( "" )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.
//
//!param: const History& history - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
History::History( const History& history )
        : mName( history.mName ), mTime( history.mTime ),
          mComment( history.mComment )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - Name of history record.
//!param: INT_4U time - Time of post-processing.
//!param: const std::string& comment - Program name and relevant comment needed to define post-processing.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
History::History( const std::string& name, INT_4U time, const std::string& comment )
        : mName( name ), mTime( time ), mComment( comment )
{
}

   
//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.   
//   
History::~History()
{
}


//-----------------------------------------------------------------------------
//
//: Gets the name of the history record.
//
//!return: const std::string& -- Name.
//
const std::string& History::getName() const
{
    return mName;
}


#ifdef not_def
//-----------------------------------------------------------------------------
//
//: FrameReaderBase Constructor.
//
// This private constructor initializes the object from an FrameReaderBase
// object.  This constructor assumes that the length, class, and instance have
// already been read.
//
//!param: FrameReaderBase & in -
//
//!exc: std::bad_alloc
//!exc: read_failure
//
History::History( FrameReaderBase& in )
        : mName( "" ), mTime( 0 ), mComment( "" )
{
    in >> mName >> mTime >> mComment;

    in.readNextNodeReference();
}
#endif


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const History& history - Object to be assigned.
//
//!return: const History& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const History& History::operator=( const History& history )
{
    if ( this != & history )
    {
        mName = history.mName;
        mTime = history.mTime;
        mComment = history.mComment;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const History& history - Object to be assigned.
//
//!return: const History& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const History& History::operator+=( const History& history )
{
    throw runtime_error( "Unimplemented History::operator+=() is called." );    

    return *this;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.
//
//!return: SH - History Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH History::getSH3()
{
    SH sh( "FrHistory", ID_HISTORY, "History Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "time", "INT_4U", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrHistory *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads a History object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: History* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
History* History::read3( FrameReaderBase& in )
{
    History& hist = *new History();

    in >> hist.mName >> hist.mTime >> hist.mComment;

    in.readNextNodeReference();

    return &hist;
}


//-----------------------------------------------------------------------------
//
//: Writes a History object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - write failure occured.
//
void History::write( Output& out ) const
{
    out.startRecord( ID_HISTORY, 22
                     + mName.size() + mComment.size(), this );
    out << mName << mTime << mComment;
    out.endRecord( this );
}



//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two History objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const History& history - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
//      
bool History::operator==( const History& history ) const
{
    if ( this == &history )
    {
        return true;
    }

    return ( mName == history.mName &&
             mTime == history.mTime &&
             mComment == history.mComment );
}

   
//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- History class ID.
//
//!exc: None.   
//      
INT_2U History::getClassId() const
{
    return ID_HISTORY;
}


   
