#ifndef FrameCppLocationHH
#define FrameCppLocationHH


#include "framecpp/config.h"
#include "types.hh"
#include "input.hh"


namespace FrameCPP {


namespace Version_3_4_5 {

//-----------------------------------------------------------------------------
//
//: Latitude & Longitude.
//
// A simple class (more like a struct) which stores a location as
// latitude/longitude.
//
class Location
{
public:

    /* Constructors */
    //!exc: None.
    Location();
    //!exc: None.   
    Location( const Location& loc );
    Location( INT_2S lonD, INT_2S lonM, REAL_4 lonS,
              INT_2S latD, INT_2S latM, REAL_4 latS );
    
    /* Operator Overloads */
//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Location& loc - Obejct to be assigned.
//
//!return: const Location& - This object.
//
//!exc: None.   
    const Location& operator=( const Location& loc );

//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two Location objects are equal.
//
//!param: const Location& loc - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
    bool operator==( const Location& loc ) const;
    
    /* Accessors */
    //!exc: None.   
    INT_2S getLatD() const;
    //!exc: None.   
    INT_2S getLatM() const;
    //!exc: None.   
    REAL_4 getLatS() const;
    //!exc: None.   
    INT_2S getLonD() const;
    //!exc: None.   
    INT_2S getLonM() const;
    //!exc: None.   
    REAL_4 getLonS() const;

    /* Friends */

//-----------------------------------------------------------------------------
//
//: Input extraction operator.
//
// This reads this object from an Input object.
//
//!param: Input& in - Input data stream object.
//!param: Location& loc - Data target.
//
//!return: Input& - Input data stream object.
//
//!exc: read_failure - Read failure occured.
//
    friend Input& operator>>( Input&, Location& );
    
private:

    INT_2S mLatD;
    INT_2S mLatM;
    REAL_4 mLatS;
    INT_2S mLonD;
    INT_2S mLonM;
    REAL_4 mLonS;
};


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a location object with a latitude of 0 degrees and a longitude
// of zero degrees.
//
//!exc: None.   
//   
inline Location::Location()
        : mLatD( 0 ), mLatM( 0 ), mLatS( 0.0 ),
          mLonD( 0 ), mLonM( 0 ), mLonS( 0.0 )
{
}

    
//-----------------------------------------------------------------------------
//
//: Copy constructor.
//
//!param: const Location& loc - Object to be copied.
//
//!exc: None.   
//      
inline Location::Location( const Location& loc )
        : mLatD( loc.mLatD ), mLatM( loc.mLatM ), mLatS( loc.mLatS ),
          mLonD( loc.mLonD ), mLonM( loc.mLonM ), mLonS( loc.mLonS )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: INT_2S lonD - 
//!param: INT_2S lonM -
//!param: REAL_4 lonS -
//!param: INT_2S latD -
//!param: INT_2S latM -
//!param: REAL_4 latS -
//
inline Location::Location( INT_2S lonD, INT_2S lonM, REAL_4 lonS,
                                     INT_2S latD, INT_2S latM, REAL_4 latS )
        : mLatD( latD ), mLatM( latM ), mLatS( latS ),
          mLonD( lonD ), mLonM( lonM ), mLonS( lonS )
{
}
    


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get Latitude Degrees.
//
//!return: INT_2S
//
//!exc: None.   
//      
inline INT_2S Location::getLatD() const
{
    return mLatD;
}


//-----------------------------------------------------------------------------
//
//: Get Latitude Minutes.
//
//!return: INT_2S
//
//!exc: None.   
//      
inline INT_2S Location::getLatM() const
{
    return mLatM;
}


//-----------------------------------------------------------------------------
//
//: Get Latitude Seconds.
//
//!return: REAL_4
//
//!exc: None.   
//      
inline REAL_4 Location::getLatS() const
{
    return mLatS;
}


//-----------------------------------------------------------------------------
//
//: Get Longitude Degrees.
//
//!return: INT_2S
//
//!exc: None.   
//      
inline INT_2S Location::getLonD() const
{
    return mLonD;
}


//-----------------------------------------------------------------------------
//
//: Get Longitude Minutes.
//
//!return: INT_2S
//
//!exc: None.   
//      
inline INT_2S Location::getLonM() const
{
    return mLonM;
}


//-----------------------------------------------------------------------------
//
//: Get Longitude Seconds.
//
//!return: REAL_4
//
//!exc: None.   
//      
inline REAL_4 Location::getLonS() const
{
    return mLonS;
}
 

//!exc: read_failure - Read failed.
Input& operator>>( Input& in, Location& loc );

//!exc: write_failure - Write failed.   
Output& operator<<(
                   Output& out, const Location& loc );


} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppLocationHH
