#ifndef FrameCppMsgHH
#define FrameCppMsgHH


// Local Includes
#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/time.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrMsg;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: Message Log Data
//
// This class represents message log data.  It is modeled after the FrMsg
// structure in the Frame Specification, LIGO-T970130-B-E.
//
// <P>The Msg class has the following attributes:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B>                        </td></tr>
//   <tr><td> alarm</td><td>Name of message, error flag or alarm state           </td></tr>
//   <tr><td> message</td><td>Message body                                       </td></tr>
//   <tr><td> severity</td><td>Message severity level                            </td></tr>
//   <tr><td> GTime</td><td>GPS time in seconds and nanoseconds                  </td></tr>
// </table>
//
class Msg : public Base
{
public:

    /* Constructors/Destructor */

    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a Msg object with the following properties:
    // <ul type=disc>
    //     <li> alarm = "none" </li>
    //     <li> message = "" </li>
    //     <li> severity = 0 </li>
    //     <li> gtime = 0 seconds and 0 nanoseconds </li>
    // </ul>
    //
    //!exc: None.
    //      
    Msg();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    //!param: const Msg& msg - Object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    Msg( const Msg& msg );

    //-----------------------------------------------------------------------------
    //
    //: Constructor
    //
    // This creates a log message with the specified values.
    //
    //!param: const std::string& alarm - Name of message, error flag or alarm state
    //!param: const std::string& message - Message body
    //!param: INT_4U severity - Message severity level
    //!param: const Time& gtime - GPS time in seconds and nanoseconds
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    Msg( const std::string& alarm, const std::string& message, INT_4U severity, const Time& gtime );
 
    /* Operator overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const Msg& msg - Object to be assigned.
    //
    //!return: const Msg& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Msg& operator=( const Msg& );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const Msg& msg - Object to be added.
    //
    //!return: const Msg& -- This object.
    //
    const Msg& operator+=( const Msg& );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two Msg objects are equal. For this comparison, all of
    // the data is compared.
    //
    //!param: const Msg& msg - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.
    //         
    bool operator==( const Msg& msg ) const;
    bool operator!=( const Msg& msg ) const;

    /* Required Methods */
    //!exc: None.
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Get Structure Header
    //
    // Constructs the Structure Header for this class.
    //
    //!return: SH -- Msg structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    static SH getSH();

    //: Frame data version 3 and 4 definition
    //!exc: std::bad_alloc - Memory allocation failed   
    static SH getSH3();

    /* Accessors */
    const std::string& getAlarm() const;
    //!exc: None.   
    const std::string& getMessage() const;
    //!exc: None.   
    INT_4U getSeverity() const;
    //!exc: None.   
    const Time& getGTime() const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrMsg;

    /* This method will be accessed by the Dictionary class but is
       not available for public use. */

    //-----------------------------------------------------------------------------
    //
    //: Reads a Msg object from an Input object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: Input& in - Input data stream object.
    //
    //!return: Msg* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static Msg* read( FrameReaderBase& in );

    //: Frame data versions 3 and 4 reader
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static Msg* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes a Msg object to an Output object.
    //
    //!param: Output& out - Output data object.
    //
    //!exc: write_failure - Write failure occured.
    //
    //!kw: const
    //
    virtual void write( Output& out ) const;

private:
    //: Name of message, error flag or alarm state
    std::string mAlarm;
    //: Message body
    std::string mMessage;
    //: Message severity level
    INT_4U mSeverity;
    //: GPS time
    Time mGTime;
};




/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const Msg& msg - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.      
inline bool Msg::operator!=( const Msg& msg ) const 
{
    return !( *this == msg );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get Message
//
// Get the message body.
//
//!return: const std::string& -- The message body
//
//!exc: None.      
inline const std::string& Msg::getMessage() const 
{
    return mMessage;
}

    
//-----------------------------------------------------------------------------
//
//: Get Severity
//
// Get the message severity level.
//
//!return: INT_4U -- The severity
//
//!exc: None.      
inline INT_4U Msg::getSeverity() const 
{
    return mSeverity;
}


//-----------------------------------------------------------------------------
//
//: Get message GPS time.
//
//!return: const Time& -- GPS time.
//
//!exc: None.      
inline const Time& Msg::getGTime() const 
{
    return mGTime;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppMsgHH
