#ifndef FrameCppOutputHH
#define FrameCppOutputHH


#include <map>
#include <iostream>
#include "container.hh"
#include "errors.hh"
#include "reference.hh"
#include <set>
#include "frame.hh"


namespace FrameCPP {

namespace Version_3_4_5 {

class Base;
class ContainerBase;
class Dictionary;
class ObjectInfo;


//-----------------------------------------------------------------------------
//
//: Object Information.
//
// Convenience class for internal library use.
//

class ObjectInfo
{
public:
  typedef std::vector< const Base* >::iterator so_iterator;
  typedef std::vector< const Base* >::iterator const_so_iterator;
  typedef std::vector< const ContainerBase* >::iterator c_iterator;
  typedef std::vector< const ContainerBase* >::iterator const_c_iterator;
  typedef std::vector< bool >::iterator i_iterator;
  typedef std::vector< bool >::iterator const_i_iterator;

  
  ObjectInfo() : subobjects(), containers(), isContainer() {};
  virtual ~ObjectInfo();

  std::vector< const Base* > subobjects;
  std::vector< const ContainerBase* > containers;
  std::vector< bool > isContainer;
};


//-----------------------------------------------------------------------------
//
//: Output.
//
// Creates a frame file as an output stream.
//
class Output
{
public:

    //: Constructs an output object and writes the fileheader.
    //   
    //!exc: write_failure - Write failed.
    //      
    Output( std::ostream& out );

    //: Destructs the output object.
    //   
    //!exc: write_failure - Write failed.
    //         
    virtual ~Output();

    /* I/O */
    //--------------------------------------------------------------------------------
    //
    //: Serialize a frame
    //
    // Outputs a frame object in the format defined by the frame spec.
    //
    //!exc: write_failure - Write failed.
    //         
    virtual void writeFrame( const Frame& frame );
    //: Write bytes to the stream.
    //   
    //!exc: write_failure - Write failed.
    //         
    void write( const void* buffer, unsigned int size );
    //--------------------------------------------------------------------------------
    //
    //: Begins a record.
    //
    // This writes out the length, class and instance id of the
    // object and also initializes the ObjectInfo structure.
    //
    //!exc: write_failure - Write failed.
    //         
    virtual void startRecord( INT_2U classId, INT_4U length, const Base* b );
    //--------------------------------------------------------------------------------
    //
    //: Ends a record.
    //
    // This method is executed when an object is finished writing
    // itself.  This method calls the write function for any objects it referred
    // to and appends a pointer to the next object if it was a linked-list node.
    //   
    //!exc: write_failure - Write failed.
    //         
    void endRecord( const Base* b );
    //: Finish frame file and close
    virtual void close( unsigned long tocOffset = 0 );
    //: Place an object into the subobjects queue
    void queueObject( const Base& o );
    
    /* Operator overloads */

    //: Output a character
    //!exc: write_failure - Write failed.
    Output& operator<<( CHAR v );
    //: Output an unsigned character
    //!exc: write_failure - Write failed.   
    Output& operator<<( CHAR_U v );
    //: Output short integer
    //!exc: write_failure - Write failed.   
    Output& operator<<( INT_2S v );
    //: Output short unsigned integer
    //!exc: write_failure - Write failed.   
    Output& operator<<( INT_2U v );
    //: Output signed integer
    //!exc: write_failure - Write failed.   
    Output& operator<<( INT_4S v );
    //: Output unsigned integer
    //!exc: write_failure - Write failed.   
    Output& operator<<( INT_4U v );
    //: Output long integer
    //!exc: write_failure - Write failed.   
    Output& operator<<( const INT_8S& v );
    //: Output long unsigned integer
    //!exc: write_failure - Write failed.   
    Output& operator<<( const INT_8U& v );
    //: Output real number
    //!exc: write_failure - Write failed.   
    Output& operator<<( const REAL_4& v );
    //: Output double precision real number
    //!exc: write_failure - Write failed.   
    Output& operator<<( const REAL_8& v );
    //: Output a std::string
    //!exc: write_failure - Write failed.   
    Output& operator<<( const std::string& s );
    //: Output an object
    //!exc: write_failure - Write failed.   
    Output& operator<<( const Base* b );
    //: Output a container.
    //!exc: write_failure - Write failed.   
    template< class T > Output& operator<<( const Container< T >& c );
    //: Output a frame
    //!exc: write_failure - Write failed.   
    Output& operator<<( const Frame& );
    //: Output a reference
    //!exc: write_failure - Write failed.   
    Output& operator<<( const Reference& ref );

    /* Accessors  */
    //: Get number of bytes written
    INT_4U getBytes() const;
    //: Get the number of frame written
    INT_4U getFrames() const;

    typedef std::map< const Base*, ObjectInfo* > ObjectQueue;

protected:
    //: Get object instance.
    INT_2U getInstance( const Base* b );
    //: Frame dictionary.
    Dictionary &dictionary;
    //: Output data stream.
    std::ostream& mOut;
    //: Whether frame file was already closed.
    bool mClosed;

private:
    //: Write file header out
    void open();
   
    //: Number of frames written.
    INT_4U mNFrames;
    //: Number of bytes written.
    INT_4U mNBytes;
    //: Whether the currently written object is a tree node.
    bool mIsNode;
    //: Next tree node.
    Reference mNextNode;
    //: Object output queue.
    ObjectQueue mObjectQueue;
    ObjectInfo* mCurrentInfo;

    //: Set of object already written.
    std::set< const Base* > mWritten;
    //: List of structure instances.
    std::map< const Base*, INT_2U > mInstance;
    //: Maximum instance values for structure classes.
    std::map< INT_2U, INT_2U > mMaxInstance;

protected:
    //:    Set of structure class IDs for which SH and SEs were already written out.
    //
    //    This is NOT reset for every frame. Kept throughout file output
    //    to make sure just one copy of SH (and one set of corresponding SEs)
    //    is output per structure type PER FILE.
    std::set< INT_2U > shWrittenSet;
};

// This no longer needs to be a template
//!exc: write_failure - Write failed.   
template< class T >
Output& Output::operator<<(
    const Container< T >& c )
{
    if ( mCurrentInfo == 0 )
    {
        throw write_failure( "Logic error.  Attempt to write before "
                             "startRecord called." );
    }
    
    if ( c.size() == 0 )
    {
        (*this) << (Base*)0;
        return *this;
    }
    
    (*this) << Reference( c[ 0 ]->getClassId(), getInstance( c[ 0 ] ) );
    mCurrentInfo->containers.push_back( &c );
    mCurrentInfo->isContainer.push_back( true );
    
    return *this;
}

inline INT_4U Output::getBytes() const
{
    return mNBytes;
}

inline INT_4U Output::getFrames() const
{
    return mNFrames;
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppOutputHH
