#include "framecpp/config.h"

#include "rawdata.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::RawData;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;



//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an RawData object with the following properties:
// <ul>
//     <li> name = "none" </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.   
//      
RawData::RawData()
        : Base(), mName( "none" ), mSerData(), mAdcData(), mTable(), mMsg(), mMore()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.
//
//!param: const RawData& rawData - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
RawData::RawData( const RawData& rawData )
        : Base( rawData ), mName( rawData.mName ),
          mSerData( rawData.mSerData ), mAdcData( rawData.mAdcData ),
	  mTable( rawData.mTable ),
          mMsg( rawData.mMsg ), mMore( rawData.mMore )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - Raw data name.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
RawData::RawData( const std::string& name ) 
        : Base(), mName( name ), mSerData(), mAdcData(), mTable(), mMsg(), mMore()
{
}

   
//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.   
//   
RawData::~RawData()
{
}


//-----------------------------------------------------------------------------
//
//: Get the raw data name.
//
//!return: const std::string& -- Name.
//
//!exc: None.   
//      
const std::string& RawData::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const RawData& rawData - Object to be assigned.
//
//!return: const RawData& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const RawData& RawData::operator=( const RawData& rawData )
{
    if ( this != &rawData )
    {
        mName = rawData.mName;
        mSerData = rawData.mSerData;
        mAdcData = rawData.mAdcData;
        mTable = rawData.mTable;
        mMsg = rawData.mMsg;
        mMore = rawData.mMore;
    }

    return *this;
}

//-----------------------------------------------------------------------------
//
//: += Assignment Operator.
//
//!param: const RawData& rawData -
//
//!return: const RawData&
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: frame_mismatch - Raw Data or ADC structures are different
//
const RawData& RawData::operator+=( const RawData& rawData ) 
{
  // Raw Data names must be the same
  //
  if ( mName != rawData.mName )
    {
      throw frame_mismatch( "Raw Data Name different" );
    }

  mAdcData += rawData.mAdcData;

#ifdef not_def
    if ( this != &rawData )
    {
        mSerData = rawData.mSerData;
        mTable = rawData.mTable;
        mMsg = rawData.mMsg;
        mMore = rawData.mMore;
    }
#endif

    return *this;
}




//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.  
//
//!return: SH - RawData structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH RawData::getSH() 
{
  SH sh( "FrRawData", ID_RAWDATA, "Raw Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "firstSer", "PTR_STRUCT(FrSerData *)", "" ) );
  sh.appendSE( SE( "firstAdc", "PTR_STRUCT(FrAdcData *)", "" ) );
  sh.appendSE( SE( "firstTable", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "logMsg", "PTR_STRUCT(FrMsg *)", "" ) );
  sh.appendSE( SE( "more", "PTR_STRUCT(FrVect *)", "" ) );

  return sh;
}

//-----------------------------------------------------------------------------
//
//: Constructs frame data vesion 3 Structure Header for this class.  
//
//!return: SH - RawData structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH RawData::getSH3()
{
    SH sh( "FrRawData", ID_RAWDATA, "Raw Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "firstSer", "PTR_STRUCT(FrSerData *)", "" ) );
    sh.appendSE( SE( "firstAdc", "PTR_STRUCT(FrAdcData *)", "" ) );
    sh.appendSE( SE( "logMsg", "PTR_STRUCT(FrMsg *)", "" ) );
    sh.appendSE( SE( "more", "PTR_STRUCT(FrVect *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an RawData object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: RawData* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure happened.
//
RawData* RawData::read( FrameReaderBase& in )
{
  RawData& raw = *new RawData();
  in >> raw.mName >> raw.mSerData >> raw.mAdcData
     >> raw.mTable >> raw.mMsg >> raw.mMore;
  return &raw;
}

//: Version 3 frame input method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure happened.
//   
RawData* RawData::read3( FrameReaderBase& in )
{
    RawData& raw = *new RawData();
    in >> raw.mName >> raw.mSerData >> raw.mAdcData >> raw.mMsg >> raw.mMore;
    return &raw;
}


//-----------------------------------------------------------------------------
//
//: Writes an RawData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void RawData::write( Output& out ) const 
{
    out.startRecord( ID_RAWDATA,
		     27
		     + 4 // for the pointer to Table
                     + mName.size(), this );
    out << mName << mSerData << mAdcData << mTable << mMsg << mMore;
    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Find an ADC object by channel number.
//
//!param: INT_4U channel -
//
//!return: const AdcData*
//
//!exc: None.   
//   
const FrameCPP::Version_3_4_5::AdcData* RawData::findAdc( INT_4U channel ) const
{
    const_adcData_iterator iter = mAdcData.begin();
    while( ( iter != mAdcData.end() ) && ( channel != (*iter)->getChannelNumber() ) )
    {
        ++iter;
    }
    
    if ( iter == mAdcData.end() )
    {
        return 0;
    }

    return *iter;
}


//-----------------------------------------------------------------------------
//
//: Find an ADC object by channel number.
//
//!param: INT_4U channel -
//
//!return: AdcData*
//
//!exc: None.   
//      
FrameCPP::Version_3_4_5::AdcData* RawData::findAdc( INT_4U channel )
{
    adcData_iterator iter = mAdcData.begin();
    while( ( iter != mAdcData.end() ) && ( channel != (*iter)->getChannelNumber() ) )
    {
        ++iter;
    }
    
    if ( iter == mAdcData.end() )
    {
        return 0;
    }

    return *iter;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two RawData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const RawData& rawData - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
//      
bool RawData::operator==( const RawData& rawData ) const 
{
    if ( this == &rawData )
    {
        return true;
    }

    return ( ( mName == rawData.mName ) &&
             ( mSerData == rawData.mSerData ) &&
             ( mAdcData == rawData.mAdcData ) &&
             ( mTable == rawData.mTable ) &&
             ( mMsg == rawData.mMsg ) &&
             ( mMore == rawData.mMore ) );
}

   
//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- RawData class ID.
//   
//!exc: None.   
//   
INT_2U RawData::getClassId() const
{
    return ID_RAWDATA;
}

   
