#ifndef FrameCppRawDataHH
#define FrameCppRawDataHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/serdata.hh"
#include "framecpp/Version3_4_5/adcdata.hh"
#include "framecpp/Version3_4_5/msg.hh"
#include "framecpp/Version3_4_5/table.hh"

#include "framecpp/SearchContainer.hh"

namespace FrameCPP {

namespace Version_6
{
  class FrRawData;
}

namespace Version_3_4_5 {


class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: Raw Data.
//
// This class represents raw data.  It is modeled after the FrRawData
// structure in the Frame Specification, LIGO-T970130-B-E.
//
class RawData : public Base
{
public:

    //!ignore_begin:
    DECLARE_CONTAINER( SerData, serData, SerData, getName );
    DECLARE_CONTAINER( AdcData, adcData, AdcData, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    DECLARE_CONTAINER( Msg, msg, Msg, getAlarm );
    DECLARE_CONTAINER( More, more, Vect, getName );
    //!ignore_end:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates an RawData object with the following properties:
    // <ul>
    //     <li> name = "none" </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.
    //
    RawData();

    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of the object.
    //
    //!param: const RawData& rawData - Object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    RawData( const RawData& rawData );


    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - Raw data name.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    RawData( const std::string& name );
    
    /* Destructor */
    //!exc: None.
    virtual ~RawData();

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const RawData& rawData - Object to be assigned.
    //
    //!return: const RawData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const RawData& operator=( const RawData& rawData );
    //!exc: None.
    bool operator==( const RawData& rawData ) const;
    //!exc: None.   
    bool operator!=( const RawData& rawData ) const;
   
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: frame_mismatch - Raw Data or ADC structures are different   
    const RawData& operator+=( const RawData& rawData );

    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.  
    //
    //!return: SH -- RawData structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //-----------------------------------------------------------------------------
    //
    //: Constructs frame data vesion 3 Structure Header for this class.  
    //
    //!return: SH -- RawData structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH3();

    /* Accessors */
    //!exc: None.      
    const std::string& getName() const;
    
    //!exc: None.      
    SerDataContainer& refSerial();
    //!exc: None.      
    const SerDataContainer& refSerial() const;

    //
    //: Get the container holding the ADC data.
    //
    //!return: AdcDataContainer&
    //
    //!exc: None.      
    AdcDataContainer& refAdc();

    //
    //: Get the container holding the ADC data.
    //
    //!return: const AdcDataContainer&
    //
    //!exc: None.      
    const AdcDataContainer& refAdc() const;
    //!exc: None.      
    const AdcData* findAdc( INT_4U channel ) const;
    //!exc: None.      
    AdcData* findAdc( INT_4U channel );


    //
    // Get the container holding the user-define table data.
    //
    //!return: TableContainer&
    //
    //!exc: None.      
    TableContainer& refTable();

    //
    // Get the container holding the user-define table data.
    //
    //!return: const TableContainer&
    //
    //!exc: None.      
    const TableContainer& refTable() const;

    //
    // Get the container holding the error messages.
    //
    //!return: MsgContainer&
    //
    //!exc: None.      
    MsgContainer& refMsg();

    //
    // Get the container holding the error messages.
    //
    //!return: const MsgContainer&
    //
    //!exc: None.      
    const MsgContainer& refMsg() const;

    //
    // Get the container holding the user-define data.
    //
    //!return: MoreContainer&
    //
    //!exc: None.      
    MoreContainer& refMore();

    //
    // Get the container holding the user-define data.
    //
    //!return: const MoreContainer< Vect, &Vect::getName >&
    //
    //!exc: None.      
    const MoreContainer& refMore() const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrRawData;


    //-----------------------------------------------------------------------------
    //
    //: Reads an RawData object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: RawData* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure happened.
    //
    static RawData* read( FrameReaderBase& in );
    //: Version 3 frame input method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure happened.
    //   
    static RawData* read3( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    //: Writes an RawData object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;

private:
  
    //: Name of raw data.
    std::string mName;
    //: Serial data.
    SerDataContainer mSerData;
    //: Adc data.
    AdcDataContainer mAdcData;
    //: Table of some data.
    TableContainer mTable;
    //: Error messages.
    MsgContainer mMsg;
    //: User defined additional data.
    MoreContainer mMore;
};
    

/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const RawData& rawData - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.   
//      
inline bool RawData::operator!=( const RawData& rawData ) const
{
    return !( *this == rawData );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get the container holding the serial data.
//
//!return: SerDataContainer& -- Serial data.
//
//!exc: None.   
//      
inline
RawData::SerDataContainer&
RawData::refSerial()
{
    return mSerData;
}


//-----------------------------------------------------------------------------
//
//: Get the container holding the serial data.
//
//!return: const SerDAtaContainer& -- Serial data.
//
//!exc: None.   
//      
inline const
RawData::SerDataContainer&
RawData::refSerial() const
{
    return mSerData;
}


//-----------------------------------------------------------------------------
//
//: Get the container holding the ADC data.
//
//!return: AdcDataContainer& -- Adc data.
//
//!exc: None.   
//      
inline
RawData::AdcDataContainer&
RawData::refAdc()
{
    return mAdcData;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the ADC data.
//
//!return: const AdcDataContainer& -- Adc data.
//
//!exc: None.   
//      
inline const
RawData::AdcDataContainer&
RawData::refAdc() const
{
    return mAdcData;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the user-define table data.
//
//!return: TableContainer& -- Table.
//
//!exc: None.   
//      
inline
RawData::TableContainer&
RawData::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the user-define table data.
//
//!return: const TableContainer& -- Table.
//
//!exc: None.   
//      
inline const
RawData::TableContainer&
RawData::refTable() const
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the error messages.
//
//!return: MsgContainer& -- Messages.
//
//!exc: None.   
//      
inline RawData::MsgContainer&
RawData::refMsg()
{
    return mMsg;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the error messages.
//
//!return: const MsgContainer& -- Messages.
//
//!exc: None.   
//      
inline
const RawData::MsgContainer&
RawData::refMsg() const
{
    return mMsg;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the user-define data.
//
//!return: MoreContainer& -- User defined data.
//
//!exc: None.   
//      
inline RawData::MoreContainer&
RawData::refMore()
{
    return mMore;
}


//-----------------------------------------------------------------------------
//
// Get the container holding the user-define data.
//
//!return: const SearchContainer< Vect, &Vect::getName >& -- User defined data.
//
//!exc: None.   
//      
inline const
RawData::MoreContainer&
RawData::refMore() const 
{
    return mMore;
}

} // namespace - Version_3_4_5
 
} // namespace - FrameCPP


#endif // FrameCppRawDataHH
