#ifndef FrameCppSeHH
#define FrameCppSeHH


#include <memory>
#include "framecpp/config.h"
#include "base.hh"
#include <string>


namespace FrameCPP {

namespace Version_3_4_5 {

  class Input;


//-----------------------------------------------------------------------------
//
//: Frame Structure Element.
//
// This class represents the Frame Structure Element object as given in the
// Frame Specification.  It stores information about a single field in a frame
// structure.  The information stored is:
// <p><ol>
//   <li> name - The name of the field. </li>
//   <li> class - The class type, e.g. "INT_2U" or "REAL_8". </li>
//   <li> comment </li>
// </ol>
//
// <p>This is implemented as a simple class without mutators.  The class has no
// default constructor.
//
//!friend: Input - Input is a friend so that the read method can be accessed.
//!friend:    This is the only private method or data accessed by Input.  The
//!friend:    method is private because it is only meant to be called by the
//!friend:    Input class.
//
class SE : public Base
{
public:

    /* Constructors */
//-----------------------------------------------------------------------------
//
//: Copy constructor.
//
//!param: const SE& se - The object to copy from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
    SE( const SE& se );

//-----------------------------------------------------------------------------
//
//: Constructor.
//
// This creates a Structure Element with the given attributes.
//
//!param: const std::string& name - The field name.
//!param: const std::string& type - The Class type
//!param: const std::string& comment - A comment.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
   SE( const std::string& name, const std::string& type, const std::string& comment );

    /* Operator Overloads */
//-----------------------------------------------------------------------------
//
//: Assignment Operator..
//
//!param: const SE& se - The object to assign from.
//
//!return: const SE& -- This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
    const SE& operator=( const SE& se );

//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two SE objects are equal. For this comparison, the
// comments are not compared.
//
//!param: const SE& se - The object to compare with.
//
//!return: bool -- true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//      
    bool operator==( const SE& se ) const;
    
    /* Required Methods */
    //!exc: None.
    virtual INT_2U getClassId() const;

    /* Accessors */
    //!exc: None.   
    const std::string& getName() const;
    //!exc: None.   
    const std::string& getClass() const;
    //!exc: None.   
    const std::string& getComment() const;
    
private: friend class Input;


//-----------------------------------------------------------------------------
//
//: Instantiates a new SE object from a stream.
//
// This method is used only by the Input object.
//
//!param: Input& in - The object to read from.
//
//!return: SE* -- The new object.  The user must destruct this object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//
//!kw: static
//
    static SE* read( Input& in );

private:


//-----------------------------------------------------------------------------
//
//: Input Constructor.
//
// This creates a Structure Element by reading it from an Input object.
//
//!param: Input& in - The Input object to read from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failed.
//
//!kw: private
//
//!todo: Check about changing this to reading from a FrameReaderBase.
//
    SE( Input& in );


//-----------------------------------------------------------------------------
//
//: Write this object.
//
// Writes this SE to an output object.
//
//!param: Output& out - The object to write to.
//
//!exc: write_failure - Write failed.
//
//!kw: const
//
    virtual void write( Output& out ) const;

    //: Name of the element described.
    std::string mName;
    //: Element data type.
    std::string mClass;
    //: Comment.
    std::string mComment;
};



/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Get the element type.
//
//!return: const std::string& -- The type.
//
//!kw: inline const
//
//!exc: None.   
//      
inline const std::string& SE::getClass() const
{
    return mClass;
}


//-----------------------------------------------------------------------------
//
//: Get the comment.
//
//!return: const std::string& -- The comment.
//
//!kw: inline const
//
//!exc: None.   
//      
inline const std::string& SE::getComment() const
{
    return mComment;
}


} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppSeHH
