#ifndef FrameCppSerDataHH
#define FrameCppSerDataHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/time.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"

#include "framecpp/SearchContainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrSerData;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: Serial Data
//
// This class represents serial data.
//
// It is modeled after the FrSerData structure in the Frame Specification,
// LIGO-T970130-B-E.
//
// <P>The SerData class has the following attributes:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B> </td></tr>
//   <tr><td> name</td><td>Name of station producing serial data stream. </td></tr>
//   <tr><td> time</td><td>Time of data acquisition. </td></tr>
//   <tr><td> sampleRate</td><td>Sample rate (samples/sec). </td></tr>
//   <tr><td> data</td><td>ASCII-based data. </td></tr>
// </table>
//
// <p>The SerData class also contains the following data:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Data</B></td><td><B>Description</B> </td></tr>
//   <tr><td> serial</td><td>Vect container for serial data </td></tr>
//   <tr><td> table</td><td>Additional user-defined data </td></tr> 
// </table>
//
// <p>The data is accessed through the container objects inside SerData.  These
// are accessed via the <code>refData</code> and <code>refTable</code> methods.
//
class SerData : public Base
{
public:


    //!ignore_begin:
    // DECLARE_CONTAINER( Data, data, Vect, getName );
    DECLARE_DEFAULT_CONTAINER( Serial, serial, Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    //!ignore_end:

    /* Constructors/Destructor */
//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a SerData object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> time = 0 seconds, 0 nanoseconds </li>
//     <li> sampleRate = 0.0 </li>
//     <li> ASCII data = "" </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.   
//      
    SerData();

//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy for owned data.
//
//!param: const SerData& serData - The object to copy from
//
//!exc: std::bad_alloc - Memory allocation failed
//
    SerData( const SerData& serData );

//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - Name of station producing serial data stream.
//!param: const Time& time - Time of data acquisition.
//!param: REAL_4& sampleRate - Sampling rate, sample / s.
//
//!exc: std::bad_alloc - Memory allocation failed
//   
   SerData( const std::string& name, const Time& time, REAL_4 sampleRate );

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    // This performs a deep copy for owned data.
    //
    //!param: const SerData& serData - The object to assign from.
    //
    //!return: const SerData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const SerData& operator=( const SerData& serData );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    // This performs a deep copy for owned data.
    //
    //!param: const SerData& serData - The object to be added.
    //
    //!return: const SerData& -- This object.
    //
    const SerData& operator+=( const SerData& serData );

//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two SerData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const SerData& serData - The object to compare with.
//
//!return: bool -- true if the objects are equal.
//
//!exc: None.
//   
    bool operator==( const SerData& serData ) const;
    //!exc: None.   
    bool operator!=( const SerData& serData ) const;

    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;
   
//-----------------------------------------------------------------------------
//
//: Get Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH -- Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
    static SH getSH();


//-----------------------------------------------------------------------------
//
//: Get frame data format version 3 Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH -- Structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
    static SH getSH3();

    /* Accessors */
    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g.
       typedef SearchContainer< SerData, &SerData::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
    //!exc: None.   
    const Time& getTime() const;
    //!exc: None.      
    REAL_8 getSampleRate() const;
    //!exc: None.      
    const std::string& getData() const;

    /* Serial Data accessors */
    //!exc: None.      
    SerialContainer& refData();
    //!exc: None.      
    const SerialContainer& refData() const;

    /* Table Accessors */
    //!exc: None.      
    TableContainer& refTable();
    //!exc: None.      
    const TableContainer& refTable() const;

    /* Mutators */
    //!exc: std::bad_alloc - Memory allocation failed.
    void setData( const std::string& data );
    
private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrSerData;

    /* This method will be accessed by the Dictionary class but is
       not available for public use. */

//-----------------------------------------------------------------------------
//
//: Reads a SerData object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: SerData* -- Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
    static SerData* read( FrameReaderBase& in );

//: Version 3 frame input function.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
    static SerData* read3( FrameReaderBase& in );

//-----------------------------------------------------------------------------
//
//: Writes a SerData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
    virtual void write( Output& out ) const;
    
private:

    //: Name of station producing serial data stream.
    std::string mName;
    //: Time of data acquisition.
    Time mTime;
    //: Sampling rate, samples / s.
    REAL_4 mSampleRate;
    //: ASCII-based data.
    std::string mData;
    //: Serial data vector.
    SerialContainer mSerial;
    //: User-defined information.
    TableContainer mTable;
};



/******************/
/* Inline Methods */
/******************/



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
// Determines whether two objects are not equal.  All contained data is
// compared.
//
//!param: const SerData& serData - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!exc: None.
//      
inline bool SerData::operator!=( const SerData& serData ) const
{
    return !( *this == serData );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
//
//: Get Acquisition Time
//
// Gets the time of data acquisition.
//
// <p>This is the GPS time in integer seconds
// and integer nanoseconds sonce the GPS standard epoch.  This is valid for
// apporximately 143 years from time origin.
//
//!return: const Time& -- Time.
//
//!exc: None.
//      
inline const Time& SerData::getTime() const
{
    return mTime;
}


//-----------------------------------------------------------------------------
//
//: Get Sample Rate
//
// Gets the sample rate, samples / sec.
//
//!return: REAL_8 -- Sampling rate.
//
//!exc: None.
//      
inline REAL_8 SerData::getSampleRate() const
{
    return mSampleRate;
}


//-----------------------------------------------------------------------------
//
//: Get ASCII data
//
// Gets ASCII-based data.
//
//!return: const std::string& -- Data.
//
//!exc: None.
//      
inline const std::string& SerData::getData() const
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: Get Table Container
//
// Gets the container holding additional user-defined table data.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Table.
//
//!exc: None.
//      
inline SerData::TableContainer& SerData::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Get const Table Container
//
// Gets the container holding additional user-defined table data.
//
//!return: const SearchContainer< Table, &Table::getName >& -- Table.
//
//!exc: None.
//      
inline const
SerData::TableContainer& SerData::refTable() const
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Set ASCII Data
//
// Set the ASCII-based data.
//
//!param: const std::string& data - Data to be set.
//
//!exc: std::bad_alloc - Memory allocation failed.
//      
inline void SerData::setData( const std::string& data )
{
    mData = data;
}


//-----------------------------------------------------------------------------
//
//: Get Data Container
//
// Get the container which contains the data.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Data.
//
//!kw: inline
//
//!exc: None.
//      
inline SerData::SerialContainer& SerData::refData()
{
    return mSerial;
}


//-----------------------------------------------------------------------------
//
//: Get const Data Container
//
// Get the container which contains the data.
//
//!return: const SearchContainer< Vect, &Vect::getName >& -- Data.
//
//!kw: inline const
//
//!exc: None.
//      
inline const
SerData::SerialContainer& SerData::refData() const
{
    return mSerial;
}

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppSerDataHH
