#ifndef FrameCppSHHH
#define FrameCppSHHH


#include "base.hh"
#include "se.hh"
#include "container.hh"


namespace FrameCPP {

namespace Version_3_4_5 {


//-----------------------------------------------------------------------------
//
//: Frame Structure Header.
//
class SH : public Base
{
public:

    /* Constructors/Destructor */
    //!exc: None.
    SH();
    //!exc: std::bad_alloc - Memory allocation failed.
    SH( const SH& sh );
    SH( const std::string& name, INT_2U classId, const std::string& comment );
    //!exc: None.   
    virtual ~SH();

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const SH& sh - Object to be assigned.
    //
    //!return: const SH& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const SH& operator=( const SH& sh );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two SH objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const SH& sh - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.
    //   
    bool operator==( const SH& sh ) const;
    
    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;

    /* Accessors */
    //!exc: None.   
    const std::string& getName() const;
    //!exc: None.   
    INT_2U getClass() const;
    //!exc: None.   
    const std::string& getComment() const;
    //!exc: None.   
    Container< SE >& getSE();
    //!exc: None.   
    const Container< SE >& getSE() const;

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison (all data except class ID number).
    //
    // Determines whether two SH objects are equal. 
    //
    //!param: const SH& sh - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.   
    bool equal( const SH& sh ) const;

    /* Mutators */
    //!exc: std::bad_alloc - Memory allocation failed.
    void appendSE( const SE& );
    
    /* Typedef's */
    typedef Container< SE >::iterator se_iterator;
    typedef Container< SE >::const_iterator const_se_iterator;


private: friend class Input;


    //-----------------------------------------------------------------------------
    //
    //: Read the SH from an Input object.
    //
    //!param: Input& in - Input data stream object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static SH* read( Input& in );

private:

    //-----------------------------------------------------------------------------
    //
    //: Input constructor.
    //
    //!param: Input& in - Input data stream object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Data read failure.
    //
    SH( Input& in );


    //-----------------------------------------------------------------------------
    //
    //: Writes the SH to an Output object.
    //
    //!param: Output& out - Output data stream object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;

    //: Name of structure beeing described by this object.
    std::string mName;
    //: Class number of structure beeing described.
    INT_2U mClass;
    //: Comment.
    std::string mComment;
    //: Structure elements for the described structure.
    Container< SE > mSE;
};


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
//!exc: None.   
//   
inline SH::SH()
        : mName( "none" ), mClass( 0 ), mComment( "" ), mSE()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
//!param: const SH& sh - Object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
inline SH::SH( const SH& sh ) 
        : mName( sh.mName ), mClass( sh.mClass ), mComment( sh.mComment ),
          mSE( sh.mSE )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - Structure name.
//!param: INT_2U classId - Structure class ID.
//!param: const std::string& comment - Comment.
//
inline
SH::SH( const std::string& name, INT_2U classId, const std::string& comment )
        : mName( name ), mClass( classId ), mComment( comment )
{
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Get the class ID of the class which this SH describes.
//
//!param: INT_2U - Described structure class ID.
//
//!exc: None.   
//      
inline INT_2U SH::getClass() const
{
    return mClass;
}


//-----------------------------------------------------------------------------
//
//: Get the comment.
//
//!return: const std::string& -- Comment.
//
//!exc: None.   
//      
inline const std::string& SH::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Get the SE container.
//
//!return: Container< SE >& -- Structure element container.
//
//!exc: None.   
//      
inline Container< SE >& SH::getSE()
{
    return mSE;
}


//-----------------------------------------------------------------------------
//
//: Get the SE container.
//
//!return: const Container< SE >& -- Structure element container.
//
//!exc: None.   
//      
inline const Container< SE >& SH::getSE() const
{
    return mSE;
}



//-----------------------------------------------------------------------------
// Mutators
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Add a SE to this header.
//
//!param: const SE& se - Structure element.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
inline void SH::appendSE( const SE& se )
{
    mSE.append( se );
}

 
} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppSHHH
