#include "framecpp/config.h"

#include <stdexcept>

#include "simdata.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::SimData;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;

using std::runtime_error;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a SimData object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> sampleRate = 0.0 </li>
//     <li> fShift = 0.0 </li>
//     <li> phase = 0.0 </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
SimData::SimData()
        : Base(), 
          mName( "none" ), mComment( "" ),  mSampleRate( 0.0 ),
          mFShift( 0.0 ), mPhase ( 0. ),
          mData(), mInput(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of all owned data.
//
//!param: const SimData& simData - the object to copy from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SimData::SimData( const SimData& simData )
        : Base( simData ),
          mName( simData.mName ),
          mComment( simData.mComment ), mSampleRate( simData.mSampleRate ),
	  mFShift( simData.mFShift ), mPhase( simData.mPhase ),
          mData( simData.mData ), mInput( simData.mInput ),
	  mTable( simData.mTable )
{
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - The name for the simulated data.
//!param: const std::string& comment - A comment.
//!param: REAL_4 sampleRate - The sample rate (samples/sec).
//!param: const REAL_8& fShift - Frequency shift.
//!param: const REAL_4& phase - Phase.
//
//!exc: None.
//      
SimData::SimData( const std::string& name, const std::string& comment,
                  REAL_4 sampleRate, const REAL_8 fShift, const REAL_4 phase )
        : Base(), mName( name ), mComment( comment ),
          mSampleRate( sampleRate ),
	  mFShift( fShift ), mPhase( phase ),
	  mData(), mInput(), mTable()
{
}




//-----------------------------------------------------------------------------
//
//: Simulated data name.
//
// Get the name of the simulated data.
//
//!return: const std::string& -- SimData name.
//
const std::string& SimData::getName() const
{
    return mName;
}


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
// This performs a deep copy of all owned data.
//
//!param: const SimData& simData - The object to assign from.
//
//!return: const SimData& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const SimData& SimData::operator=( const SimData& simData )
{
    if ( this != &simData )
    {
        mName = simData.mName;
        mComment = simData.mComment;
        mSampleRate = simData.mSampleRate;
        mFShift = simData.mFShift;
        mPhase = simData.mPhase;
        mData = simData.mData;
        mInput = simData.mInput;
        mTable = simData.mTable;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
// This performs a deep copy of all owned data.
//
//!param: const SimData& simData - The object to be added.
//
//!return: const SimData& - This object.
//
const SimData& SimData::operator+=( const SimData& simData )
{
    throw runtime_error( "Unimplemented SimData::operator+=() is called." );

    return *this;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs the Structure Header for this class.
//
//!return: SH - SimData structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH SimData::getSH() 
{
  SH sh( "FrSimData", ID_SIMDATA, "Simulated Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_4", "" ) );
  sh.appendSE( SE( "fShift", "REAL_8", "" ) );
  sh.appendSE( SE( "phase", "REAL_4", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "input", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrSimData *)", "" ) );
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Get frame data format version 4 Structure Header
//
// Constructs the Structure Header for this class.
//
//!return: SH - SimData structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH SimData::getSH4() 
{
  SH sh( "FrSimData", ID_SIMDATA, "Simulated Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "sampleRate", "REAL_4", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "input", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrSimData *)", "" ) );
  return sh;
}

//-----------------------------------------------------------------------------
//
//: Get frame data format version 3 Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH - SimData structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH SimData::getSH3() 
{
    SH sh( "FrSimData", ID_SIMDATA, "Simulated Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "sampleRate", "REAL_4", "" ) );
    sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "input", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrSimData *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads a SimData object from an Input object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: SimData* - Newly created SimData object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
SimData* SimData::read( FrameReaderBase& in )
{
  SimData& sim = *new SimData();
  in >> sim.mName >> sim.mComment >> sim.mSampleRate
     >> sim.mFShift >> sim.mPhase
     >> sim.mData >> sim.mInput >> sim.mTable;
  
  in.readNextNodeReference();    
  return &sim;
}

//: Version 4 frame SumData read function. 
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
SimData* SimData::read4( FrameReaderBase& in )
{
  SimData& sim = *new SimData();
  in >> sim.mName >> sim.mComment >> sim.mSampleRate
     >> sim.mData >> sim.mInput >> sim.mTable;
  
  in.readNextNodeReference();    
  return &sim;
}

//: Version 3 frame SumData read function. 
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
SimData* SimData::read3( FrameReaderBase& in )
{
    SimData& sim = *new SimData();
    in >> sim.mName >> sim.mComment >> sim.mSampleRate
       >> sim.mData >> sim.mInput;

    in.readNextNodeReference();    
    return &sim;
}


//-----------------------------------------------------------------------------
//
// Writes a SimData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Data write failed.
//
void SimData::write( Output& out ) const
{
    out.startRecord( ID_SIMDATA,
		     46
                     + mName.size() + mComment.size(), this );

    out << mName << mComment << mSampleRate
	<< mFShift << mPhase
        << mData << mInput << mTable;

    out.endRecord( this );
}



//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two SimData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const SimData& simData - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
bool SimData::operator==( const SimData& simData ) const 
{
    return ( mName == simData.mName &&
	     mComment == simData.mComment &&
             mSampleRate == simData.mSampleRate &&
             mFShift == simData.mFShift &&
             mPhase == simData.mPhase &&
	     mData == simData.mData &&
             mInput == simData.mInput &&
	     mTable == simData.mTable );
}

   

//-----------------------------------------------------------------------------
//
//: Class ID.
//
// Gets the class ID.
//
//!return: INT_2U -- SimData class ID.
//
//!exc: None.
//      
INT_2U SimData::getClassId() const
{
    return ID_SIMDATA;
}

   
