#ifndef FrameCppSimDataHH
#define FrameCppSimDataHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"

#include "framecpp/SearchContainer.hh"


namespace FrameCPP {

namespace Version_6
{
  class FrSimData;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: Simulated data.
//
// This class represents simulated ADC data.
// It is modeled after the FrSimData structure in the Frame Specification,
// LIGO-T970130-B-E.
//
// <P>SimData possess the following attributes:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Attribute</B></td><td><B>Description</B> </td></tr>
//   <tr><td> name      </td><td>Name of simulated data </td></tr>
//   <tr><td> comment   </td><td>Comment </td></tr>
//   <tr><td> sampleRate</td><td>Data acquisistion rate, samples / sec </td></tr>
//   <tr><td> fShift       </td><td>Frequency shift if signal has been heterodyned before </td></tr>
//   <tr><td> phase        </td></td> Phase of heterodyning signal at start of data set   </td></tr>
// </table>
//                
// <p>The SimData object is a container for two types of data:
//
// <table border=2 cellpadding=4>
//   <tr><td> <B>Data</B></td><td><B>Description</B>                             </td></tr>
//   <tr><td> data </td><td>A container of Vect objects which contain the sampled
//                  data. </td></tr>
//   <tr><td> input</td><td>A container of Vect objects holding seed MC input data
//                  for the simulation. </td></tr>
//   <tr><td> table</td><td>A container of Table objects with user data. </td></tr>
// </table>
//
// <p>The data is available through the container objects inside SimData.
// These are accessed via the <code>refData</code> and <code>refAux</code>
// methods.
//
class SimData : public Base
{
public:

    //!ignore_begin:
    DECLARE_DEFAULT_CONTAINER( Data, data, Vect, getName );
    DECLARE_CONTAINER( Input, input, Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    // !ignore_end:

    /* Constructors/Destructor */
    //-----------------------------------------------------------------------------
    //: Default Constructor.
    //
    // This creates a SimData object with the following properties:
    // <ul type=disc>
    //     <li> name = "none" </li>
    //     <li> comment = "" </li>
    //     <li> sampleRate = 0.0 </li>
    //     <li> fShift = 0.0 </li>
    //     <li> phase = 0.0 </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.   
    //      
    SimData();


    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.
    //
    // This performs a deep copy of all owned data.
    //
    //!param: const SimData& simData - the object to copy from.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    SimData( const SimData& simData );
    //-----------------------------------------------------------------------------
    //
    //: Constructor.
    //
    //!param: const std::string& name - The name for the simulated data.
    //!param: const std::string& comment - A comment.
    //!param: REAL_4 sampleRate - The sample rate (samples/sec).
    //!param: const REAL_8& fShift - Frequency shift.
    //!param: const REAL_4& phase - Phase.
    //
    //!exc: None.   
    //         
    SimData( const std::string& name, const std::string& comment, REAL_4 sampleRate, 
            REAL_8 fshift, REAL_4 phase );

    /* Operator Overloads */

    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    // This performs a deep copy of all owned data.
    //
    //!param: const SimData& simData - The object to assign from.
    //
    //!return: const SimData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const SimData& operator=( const SimData& simData );

    //-----------------------------------------------------------------------------
    //
    //: += Operator.
    //
    //!param: const SimData& simData - The object to be added.
    //
    //!return: const SimData& -- This object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const SimData& operator+=( const SimData& simData );


    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two SimData objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const SimData& simData - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.   
    //         
    bool operator==( const SimData& simData ) const;
    //!exc: None.      
    bool operator!=( const SimData& simData ) const;

    /* Required Methods */
    //!exc: None.   
    virtual INT_2U getClassId() const;

    //-----------------------------------------------------------------------------
    //
    //: Structure Header
    //
    // Constructs the Structure Header for this class.
    //
    //!return: SH -- SimData structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    static SH getSH();

    //: Get frame data format version 4 Structure Header
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    static SH getSH4();

    //: Get frame data format version 3 Structure Header
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    static SH getSH3();

    /* Accessors */

    /* gcc 2.95.1 can't handle a typedef with this if the throw spec is
       present (e.g.
       typedef SearchContainer< SimData, &SimData::getName > dc; ).
       
       const std::string& getName() const throw(); */
    const std::string& getName() const;
    //!exc: None.      
    const std::string& getComment() const;
    //!exc: None.      
    REAL_8 getSampleRate() const;
    //!exc: None.      
    REAL_8 getFShift() const;
    //!exc: None.      
    REAL_4 getPhase() const;

    /* Data Accessors */
    //!exc: None.      
    SimData::DataContainer& refData();
    //!exc: None.      
    const SimData::DataContainer& refData() const;

    /* Input Accessors */
    //!exc: None.      
    SimData::InputContainer& refInput();
    //!exc: None.      
    const SimData::InputContainer& refInput() const;

    /* Table Accessors */
    //!exc: None.      
    SimData::TableContainer& refTable();
    //!exc: None.      
    const SimData::TableContainer& refTable() const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrSimData;

    //-----------------------------------------------------------------------------
    //
    //: Reads a SimData object from an Input object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: SimData* -- Newly created SimData object.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static SimData* read( FrameReaderBase& in );

    //: Version 3 frame SumData read function. 
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static SimData* read3( FrameReaderBase& in );

    //: Version 4 frame SumData read function. 
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static SimData* read4( FrameReaderBase& in );

    //-----------------------------------------------------------------------------
    //
    // Writes a SimData object to an Output object.
    //
    //!param: Output& out out - Output data stream object.
    //  
    //!exc: write_failure - Data write failed.
    // 
    virtual void write( Output& out ) const;

private:
    //: Name of simulated data.
    std::string mName;
    //: Comment.
    std::string mComment;
    //: Simulated data sample rate, samples / s.
    REAL_4 mSampleRate;
    //: Frequency shift.
    REAL_8 mFShift;
    //: Phase of heterodyning signal at start of dataset
    REAL_4 mPhase;
    //: Simulated data.
    DataContainer mData;
    //: Input parameters for simulation.
    InputContainer mInput;
    //: Table data structure.
    TableContainer mTable;
};    
    


/******************/
/* Inline Methods */
/******************/


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const SimData& simData - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.
//   
inline bool SimData::operator!=( const SimData& simData ) const
{
    return (*this) != simData;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//
//: Comment
//
// Get the comment.
//
//!return: const std::string& -- Comment.
//
//!exc: None.
//      
inline const std::string& SimData::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Sample Rate.
//
// Get the simulated data sample rate.
//
// The rate is in samples / sec.
//
//!return: REAL_8 -- Sample rate.
//
//!exc: None.
//      
inline REAL_8 SimData::getSampleRate() const 
{
    return mSampleRate;
}


//-----------------------------------------------------------------------------
//
//: Get the frequency shift if the signal has been heterodyned before the ADC.
//
// fShift = (f_heterodyne - fNyquist@sampleRate).
//
//!return: REAL_8 -- Frequency shift.
//
//!exc: None.
//      
inline REAL_8 SimData::getFShift() const
{
    return mFShift;
}


//-----------------------------------------------------------------------------
//
//: Get phase of heterodyning signal at start of dataset.
//
//!return: REAL_4 -- phase.
//
//!exc: None.
//      
inline REAL_4 SimData::getPhase() const
{
    return mPhase;
}


//-----------------------------------------------------------------------------
//
//: Input Container.
//
// Get the container for seed MC input data for the simulation.
//
//!return: Container< Vect >& -- Data Container.
//
//!exc: None.
//      
inline SimData::InputContainer& SimData::refInput()
{
    return mInput;
}


//-----------------------------------------------------------------------------
//
//: const Input Container.
//
// Get the container for seed MC input data for the simulation.
//
//!return: const Container< Vect >& -- Data container.
//
//!exc: None.
//      
inline
const SimData::InputContainer& SimData::refInput() const
{
    return mInput;
}


//-----------------------------------------------------------------------------
//
//: Table Container.
//
// Get table container.
//
//!return: Container< Table >& -- Table container.
//
//!exc: None.
//      
inline SimData::TableContainer& SimData::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: const Table Container.
//
// Get table container.
//
//!return: const Container< Vect >& -- Table container.
//
//!exc: None.
//      
inline
const SimData::TableContainer& SimData::refTable() const
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Data Container.
//
// Get the container which contains the data.
//
//!return: SearchContainer< Vect, &Vect::getName >& -- Data container.
//
//!kw: inline
//
//!exc: None.
//      
inline SimData::DataContainer& SimData::refData()
{
    return mData;
}


//-----------------------------------------------------------------------------
//
//: const Data Container.
//
// Get the container which contains the data.
//
//!return: const SearchContainer< Vect, &Vect::getName >& -- Data container.
//
//!kw: inline const
//
//!exc: None.
//      
inline const
SimData::DataContainer& SimData::refData() const
{
    return mData;
}


} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppSimDataHH
