#include "framecpp/config.h"

#include <stdexcept>

#include "simevent.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::SimEvent;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;

using std::runtime_error;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a SimEvent object with the following properties:
// <ul type=disc>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> inputs = "" </li>
//     <li> GTime = (0,0) </li>
//     <li> timeBefore = 0.0 </li>
//     <li> timeAfter = 0.0 </li>
//     <li> amplitude = 0.0 </li>
//     <li> No vector or table data. </li>
// </ul>
//
//!exc: None.   
//      
SimEvent::SimEvent()
  : Base(), 
    mName( "none" ), mComment( "" ), mInputs( "" ),
    mGTime(),
    mTimeBefore( 0.0 ), mTimeAfter( 0.0 ),
    mAmplitude( 0.0 ),
    mData(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of all owned data.
//
//!param: const SimEvent& SimEvent - the object to copy from.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
SimEvent::SimEvent( const SimEvent& sim )
  : Base( sim ),
    mName( sim.mName ),
    mComment( sim.mComment ), 
    mInputs( sim.mInputs ),
    mGTime( sim.mGTime ),
    mTimeBefore( sim.mTimeBefore ), mTimeAfter( sim.mTimeAfter ),
    mAmplitude( sim.mAmplitude ),
    mData( sim.mData ), mTable( sim.mTable )
{
}


//: Constructor.
//!exc: None.   
//      
SimEvent::SimEvent( const std::string& name, const std::string& inputs, 
			const Time& mGTime,
			REAL_4 timeBefore, REAL_4 timeAfter,
			REAL_4 amplitude ) 
  : Base(), mName( name ), mInputs( inputs ),
    mGTime( mGTime ),
    mTimeBefore( timeBefore ), mTimeAfter( timeAfter ),
    mAmplitude( amplitude ),
    mData(), mTable()
{
}



//-----------------------------------------------------------------------------
//
//: Simulated event data name.
//
// Get the name of the simulated event data.
//
//!return: const std::string& -- Name.
//
const std::string& SimEvent::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
// This performs a deep copy of all owned data.
//
//!param: const SimEvent& SimEvent - The object to assign from.
//
//!return: const SimEvent& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const SimEvent& SimEvent::operator=( const SimEvent& sim )
{
  if ( this != &sim )
  {
    mName = sim.mName;
    mComment = sim.mComment;
    mInputs = sim.mInputs;
    mGTime = sim.mGTime;
    mTimeBefore = sim.mTimeBefore;
    mTimeAfter = sim.mTimeAfter;
    mAmplitude = sim.mAmplitude;
    mData = sim.mData;
    mTable = sim.mTable;
  }
  return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const SimEvent& SimEvent - The object to be added.
//
//!return: const SimEvent& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const SimEvent& SimEvent::operator+=( const SimEvent& sim )
{
   throw runtime_error( "Unimplemented SimEvent::operator+=() is called." );

   return *this;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
//!param: const SimEvent& sim - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
//      
bool SimEvent::operator==( const SimEvent& sim ) const
{
  return ( mName == sim.mName
	   && mComment == sim.mComment
	   && mInputs == sim.mInputs
	   && mGTime == sim.mGTime
	   && mTimeBefore == sim.mTimeBefore
	   && mTimeAfter == sim.mTimeAfter
	   && mAmplitude == sim.mAmplitude
	   && mData == sim.mData
	   && mTable == sim.mTable );
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Structure Header
//
// Constructs the Structure Header for this class.
//
//!return: SH  - SimEvent class ID.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH SimEvent::getSH()
{
  SH sh( "FrSimEvent", ID_SIMEVENT, "Simulated Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "inputs", "STRING", "" ) );
  sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
  sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
  sh.appendSE( SE( "timeBefore", "REAL_4", "" ) );
  sh.appendSE( SE( "timeAfter", "REAL_4", "" ) );
  sh.appendSE( SE( "amplitude", "REAL_4", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrSimEvent *)", "" ) );
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads a SimEvent object from an Input object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: SimEvent* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
SimEvent* SimEvent::read( FrameReaderBase& in ) 
{
  SimEvent& sim = *new SimEvent();
  in >> sim.mName >> sim.mComment >> sim.mInputs
     >> sim.mGTime
     >> sim.mTimeBefore >> sim.mTimeAfter >> sim.mAmplitude
     >> sim.mData >> sim.mTable;
  
  in.readNextNodeReference();    
  return &sim;
}


//-----------------------------------------------------------------------------
//
// Writes a SimEvent object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void SimEvent::write( Output& out ) const
{
    out.startRecord( ID_SIMEVENT,
		     8 * 1         // One header (eight bytes)
		     + 4 * 5       // two integers and three reals (4 bytes each)
		     + 3 * 3       // Three strings
		     + 4 * 3       // Three pointers
                     + mName.size()
		     + mComment.size()
		     + mInputs.size(),
		     this );

    out << mName << mComment << mInputs
	<< mGTime
	<< mTimeBefore << mTimeAfter << mAmplitude
        << mData << mTable;

    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Class ID
//
// Gets the class ID.
//
//!return: INT_2U -- SimEvent class ID.
//
//!exc: None.   
//      
INT_2U SimEvent::getClassId() const
{
    return ID_SIMEVENT;
}


   
