#include "framecpp/config.h"

#include <stdexcept>

#include "statdata.hh"
#include "detector.hh"
#include "framereader.hh"
#include "output.hh"
#include "table.hh"

using FrameCPP::Version_3_4_5::StatData;
using FrameCPP::Version_3_4_5::Vect;
using FrameCPP::Version_3_4_5::Container;
using FrameCPP::Version_3_4_5::Base;

using std::runtime_error;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates an StatData object with the following properties:
// <ul>
//     <li> name = "none" </li>
//     <li> comment = "" </li>
//     <li> representation = "" </li>
//     <li> timeStart = 0 </li>
//     <li> timeEnd = 0 </li>
//     <li> version = 0 </li>
//     <li> detector = none (null pointer) </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.
//      
StatData::StatData()
        : Base(), 
    mName( "none" ), mComment( "" ), mRepresentation( "" ), mTimeStart( 0 ),
    mTimeEnd( 0 ), mVersion( 0 ), mDetector( 0 ), mData(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.
//
// This performs a deep copy of the object.  However,
// since the StatData object does not own the Detector, a deep copy is NOT
// performed on this object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//      
StatData::StatData( const StatData& statData )
        : Base( statData ),  mName( statData.mName ),
          mComment( statData.mComment ), mRepresentation( statData.mRepresentation ),
	  mTimeStart( statData.mTimeStart ),
          mTimeEnd( statData.mTimeEnd ), mVersion( statData.mVersion ),
          mDetector( statData.mDetector ), mData( statData.mData ),
	  mTable( statData.mTable )
{
}


//: Constructor
//
//!exc: std::bad_alloc - Memory allocation failed.
//         
StatData::StatData( const std::string& name, const std::string& comment, const std::string& representation,
                    INT_4U timeStart, INT_4U timeEnd, INT_4U version )
        : Base(), 
          mName( name ), mComment( comment ), mRepresentation( representation ),
	  mTimeStart( timeStart ),
    mTimeEnd( timeEnd ), mVersion( version ), mDetector( 0 ), mData(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.
//         
StatData::~StatData()
{
}


//-----------------------------------------------------------------------------
//
//: Gets the static data name.
//
//!return: const std::string&
//
const std::string& StatData::getName() const
{
    return mName;
}


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const StatData& statData - Object to be assigned.
//
//!return: const StatData& - This object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const StatData& StatData::operator=( const StatData& statData )
{
    if ( this != &statData )
    {
        mName = statData.mName;
        mComment = statData.mComment;
        mRepresentation = statData.mRepresentation;
        mTimeStart = statData.mTimeStart;
        mTimeEnd = statData.mTimeEnd;
        mVersion = statData.mVersion;
        mDetector = statData.mDetector;
        mData = statData.mData;
        mTable = statData.mTable;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const StatData& statData - Object to be added.
//
//!return: const StatData& - This object.
//
const StatData& StatData::operator+=( const StatData& statData )
{
    throw runtime_error( "Unimplemented StatData::operator+=() is called." );

    return *this;
}



//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.  
//
//!return: SH - Static Data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH StatData::getSH() 
{
  SH sh( "FrStatData", ID_STATDATA, "Static Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "representation", "STRING", "" ) );
  sh.appendSE( SE( "timeStart", "INT_4U", "" ) );
  sh.appendSE( SE( "timeEnd", "INT_4U", "" ) );
  sh.appendSE( SE( "version", "INT_4U", "" ) );
  sh.appendSE( SE( "detector", "PTR_STRUCT(FrDetector *)", "" ) );
  sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Get frame data format version 3 Structure Header
//
// Constructs the Structure Header for this class.  
//
//!return: SH - Static Data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH StatData::getSH3()
{
    SH sh( "FrStatData", ID_STATDATA, "Static Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "timeStart", "INT_4U", "" ) );
    sh.appendSE( SE( "timeEnd", "INT_4U", "" ) );
    sh.appendSE( SE( "version", "INT_4U", "" ) );
    sh.appendSE( SE( "detector", "PTR_STRUCT(FrDetector* )", "" ) );
    sh.appendSE( SE( "data", "PTR_STRUCT(FrVect *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads a StatData object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: StatData* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
StatData* StatData::read( FrameReaderBase& in )
{
  StatData&	stat = *new StatData();
  void*		detector = &(stat.mDetector);

  in >> stat.mName >> stat.mComment >> stat.mRepresentation
     >> stat.mTimeStart >> stat.mTimeEnd >> stat.mVersion
     >> (Base**)detector >> stat.mData >> stat.mTable;
  return &stat;
}

//: Read framecpp version 3 static data object.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
StatData* StatData::read3( FrameReaderBase& in )
{

    StatData&	stat = *new StatData();
    void*	detector = &(stat.mDetector);

    in >> stat.mName >> stat.mComment >> stat.mTimeStart >> stat.mTimeEnd >> stat.mVersion
       >> (Base**)detector >> stat.mData;
    return &stat;
}


//-----------------------------------------------------------------------------
//
//: Writes an StatData object to an Output object.
//
//!param: Output& out out - Output data stream object.
//
//!exc: write_failure - Write failure occured.
//
void StatData::write( Output& out ) const 
{
    out.startRecord( ID_STATDATA,
                     41 + mRepresentation.size() + mName.size() + mComment.size(), this );
    out << mName << mComment << mRepresentation
	<< mTimeStart << mTimeEnd << mVersion
        << mDetector << mData << mTable;

    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two StatData objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const StatData& statData - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.
//   
bool StatData::operator==( const StatData& statData ) const
{
    return ( mName == statData.mName &&
             mComment == statData.mComment &&
             mRepresentation == statData.mRepresentation &&
             mTimeStart == statData.mTimeStart &&
             mTimeEnd == statData.mTimeEnd &&
             mVersion == statData.mVersion &&
             mData == statData.mData &&
             mTable == statData.mTable );
}

   

//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- Static Data class ID.
//
//!exc: None.
//   
INT_2U StatData::getClassId() const
{
    return ID_STATDATA;
}

   
