#include "framecpp/config.h"

#include <stdexcept>

#include "summary.hh"
#include "framereader.hh"
#include "output.hh"


using FrameCPP::Version_3_4_5::Summary;

using std::runtime_error;



//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a Summary object with the following properties:
// <ul>
//     <li> none = "none" </li>
//     <li> comment = "" </li>
//     <li> test = "" </li>
//     <li> gtimes = 0 seconds and 0 nanoseconds </li>
//     <li> No other data. </li>
// </ul>
//
//!exc: None.      
Summary::Summary()
        : Base(), mName( "none" ), mComment( "" ), mTest( "" ),
	  mGTime( 0, 0 ),
	  mMoments(), mTable()
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.  
//
//!param: const Summary& summary - Summary data object to be copied.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Summary::Summary( const Summary& summary )
        : Base( summary ), mName( summary.mName ),
          mComment( summary.mComment ),
	  mTest( summary.mTest ),
	  mGTime( summary.mGTime ),
          mMoments( summary.mMoments ),
	  mTable( summary.mTable )
{
}

//: Constructor
//
//!exc: std::bad_alloc - Memory allocation failed.
//   
Summary::Summary(
    const std::string& name, const std::string& comment, const std::string& test, const Time& gtime )
        : Base(), mName( name ), mComment( comment ), mTest( test ),
	  mGTime( gtime ),
	  mMoments(), mTable()
{
}

   
//-----------------------------------------------------------------------------
//
//: Destructor.
//
// This method currently does nothing.
//
//!exc: None.   
Summary::~Summary()
{
}



//-----------------------------------------------------------------------------
//
//: Gets the summary statistic name.
//
//!return: const std::string& -- Summary data name.
//
const std::string& Summary::getName() const
{
    return mName;
}



//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Summary& summary -
//
//!return: const Summary&
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Summary& Summary::operator=( const Summary& summary )
{
    if ( this != & summary )
    {
        mName = summary.mName;
        mComment = summary.mComment;
        mTest = summary.mTest;
        mGTime = summary.mGTime;
        mMoments = summary.mMoments;
        mTable = summary.mTable;
    }

    return *this;
}


//-----------------------------------------------------------------------------
//
//: += Operator.
//
//!param: const Summary& summary - Object to be added.
//
//!return: const Summary& - This object.
//
const Summary& Summary::operator+=( const Summary& summary )
{
    throw runtime_error( "Unimplemented Summary::operator+=() is called." );
 
    return *this;
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.  
//
//!return: SH - Summary data structure header.
//
//!exc: std::bad_alloc - Memory allocation failed
//
FrameCPP::Version_3_4_5::SH Summary::getSH()
{
  SH sh( "FrSummary", ID_SUMMARY, "Summary Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "test", "STRING", "" ) );
  sh.appendSE( SE( "GTimeS", "INT_4U", "" ) );
  sh.appendSE( SE( "GTimeN", "INT_4U", "" ) );
  sh.appendSE( SE( "moments", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrSummary *)", "" ) );
  return sh;
}

//: Get frame data format version 4 Structure Header
//
//!exc: std::bad_alloc - Memory allocation failed
//   
FrameCPP::Version_3_4_5::SH Summary::getSH4()
{
  SH sh( "FrSummary", ID_SUMMARY, "Summary Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "comment", "STRING", "" ) );
  sh.appendSE( SE( "test", "STRING", "" ) );
  sh.appendSE( SE( "moments", "PTR_STRUCT(FrVect *)", "" ) );
  sh.appendSE( SE( "table", "PTR_STRUCT(FrTable *)", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrSummary *)", "" ) );
  return sh;
}

//: Get frame data format version 3 Structure Header
//
//!exc: std::bad_alloc - Memory allocation failed
//   
FrameCPP::Version_3_4_5::SH Summary::getSH3()
{
    SH sh( "FrSummary", ID_SUMMARY, "Summary Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "comment", "STRING", "" ) );
    sh.appendSE( SE( "test", "STRING", "" ) );
    sh.appendSE( SE( "moments", "PTR_STRUCT(FrVect *)", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrSummary *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads a Summary object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Input data stream object.
//
//!return: Summary* - Newly created.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//
Summary* Summary::read( FrameReaderBase& in ) 
{
  Summary& sum = *new Summary();
  in >> sum.mName >> sum.mComment >> sum.mTest
     >> sum.mGTime
     >> sum.mMoments >> sum.mTable;

  in.readNextNodeReference();
  return &sum;
}

//: Frame version 4 read method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
Summary* Summary::read4( FrameReaderBase& in )
{
  Summary& sum = *new Summary();
  in >> sum.mName >> sum.mComment >> sum.mTest >> sum.mMoments >> sum.mTable;

  in.readNextNodeReference();
  return &sum;
}

//: Frame version 3 read method.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Read failure occured.
//   
Summary* Summary::read3( FrameReaderBase& in )
{
    Summary& sum = *new Summary();
    in >> sum.mName >> sum.mComment >> sum.mTest >> sum.mMoments;

    in.readNextNodeReference();
    return &sum;
}


//-----------------------------------------------------------------------------
//
//: Writes a Summary object to an Output object.
//
//!param: Output& out out - Data output object.
//
//!exc: write_failure - Write failure occured.
//
void Summary::write( Output& out ) const
{
    out.startRecord( ID_SUMMARY,
		     37
                     + mName.size() + mComment.size()
                     + mTest.size(), this );
    out << mName << mComment << mTest
	<< mGTime
	<< mMoments << mTable;
    out.endRecord( this );
}


//-----------------------------------------------------------------------------
//
//: Equal comparison.
//
// Determines whether two Summary objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const Summary& summary - The object to compare with.
//
//!return: bool - true if the objects are equal.
//
//!kw: const
//
//!exc: None.   
bool Summary::operator==( const Summary& summary ) const
{
    if ( this == &summary )
    {
        return true;
    }

    return ( mName == summary.mName &&
             mComment == summary.mComment &&
             mTest == summary.mTest &&
             mGTime == summary.mGTime &&
             mMoments == summary.mMoments &&
             mTable == summary.mTable );
}

   

//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U Summary data class ID.
//
//!exc: None.      
INT_2U Summary::getClassId() const
{
    return ID_SUMMARY;
}

   
