#ifndef FrameCppSummaryHH
#define FrameCppSummaryHH


#include "framecpp/config.h"
#include "framecpp/Version3_4_5/base.hh"
#include "framecpp/Version3_4_5/sh.hh"
#include "framecpp/Version3_4_5/vect.hh"
#include "framecpp/Version3_4_5/table.hh"
#include "framecpp/Version3_4_5/time.hh"

#include "framecpp/SearchContainer.hh"

namespace FrameCPP {

namespace Version_6
{
  class FrSummary;
}

namespace Version_3_4_5 {

class FrameReaderBase;


//-----------------------------------------------------------------------------
//
//: This class represents summary data.
//
// It is modeled after the FrSummary structure in the Frame Specification,
// LIGO-T970130-B-E.
//
class Summary : public Base
{
public:

    //!ignore_begin:
    DECLARE_DEFAULT_CONTAINER( Moments, moments, Vect, getName );
    DECLARE_CONTAINER( Table, table, Table, getName );
    //!ignore_end:

    /* Constructors */
    //-----------------------------------------------------------------------------
    //
    //: Default Constructor.
    //
    // This creates a Summary object with the following properties:
    // <ul>
    //     <li> none = "none" </li>
    //     <li> comment = "" </li>
    //     <li> test = "" </li>
    //     <li> gtimes = 0 seconds and 0 nanoseconds </li>
    //     <li> No other data. </li>
    // </ul>
    //
    //!exc: None.   
    Summary();


    //-----------------------------------------------------------------------------
    //
    //: Copy Constructor.  
    //
    //!param: const Summary& summary - Summary data object to be copied.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    Summary( const Summary& summary );

    // Constructor
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //   
    Summary( const std::string& name, const std::string& comment, const std::string& test,
	     const Time& gtime );
    
    /* Destructor */
    //!exc: None.      
    virtual ~Summary();

    /* Operator Overloads */
    //-----------------------------------------------------------------------------
    //
    //: Assignment Operator.
    //
    //!param: const Summary& summary -
    //
    //!return: const Summary&
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //
    const Summary& operator=( const Summary& summary );

    //-----------------------------------------------------------------------------
    //
    //: Assignment operator.
    //
    //!param: const Summary& summary - The object to be added.
    //
    //!return: const Summary& - This object.
    //
    const Summary& operator+=( const Summary& summary );

    //-----------------------------------------------------------------------------
    //
    //: Equal comparison.
    //
    // Determines whether two Summary objects are equal. For this comparison,
    // all of the data is compared.
    //
    //!param: const Summary& summary - The object to compare with.
    //
    //!return: bool -- true if the objects are equal.
    //
    //!kw: const
    //
    //!exc: None.         
    bool operator==( const Summary& summary ) const;
    //!exc: None.         
    bool operator!=( const Summary& summary ) const;
    
    /* Required Methods */
    //!exc: None.         
    virtual INT_2U getClassId() const;


    //-----------------------------------------------------------------------------
    //
    //: Constructs the Structure Header for this class.  
    //
    //!return: SH -- Summary data structure header.
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //
    static SH getSH();

    //: Get frame data format version 4 Structure Header
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //   
    static SH getSH4();

    //: Get frame data format version 3 Structure Header
    //
    //!exc: std::bad_alloc - Memory allocation failed
    //   
    static SH getSH3();

    /* Accessors */
    const std::string& getName() const;
    //!exc: None.            
    const std::string& getComment() const;
    //!exc: None.            
    const std::string& getTest() const;
    //!exc: None.            
    const Time& getGTime() const;
    //!exc: None.            
    MomentsContainer& refMoments();
    //!exc: None.            
    const MomentsContainer& refMoments() const;
    //!exc: None.            
    TableContainer& refTable();
    //!exc: None.            
    const TableContainer& refTable() const;

private:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrSummary;

    /* This method will be accessed by the Dictionary class but is
       not available for public use. */

    //-----------------------------------------------------------------------------
    //
    //: Reads a Summary object from an FrameReaderBase object.
    //
    // This method creates a dynamically allocated object.
    //
    // <p>*NOTE* This method allocates memory and returns a pointer to it.  The
    // user is responsible for freeing this memory.
    //
    //!param: FrameReaderBase& in - Input data stream object.
    //
    //!return: Summary* -- Newly created.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //
    static Summary* read( FrameReaderBase& in );

    //: Frame version 4 read method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static Summary* read4( FrameReaderBase& in );

    //: Frame version 3 read method.
    //
    //!exc: std::bad_alloc - Memory allocation failed.
    //!exc: read_failure - Read failure occured.
    //   
    static Summary* read3( FrameReaderBase& in );


    //-----------------------------------------------------------------------------
    //
    //: Writes a Summary object to an Output object.
    //
    //!param: Output& out out - Data output object.
    //
    //!exc: write_failure - Write failure occured.
    //
    virtual void write( Output& out ) const;
    
private:

    //: Name of summary statistics.
    std::string mName;
    //: Comment.
    std::string mComment;
    //: Statistical test(s) used on raw data.
    std::string mTest;
    //: GPS time
    Time mGTime;
    //: Statistical descriptors.
    MomentsContainer mMoments;
    //: Additional summary information.
    TableContainer mTable;
};



/******************/
/* Inline Methods */
/******************/

//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Not-equal comparison.
//
//!param: const Summary& summary - The object to compare with.
//
//!return: bool -- True if the objects are not equal.
//
//!kw: inline const
//
//!exc: None.      
inline bool Summary::operator!=( const Summary& summary ) const
{
    return !( *this == summary );
}



//-----------------------------------------------------------------------------
// Accessors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Gets the comment.
//
//!return: const std::string& -- Summary data comment.
//
//!exc: None.      
inline const std::string& Summary::getComment() const
{
    return mComment;
}


//-----------------------------------------------------------------------------
//
//: Gets the statistical test(s) used on raw data.
//
//!return: const std::string& -- Tests.
//
//!exc: None.      
inline const std::string& Summary::getTest() const
{
    return mTest;
}


//-----------------------------------------------------------------------------
//
//: Get message GPS time.
//
//!return: const Time& -- GPS time.
//
//!exc: None.      
inline const Time& Summary::getGTime() const
{
    return mGTime;
}


//-----------------------------------------------------------------------------
//
//: Gets the container holding the statistical descriptors.
//
//!return: MomentsContainer& -- Moments.
//
//!exc: None.      
inline Summary::MomentsContainer&
Summary::refMoments() 
{
    return mMoments;
}


//-----------------------------------------------------------------------------
//
//: Gets the container holding the statistical descriptors.
//
//!return: const MomentsContainer& -- Moments.
//
//!exc: None.      
inline const
Summary::MomentsContainer&
Summary::refMoments() const
{
    return mMoments;
}


//-----------------------------------------------------------------------------
//
//: Gets the container holding table data.
//
//!return: TableContainer< Table >& -- Table.
//
//!exc: None.      
inline Summary::TableContainer&
Summary::refTable()
{
  return mTable;
}


//-----------------------------------------------------------------------------
//
//: Gets the container holding table data.
//
//!return: const TableContainer& -- Table.
//
//!exc: None.      
inline const
Summary::TableContainer&
Summary::refTable() const
{
  return mTable;
}


} // namespace - Version_3_4_5

} // namespace - FrameCPP

#endif // FrameCppSummaryHH
