#ifndef TOCHH
#define TOCHH

#include <stdexcept>
#include <functional>
#include <algorithm>
#include <vector>
#include <map>

#include "general/hash_map.hh"

#include "base.hh"
#include "sh.hh"
#include "input.hh"
#include "types.hh"
#include "time.hh"

namespace FrameCPP {

namespace Version_6
{
  class FrTOC;
}

namespace Version_3_4_5 {

//-----------------------------------------------------------------------------
//
//: This class represents frame file table of contents .
//

class TOC : public Base
{
public:

  /* Constructors */
  //-----------------------------------------------------------------------------
  //
  //: Default Constructor.
  //
  //!exc: None.
  TOC();


  //-----------------------------------------------------------------------------
  //
  //: Copy Constructor.
  //
  // This performs a deep copy of the object.
  //
  //!param: const TOC& toc - Object to be copied.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
  TOC( const TOC& eof );

  /* Destructor */
  virtual ~TOC();

  /* Operator Overloads */

  //-----------------------------------------------------------------------------
  //
  //: Assignment Operator.
  //
  //!param: const TOC& toc - Object to be assigned.
  //
  //!return: const TOC& -- This TOC object.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
  const TOC& operator=( const TOC& eof );

  /* Required Methods */
  //!exc: None.   
  virtual INT_2U getClassId() const;


  //-----------------------------------------------------------------------------
  //
  //: Constructs the Structure Header for this class.
  //
  //!return: SH -- TOC structure header.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //
  static SH getSH();

  //: Constructs frame data version 4 Structure Header for this class
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //   
  static SH getSH4();

  /* Accessors */

  //: Get the total number of frames in file
  inline unsigned long getFrameNumber() const { return GTime.size(); }
  //: Get leap seconds from the first frame in file
  inline INT_2S getULeapS() const { return ULeapS; }
  //: Get run numbers
  inline const std::vector<INT_4S>& getRuns(unsigned long idx) const { return runs; }
  //: Get frame numbers
  inline const std::vector<INT_4U>& getFrame() const { return frame; }
  //: Get frame time stamps
  inline const std::vector<Time>& getGTime() const { return GTime; }
  //: Get frame durations
  inline const std::vector<REAL_8>& getDt() const { return dt; }
  //: Get frame header positions
  inline const std::vector<INT_8U>& getPositionH() const { return position; }
  //: Get first ADC positions
  inline const std::vector<INT_8U>& getFirstADCPosition() const { return positionADC; }
  //: Get first serial data structure positions
  inline const std::vector<INT_8U>& getFirstSerPosition() const { return positionSer; }
  //: Get first raw data referenced table positions
  inline const std::vector<INT_8U>& getFirstTablePosition() const { return positionTable; }
  //: Get first message structure positions
  inline const std::vector<INT_8U>& getFirstMsgPosition() const { return positionMsg; }

protected:
  friend class Dictionary;
  friend class FrameCPP::Version_6::FrTOC;
  

  //-----------------------------------------------------------------------------
  //
  //: Skips TOC structure in an Input object.
  //
  // This method skips over the TOC. It is not read here.
  //
  //!param: Input& in - Input data stream object.
  //
  //!return: TOC* -- Always return zero.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //!exc: read_failure - Failed to skip over the TOC.
  //
  static TOC* read( Input& in );

  //: Skips TOC structure in an Input object.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //!exc: read_failure - Failed to skip over the TOC.
  //   
  static TOC* read4( Input& in ) { return read( in ); }


  //-----------------------------------------------------------------------------
  //
  //: Writes an TOC object to an Output object.
  //
  //!param: Output& out out - Outpu stream object.
  //
  //!exc: write_failure - Write failed.
  //
  virtual void write( Output& out ) const;
    
protected:


  //-----------------------------------------------------------------------------
  //
  //: Reads an TOC object from an Input object.
  //
  // This method is used to read the table of contents.
  //
  //!param: Input& in - Input stream object.
  //
  //!return: 0 - absent or corrupt TOC; this -- everything was fine.
  //
  //!exc: std::bad_alloc - Memory allocation failed.
  //!exc: read_failure - Read failed.
  //
  TOC* reallyRead( Input& in );

  /* Frame header */
  //: from the first FrameH in the file
  INT_2S ULeapS;
  //: from the first FrameH in the file
  std::vector<INT_4U> dataQuality;
  //: frame times
  std::vector<Time> GTime;
  //: frame durations
  std::vector<REAL_8> dt;
  //: frame run numbers
  std::vector<INT_4S> runs;
  //: frame numbers
  std::vector<INT_4U> frame;
  //: offset in a file to frame header structure's first byte
  std::vector<INT_8U> position;
  //: for the first ADC structure in each frame
  std::vector<INT_8U> positionADC;
  //: for the first serial data structure in each frame
  std::vector<INT_8U> positionSer;
  //: for the first refernced from RawData table in each frame
  std::vector<INT_8U> positionTable;
  //: for the first message in each frame
  std::vector<INT_8U> positionMsg;

  /* Structure Header */
  //: structure IDs
  std::vector<INT_2U> SHid;
  //: structure names
  std::vector<std::string> SHname;
 
public:
  //
  //: Detector
  //
  // Maps Detector name into its position
  //
  std::map< std::string, INT_8U > detectorNamePositionMap;
  typedef std::map<std::string, INT_8U>::const_iterator DNPI;
  typedef std::map<std::string, INT_8U>::iterator DNPINC;

  //
  //: Static data
  //
  // Maps a Stat Data structure name on the following construct
  // (detector_name, ([tStart, tEnd, version], position)*), where
  // vars in [] are represented as std::vector< INT_4U > with three elements.
  //
  typedef std::pair< std::string, std::vector< std::pair< std::vector< INT_4U >, INT_8U > > > SNP_value;
  typedef std::map< std::string, SNP_value > statNamePositionMap_type;
  statNamePositionMap_type statNamePositionMap;
  typedef statNamePositionMap_type::value_type SNP_value_type; 
  typedef std::map< std::string, std::pair< std::string, std::vector< std::pair< std::vector< INT_4U >, INT_8U > > > >::const_iterator SNPI;
  typedef std::map< std::string, std::pair< std::string, std::vector< std::pair< std::vector< INT_4U >, INT_8U > > > >::iterator SNPINC;

  //
  //: Adc Data
  //
  // Maps an ADC channel name on a std::pair of std::vectors.
  // First std::vector is of FrAdcData structure positions.
  // First std::vectors's length equals to the number of frames in
  // the frame file. If an ADC channel is not present in a frame,
  // corresponding position std::vector element is 0.
  // Second std::vector in a std::pair is a _std::pair_ (a std::vector with 2 elements here)
  // of IDs. First element is channel ID. Second element is channel group ID.
  //
  typedef HASH_MAP_NAMESPACE::hash_map< std::string, std::pair< std::vector<INT_8U>, std::vector<INT_4U> > > adcNamePositionMap_type;
  typedef adcNamePositionMap_type::const_iterator ANPI;
  typedef adcNamePositionMap_type::iterator ANPINC;
  adcNamePositionMap_type adcNamePositionMap;

  //
  //: Post-processed data
  //
  // Maps proc name onto the std::vector of positions.
  // One position per name per frame.
  //
  std::map< std::string, std::vector<INT_8U> > procNamePositionMap;
  typedef std::map<std::string, std::vector<INT_8U> >::const_iterator PNPI;
  typedef std::map<std::string, std::vector<INT_8U> >::iterator PNPINC;  


  //
  //: Simulated data
  //
  // Maps simulated data name on the std::vector of positions
  // One position per name per frame
  //
  std::map< std::string, std::vector<INT_8U> > simNamePositionMap;
  typedef std::map<std::string, std::vector<INT_8U> >::const_iterator SIMNPI;
  typedef std::map<std::string, std::vector<INT_8U> >::iterator SIMNPINC;    

  //
  //: Serial data
  //
  // Maps Serial name onto the std::vector of positions.
  // One position per name per frame.
  //
  std::map< std::string, std::vector<INT_8U> > serNamePositionMap;
  typedef std::map<std::string, std::vector<INT_8U> >::const_iterator SERNPI;
  typedef std::map<std::string, std::vector<INT_8U> >::iterator SERNPINC;    

  //
  //: Summary data
  //
  // Maps Serial name onto the std::vector of positions.
  // One position per name per frame.
  //
  std::map< std::string, std::vector<INT_8U> > sumNamePositionMap;
  typedef std::map<std::string, std::vector<INT_8U> >::const_iterator SUMNPI;
  typedef std::map<std::string, std::vector<INT_8U> >::iterator SUMNPINC;    

  //
  //: Event Data
  //
  class eventAttr {
  public:
    eventAttr(): gtime(0,0), amplitude(.0), position(0) {}
    eventAttr(Time t, REAL_4 a, INT_8U p): gtime(t), amplitude(a), position(p) {}
    Time gtime; // event time
    REAL_4 amplitude; // event ampliture
    INT_8U position; // event structure file position
  };

  std::map< std::string, std::vector< TOC::eventAttr > > eventNamePositionMap;
  typedef std::map< std::string, std::vector< TOC::eventAttr > >::const_iterator EVENTNPI;
  typedef std::map< std::string, std::vector< TOC::eventAttr > >::iterator EVENTPINC;

  //
  //: Simulated Event Data
  //
  std::map< std::string, std::vector< TOC::eventAttr > > simEventNamePositionMap;
  typedef std::map< std::string, std::vector< TOC::eventAttr > >::const_iterator SIMEVENTNPI;
  typedef std::map< std::string, std::vector< TOC::eventAttr > >::iterator SIMEVENTPINC;
};
    

} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // TOCHH
