#include "framecpp/config.h"

#include <sstream>
   
#include "tocreader.hh"

using namespace FrameCPP::Version_3_4_5;
using namespace std;   
   

// Constructor
//
// Frame file header is read in during Input() initialization.
// Stream `in' should be positioned at the beginning of the File Header structure.
// Stream must be positioned to the beginning of the frame file.
//
//!exc: read_failure - Read failed.
//!exc: bad_alloc - Memory allocation failed.   
//   
TOCReader::TOCReader( std::istream& in, bool buildToc )
  : TOC(), FrameReaderBase( in ), mHasToc(true)
{
  try {
    // Initialize the file header object if necessary
    if ( mFileHeader == 0 )
      {
	//      mStream->seekg( 0, std::ios::beg );
	mFileHeader = new FileHeader( *this );
	dictionary = mFileHeader->getDictionary();
      }

    // See if this file is too old to have TOC
    //
    if ( mFileHeader->getDataFormatVersion() < 4 )
      {
	throw read_failure( "Old data format version: no Table of Contents", NOTOC );
      }

    // Get the TOC offset from the end of file structure
    //
    mStream->seekg( -4, std::ios::end );
    INT_4U seekTOC = 0;
    *this >> seekTOC;

    // TOC is not present
    //
    if ( !seekTOC ) 
      {
	throw read_failure( "TOC not present", NOTOC );
      }
    
    // Position to the beginning of the TOC structure
    // and read structure header variables
    //
    mStream->seekg( - ((long) seekTOC), std::ios::end );


    INT_4U length;   // length of structure in bytes
    INT_2U classId;  // File Structure class
    INT_2U instance; // Instance count for this class 
    *this >> length >> classId >> instance;


    // Sanity checks
    //
    if ( instance != 0 ) // There must be only one TOC in a file!
      {
	throw read_failure( "Invalid TOC instance", BADTOC );
      }

    // Read the TOC variables
    TOC::reallyRead( *this );

    // Construct class ID translation map
    for ( unsigned int i = 0; i < SHid.size(); i++ )
      {
	if ( dictionary->find( SHname[ i ] ) == dictionary->end() )
	  {
	    throw read_failure( "Unknown class type.  The structure header "
				"for an unknown class appeared.", BADTOC );
	  }
	else
	  {
	    const SH& localSH( dictionary->find( SHname[ i ] )->second.getSH() );
            
	    mTransMap.insert( std::pair< INT_2U, INT_2U > ( SHid[ i ], localSH.getClass() ));
	  }
      }
  } catch ( read_failure ) {
    if ( !buildToc )
      throw;

    mHasToc = false;
    // :TODO: do not read actual data, just read structure headers and whatever is needed from the structures.
    // For ADCs read just the name and nothing else.
    // :FIXME: has to actually parse whole file here (i.e. read all the frames). This is required
    // to avoid problems if the frames are different. They could be different if they have different ADCs 
    // or some of the ADCs have different attributes (different data type or sampling rate).
    //

    delete FrameReaderBase::readFrame(); // Read first frame
    if ( !mEndOfFile ) {
      delete FrameReaderBase::readFrame(); // Read second frame
    }
  }
}

   

//: Destructor
TOCReader::~TOCReader() 
{}
   

// Selectively read a frame.
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//   
Frame* TOCReader::readFrame( unsigned long frameNumber )
{
  // Check the index falls into the range of available frames
  //
  if ( frameNumber >= position.size() )
    {
      throw not_found_error( "bad index" );
    }

  // Initialize the file header object if necessary
  if ( mFileHeader == 0 )
    {
      mStream->seekg( 0, std::ios::beg );
      mFileHeader = new FileHeader( *this );
      dictionary = mFileHeader->getDictionary();
    }

  // Position to the start of requested frame
  //
  mStream->seekg( position[ frameNumber ], std::ios::beg );

  Frame* retval = FrameReaderBase::readFrame();
  skip(0);
  return retval;
}


// Read just the frame header structure
// :TODO: needs debugging; corrupts dynamic memory
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
Frame* TOCReader::readFrameHeader( unsigned long frameNumber )
{
  // Check the index falls into the range of available frames
  //
  if ( frameNumber >= position.size() )
    {
      throw not_found_error( "bad index" );
    }

  // Initialize the file header object if necessary
  if ( mFileHeader == 0 )
    {
      mStream->seekg( 0, std::ios::beg );
      mFileHeader = new FileHeader( *this );
      dictionary = mFileHeader->getDictionary();
    }

  // Position to the start of requested frame
  //
  mStream->seekg( position[ frameNumber ], std::ios::beg );

  Frame* retval = dynamic_cast< Frame* >( readStructure(false) );
  return retval;
}


// Read single channel from the specified frame
//
// :TODO: break this switched function into several small
// functions, one for every supported data type
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
FrameCPP::Version_3_4_5::Base* TOCReader::read( unsigned long frameNumber, INT_2U structureClassId, std::string name )
{
  // Check the index falls into the range of available frames
  //
  if ( frameNumber >= position.size() )
    {
      throw not_found_error( "bad frame index" );
    }

  switch ( structureClassId ) {
  case ID_ADCDATA:
    {
      // See if there exists named ADC
      //
      ANPI iter = adcNamePositionMap.find( name );
      if ( iter == adcNamePositionMap.end() )
	{
	  throw not_found_error( "bad ADC name" );    
	}
      
      // Position to the start of the ADC
      //
      unsigned long pos = iter->second.first[ frameNumber ];
      if ( pos == 0 )
	{
	  throw not_found_error( "no ADC in the frame" );
	}
      mStream->seekg( pos , std::ios::beg );
    }
    break;
  case ID_PROCDATA:
    {
      PNPI iter = procNamePositionMap.find( name );
      if ( iter == procNamePositionMap.end() )
	{
	  throw not_found_error( "bad Proc name" );
	}
      unsigned long pos = iter->second[ frameNumber ];
      if ( pos == 0 )
	{
	  throw not_found_error( "no Proc in the frame" );
	}
      mStream->seekg( pos , std::ios::beg );
    }
    break;
  case ID_SIMDATA:
    {
      SIMNPI iter = simNamePositionMap.find( name );
      if ( iter == simNamePositionMap.end() )
	{
	  throw not_found_error( "bad Sim name" );
	}
      unsigned long pos = iter->second[ frameNumber ];
      if ( pos == 0 )
	{
	  throw not_found_error( "no Sim in the frame" );
	}
      mStream->seekg( pos , std::ios::beg );
    }
    break;
  case ID_SERDATA:
    {
      SERNPI iter = serNamePositionMap.find( name );
      if ( iter == serNamePositionMap.end() )
	{
	  throw not_found_error( "bad Ser name" );
	}
      unsigned long pos = iter->second[ frameNumber ];
      if ( pos == 0 )
	{
	  throw not_found_error( "no Ser in the frame" );
	}
      mStream->seekg( pos , std::ios::beg );
    }
    break;
  case ID_SUMMARY:
    {
      SUMNPI iter = sumNamePositionMap.find( name );
      if ( iter == sumNamePositionMap.end() )
	{
	  throw not_found_error( "bad Sum name" );
	}
      unsigned long pos = iter->second[ frameNumber ];
      if ( pos == 0 )
	{
	  throw not_found_error( "no Sum in the frame" );
	}
      mStream->seekg( pos , std::ios::beg );
    }
    break;
  default:
    throw not_found_error( "unsupported structure class" );
  }

  FrameCPP::Version_3_4_5::Base* base = readStructure();
  return base;
}

// Read single channel from all the frames in the file
//
// :TODO: break this switched function into several small
// functions, one for every supported data type
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector<FrameCPP::Version_3_4_5::Base *>
TOCReader::read( INT_2U structureClassId, std::string name )
{
  std::vector<FrameCPP::Version_3_4_5::Base *> rv;

  switch ( structureClassId ) {
  case ID_ADCDATA:
    {
      ANPI iter = adcNamePositionMap.find( name );
      if ( iter == adcNamePositionMap.end() )
	{
	  throw not_found_error( "bad ADC name" );    
	}
      
      for (unsigned i = 0; i < position.size(); i++)
	{
	  unsigned long pos = iter->second.first[ i ];
	  if ( pos == 0 )
	    {
	      rv.push_back( NULL );
	    }
	  else 
	    {
	      mStream->seekg( pos , std::ios::beg );
	      rv.push_back( readStructure() );
	    }
	}      
    }
    break;
  case ID_PROCDATA:
    {
      PNPI iter = procNamePositionMap.find( name );
      if ( iter == procNamePositionMap.end() )
	{
	  throw not_found_error( "bad Proc name" );
	}
      for (unsigned i = 0; i < position.size(); i++)
	{
	  unsigned long pos = iter->second[ i ];
	  if ( pos == 0 )
	    {
	      rv.push_back( NULL );
	    }
	  else 
	    {
	      mStream->seekg( pos , std::ios::beg );
	      rv.push_back( readStructure() );
	    }
	}
    }
    break;
  case ID_SIMDATA:
    {
      SIMNPI iter = simNamePositionMap.find( name );
      if ( iter == simNamePositionMap.end() )
	{
	  throw not_found_error( "bad Sim name" );
	}
      for (unsigned i = 0; i < position.size(); i++)
	{
	  unsigned long pos = iter->second[ i ];
	  if ( pos == 0 )
	    {
	      rv.push_back( NULL );
	    }
	  else 
	    {
	      mStream->seekg( pos , std::ios::beg );
	      rv.push_back( readStructure() );
	    }
	}
    }
    break;
  case ID_SERDATA:
    {
      SERNPI iter = serNamePositionMap.find( name );
      if ( iter == serNamePositionMap.end() )
	{
	  throw not_found_error( "bad Ser name" );
	}
      for (unsigned i = 0; i < position.size(); i++)
	{
	  unsigned long pos = iter->second[ i ];
	  if ( pos == 0 )
	    {
	      rv.push_back( NULL );
	    }
	  else 
	    {
	      mStream->seekg( pos , std::ios::beg );
	      rv.push_back( readStructure() );
	    }
	}
    }
    break;
  case ID_SUMMARY:
    {
      SUMNPI iter = sumNamePositionMap.find( name );
      if ( iter == sumNamePositionMap.end() )
	{
	  throw not_found_error( "bad Sum name" );
	}
      for (unsigned i = 0; i < position.size(); i++)
	{
	  unsigned long pos = iter->second[ i ];
	  if ( pos == 0 )
	    {
	      rv.push_back( NULL );
	    }
	  else 
	    {
	      mStream->seekg( pos , std::ios::beg );
	      rv.push_back( readStructure() );
	    }
	}
    }
    break;
  default:
    throw not_found_error( "unsupported structure class" );
  }
  return rv;
}

// Get all statics from the file
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< StatData* > TOCReader::readStat()
{
  std::vector< StatData* > v;

  // for every stat name
  for ( SNPI i = statNamePositionMap.begin(); i != statNamePositionMap.end(); i++ )
    {
      // for every timestamp/version
      for ( unsigned int j = 0; j < i->second.second.size(); j++ )
	{
	  INT_8U pos = i->second.second[j].second;
	  mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< StatData* >( readStructure() ));
	}
    }
  return v;
}

// Get all named statics from the file
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< StatData* > TOCReader::readStatName( 
   const std::string& statName )
{
  std::vector< StatData* > v;

  SNPI iter = statNamePositionMap.find( statName );
  if ( iter == statNamePositionMap.end() )
    {
      throw not_found_error( "bad Stat name" );
    }

  // for every timestamp/version
  const size_t size( iter->second.second.size() );

  for ( unsigned int j = 0; j < size; ++j )
    {
      INT_8U pos = iter->second.second[j].second;
      mStream->seekg( pos , std::ios::beg );
      v.push_back( dynamic_cast< StatData* >( readStructure() ));
    }

  return v;
}

// Get all named statics from the file for the time 
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< StatData* > TOCReader::readStatTime( 
   const std::string& statName, INT_4U time )
{
  std::vector< StatData* > v;

  SNPI iter = statNamePositionMap.find( statName );
  if ( iter == statNamePositionMap.end() )
    {
      throw not_found_error( "bad Stat name" );
    }

  // for every timestamp/version
  for ( unsigned int j = 0; j < iter->second.second.size(); j++ )
    {
      // check the time
      //
      if ( time >= iter->second.second[j].first[0] // tStart
	   && time < iter->second.second[j].first[1] ) // tEnd
	{
	  INT_8U pos = iter->second.second[j].second;	  
	  mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< StatData* >( readStructure() ));
	}
    }

  if ( v.empty() )
    {
      throw not_found_error( "no static structure for the time" );
    }

  return v;
}

// Get a statics from the file for the time with the highest version number
//
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
StatData* TOCReader::readStat( const std::string& statName, INT_4U time )
{
  SNPI iter = statNamePositionMap.find( statName );
  if ( iter == statNamePositionMap.end() )
    {
      throw not_found_error( "bad Stat name" );
    }
  
  bool found = false;
  INT_4U idx = 0;
  
  // for every timestamp/version
  const size_t size( iter->second.second.size() );

  for ( unsigned int j = 0; j < size; ++j )
    {
      // check the time and find the highest vesion
      //
      if ( time >= iter->second.second[j].first[0] // tStart
	   && time < iter->second.second[j].first[1] ) // tEnd
	{
	  if ( !found )
	    {
	      found = true;
	      idx = j;
	    }
	  else
	    {
	      if ( iter->second.second[j].first[2] > // current vesion
		   iter->second.second[idx].first[2] ) // current maximum vesion
		{
		  idx = j;
		}
	    }
	}
    }

  if ( !found )
    {
      throw not_found_error( "no static structure for the time" );
    }
  
  // read the actual static structure from the frame file
  //
  INT_8U pos = iter->second.second[idx].second;
  mStream->seekg( pos , std::ios::beg );
  return dynamic_cast< StatData* >( readStructure() );
}


// Get all events from the file
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< Event* > TOCReader::readEvent()
{
  std::vector< Event* > v;

  for ( EVENTNPI i = eventNamePositionMap.begin(); i != eventNamePositionMap.end(); i++ )
    {
      for ( unsigned int j = 0; j < i->second.size(); j++ )
	{
	  INT_8U pos = i->second[j].position;
	  mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< Event* >( readStructure() ));
	}
    }
  return v;
}


// Get all named events from the file
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< Event* > TOCReader::readEvent( std::string eventName )
{
  std::vector< Event* > v;

  EVENTNPI i = eventNamePositionMap.find( eventName );
  if ( i == eventNamePositionMap.end() )
    {
      throw not_found_error( "event not found" );
    }
  for ( unsigned int j = 0; j < i->second.size(); j++ )
    {
      INT_8U pos = i->second[ j ].position;
      mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< Event* >( readStructure() ));
    }

  return v;
}


// Get all named events from the file with certain timestamp
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< Event* > TOCReader::readEvent( std::string eventName, Time time )
{
  std::vector< Event* > v;

  EVENTNPI i = eventNamePositionMap.find( eventName );
  if ( i == eventNamePositionMap.end() )
    {
      throw not_found_error( "event name not found" );
    }
  for ( unsigned int j = 0; j < i->second.size(); j++ )
    {
      // check the time
      if ( i->second[ j ].gtime == time )
	{
	  INT_8U pos = i->second[ j ].position;
	  mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< Event* >( readStructure() ));
	}
    }

  if ( v.empty() )
    {
      throw not_found_error( "no events for the time" );
    }

  return v;
}


// Get all simEvents from the file
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< SimEvent* > TOCReader::readSimEvent()
{
  std::vector< SimEvent* > v;

  for ( SIMEVENTNPI i = simEventNamePositionMap.begin(); i != simEventNamePositionMap.end(); i++ )
    {
      for ( unsigned int j = 0; j < i->second.size(); j++ )
	{
	  INT_8U pos = i->second[j].position;
	  mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< SimEvent* >( readStructure() ));
	}
    }
  return v;
}


// Get all named simEvents from the file
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< SimEvent* > TOCReader::readSimEvent( std::string simEventName )
{
  std::vector< SimEvent* > v;

  SIMEVENTNPI i = simEventNamePositionMap.find( simEventName );
  if ( i == simEventNamePositionMap.end() )
    {
      throw not_found_error( "simEvent not found" );
    }
  for ( unsigned int j = 0; j < i->second.size(); j++ )
    {
      INT_8U pos = i->second[ j ].position;
      mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< SimEvent* >( readStructure() ));
    }

  return v;
}


// Get all named simEvents from the file with certain timestamp
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//!exc: not_found_error - Data is not found.
//      
std::vector< SimEvent* > TOCReader::readSimEvent( std::string simEventName, Time time )
{
  std::vector< SimEvent* > v;

  SIMEVENTNPI i = simEventNamePositionMap.find( simEventName );
  if ( i == simEventNamePositionMap.end() )
    {
      throw not_found_error( "simEvent name not found" );
    }
  for ( unsigned int j = 0; j < i->second.size(); j++ )
    {
      // check the time
      if ( i->second[ j ].gtime == time )
	{
	  INT_8U pos = i->second[ j ].position;
	  mStream->seekg( pos , std::ios::beg );
	  v.push_back( dynamic_cast< SimEvent* >( readStructure() ));
	}
    }

  if ( v.empty() )
    {
      throw not_found_error( "no simEvents for the time" );
    }

  return v;
}


// :TODO: have it read all tables and std::vectors linked in the "next" chain
//
//!exc: std::bad_alloc - Memory allocation failed,
//!exc: read_failure - Read failed.
//      
FrameCPP::Version_3_4_5::Base* TOCReader::readStructure(bool pick)
{
  // Erase any errors which were left over from a previous read
  clearErrors();
  mEndOfFile = false;
  FrameCPP::Version_3_4_5::Base* root;

  // read root object first
  mCurrentObjectHeader = readNextObjectHeader();
  if ( mCurrentObjectHeader.getLocalClassId() == ID_FRAME ) {
    root = dynamic_cast< Frame* >( dictionary->readElement( *this, ID_FRAME ) );	
  }  else  {
    bool endOfFrame = readBase();
    if ( endOfFrame )
      {
	addError( "TOC reader saw end of file structure" );
      }

    Reference rootReference = mCurrentObjectHeader.getReference();
    root = mObjectStorage.find( rootReference )->second;


    if ( pick && mLinkedListStart.size() > 0 ) {

      // Read single level of referenced objects from mLinkedListNodes
      // Read as many as possible: continue reading until the next object is not
      // in mLinkedListNodes
      //
      for(;;) {
	// Read the next header, this is the 8 byte header in front of each
	// frame structure
	mCurrentObjectHeader = readNextObjectHeader();

	// Pick up only table and vectors
	//
	if ( mCurrentObjectHeader.getLocalClassId() != ID_VECT
	     && mCurrentObjectHeader.getLocalClassId() != ID_TABLE ) {
	  break;
	}

	// Read the structure
	bool endOfFrame = readBase();

	if ( endOfFrame )
	  {
	    addError( "TOC reader saw end of file structure" );
	  }
      }
    }

    if ( pick )
      {
	// connect references
	reconnect( true );
      }

  }
  // cleanup 

    // Delete the linked list start information
  for ( LinkedListStart::iterator iter = mLinkedListStart.begin();
	iter != mLinkedListStart.end(); ++iter )
  {
      delete iter->first;
      iter->first = 0;
  }
  mLinkedListStart.erase( mLinkedListStart.begin(),
			  mLinkedListStart.end() );
      
  // Delete the list of nodes
  mLinkedListNodes.erase( mLinkedListNodes.begin(),
			  mLinkedListNodes.end() );
  
  // Delete the list of pointers
  mPointers.erase( mPointers.begin(), mPointers.end() );
  mStatDetectorPointers.erase( mStatDetectorPointers.begin(),
			       mStatDetectorPointers.end() );
  
  // Delete any orphaned objects.  These are objects which existed in the
  // frame but were not referenced by any other objects.
  for ( ObjectStorage::iterator iter = mObjectStorage.begin();
	iter != mObjectStorage.end(); ++iter )
    {
      if ( iter->second != 0 && iter->second != root)
	{
	  ostringstream ss;
	  ss << "TCOReader: Orphan found: (" << iter->first.getClass()
	     << "," << iter->first.getInstance() << ")";
	  addError( ss.str() );
	  //	  delete iter->second;
	}
    }  
  mObjectStorage.erase( mObjectStorage.begin(), mObjectStorage.end() );

  skip(0);
  return root;
}


void TOCReader::readBaseHook( Base *base, INT_2U classId )
{}
