#ifndef FrameCppTOCReaderHH
#define FrameCppTOCReaderHH

#include "base.hh"
#include "adcdata.hh"
#include "procdata.hh"
#include "simdata.hh"
#include "serdata.hh"
#include "summary.hh"
#include "toc.hh"
#include "framereaderbase.hh"
#include "time.hh"
#include "location.hh"
#include "frame.hh"

namespace FrameCPP {

namespace Version_3_4_5 {

class Time;
class Location;


//-----------------------------------------------------------------------------
//
//: TOCReader
//
// This class is used to access a frame file with table of contents.
//
class TOCReader : public virtual TOC, protected FrameReaderBase 
{
public:

  enum ErrorCode
  {
    BADSTREAM = FrameReaderBase::BADSTREAM,
    SKIPFAILED = FrameReaderBase::SKIPFAILED,
    ENDOFFILE = FrameReaderBase::ENDOFFILE,
    NOTOC,
    BADTOC
  };

  /* Error Methods */
  Input::getError;
  Input::errorBegin;
  Input::errorEnd;
  Input::errorQuantity;
  Input::errorFirst;
  Input::errorLast;
  Input::clearErrors;
  Input::setErrorWatch;

  //: Constructor
  //!exc: read_failure - Read failed.
  TOCReader( std::istream& in, bool buildToc = false  );

  //: Destructor
  virtual ~TOCReader();

  /*
    Table of contents data accessors
  */
  
  /* Frame header */
  //: Get leap second from the first frame
  //!exc: None. 
  inline const INT_2S getULeapS() const { return TOC::ULeapS; }
  //: Get frame time stamps
  //!exc: None.    
  inline const std::vector<Time>& getGTime() const { return TOC::GTime; }
  //: Get frame durations
  //!exc: None.    
  inline const std::vector<REAL_8>& getDt() const { return TOC::dt; }
  //: Get frame run numbers
  //!exc: None.    
  inline const std::vector<INT_4S>& getRuns() const { return TOC::runs; }
  //: Get frame numbers
  //!exc: None.    
  inline const std::vector<INT_4U>& getFrame() const { return TOC::frame; }
  //: Get frame header positions
  //!exc: None.    
  inline const std::vector<INT_8U>& getPosition() const { return TOC::position; }
  //: Get first ADC structure positions
  //!exc: None.    
  inline const std::vector<INT_8U>& getPositionADC() const { return TOC::positionADC; }
  //: Get first serial data structure positions
  //!exc: None.    
  inline const std::vector<INT_8U>& getPositionSer() const { return TOC::positionSer; }
  //: Get first referenced from raw data structure table structure positions
  //!exc: None.    
  inline const std::vector<INT_8U>& getPositionTable() const { return TOC::positionTable; }
  //: Get message structure positions
  //!exc: None.    
  inline const std::vector<INT_8U>& getPositionMsg() const { return TOC::positionMsg; }

  /* Structure header */
  //: Get structure IDs
  //!exc: None.    
  inline const std::vector<INT_2U>& getSHid() const { return TOC::SHid; }
  //: Get structure names
  //!exc: None.    
  inline const std::vector<std::string>& getSHname() const { return TOC::SHname; }

  //: Get static data std::map
  //!exc: None.    
  inline const std::map< std::string, std::pair< std::string, std::vector< std::pair< std::vector< INT_4U >, INT_8U > > > >&
      getStatNamePositionMap() const { return TOC::statNamePositionMap; }
  typedef TOC::SNPI SNPI;

  //: Get ADC std::map
  //!exc: None.    
  inline const TOC::adcNamePositionMap_type&
     getAdcNamePositionMap() const { return TOC::adcNamePositionMap; }
  typedef TOC::ANPI ANPI;

  //: Get post-processed data std::map
  //!exc: None.    
  inline const std::map< std::string, std::vector<INT_8U> >& getProcNamePositionMap() const {
    return TOC::procNamePositionMap; 
  }
  typedef TOC::PNPI PNPI;

  //: Get serial data std::map
  //!exc: None.    
  inline const std::map< std::string, std::vector<INT_8U> >& getSerNamePositionMap() const {
    return TOC::serNamePositionMap;
  }
  typedef TOC::SERNPI SERNPI;

  //: Get event data std::map
  //!exc: None.    
  inline const std::map< std::string, std::vector<TOC::eventAttr> >& getEventNamePositionMap() const {
    return TOC::eventNamePositionMap;
  }
  typedef TOC::EVENTNPI EVENTNPI;

  //: Get simulated event data std::map
  //!exc: None.    
  inline const std::map< std::string, std::vector<TOC::eventAttr> >& getSimEventNamePositionMap() const {
    return TOC::simEventNamePositionMap;
  }
  typedef TOC::SIMEVENTNPI SIMEVENTTNPI;

  //: Get summary data std::map
  //!exc: None.    
  inline const std::map< std::string, std::vector<INT_8U> >& getSumNamePositionMap() const {
    return TOC::sumNamePositionMap;
  }
  typedef TOC::SUMNPI SUMNPI;

  /*
   Structure read methods
  */

  //: Fully read one frame
  //!exc: std::bad_alloc, read_failure, not_found_error
  Frame* readFrame( unsigned long frameNumber ); 

  //: Read one frame header
  //!exc: std::bad_alloc, read_failure, not_found_error
  Frame* readFrameHeader( unsigned long frameNumber );

  //: Get all statics from the file
  //!exc: std::bad_alloc, read_failure, not_found_error
  std::vector< StatData* > readStat();

  //: Get all named statics from the file
  //!exc: std::bad_alloc, read_failure, not_found_error
  std::vector< StatData* > readStatName( const std::string& statName );

  //: Get all named statics from the file for certain time
  //!exc:  std::bad_alloc, read_failure, not_found_error
  std::vector< StatData* > readStatTime( const std::string& statName, INT_4U time );

  //: Get a static from the file for certain time with the highest version number
  //
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  StatData* readStat( const std::string& statName, INT_4U time );

  //: Get all events from the file
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  std::vector< Event* > readEvent();

  //: Get all named events
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  std::vector< Event* > readEvent( std::string eventName );

  //: Get all named events with certain time
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  std::vector< Event* > readEvent( std::string eventName, Time time );

  //: Get all simulated events from the file
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  std::vector< SimEvent* > readSimEvent();

  //: Get all named simulated events
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  std::vector< SimEvent* > readSimEvent( std::string simName );

  //: Get all named simulated events with certain time
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  std::vector< SimEvent* > readSimEvent( std::string simName, Time time );

  //: Get one ADC channel from one frame
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline AdcData* readADC( unsigned long frameNumber, std::string adcName )
  {
    return dynamic_cast< AdcData* >( read( frameNumber, ID_ADCDATA, adcName ));
  }

  //: Get one processed data structure from one frame
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline ProcData* readProc( unsigned long frameNumber, std::string procName )
  {
    return dynamic_cast< ProcData* >( read( frameNumber, ID_PROCDATA, procName ));
  }

  //: Get one simulated data structure from one frame
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline SimData* readSim( unsigned long frameNumber, std::string simName )
  {
    return dynamic_cast< SimData* >( read( frameNumber, ID_SIMDATA, simName ));
  }

  //: Get one serial data structure from one frame
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline SerData* readSer( unsigned long frameNumber, std::string serName )
  {
    return dynamic_cast< SerData* >( read( frameNumber, ID_SERDATA, serName ));
  }

  //: Get one summary data structure from one frame
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline Summary* readSummary( unsigned long frameNumber, std::string sumName )
  {
    return dynamic_cast< Summary* >( read( frameNumber, ID_SUMMARY, sumName ));
  }

  //!exc:  std::bad_alloc, read_failure, not_found_error   
  template< class T, INT_2U ID > std::vector< T* > read( std::string name )
  {
    std::vector< FrameCPP::Version_3_4_5::Base* > bv = read( ID, name );
    std::vector< T* > v;
    for ( unsigned int i = 0; i < bv.size(); i++ )
    {
       v.push_back( dynamic_cast< T* >( bv[i] ));
    }
     return v;
  }

  //: Get named ADC channel from all frames
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline std::vector<AdcData*> readADC( std::string adcName )
  {
  std::vector< FrameCPP::Version_3_4_5::Base* > bv = read( ID_ADCDATA, adcName );
  std::vector< AdcData* > v;
  for ( unsigned int i = 0; i < bv.size(); i++ )
    {
      v.push_back( dynamic_cast< AdcData* >( bv[i] ));
    }
  return v;
  }

  //: Get named processed data structure from all frames
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline std::vector<ProcData*> readProc( std::string procName )
  {
    return read<ProcData, ID_PROCDATA>( procName );
  }

  //: Get named simulated data structure from all frames
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline std::vector<SimData*> readSim( std::string simName )
  {
    return read<SimData, ID_SIMDATA>( simName );
  }

  //: Get named serial data structure from all frames
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline std::vector<SerData*> readSer( std::string serName )
  {
    return read<SerData, ID_SERDATA>( serName );
  }

  //: Get named summary data structure from all frames
  //!exc:  std::bad_alloc, read_failure, not_found_error   
  inline std::vector<Summary*> readSum( std::string sumName )
  {
    return read<Summary, ID_SUMMARY>( sumName );
  }

protected:
  FrameReaderBase::mStream;

  //: Whether this reader is using the TOC on frame files
  bool mHasToc;

  //
  // Read any structure.
  // Read all referenced Table and Vector structures, if `pick' is true.
  //
  //!exc: std::bad_alloc, read_failure
  FrameCPP::Version_3_4_5::Base* readStructure( bool pick = true ); 

  //!exc: std::bad_alloc - Memory allocation failed.
  virtual void readBaseHook( Base *base, INT_2U classId );

private:

  // Read single channel from all frames in the file
  // Makes it possible to get a std::vector of structures of particular type. For instance,
  // it is usefull to be able to get a std::vector of FrAdcData structures, where each structure
  // is an instance of an ADC channel from corresponding frame and all ADC structures have the same
  // name. In other words, it will get the data for one ADC channel from all frame in the file.
  //
  // Supports AdcData,  ProcData, SimData, SerData and Summary.
  //
  //!exc:  std::bad_alloc, read_failure, not_found_error    
  std::vector<FrameCPP::Version_3_4_5::Base*> read( INT_2U structureClassId, std::string name );

  // Read single data structure from the specified frame.
  // Supports AdcData,  ProcData, SimData, SerData and Summary.
  //
  //!exc:  std::bad_alloc, read_failure, not_found_error       
  FrameCPP::Version_3_4_5::Base* read( unsigned long frameNumber, INT_2U structureClassId, std::string name );

};


} // namespace - Version_3_4_5

} // namespace - FrameCPP


#endif // FrameCppTOCReaderHH
