#include "framecpp/config.h"

#include <stdexcept>
#include <vector>

#include "framecpp/SearchContainer.hh"

#include "vect.hh"
#include "framereader.hh"
#include "output.hh"

#include "container.hh"

extern "C" {
#define BZ_NO_STDIO
#include "bzlib.h"
}

using FrameCPP::Version_3_4_5::Vect;
using namespace std;


//-----------------------------------------------------------------------------
// Initialization
//-----------------------------------------------------------------------------

const int Vect::bzip2_blockSize100k = 9;
const int Vect::bzip2_workFactor = 30;


//-----------------------------------------------------------------------------
// Constructors
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Default Constructor.
//
// This creates a Vect object with the following properties:
// <ul>
//     <li> name = "none" </li>
//     <li> compression = 0 (none) = BIGENDIAN</li>
//     <li> type = FR_VECT_2S (INT_2S) </li>
//     <li> nData = 0 </li>
//     <li> nBytes = 0 </li>
//     <li> data = 0 (null) </li>
//     <li> nDim = 0 </li>
//     <li> No dimension information </li>
//     <li> unitY = "" </li>
// </ul>
//
//!exc: None.      
Vect::Vect()
        : Base(), mName( "none" ), mCompress( 0 ),
          mType( FR_VECT_2S ),
          mNData( 0 ), mNBytes( 0 ), mData( 0 ), mNDim( 0 ),
          mDimension(), mUnitY( "" ), mOwnsData( true ), 
	  dataNeedsByteSwap( false )
{
}


//-----------------------------------------------------------------------------
//
//: Copy Constructor.  
//
//!param: const Vect& vect - Copied vector.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Vect::Vect( const Vect& vect )
        : Base( vect ), mName( vect.mName ),
          mCompress( vect.mCompress ), mType( vect.mType ),
          mNData( vect.mNData ), mNBytes( vect.mNBytes ),
          mData( 0 ), mNDim( vect.mNDim ), mDimension( vect.mDimension ),
          mUnitY( vect.mUnitY ), mOwnsData( vect.mOwnsData ),
	  dataNeedsByteSwap( vect.dataNeedsByteSwap )
{
    if ( vect.mData != 0 )
    {
        if ( mOwnsData )
        {
            mData = new unsigned char[ mNBytes ];
            memcpy( mData, vect.mData, mNBytes );
        }
        else
        {
            mData = vect.mData;
        }
    }
}


//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - The name of the data.
//!param: INT_2U type - The data type.
//!param: INT_4U nDim - The number of dimensions.
//!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
//+     containing information about the dimensions for this data.
//!param: const ByteOrder byte_order - Byte order of the data. Default is 
//+       BIGENDIAN.      
//!param: const void* data - A pointer to the data.  Default: 0
//!param: const std::string& unitY - Units for the data.  Default: ""
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Vect::Vect( const std::string& name, INT_2U type, INT_4U nDim,
            const Dimension* dims, const ByteOrder byte_order, 
            const void* data, const std::string& unitY ) 
        : mName( name ), mCompress( byte_order == BIG ? 0 : 256 ),
          mType( type ), mNData( 1 ), mNBytes( 0 ), mData( 0 ), mNDim( nDim ),
          mDimension(), mUnitY( unitY ), mOwnsData( true ),
	  dataNeedsByteSwap(
#ifdef WORDS_BIGENDIAN   
                            byte_order == BIG ? false : true
#else
                            byte_order == LITTLE ? false : true   
#endif
                            )
{
    // Store the dimension information &
    // calculate the number of data points and bytes
    for ( size_t i = 0; i < mNDim; ++i )
    {
        mDimension.push_back( dims[ i ] );
        mNData *= dims[ i ].getNx();
    }
    mNBytes = mNData * getTypeSize( type );

    // Set the data
    if ( data != 0 )
    {
        mData = new unsigned char[ mNBytes ];
        memcpy( mData, data, mNBytes );
    }    
}

    
//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - The name of the data.
//!param: INT_2U type - The data type.
//!param: INT_4U nDim - The number of dimensions.
//!param: const Dimension* dims - A pointer to 'nDim' Dimension objects
//+     containing information about the dimensions for this data.
//!param: const ByteOrder byte_order - Byte order of the data. Default is 
//+       BIGENDIAN.         
//!param: const void* data - A pointer to the data.  Default: 0
//!param: const std::string& unitY - Units for the data.  Default: ""
//!param: bool allocate - Whether to allocate data. Default: true.
//!param: bool owns - Whether to destruct data in the destructor. Default: true.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Vect::Vect( const std::string& name, INT_2U type, INT_4U nDim,
            const Dimension* dims, const ByteOrder byte_order, void* data,
            const std::string& unitY, bool allocate, bool owns )
        : mName( name ), mCompress( byte_order == BIG ? 0 : 256 ),
          mType( type ), mNData( 1 ), mNBytes( 0 ), mData( 0 ), mNDim( nDim ),
          mDimension(), mUnitY( unitY ), mOwnsData( owns ),
	  dataNeedsByteSwap( 
#ifdef WORDS_BIGENDIAN   
                            byte_order == BIG ? false : true
#else
                            byte_order == LITTLE ? false : true   
#endif
                            )
{
    // Store the dimension information &
    // calculate the number of data points and bytes
    for ( size_t i = 0; i < mNDim; ++i )
    {
        mDimension.push_back( dims[ i ] );
        mNData *= dims[ i ].getNx();
    }
    mNBytes = mNData * getTypeSize( type );

    // Set the data
    if ( data != 0 )
    {
        if ( allocate )
        {
            mData = new unsigned char[ mNBytes ];
            memcpy( mData, data, mNBytes );
        }
        else
        {
            mData = reinterpret_cast< unsigned char* >( data );
        }
    }    
}

    
//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - The name of the data.
//!param: INT_2U type - The data type.
//!param: INT_4U nDim - The number of dimensions.
//!param: const INT_4U* dims - A pointer to 'nDim' Dimension objects
//+     containing information about the dimensions for this data.
//!param: const ByteOrder byte_order - Byte order of the data. Default is 
//+       BIGENDIAN.         
//!param: const void* data - A pointer to the data.  Default: 0
//!param: const std::string& unitY - Units for the data.  Default: ""
//
//!exc: std::bad_alloc - Memory allocation failed.
//
Vect::Vect( const std::string& name, INT_2U type, INT_4U nDim, const INT_4U* dims,
            const ByteOrder byte_order, const void* data, const std::string& unitY ) 
        : mName( name ), mCompress( byte_order == BIG ? 0 : 256 ), mType( type ),
          mNData( 1 ), mNBytes( 0 ), mData( 0 ), mNDim( nDim ),
          mDimension(), mUnitY( unitY ), 
          dataNeedsByteSwap( 
#ifdef WORDS_BIGENDIAN   
                            byte_order == BIG ? false : true
#else
                            byte_order == LITTLE ? false : true   
#endif
                            )   
{
    // Store the dimension information &
    // calculate the number of data points and bytes
    for ( size_t i = 0; i < mNDim; ++i )
    {
        mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
        mNData *= dims[ i ];
    }
    mNBytes = mNData * getTypeSize( type );

    // Set the data
    if ( data != 0 )
    {
        mData = new unsigned char[ mNBytes ];
        memcpy( mData, data, mNBytes );
    }    
}

    
//-----------------------------------------------------------------------------
//
//: Constructor.
//
//!param: const std::string& name - The name of the data.
//!param: INT_2U type - The data type.
//!param: INT_4U nDim - The number of dimensions.
//!param: const INT_4U* dims - A pointer to 'nDim' Dimension objects
//+     containing information about the dimensions for this data.
//!param: const ByteOrder byte_order - Byte order of the data. Default is 
//+       BIGENDIAN.         
//!param: const void* data - A pointer to the data.  Default: 0
//!param: const std::string& unitY - Units for the data.  Default: ""
//!param: bool allocate - Whether to allocate data. Default: true.
//!param: bool owns - Whether to destruct data in the destructor. Default: true.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
//!todo: Shouldn't the number of bytes be calculated?  How do we initialize
//!todo: with compressed data?
//
// <font color="red">This constructor sets BYTE_ORDER to BIGENDIAN!!! Fix it before
// using!</font>    
Vect::Vect( const std::string& name, INT_2U type, INT_4U nDim, const INT_4U* dims,
            const ByteOrder byte_order, void* data, const std::string& unitY,
            bool allocate, bool owns )
        : mName( name ), mCompress( byte_order == BIG ? 0 : 256 ), mType( type ),
          mNData( 1 ), mNBytes( 0 ), mData( 0 ), mNDim( nDim ),
          mDimension(), mUnitY( unitY ), mOwnsData( owns ),
	  dataNeedsByteSwap(
#ifdef WORDS_BIGENDIAN   
                            byte_order == BIG ? false : true
#else
                            byte_order == LITTLE ? false : true   
#endif
                            )   
{
    // Store the dimension information &
    // calculate the number of data points and bytes
    for ( size_t i = 0; i < mNDim; ++i )
    {
        mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
        mNData *= dims[ i ];
    }
    mNBytes = mNData * getTypeSize( type );

    // Set the data
    if ( data != 0 )
    {
        if ( allocate )
        {
            mData = new unsigned char[ mNBytes ];
            memcpy( mData, data, mNBytes );
        }
        else
        {
            mData = reinterpret_cast< unsigned char* >( data );
        }
    }    
}

//-----------------------------------------------------------------------------
// Destructor
//-----------------------------------------------------------------------------


//: Delete owned data.
//!exc: None.      
void Vect::deleteData()
{
    if ( mOwnsData )
    {
        switch( getType() )
        {
            case FR_VECT_C:
                delete[] reinterpret_cast< CHAR* >( mData );
                break;
            
            case FR_VECT_2S:
                delete[] reinterpret_cast< INT_2S* >( mData );
                break;
            
            case FR_VECT_8R:
                delete[] reinterpret_cast< REAL_8* >( mData );
                break;
            
            case FR_VECT_4R:
                delete[] reinterpret_cast< REAL_4* >( mData );
                break;
            
            case FR_VECT_4S:
                delete[] reinterpret_cast< INT_4S* >( mData );
                break;
            
            case FR_VECT_8S:
                delete[] reinterpret_cast< INT_8S* >( mData );
                break;
            
            case FR_VECT_8C:
                delete[] reinterpret_cast< REAL_4* >( mData );
                break;
            
            case FR_VECT_16C:
                delete[] reinterpret_cast< REAL_8* >( mData );
                break;
            
            case FR_VECT_2U:
                delete[] reinterpret_cast< INT_2U* >( mData );
                break;
            
            case FR_VECT_4U:
                delete[] reinterpret_cast< INT_4U* >( mData );
                break;
            
            case FR_VECT_8U:
                delete[] reinterpret_cast< INT_8U* >( mData );
                break;
            
            case FR_VECT_1U:
                delete[] reinterpret_cast< CHAR_U* >( mData );
                break;
        }
   
        mData = 0;
    }
}


//-----------------------------------------------------------------------------
//
//: Destructor.
//
// Deletes data, if it is owned.
//
//!exc: None.      
Vect::~Vect()
{
    deleteData();
}


//-----------------------------------------------------------------------------
//
//: Get Vector name.
//
//!return: const std::string& -- Vector name.
//
const std::string& Vect::getName() const
{
    return mName;
}


//-----------------------------------------------------------------------------
// Operator Overloads
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Assignment Operator.
//
//!param: const Vect& vect - Vector to be assigned.
//
//!return: const Vect& - This vector.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
const Vect& Vect::operator=( const Vect& vect )
{
    if ( this != &vect )
    {
        deleteData();
        
        mName = vect.mName;
        mCompress = vect.mCompress;
        mType = vect.mType;
        mNData = vect.mNData;
        mNBytes = vect.mNBytes;
        mNDim = vect.mNDim;
        mDimension = vect.mDimension;
        mUnitY = vect.mUnitY;
        mOwnsData = vect.mOwnsData;
	dataNeedsByteSwap = vect.dataNeedsByteSwap;

        mData = 0;
        if ( mNBytes != 0 )
        {
            if ( mOwnsData )
            {
                mData = new unsigned char[ mNBytes ];
                memcpy( mData, vect.mData, mNBytes );
            }
            else
            {
                mData = vect.mData;
            }
        }
    }

    return *this;
}

//-----------------------------------------------------------------------------
//
//: += Assignment Operator.
//
//!param: const Vect& vect - Vector with which to concatenate.
//
//!return: const Vect& - This vector.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: frame_mismatch - Vectors are incompatible.
//
const Vect& Vect::operator+=( const Vect& vect )
{
  // Check all required attributes are the same
  // Only one dimensional vectors supported
  //
  // TO BE DONE
  //

  throw runtime_error( "Unimplemented Vect::operator+= is called." );

#ifdef not_def
    if ( this != &vect )
    {
        deleteData();
        
        mName = vect.mName;
        mCompress = vect.mCompress;
        mType = vect.mType;
        mNData = vect.mNData;
        mNBytes = vect.mNBytes;
        mNDim = vect.mNDim;
        mDimension = vect.mDimension;
        mUnitY = vect.mUnitY;
        mOwnsData = vect.mOwnsData;

        mData = 0;
        if ( mNBytes != 0 )
        {
            if ( mOwnsData )
            {
                mData = new unsigned char[ mNBytes ];
                memcpy( mData, vect.mData, mNBytes );
            }
            else
            {
                mData = vect.mData;
            }
        }
    }
#endif

    return *this;
}


//-----------------------------------------------------------------------------
// Required Methods
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//: Constructs the Structure Header for this class.
//
//!return: SH - Vector class structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Vect::getSH() 
{
  SH sh( "FrVect", ID_VECT, "Vector Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "compress", "INT_2U", "" ) );
  sh.appendSE( SE( "type", "INT_2U", "" ) );
  sh.appendSE( SE( "nData", "INT_4U", "" ) );
  sh.appendSE( SE( "nBytes", "INT_4U", "" ) );
  sh.appendSE( SE( "data", "CHAR[nBytes]", "" ) );
  sh.appendSE( SE( "nDim", "INT_4U", "" ) );
  sh.appendSE( SE( "nx", "INT_4U[nDim]", "" ) );
  sh.appendSE( SE( "dx", "REAL_8[nDim]", "" ) );
  sh.appendSE( SE( "startX", "REAL_8[nDim]", "" ) );
  sh.appendSE( SE( "unitX", "STRING[nDim]", "" ) );
  sh.appendSE( SE( "unitY", "STRING", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrVect *)", "" ) );
  
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 4 Structure Header for this class.
//
//!return: SH - Vector class structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Vect::getSH4()
{
  SH sh( "FrVect", ID_VECT, "Vector Data" );
  sh.appendSE( SE( "name", "STRING", "" ) );
  sh.appendSE( SE( "compress", "INT_2U", "" ) );
  sh.appendSE( SE( "type", "INT_2U", "" ) );
  sh.appendSE( SE( "nData", "INT_4U", "" ) );
  sh.appendSE( SE( "nBytes", "INT_4U", "" ) );
  sh.appendSE( SE( "data", "*", "" ) );
  sh.appendSE( SE( "nDim", "INT_4U", "" ) );
  sh.appendSE( SE( "nx", "INT_4U *", "" ) );
  sh.appendSE( SE( "dx", "REAL_8 *", "" ) );
  sh.appendSE( SE( "startX", "REAL_8 *", "" ) );
  sh.appendSE( SE( "unitX", "STRING *", "" ) );
  sh.appendSE( SE( "unitY", "STRING", "" ) );
  sh.appendSE( SE( "next", "PTR_STRUCT(FrVect *)", "" ) );
  
  return sh;
}


//-----------------------------------------------------------------------------
//
//: Constructs frame data format version 3 Structure Header for this class.
//
//!return: SH - Legacy structure header.
//
//!exc: std::bad_alloc - Memory allocation failed.
//
FrameCPP::Version_3_4_5::SH Vect::getSH3() 
{
    SH sh( "FrVect", ID_VECT, "Vector Data" );
    sh.appendSE( SE( "name", "STRING", "" ) );
    sh.appendSE( SE( "compress", "INT_2U", "" ) );
    sh.appendSE( SE( "type", "INT_2U", "" ) );
    sh.appendSE( SE( "nData", "INT_4U", "" ) );
    sh.appendSE( SE( "nBytes", "INT_4U", "" ) );
    sh.appendSE( SE( "data", "*", "" ) );
    sh.appendSE( SE( "nDim", "INT_4U", "" ) );
    sh.appendSE( SE( "nx", "INT_4U *", "" ) );
    sh.appendSE( SE( "dx", "REAL_8 *", "" ) );
    sh.appendSE( SE( "unitX", "STRING *", "" ) );
    sh.appendSE( SE( "unitY", "STRING", "" ) );
    sh.appendSE( SE( "next", "PTR_STRUCT(FrVect *)", "" ) );

    return sh;
}


//-----------------------------------------------------------------------------
//
//: Reads an Vect object from an FrameReaderBase object.
//
// This method creates a dynamically allocated object.
//
// <p>*NOTE* This method allocates memory and returns a pointer to it.  The
// user is responsible for freeing this memory.
//
//!param: FrameReaderBase& in - Data source.
//
//!return: Vect* - Newly created vector.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Data read failure occured.
//
Vect* Vect::read4( FrameReaderBase& in )
{
  Vect& vect = *new Vect();

  in >> vect.mName >> vect.mCompress >> vect.mType >> vect.mNData >> vect.mNBytes;

  vect.mData = new unsigned char[ vect.mNBytes ];
  in.read( vect.mData, vect.mNBytes );
  if ( vect.mCompress == NONE || vect.mCompress == NONE + 0x100)
    {
#ifdef WORDS_BIGENDIAN
      vect.dataNeedsByteSwap = vect.mCompress > 0xff;
#else
      vect.dataNeedsByteSwap = vect.mCompress <= 0xff;
#endif
    }

  in >> vect.mNDim;

  if (vect.mNDim >= 256) { 
       throw( read_failure("Invalid vector dimension", Input::BADSTREAM) );
  }
  INT_4U *nx = new INT_4U[ vect.mNDim ];
  REAL_8 *dx = new REAL_8[ vect.mNDim ];
  REAL_8 *startX = new REAL_8[ vect.mNDim ];
  std::string *unitX = new std::string[ vect.mNDim ];

  for ( INT_4U i = 0; i < vect.mNDim; ++i )
    {
      in >> nx[ i ];
    }
    
  for ( INT_4U i = 0; i < vect.mNDim; ++i )
    {
      in >> dx[ i ];
    }

  for ( INT_4U i = 0; i < vect.mNDim; ++i )
    {
      in >> startX[ i ];
    }

  for ( INT_4U i = 0; i < vect.mNDim; ++i )
    {
      in >> unitX[ i ];
    }

  for ( INT_4U i = 0; i < vect.mNDim; ++i )
    {
      vect.mDimension.push_back( Dimension( nx[ i ], dx[ i ], unitX[ i ], startX[ i ] ) );
    }

  delete[] nx;
  nx = 0;
   
  delete[] dx;
  dx = 0;
   
  delete[] startX;
  startX = 0;
   
  delete[] unitX;
  unitX = 0;

  in >> vect.mUnitY;
  in.readNextNodeReference();

  return &vect;
}

//: Read legacy data vector.
//
//!exc: std::bad_alloc - Memory allocation failed.
//!exc: read_failure - Data read failure occured.
//   
Vect* Vect::read3( FrameReaderBase& in )
{
    Vect& vect = *new Vect();

    in >> vect.mName >> vect.mCompress >> vect.mType >> vect.mNData >> vect.mNBytes;

    vect.mData = new unsigned char[ vect.mNBytes ];
    in.read( vect.mData, vect.mNBytes );
    if ( vect.mCompress == NONE || vect.mCompress == NONE + 0x100)
      {
#ifdef WORDS_BIGENDIAN
        vect.dataNeedsByteSwap = vect.mCompress > 0xff;
#else
        vect.dataNeedsByteSwap = vect.mCompress <= 0xff;
#endif
      }

    in >> vect.mNDim;

    if (vect.mNDim >= 256) { 
       throw( read_failure( "Invalid vector dimension", Input::BADSTREAM ));
    }


    vector< INT_4U > nx( vect.mNDim );
    for( vector< INT_4U >::iterator iter = nx.begin(), end_iter = nx.end();
	 iter != end_iter; ++iter )
    {
       in >> ( *iter );
    }

    vector< REAL_8 > dx( vect.mNDim );
    for( vector< REAL_8 >::iterator iter = dx.begin(), end_iter = dx.end();
         iter != end_iter; ++iter )
    {
      in >> ( *iter );
    }

    vector< string > unitX( vect.mNDim );
    for( vector< string >::iterator iter = unitX.begin(), end_iter = unitX.end();
         iter != end_iter; ++iter )
    {
      in >> ( *iter );
    }

    for ( INT_4U i = 0; i < vect.mNDim; ++i )
    {
        vect.mDimension.push_back( Dimension( nx[ i ], dx[ i ], unitX[ i ] ) );
    }

    in >> vect.mUnitY;
    in.readNextNodeReference();

    return &vect;
}



//-----------------------------------------------------------------------------
//
//: Writes a Vect object to an Output object.
//
//!param: Output& out out - Output data stream.
//
//!exc: write_failure - Data write failure occured.
//
void Vect::write( Output& out ) const
{
    INT_4U length = 34
        + mName.size()
        + mNBytes + mUnitY.size()
        + mNDim*( 15 + 8 );

    for ( INT_4U i = 0; i < mNDim; i++ )
    {
        length += mDimension[ i ].getUnitX().size();
    }
    
    out.startRecord( ID_VECT, length, this );
    out << mName << mCompress << mType << mNData << mNBytes;
    out.write( mData, mNBytes );
    out << mNDim;
    
    for ( INT_4U i = 0; i < mNDim; ++i )
    {
        out << mDimension[ i ].getNx();
    }
    
    for ( INT_4U i = 0; i < mNDim; ++i )
    {
        out << mDimension[ i ].getDx();
    }

    for ( INT_4U i = 0; i < mNDim; ++i )
    {
        out << mDimension[ i ].getStartX();
    }

    for ( INT_4U i = 0; i < mNDim; ++i )
    {
        out << mDimension[ i ].getUnitX();
    }

    out << mUnitY;
    out.endRecord( this );
}

//-----------------------------------------------------------------------------
//
//: Get size of single data point of particular size.
//
//!param: INT_2U type - Data type ID.
//
//!return: INT_2U - Data size.
//
//!exc: None.   
size_t Vect::getTypeSize( INT_2U type ) 
{
    switch( type )
    {
        case FR_VECT_C:
            return sizeof( CHAR );
        case FR_VECT_2S:
            return sizeof( INT_2S );
        case FR_VECT_8R:
            return sizeof( REAL_8 );
        case FR_VECT_4R:
            return sizeof( REAL_4 );
        case FR_VECT_4S:
            return sizeof( INT_4S );
        case FR_VECT_8S:
            return sizeof( INT_8S );
        case FR_VECT_8C:
            return ( sizeof( REAL_4 ) * 2 );
        case FR_VECT_16C:
            return ( sizeof( REAL_8 ) * 2 );
/*
 ????
        case FR_VECT_STR:
            return sizeof( std::std::string );
 ????
*/
        case FR_VECT_2U:
            return sizeof( INT_2U );
        case FR_VECT_4U:
            return sizeof( INT_4U );
        case FR_VECT_8U:
            return sizeof( INT_8U );
        case FR_VECT_1U:
            return sizeof( CHAR_U );
    }
    // None of above: Perhaps add type for throw?
    return 0;
}


//-----------------------------------------------------------------------------
//: Perform byte swapping operation on vector data
//
void Vect::byteSwapData()
{
  switch (mNBytes/mNData)
    {
    case (1):
      break;
    case (2):
      reverse< 2 >( mData, mNData );
      break;
    case (4):
      reverse< 4 >( mData, mNData );
      break;
    case (8):
      if (mType == FR_VECT_8C)
	reverse< 4 >( mData, mNData * 2 );
      else
        reverse< 8 >( mData, mNData );
      break;
    case (16):
      // This case only works for COMPLEX_16 data
      reverse< 8 >( mData, mNData * 2 );
      break;
    default:
      abort();
    }
}


//-----------------------------------------------------------------------------
//
//: Compress the data for this Vect.
//
//!param: Compression c - Compressions type.
//!param: size_t level - Compressions level.
//
void Vect::compress( Compression c, size_t level )
{
  if ( level > 9 )
    {
      level = 9;
    }
    
  switch( c )
    {
    case NONE:
      if ( getCompress() != NONE )
	{
	  expand();
	}
      break;

    case GZIP:
      if ( getCompress() != NONE )
	{
	  expand();
	}
      gzipCompress( level );
      break;

    case DIFF:
#           ifdef WORDS_BIGENDIAN
      if ( getCompress() != DIFF || getByteOrder() != BIG )
#           else
	if ( getCompress() != DIFF || getByteOrder() != LITTLE )  
#           endif
	  {
	    if ( getCompress() != NONE )
	      {
		expand();
	      }
	    diffCompress();
	  }
      break;

    case GZIP_DIFF:
      if ( getCompress() != NONE )
	{
	  expand();
	}
      diffCompress();
      gzipCompress( level );
      break;
    case DIFF_ZERO_SUPPRESS_SHORT:
      if ( getCompress() != NONE )
	{
	  expand();
	}
      // Compression is done for short data only
      //
      if ( getType() == FR_VECT_2S || getType() == FR_VECT_2U )
	{
	  diffCompress();
	  zCompress();
	}
      break;
    case DIFF_ZERO_SUPPRESS_SHORT_DIFF_GZIP_OTHER:
      if ( getCompress() != NONE )
	{
	  expand();
	}
      diffCompress();
      if ( getType() == FR_VECT_2S || getType() == FR_VECT_2U )
	{
	  zCompress();
	}
      else
	{
	  gzipCompress( level );
	}
      break;
    case ZERO_SUPPRESS_INT_FLOAT:
      if ( getCompress() != NONE )
	{
	  expand();
	}
      
      // :TODO: need to write this
      break;
    case BZIP2:
#if HAVE_LIBBZ2
      if ( getCompress() != NONE )
	{
	  expand();
	}
      {
	char *dest = new char[mNBytes];
	unsigned int destLen = mNBytes;
	
	int res = 
#if defined(HAVE_NEW_LIBZ2)
		BZ2_bzBuffToBuffCompress
#else
		bzBuffToBuffCompress
#endif
					( dest,
					&destLen,
					(char *)mData,
					mNBytes,
					bzip2_blockSize100k,
					0, /* verbosity - 0 - silent */
					bzip2_workFactor );
	if (BZ_OK  != res ) {
	  delete dest;
          dest = 0;

	  switch (res) { 
	  case BZ_MEM_ERROR:
	    // if insufficient memory is available 
	    throw std::bad_alloc();

	  case BZ_OUTBUFF_FULL:
	    // if the size of the compressed data exceeds *destLen
	    throw unknown_error( "BZIP2 library is not compressing" );	    
#if 0
	  case BZ_CONFIG_ERROR:  // if the library has been mis-compiled
#endif
	  case BZ_PARAM_ERROR :
	  default:
	    // if dest is NULL or destLen is NULL
	    // or blockSize100k < 1 or blockSize100k > 9
	    // or verbosity < 0 or verbosity > 4 
	    // or workFactor < 0 or workFactor > 250
	    throw unknown_error( "BZIP2 library is bad or used badly" );
	  }
	}

	delete[] mData;
        mData = 0;
   
	mData = new unsigned char[ destLen ];
	memcpy( mData, dest, destLen );
	mNBytes = destLen;
	delete[] dest;
        dest = 0;
	break;
      }
#else /* HAVE_LIBBZ2 */
      throw format_error( "BZip2 compression not supported by this library" );
#endif /* HAVE_LIBBZ2 */
      break;
    }
}


//-----------------------------------------------------------------------------
//
//: Differentially compress the data.
//
// The data must be in an uncompressed state before this is called.
//
void Vect::diffCompress()
{
    switch( mType )
    {
        case FR_VECT_C:
            diffCompressType< CHAR >();
            break;
                
        case FR_VECT_2S:
            diffCompressType< INT_2S >();
            break;
                
        case FR_VECT_8R:
            diffCompressType< REAL_8 >();
            break;
                
        case FR_VECT_4R:
            diffCompressType< REAL_4 >();
            break;
                
        case FR_VECT_4S:
            diffCompressType< INT_4S >();
            break;
                
        case FR_VECT_8S:
            diffCompressType< INT_8S >();
            break;
                
        case FR_VECT_8C:
            diffCompressType< COMPLEX_8 >();
            break;
                
        case FR_VECT_16C:
            diffCompressType< COMPLEX_16 >();
            break;
                
        case FR_VECT_2U:
            diffCompressType< INT_2U >();
            break;
                
        case FR_VECT_4U:
            diffCompressType< INT_4U >();
            break;
                
        case FR_VECT_8U:
            diffCompressType< INT_8U >();
            break;
                
        case FR_VECT_1U:
            diffCompressType< CHAR_U >();
            break;
    }
}


//-----------------------------------------------------------------------------
//
//: Differentially compress the data.
//
// The data must be in an uncompressed state before this is called.
//
template< class T >
void Vect::diffCompressType()
{
    if ( mNData > 1 )
    {
        T* first( reinterpret_cast< T* >(mData) + mNData - 2 );
        T* second( first + 1 );
        
        for ( int i = mNData - 1; i != 0; --i )
        {
            *(second--) -= *(first--);
        }

#       ifdef WORDS_BIGENDIAN
        {
            mCompress = DIFF;
        }
#       else
        {
            mCompress = DIFF + 256;
        }
#       endif
    }
}


template<>
void Vect::diffCompressType< COMPLEX_8 >()
{
   // Double number of data points
   mNData *= 2;
   diffCompressType< REAL_4 >();
   mNData /= 2;

   return;
}

   
template<>
void Vect::diffCompressType< COMPLEX_16 >()
{
   // Double number of data points
   mNData *= 2;
   diffCompressType< REAL_8 >();
   mNData /= 2;

   return;
}
   

//-----------------------------------------------------------------------------
//
//: Compress the data with GZip.
//
// The data must be in an uncompressed state before this is called.
//
//!param: size_t level - Compression level.
//
void Vect::gzipCompress( size_t level )
{
    if ( level == 0 )
    {
        return;
    }
    
    unsigned char* buffer = 0;
    unsigned long bufferSize = static_cast< unsigned long >(
        mNBytes * 1.001 + 13 );
    
    buffer = new unsigned char[ bufferSize ];
#if HAVE_LIBZ_COMPRESS2
    ::compress2( buffer, &bufferSize, mData, mNBytes, level );
#else
    ::compress( buffer, &bufferSize, mData, mNBytes );
#endif

    delete[] mData;
    mData = 0;
   
    mData = new unsigned char[ bufferSize ];
    memcpy( mData, buffer, bufferSize );
    mNBytes = bufferSize;
    delete[] buffer;
    buffer = 0;

    Compression compress_now( getCompress() );
    if( compress_now == NONE )
    {
       mCompress += GZIP;
    }
    else if( compress_now == DIFF )
    {
       mCompress -= DIFF;
       mCompress += GZIP_DIFF;
    }
   
    return;
}


//-----------------------------------------------------------------------------
//
//: Suppress zeroes.
//
void Vect::zCompress()
{
    unsigned short* buffer = new unsigned short[ getNBytes() ];
    unsigned long bufferSize = getNBytes();

    int res = FrVectZComp( buffer, &bufferSize, (short int *) mData, (int) getNData(), 6 );
    if ( res != 0 )
      {
	throw unknown_error( "logical error: zero suppression failed" );
      }

    delete[] mData;
    mData = 0;
   
    mData = new unsigned char[ bufferSize ];
    memcpy( mData, buffer, bufferSize );
    mNBytes = bufferSize;
    delete[] buffer;
    buffer = 0;

    Compression compress_now( getCompress() );
    if( compress_now == NONE )
    {
       mCompress += DIFF_ZERO_SUPPRESS_SHORT;
    }
    else if( compress_now == DIFF )
    {
       mCompress -= DIFF;
       mCompress += DIFF_ZERO_SUPPRESS_SHORT_DIFF_GZIP_OTHER;
    }
    
    return;
}


//-----------------------------------------------------------------------------
//: Suppress zeroes.
//
// Borrowed from FrameL.c, v3r85
//
int Vect::FrVectZComp(unsigned short *out, 
		      unsigned long *compL,
		      short *data,
		      int nData,
		      int bSize)
{unsigned short nBits, uData, limit;
 short max, pOut;
 int  iIn, iOut, i, maxOut;
 short wMax[] = {0,0,1,3,7,0xf,0x1f,0x3f,0x7f,0xff,0x1ff,
                   0x3ff,0x7ff,0xfff,0x1fff,0x3fff,0x7fff};

  maxOut = *compL/2;

 /*---- store the bloc size -----*/

 out[0] = bSize;
 iOut = 0;
 pOut = 16;

 /*--------- store the data  --------------*/

 iIn = 0;
 while(iIn<nData)

   { /*-------- tune the size of the last bloc -----*/

     if(iIn + bSize > nData) {bSize = nData - iIn;}

      /*---- get the maximum amplitude --------------*/
      
     max = 0;
     for(i=0; i<bSize; i++)  
          {if(data[iIn+i] > 0)
                {max = max |  data[iIn + i];}
           else {max = max | -data[iIn + i];}}

     /*----- determine the number of bits needed -----*/
     /*          (2*(max-1)  < 2**nBits)              */

     if(max > 127) 
       {if(max > 2047) 
          {if(max > 8191)
               {if(max > 16383) {nBits = 16;}
                else            {nBits = 15;}}
           else{if(max > 4095)  {nBits = 14;}
                else            {nBits = 13;}}}
        else
          {if(max > 511)
               {if(max > 1023)  {nBits = 12;}
                else            {nBits = 11;}}
           else{if(max > 255)   {nBits = 10;}
                else            {nBits = 9;}}}}
     else
       {if(max > 7) 
          {if(max > 31)
               {if(max > 63)    {nBits = 8;}
                else            {nBits = 7;}}
           else{if(max > 15)    {nBits = 6;}
                else            {nBits = 5;}}}
        else
          {if(max > 1)
               {if(max > 3)     {nBits = 4;}
                else            {nBits = 3;}}
           else{if(max > 0)     {nBits = 2;}
                else            {nBits = 1;}}}}

       /*---- encode the data size ----------*/
       /*---- we store nBits - 1 in 4 bits---*/

     out[iOut] = out[iOut] | ((nBits-1) << pOut);
     if(pOut > 12)
        {iOut ++;
         if(iOut >= maxOut) return(-1);
        pOut = pOut - 16;
        out[iOut] = (nBits-1) >> - pOut;}
     pOut  = pOut +4;

       /*--- encode the data itself ------*/

     if(nBits > 1)
       {limit = 16 - nBits;
        for(i = 0; i<bSize; i++)
          {uData = (unsigned short) (data[iIn+i] + wMax[nBits]);

           out[iOut] = out[iOut] | (uData << pOut);
           if(pOut > limit)
             {iOut ++;
              if(iOut >= maxOut) return(-1);
              pOut = pOut - 16;
              out[iOut] = uData >> -pOut;}
           pOut  = pOut + nBits;
        }  } 


     /*-------- increase pointer -------*/
  
     iIn = iIn + bSize;}

  *compL = 2*(iOut+1);

return(0);}

//-----------------------------------------------------------------------------
//
//: Expand zero suppressed data.
//
// Borrowed from FrameL.c, v3r85
//
void Vect::FrVectZExpand(short *out, 
			 unsigned short *data, 
			 int nData)
/*--------------------------------------------------------------------*/
{
  union {
    INT_4U	u_int;
    INT_2U	u_short[2];
  } iAlpha, iBuf;

 unsigned short nBits, pIn, uData;
 unsigned short wMax[] = {0,0,1,3,7,0xf,0x1f,0x3f,0x7f,0xff,0x1ff,
                      0x3ff,0x7ff,0xfff,0x1fff,0x3fff,0x7fff};
 unsigned short mask[] = {0,1,3,7,0xf,0x1f,0x3f,0x7f,0xff,0x1ff,
                      0x3ff,0x7ff,0xfff,0x1fff,0x3fff,0x7fff,0xffff};
 int  i, iIn, iOut, bSize;

 /*----- check which indian --------*/

 iAlpha.u_int = 0x12345678;

 /*---- retrieve the bloc size -----*/

 bSize = data[0];
 iIn = 1;
 pIn = 0;

 /*--------- retrieve the data  --------------*/

 iOut = 0;
 do
    {/*-------- extract nBits -------------
        -----(we check if data is in 1 or 2 words) ------*/
        
    if(pIn <= 12)
       {uData = data[iIn] >> pIn;
        pIn = pIn + 4;}
     else
       {iBuf.u_short[0] = data[iIn];
        iIn++;
        if(iAlpha.u_short[0] == 0x1234)
             {iBuf.u_short[1] = iBuf.u_short[0];
              iBuf.u_short[0] = data[iIn];}
        else {iBuf.u_short[1] = data[iIn];}
        uData = iBuf.u_int >> pIn;
        pIn = pIn - 12;}

    nBits = 1 + (mask[4] & uData);
    if(nBits == 1) nBits = 0;

     /*----extract data ----------*/

     for(i=0; i<bSize; i++)
       {if(iOut >= nData) break;

        if(pIn + nBits <= 16)
          {uData = data[iIn] >> pIn;
           pIn = pIn + nBits;}
        else
          {iBuf.u_short[0] = data[iIn];
           iIn++;
           if(iAlpha.u_short[0] == 0x1234)
                {iBuf.u_short[1] = iBuf.u_short[0];
                 iBuf.u_short[0] = data[iIn];}
           else {iBuf.u_short[1] = data[iIn];}  
           uData = iBuf.u_int >> pIn;
           pIn = pIn + nBits - 16;}

        out[iOut] = (mask[nBits] & uData) - wMax[nBits];
        iOut++;}}
 while(iOut<nData);

return;}



//-----------------------------------------------------------------------------
//
//: Expand compressed data.
//
// In general, uncompressing follows three steps:
//
// 1) Gunzip the data.
// 2) Fix byte-ordering.
// 3) Integrate.
//
// Depending upon the compression type and byte-ordering differences, not all
// of these steps are needed.
//
void Vect::expand()
{
    // Current compression level:
    Compression compress_now( getCompress() );
   
    // If it's not compressed, then it might need to be byte swapped   
    if( compress_now == NONE )
      {
	if ( dataNeedsByteSwap ) 
	  {
	    byteSwapData();
	    dataNeedsByteSwap = false;
	  }
	return;
      }

    // Calculate the length of the uncompressed data.
    uLongf length = getTypeSize( getType() ) * getNData();

    // See if the data should be z-uncompressed (zero suppress method)
    // for short data only
    //
    if( ( compress_now == DIFF_ZERO_SUPPRESS_SHORT_DIFF_GZIP_OTHER ||
	  compress_now == DIFF_ZERO_SUPPRESS_SHORT ) &&
        ( getType() == FR_VECT_2S || getType() == FR_VECT_2U ) )
      {
        // This may throw std::bad_alloc
        unsigned char* buffer = new unsigned char[ length ];
        FrVectZExpand(( short * ) buffer, (short unsigned int *) mData, (int) getNData() );

        // Assign the data
        delete mData;
        mData = buffer;
        mNBytes = length;
      }
    else

    // Determine if we need to gunzip the data.
    if ( compress_now == GZIP ||
         compress_now == GZIP_DIFF || 
         ( compress_now == DIFF_ZERO_SUPPRESS_SHORT_DIFF_GZIP_OTHER &&
           ( getType() != FR_VECT_2S || getType() != FR_VECT_2U ) ) )
    {
        // This may throw std::bad_alloc
        unsigned char* buffer = new unsigned char[ length ];
        uLongf uncompressedLength;
        int error = uncompress( buffer, &uncompressedLength, mData, mNBytes );

        // Check the return status from the uncompress
        if ( error != Z_OK )
        {
            delete[] buffer;
            buffer = 0;
            
            switch( error )
            {
                case Z_MEM_ERROR:
                    throw std::bad_alloc();
                    break;
                    
                case Z_BUF_ERROR:
                case Z_DATA_ERROR:
                    throw format_error( "Compressed data is corrupt." );
                    break;
                    
                default:
                    throw unknown_error( "An unknown error occurred during "
                                         "uncompression." );
            }
        }

        // Make sure we uncompressed the correct number of bytes.
        if ( uncompressedLength != length )
        {
	    delete[] buffer;
            buffer = 0;
   
            throw format_error( "Compressed data is corrupt" );
        }
        
        // Assign the data
        delete mData;
        mData = buffer;
        mNBytes = length;
    }

    // Determine if we need to uncompress with bzip2.
    if( compress_now == BZIP2 )	   
    {
#if HAVE_LIBBZ2
        unsigned char* buffer = new unsigned char[ length ];
        unsigned int uncompressedLength;
  
	int res =
#if defined(HAVE_NEW_LIBZ2)
		BZ2_bzBuffToBuffDecompress
#else
		bzBuffToBuffDecompress
#endif
					( (char *)buffer,
					   &uncompressedLength,
					   (char *)mData,
					   mNBytes,
					   0, /* small */
					   0 /* verbosity */ );

        // Check the return status from the uncompress
	if ( res != BZ_OK )
        {
            delete[] buffer;
            buffer = 0;
            
            switch( res )
	      {
	      case BZ_MEM_ERROR:
		throw std::bad_alloc();
		break;
		
	      case BZ_DATA_ERROR:       // if a data integrity error was detected in the compressed data
	      case BZ_DATA_ERROR_MAGIC: // if the compressed data doesn't begin with the right magic bytes
	      case BZ_UNEXPECTED_EOF:   // if the compressed data ends unexpectedly
		throw format_error( "Compressed bzip2 data is corrupt." );
		break;
		
	      default:
		throw unknown_error( "An unknown error occurred during bzip2 uncompression." );
	      }
        }

        // Make sure we uncompressed the correct number of bytes.
        if ( uncompressedLength != length )
        {
	    delete[] buffer;
            buffer = 0;
            throw format_error( "Compressed data was corrupt" );
        }
        
        // Assign the data
        delete mData;
        mData = buffer;
        mNBytes = length;
#else /* HAVE_LIBBZ2 */
	throw format_error( "BZip2 decompression not supported" );
#endif /* HAVE_LIBBZ2 */
    }
    
    // Reverse the byte order if necessary
#   ifdef WORDS_BIGENDIAN
    if ( getTypeSize(getType()) > 1 && ( mCompress > 255 ) )
#   else
    if ( getTypeSize(getType()) > 1 && ( mCompress < 256 ) )
#   endif
    {
	switch (getType()) {
	case FR_VECT_8C:
        	reverse<4>( mData, mNData*2 );
		break;
	case FR_VECT_16C:
        	reverse<8>( mData, mNData*2 );
		break;
	default:
		switch (getTypeSize(getType())) {
		case 2:
        		reverse< 2 >( mData, mNData );
			break;
		case 4:
			reverse< 4 >( mData, mNData );
                        break;
		case 8:
			reverse< 8 >( mData, mNData );
                        break;
		}
	}
   
        dataNeedsByteSwap = false;   
    }

    // Integrate the data if necessary.
    if ( compress_now == DIFF ||
         compress_now == GZIP_DIFF ||
         compress_now == DIFF_ZERO_SUPPRESS_SHORT_DIFF_GZIP_OTHER ||
	 ( compress_now == DIFF_ZERO_SUPPRESS_SHORT &&
	   ( getType() == FR_VECT_2S || getType() == FR_VECT_2U ) ) )
    {
        if ( mNData > 1 )
        {
	  switch( getType() )
	    {
	    case FR_VECT_C:
	      {
		CHAR* start = reinterpret_cast< CHAR* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_1U:
	      {
		CHAR_U* start = reinterpret_cast< CHAR_U* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_2S:
	      {
		INT_2S* start = reinterpret_cast< INT_2S* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_2U:
	      {
		INT_2U* start = reinterpret_cast< INT_2U* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_4S:
	      {
		INT_4S* start = reinterpret_cast< INT_4S* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_4U:
	      {
		INT_4U* start = reinterpret_cast< INT_4U* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_8S:
	      {
		INT_8S* start = reinterpret_cast< INT_8S* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_8U:
	      {
		INT_8U* start = reinterpret_cast< INT_8U* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_4R:
	      {
		REAL_4* start = reinterpret_cast< REAL_4* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_8R:
	      {
		REAL_8* start = reinterpret_cast< REAL_8* >( mData ) + 1;
		for ( int i = mNData - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_8C:
	      {
		REAL_4* start = reinterpret_cast< REAL_4* >( mData ) + 1;
		for ( int i = mNData * 2 - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    case FR_VECT_16C:
	      {
		REAL_8* start = reinterpret_cast< REAL_8* >( mData ) + 1;
		for ( int i = mNData * 2 - 1; i != 0; --i )
		  {
		    *start += *(start - 1);
		    ++start;
		  }
		break;
	      }
	    }
        }
    }
    
#ifdef WORDS_BIGENDIAN
    mCompress = 0;
#else
    mCompress = 0x100;
#endif
}

//----------------------------------------------------------------------------
//
//: Set the data.
//
//!param: const void* data - Data to be copied to the vector.
//
//!exc: bad alloc - Memory allocation failed.
//
void Vect::setData( const void* data )
{

    // release an existing data vector
    if (mData != 0) deleteData();

    // Set the data
    if ( data != 0 )
    {
        mData = new unsigned char[ mNBytes ];
        memcpy( mData, data, mNBytes );
	mOwnsData = true;
    }    
}

//----------------------------------------------------------------------------
//
//: Set or adopt the data.
//
//!param: const void* data - Data to be copied to the vect.
//!param: bool allocate - Whether to allocate data. Default: true.
//!param: bool owns - Whether to destruct data in the destructor. Default: true.
//
//!exc: bad alloc - Memory allocation failed.
//
void Vect::setData( void* data, bool allocate, bool owns )
{

    // release an existing data vector
    if (mData != 0) deleteData();

    mOwnsData = owns;

    // Set the data
    if ( data != 0 )
    {
        if ( allocate )
        {
            mData = new unsigned char[ mNBytes ];
            memcpy( mData, data, mNBytes );
        }
        else
        {
            mData = reinterpret_cast< unsigned char* >( data );
        }
    }    
}

//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a CHAR.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< CHAR >()
{
    return FR_VECT_C;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to an INT_2S.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type FrameCPP::Version_3_4_5::Vect::getDataType< INT_2S >()
{
    return FR_VECT_2S;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a REAL_8.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< REAL_8 >()
{
    return FR_VECT_8R;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a REAL_4.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< REAL_4 >()
{
    return FR_VECT_4R;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to an INT_4S.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< INT_4S >()
{
    return FR_VECT_4S;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a INT_8S.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< INT_8S >()
{
    return FR_VECT_8S;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a COMPLEX_8.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type FrameCPP::Version_3_4_5::Vect::getDataType< COMPLEX_8 >()
{
    return FR_VECT_8C;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a COMPLEX_16.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< COMPLEX_16 >()
{
    return FR_VECT_16C;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a INT_2U.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< INT_2U >()
{
    return FR_VECT_2U;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a INT_4U.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< INT_4U >()
{
    return FR_VECT_4U;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a INT_8U.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< INT_8U >()
{
    return FR_VECT_8U;
}


//-----------------------------------------------------------------------------
//
//: Template specialization to get the type ID corresponding to a CHAR_U.
//
//!return: Type
//
template<> FrameCPP::Version_3_4_5::Vect::Type
FrameCPP::Version_3_4_5::Vect::getDataType< CHAR_U >()
{
    return FR_VECT_1U;
}


//-----------------------------------------------------------------------------
//
//: Equal comparison operator.
//
// Determines whether two Vect objects are equal. For this comparison,
// all of the data is compared.
//
//!param: const Vect& v - The object with which to compare.
//
//!return: bool - True if the objects are equal.
//
//!exc: None.      
bool Vect::operator==( const Vect& v ) const 
{
    bool eq = true;
    
    if ( this != &v )
    {
        eq = ( ( mName == v.mName ) && ( mCompress == v.mCompress ) &&
                ( mType == v.mType ) && ( mNData == v.mNData ) &&
                ( mNBytes == v.mNBytes ) && ( mNDim == v.mNDim ) &&
                ( mDimension == v.mDimension ) && ( mUnitY == v.mUnitY ) );

        if ( eq )
        {
            eq = ( memcmp( mData, v.mData, mNBytes ) == 0 );
        }
    }

    return eq;
}


//!exc: std::bad_alloc - Memory allocation failed.   
template<>
Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   const COMPLEX_8* data, const std::string& unitY )
   : mName( name ), 
     mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
      ),
      mType( FR_VECT_8C ), mNData( 1 ),
      mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
      mUnitY( unitY ), mOwnsData( true ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData
        for ( size_t i = 0; i < nDim; i++ )
        {
            mNData *= dims[ i ].getNx();
        }

        // add the dimension information
        for ( size_t i = 0; i < mNDim; ++i )
        {
            mDimension.push_back( dims[ i ] );
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

    // Allocate memory for the data
    const INT_4U real4_ndata( mNData * 2 ); 
    REAL_4* buf( new REAL_4[ real4_ndata ] );
   
    // if data was provided then copy it
    if ( data != 0 )
    {
        const COMPLEX_8* data_elem( data );
        // There's at least one array element
        const COMPLEX_8* const data_end( data + mNData - 1 ); 

        REAL_4* buf_elem( buf );
        while( data_elem <= data_end ) 
        {
           *buf_elem = data_elem->real();
           ++buf_elem;
           *buf_elem = data_elem->imag();
           ++buf_elem;
           ++data_elem;
        }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );   
}

   
//!exc: std::bad_alloc - Memory allocation failed.      
template<>
Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   const COMPLEX_16* data, const std::string& unitY )
  : mName( name ), 
    mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
     ),
     mType( FR_VECT_16C ), mNData( 1 ),
     mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
     mUnitY( unitY ), mOwnsData( true ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData
        for ( size_t i = 0; i < nDim; i++ )
        {
            mNData *= dims[ i ].getNx();
        }

        // add the dimension information
        for ( size_t i = 0; i < mNDim; ++i )
        {
            mDimension.push_back( dims[ i ] );
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

    // Allocate memory for the data
    const INT_4U real8_ndata( mNData * 2 ); 
    REAL_8* buf( new REAL_8[ real8_ndata ] );
    
    // if data was provided then copy it
    if ( data != 0 )
    {
        const COMPLEX_16* data_elem( data );
        // There's at least one array element
        const COMPLEX_16* const data_end( data + mNData - 1 ); 

        REAL_8* buf_elem( buf );
        while( data_elem <= data_end ) 
        {
           *buf_elem = data_elem->real();
           ++buf_elem;
           *buf_elem = data_elem->imag();
           ++buf_elem;
           ++data_elem;
        }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );   
}
   

//!exc: std::bad_alloc - Memory allocation failed.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_8> constructor argument: allocate = false.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_8> constructor argument: owns = false.      
template<>
Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   COMPLEX_8* data, const std::string& unitY, bool allocate, bool owns )
   : mName( name ), 
     mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
      ),
      mType( FR_VECT_8C ), mNData( 1 ),
      mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
      mUnitY( unitY ), mOwnsData( owns ), dataNeedsByteSwap( false )
{
    if( allocate == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_8> constructor argument: allocate = false" ); 
    }
    if( owns == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_8> constructor argument: owns = false" ); 
    }   
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData &
        // add the dimension information
        for ( size_t i = 0; i < nDim; i++ )
        {
            mDimension.push_back( dims[ i ] );
            mNData *= dims[ i ].getNx();
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

    // Allocate memory for the data
    const INT_4U real4_ndata( mNData * 2 );    
    REAL_4* buf( new REAL_4[ real4_ndata ] );

    // if data was provided then copy it
    if ( data != 0 )
    {
       const COMPLEX_8* data_elem( data );
       // There's at least one array element
       const COMPLEX_8* const data_end( data + mNData - 1 ); 

       REAL_4* buf_elem( buf );
       while( data_elem <= data_end ) 
       {
          *buf_elem = data_elem->real();
          ++buf_elem;
          *buf_elem = data_elem->imag();
          ++buf_elem;
          ++data_elem;
       }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );   
}

   
//!exc: std::bad_alloc - Memory allocation failed.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_16> constructor argument: allocate = false.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_16> constructor argument: owns = false.         
template<>
Vect::Vect( const std::string& name, INT_4U nDim, const Dimension* dims,
   COMPLEX_16* data, const std::string& unitY, bool allocate, bool owns )
   : mName( name ), 
     mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
      ),
      mType( FR_VECT_16C ), mNData( 1 ),
      mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
      mUnitY( unitY ), mOwnsData( owns ), dataNeedsByteSwap( false )
{
    if( allocate == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // an invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_16> constructor argument: allocate = false");
    }
    if( owns == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // an invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_16> constructor argument: owns = false");
    }   
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData &
        // add the dimension information
        for ( size_t i = 0; i < nDim; i++ )
        {
            mDimension.push_back( dims[ i ] );
            mNData *= dims[ i ].getNx();
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

    // Allocate memory for the data
    const INT_4U real8_ndata( mNData * 2 );    
    REAL_8* buf( new REAL_8[ real8_ndata ] );

    // if data was provided then copy it
    if ( data != 0 )
    {
       const COMPLEX_16* data_elem( data );
       // There's at least one array element
       const COMPLEX_16* const data_end( data + mNData - 1 ); 

       REAL_8* buf_elem( buf );
       while( data_elem <= data_end ) 
       {
          *buf_elem = data_elem->real();
          ++buf_elem;
          *buf_elem = data_elem->imag();
          ++buf_elem;
          ++data_elem;
       }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );   
}
   

//!exc: std::bad_alloc - Memory allocation failed.   
template<>
Vect::Vect(
    const std::string& name, INT_4U nDim, const INT_4U* dims,
    const COMPLEX_8* data, const std::string& unitY ) 
        : mName( name ),
	  mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( FR_VECT_8C ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( true ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData
        for ( size_t i = 0; i < nDim; i++ )
        {
            mNData *= dims[ i ];
        }

        // add the dimension information
        for ( size_t i = 0; i < mNDim; ++i )
        {
            mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

     // Allocate memory for the data
    const INT_4U real4_ndata( mNData * 2 ); 
    REAL_4* buf( new REAL_4[ real4_ndata ] );
   
    // if data was provided then copy it
    if ( data != 0 )
    {
        const COMPLEX_8* data_elem( data );
        // There's at least one array element
        const COMPLEX_8* const data_end( data + mNData - 1 ); 

        REAL_4* buf_elem( buf );
        while( data_elem <= data_end ) 
        {
           *buf_elem = data_elem->real();
           ++buf_elem;
           *buf_elem = data_elem->imag();
           ++buf_elem;
           ++data_elem;
        }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );      
}

   
//!exc: std::bad_alloc - Memory allocation failed.   
template<>
Vect::Vect(
    const std::string& name, INT_4U nDim, const INT_4U* dims,
    const COMPLEX_16* data, const std::string& unitY )
        : mName( name ),
	  mCompress( 
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( FR_VECT_16C ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( true ), dataNeedsByteSwap( false )
{
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData
        for ( size_t i = 0; i < nDim; i++ )
        {
            mNData *= dims[ i ];
        }

        // add the dimension information
        for ( size_t i = 0; i < mNDim; ++i )
        {
            mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

     // Allocate memory for the data
    const INT_4U real8_ndata( mNData * 2 ); 
    REAL_8* buf( new REAL_8[ real8_ndata ] );
   
    // if data was provided then copy it
    if ( data != 0 )
    {
        const COMPLEX_16* data_elem( data );
        // There's at least one array element
        const COMPLEX_16* const data_end( data + mNData - 1 ); 

        REAL_8* buf_elem( buf );
        while( data_elem <= data_end ) 
        {
           *buf_elem = data_elem->real();
           ++buf_elem;
           *buf_elem = data_elem->imag();
           ++buf_elem;
           ++data_elem;
        }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );      
}
   

//!exc: std::bad_alloc - Memory allocation failed.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_8> constructor argument: allocate = false.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_8> constructor argument: owns = false.            
template<>
Vect::Vect(
    const std::string& name, INT_4U nDim, const INT_4U* dims,
    COMPLEX_8* data, const std::string& unitY, bool allocate, bool owns )
        : mName( name ), 
	  mCompress(
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( FR_VECT_8C ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( owns ), dataNeedsByteSwap( false )
{
    if( allocate == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // an invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_8> constructor argument: allocate = false" ); 
    }   
    if( owns == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // an invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_8> constructor argument: owns = false" ); 
    }      
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData &
        // add the dimension information
        for ( size_t i = 0; i < nDim; i++ )
        {
            mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
            mNData *= dims[ i ];
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

    // Allocate memory for the data
    const INT_4U real4_ndata( mNData * 2 );    
    REAL_4* buf( new REAL_4[ real4_ndata ] );

    // if data was provided then copy it
    if ( data != 0 )
    {
       const COMPLEX_8* data_elem( data );
       // There's at least one array element
       const COMPLEX_8* const data_end( data + mNData - 1 ); 

       REAL_4* buf_elem( buf );
       while( data_elem <= data_end ) 
       {
          *buf_elem = data_elem->real();
          ++buf_elem;
          *buf_elem = data_elem->imag();
          ++buf_elem;
          ++data_elem;
       }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );   
}


//!exc: std::bad_alloc - Memory allocation failed.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_16> constructor argument: allocate = false.   
//!exc: std::runtime_error - Invalid Vect<COMPLEX_16> constructor argument: owns = false.               
template<>
Vect::Vect(
    const std::string& name, INT_4U nDim, const INT_4U* dims,
    COMPLEX_16* data, const std::string& unitY, bool allocate, bool owns )
        : mName( name ), 
	  mCompress(
#ifdef WORDS_BIGENDIAN
		0
#else
		256
#endif
	  ),
          mType( FR_VECT_16C ), mNData( 1 ),
          mNBytes( 0 ), mData( 0 ), mNDim( nDim ), mDimension(),
          mUnitY( unitY ), mOwnsData( owns ), dataNeedsByteSwap( false )
{
    if( allocate == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // an invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_16> constructor argument: allocate = false" ); 
    }   
    if( owns == false )
    {
       // In case of complex datatype, need to allocate data buffer always:
       // data buffer can't store objects ==> it need to store data of these
       // objects   
   
       // an invalid argument value
       throw std::runtime_error( "Invalid Vect<COMPLEX_16> constructor argument: owns = false" ); 
    }      
    // if dims is specified, copy them
    if ( dims != 0 )
    {
        // calculate mNData &
        // add the dimension information
        for ( size_t i = 0; i < nDim; i++ )
        {
            mDimension.push_back( Dimension( dims[ i ], 1.0, "" ) );
            mNData *= dims[ i ];
        }
    }
    // otherwise assume that there is just one data point
    else
    {
        mNData = 1;

        for( size_t i = 0; i < nDim; ++i )
        {
            // Dimension size = 1, scale factor = 1.0, no units
            mDimension.push_back( Dimension( 1, 1.0, "" ) );
        }
    }

    // calculate number of bytes
    mNBytes = mNData * getTypeSize( mType );

    // Allocate memory for the data
    const INT_4U real8_ndata( mNData * 2 );    
    REAL_8* buf( new REAL_8[ real8_ndata ] );

    // if data was provided then copy it
    if ( data != 0 )
    {
       const COMPLEX_16* data_elem( data );
       // There's at least one array element
       const COMPLEX_16* const data_end( data + mNData - 1 ); 

       REAL_8* buf_elem( buf );
       while( data_elem <= data_end ) 
       {
          *buf_elem = data_elem->real();
          ++buf_elem;
          *buf_elem = data_elem->imag();
          ++buf_elem;
          ++data_elem;
       }
    }
   
    mData = reinterpret_cast< unsigned char* >( buf );   
}
   



//-----------------------------------------------------------------------------
//
//: Gets the class ID.
//
//!return: INT_2U -- Vector class ID.
//
//!exc: None.   
INT_2U Vect::getClassId() const 
{
    return ID_VECT;
}

   
