#include "framecpp/config.h"

#include <algorithm>
#include <cassert>
#include <sstream>


#include "framecpp/Common/MD5Sum.hh"
#include "framecpp/Version6/Dictionary.hh"
#include "framecpp/Version6/IFrameStream.hh"
#if 1
#include "framecpp/Version6/OFrameStream.hh"
#endif /* 0 */
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/FrSE.hh"

#define	DEBUG_MD5SUM	0
#define	LM_DEBUG 	0

#if LM_DEBUG
#define MESSAGE( a ) std::cerr << a << __FILE__ << " " << __LINE__ << std::endl
#else
#define MESSAGE( a )
#endif
#define AT( ) MESSAGE ( "" )

using namespace ::FrameCPP::Version_6;
using FrameCPP::Common::MD5Sum;

namespace
{
  inline bool
  class_id_range_check( INT_4U ClassId )
  {
    // Check if the ClassID is out of range
    return ( ClassId < FrameCPP::Version_6::CLASS_LAST_ENTRY );
  }

  inline void
  validate_stream_position( INT_8U Current, INT_8U Expected)
  {
    if ( Current != Expected )
    {
      std::ostringstream oss;

      oss << "Invalid file position: " << Current
	  << " (Expected: " << Expected << ")";
#if LM_DEBUG
      std::cerr << oss.str( ) << std::endl;
      assert( 0 );
#else
      throw std::runtime_error( oss.str( ) );
#endif
    }
  }
} // anonymous namespace

#if 1
// This functor allows searching of callbacks to find the one that should be
// removed from the list.
class Dictionary::key_is
{
public:
  //: construtor
  key_is( Dictionary::key_id_type Key );
  //: Routine to match source to key.
  //
  // If the Source has the same key registered with this objects creation,
  // then return true.
  //
  //!return: If the Source object has the same key as m_key, return true,
  //+	false otherwise.
  bool operator()( Dictionary::callback_type Source );
private:
  //: default constructor - not allowed.
  key_is( );

  //: Key to find
  Dictionary::key_id_type	m_key;
};

inline Dictionary::key_is::
key_is( Dictionary::key_id_type Key )
  : m_key( Key )
{
}

inline bool Dictionary::key_is::
operator()( Dictionary::callback_type Source )
{
  return Source.first == m_key;
}

Dictionary::element_type::
element_type( const std::string& Name,
	   const std::string& ClassType,
	   const std::string& Comment )
  : name( Name ),
    classType( ClassType ),
    comment( Comment )
{
}

Dictionary::definition_type::
definition_type( const std::string& Name,
   INT_2U ClassId, const std::string& Comment )
  : name( Name ),
    classId( ClassId ),
    comment( Comment )
{
}

Dictionary::definition_type::
~definition_type( )
{
}

Dictionary::io_calls_type::
io_calls_type( const std::string& Name,
	    INT_2U ClassId,
	    read_func_type Reader,
	    write_func_type Writer )
  : name( Name ),
    classId( ClassId ),
    read( Reader ),
    write( Writer )
{}


// The default constructor initializes the class to no reader, no writer and 
// disabled
Dictionary::library_io::
library_io( )
  : m_enabled( false ),
    m_read( (read_func_type)0 ),
    m_write((write_func_type)0 )
{
}

Dictionary::
Dictionary( const std::vector< Dictionary::io_calls_type >& LibraryDefs,
	    INT_4U NumberOfClassTypes )
  : m_library_defs( CLASS_LAST_ENTRY ), // Allocate space for max entries
    m_reference_info( NumberOfClassTypes, *this )
{
  INT_4U	classid;

  // Go through list of supplied routines and add them to m_library_defs
  for( std::vector< Dictionary::io_calls_type >::const_iterator
	 ld( LibraryDefs.begin( ) ),
	 end( LibraryDefs.end( ) );
       ld != end;
       ++ld )
  {
    classid = (*ld).classId;

    if ( class_id_range_check( classid ) == false )
    {
      std::ostringstream	err;
      err << "Unable to initialize the Dictionary because the value "
	  << classid << " exceeds the maximum class id of "
	  << CLASS_LAST_ENTRY;
      throw std::range_error( err.str( ) );
    }
    //:TODO: Should change m_enabled to false if no read routine supplied
    m_library_defs[ classid ].m_enabled = true;
    m_library_defs[ classid ].m_read = (*ld).read;
    m_library_defs[ classid ].m_write = (*ld).write;
    m_library_map[ (*ld).name ] = classid;
  }
  // Add entries for FrSH and FrSE since they will never be defined in file
  AddDefinition( new definition_type( FrSH::getStaticName(),
				      FrBase::CLASS_FR_SH,
				      "Frame Structure Header" ) );
  AddDefinition( new definition_type( FrSE::getStaticName(),
  				      FrBase::CLASS_FR_SE,
				      "Frame Structure Element" ) );
}

Dictionary::
~Dictionary( )
{
  for ( HASH_MAP_NAMESPACE::hash_map< INT_2U, const definition_type* >::iterator
	  sd( m_stream_defs.begin( ) );
	sd != m_stream_defs.end( );
	++sd )
  {
    delete (*sd).second;
    (*sd).second = (const definition_type*)NULL;
  }

  m_stream_defs.clear();
}

const Dictionary::callback_key_type Dictionary::
AddCallback( INT_2U ClassId, Callback& Callback )
{
  // Get key to uniquely identify this record
  callback_key_type	ret( ClassId, key_checkout( ) );
  // Put callback into list of callbacks for the function
  m_callbacks[ ClassId ].push_back( callback_type( ret.m_key_id, &Callback ) );

  return ret;
}

void Dictionary::
AddDefinition( const definition_type* Definition )
{
  if ( m_stream_defs.find( Definition->classId ) == m_stream_defs.end( ) )
  {
    // Update list of definitions
    m_stream_defs[ Definition->classId ] = Definition;
    const library_map_type::const_iterator
      lmi( m_library_map.find( Definition->name ) );
    if ( lmi != m_library_map.end( ) )
    {
      // Update stream to lib mapping
      m_stream_to_lib[ Definition->classId ] = (*lmi).second;
      // Update lib to stream mapping
      m_lib_to_stream[ (*lmi).second ] = Definition->classId;
    }
  }
  else
  {
    // Duplicate entry. Quietly ignore and release space.
    delete Definition;
  }
}

void Dictionary::
DisableLibraryReaderAll( )
{
  for ( library_defs_type::iterator l( m_library_defs.begin( ) );
	l != m_library_defs.end( );
	l++ )
  {
    (*l).m_enabled = false;
  }
}

void Dictionary::
EnableLibraryReader( INT_4U ClassId )
{
  if ( (class_id_range_check( ClassId ) ) &&
       ( m_library_defs[ ClassId ].m_read ) )
  {
    m_library_defs[ ClassId ].m_enabled = true;
  }
}

void Dictionary::
EnableLibraryReaderAll( )
{
  for ( library_defs_type::iterator l( m_library_defs.begin( ) );
	l != m_library_defs.end( );
	l++ )
  {
    // Only set if there is a read and write routine defined
    if ( (*l).m_read )
    {
      (*l).m_enabled = true;
    }
  }
}

INT_2U Dictionary::
GetLibraryClass( const std::string& Name ) const
{
  library_map_type::const_iterator c( m_library_map.find( Name ) );
  if ( c == m_library_map.end( ) )
  {
    return FrBase::CLASS_NULL;
  }
  return (*c).second;
}

bool Dictionary::
IsReadable( IFrameStream& Stream, const FrBase& Base ) const
{
  const id_map_type::const_iterator
    lib_id( m_stream_to_lib.find( Base.GetClassId( ) ) );
  if ( lib_id != m_stream_to_lib.end( ) )
  {
    // Check if the classid is valid and enabled
    INT_4U	classid( (*lib_id).second );
    if ( ( class_id_range_check( classid ) ) &&
	 ( m_library_defs[ classid ].m_enabled ) )
    {
      return true;
    }
  }
  return false;
}

INT_2U Dictionary::
MapStreamClassToLibraryClass( INT_2U StreamClass ) const
{
  INT_2U	ret( FrBase::CLASS_NULL );

  AT( );
  id_map_type::const_iterator	p( m_stream_to_lib.find( StreamClass ) );
  if ( p != m_stream_to_lib.end( ) )
  {
    AT( );
    ret = (*p).second;
  }

  AT( );
  return ret;
}

FrBase* Dictionary::
Read( IFrameStream& Stream, const FrBase& Base )
try
{
  if ( Base.GetLength( ) < Base.Bytes( Stream.GetVersion( ) ) )
  {
    throw FrBase::TooSmallException( Base, Stream.GetVersion( ) );
  }
  MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " << Base << " " );
  INT_8U start_pos( (INT_8U)( Stream.tellg( ) ) );
  INT_8U dest_pos( start_pos + Base.Skip( Stream ) );
  const id_map_type::const_iterator
    lib_id( m_stream_to_lib.find( Base.GetClassId( ) ) );
   
  if ( ( lib_id == m_stream_to_lib.end( ) ) ||
       ( class_id_range_check( (*lib_id).second ) == false ) ||
       ( m_library_defs[ (*lib_id).second ].m_enabled == false ) )
  {
    MESSAGE( "Check: lib_id == m_stream_to_lib.end( ): " \
	     << (lib_id == m_stream_to_lib.end( ) ) );
    if ( lib_id != m_stream_to_lib.end( ) )
    {
      MESSAGE( "Check: class_id_range_check( (*lib_id).second ) == false: " \
	       << (class_id_range_check( (*lib_id).second ) == false) );
      MESSAGE( "Check: m_library_defs[ (*lib_id).second ].m_enabled == false: " \
	       << (m_library_defs[ (*lib_id).second ].m_enabled == false) );
    }
      
    Stream.SkipObject( Base );
    MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
    validate_stream_position( Stream.tellg( ), dest_pos );
    return (FrBase*)NULL;
  }
  //---------------------------------------------------------------------
  // Use application supplied callback for reading in of object
  //---------------------------------------------------------------------
  FrBase* base( (FrBase*)NULL );

  try
  {
    do
    {
      MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
      Stream.SetReadBoundry( Stream.tellg( ) + Base.Skip( Stream ) );
   
      INT_4U	classid( (*lib_id).second );

      if ( class_id_range_check( classid ) == false )
      {
         std::ostringstream msg;
         msg << "Dictionary::Read(): missing m_library_defs entry for class id="
             << Base.GetClassId();
         throw std::runtime_error( msg.str() );
      }
   
      base = (*(m_library_defs[ classid ].m_read))( Stream, Base );
    }
    while ( (  base == (FrBase*)NULL ) &&
	    ( Stream.good( ) ) && 
	    ( start_pos == Stream.tellg( ) ) );
  }
  catch( ... )
  {
    Stream.SetReadBoundry( );
    throw;
  }
  Stream.SetReadBoundry( );
  MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
  validate_stream_position( Stream.tellg( ), dest_pos );
  MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
  Stream.m_length_of_last_object = Base.GetLength( );
  //---------------------------------------------------------------------
  // Keep track of this object when needing to link
  //---------------------------------------------------------------------
  if ( base )
  {
    MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
    m_reference_info.AddObject( base );
  }
  //---------------------------------------------------------------------
  // Try to find application registered callback for the object
  //---------------------------------------------------------------------
  MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
  callback_map_type::const_iterator
    cbl( m_callbacks.find( base->GetClassId( ) ) );
  if ( cbl != ( m_callbacks.end( ) ) )
  {
    for( callback_list_type::const_iterator i( (*cbl).second.begin( ) );
	 i != (*cbl).second.end( );
	 i++ )
    {
      (*i).second->operator()( *base );
    }
  }
  //---------------------------------------------------------------------
  // Return object
  //---------------------------------------------------------------------
  MESSAGE( "Dictionary::Read: Pos: " << Stream.tellg( ) << " " );
  return base;
}
catch( std::exception& e )
{
  std::ostringstream		oss;
  oss << "Error reading object: ";
  stream_defs_type::iterator
    def( m_stream_defs.find( Base.GetClassId( ) ) );

  if ( def != m_stream_defs.end( ) )
  {
    oss << (*def).second->name;
  }
  else
  {
    oss << Base.GetClassId( );
  }
  oss << " - " << Base.GetInstance( );
  oss << ": ";
  Rethrow( e, oss.str( ) );
  return (FrBase*)NULL;
}

void Dictionary::
RemoveCallback( const callback_key_type Key )
{
  //---------------------------------------------------------------------
  // locate the position of the callback associated with the key
  //---------------------------------------------------------------------
  callback_list_type::iterator
    p( std::find_if( m_callbacks[ Key.m_class_id ].begin( ),
		     m_callbacks[ Key.m_class_id ].end( ),
		     key_is( Key.m_key_id ) ) );
  //---------------------------------------------------------------------
  // Remove the callback
  //---------------------------------------------------------------------
  m_callbacks[ Key.m_class_id ].erase( p );
  //---------------------------------------------------------------------
  // Release the key
  //---------------------------------------------------------------------
  key_checkin( Key.m_key_id );
}

void Dictionary::
Write( OFrameStream& Stream, FrBase& Base )
{
#if LM_DEBUG
  std::cerr << "DEBUG: Dictionary::Write: Writing: "
	    << Base.GetClassId( )
	    << "-"
	    << Base.GetInstance( )
	    << std::endl
    ;
#endif /* LM_DEBUG */
  //---------------------------------------------------------------------
  // See if the description for this object has been written to the
  //    stream.
  //---------------------------------------------------------------------
  const described_instance_type::const_iterator
    ici( m_described_instance.find( Base.GetClassId( ) ) );
  if ( ici == m_described_instance.end( ) )
  {
    //-------------------------------------------------------------------
    // Write the description of the object to the stream
    //-------------------------------------------------------------------
    Base.WriteDescription( Stream );
    m_described_instance.insert( Base.GetClassId( ) );
  }
  //---------------------------------------------------------------------
  // Look up and evaluate the callback routine
  //---------------------------------------------------------------------
  callback_map_type::const_iterator
    cbl( m_callbacks.find( Base.GetClassId( ) ) );
  if ( cbl != ( m_callbacks.end( ) ) )
  {
    for( callback_list_type::const_iterator i( (*cbl).second.begin( ) );
	 i != (*cbl).second.end( );
	 i++ )
    {
      (*i).second->operator()( Base );
    }
  }
  //---------------------------------------------------------------------
  // Write the object to the stream
  //---------------------------------------------------------------------
  INT_8U	pos( Stream.tellp( ) );

  Base.FrBase::Write( Stream );
  Base.Write( Stream );
#if DEBUG_MD5SUM
      MD5Sum	md5_here( Stream.GetMD5Sum( ) );
      std::cerr << "DEBUG: pos: " << Stream.tellp( )
		<< " md5sum: " << md5_here
		<< std::endl
		<< std::flush;
#endif /* DEBUG_MD5SUM */


  if ( ( Stream.tellp( ) - pos ) != ( Base.GetLength( ) ) )
  {
#if LM_DEBUG
    std::cerr << "FATAL:"
	      << " Object: " << Base
	      << " Wrote: " << ( Stream.tellp( ) - pos )
	      << " Instead of: " << Base.GetLength( ) 
	      << std::endl;
    assert( 0 );
#else
    std::ostringstream oss;
    
    oss << Base.ErrorInfo( )
	<< " Object: " << Base
	<< " [" << GetClassName( Base.GetClassId( ) ) << "]"
	<< " Invalid number of bytes written: " << ( Stream.tellp( ) - pos )
	<< " (Expected: " << Base.GetLength( ) << ")";
    throw std::runtime_error( oss.str( ) );
#endif
  }
}

// This function is responsible for making keys uniquely available.
//
//!return: Dictionary::key_id_type - Unique key
Dictionary::key_id_type Dictionary::
key_checkout( )
{
  key_id_type	key;
  key_chain_type::iterator
    key_iter( std::find( m_key_chain.begin( ), m_key_chain.end( ), false ) );

  if ( key_iter == m_key_chain.end( ) )
  {
    // All keys currently in use, create a new one.
    m_key_chain.push_back( true );
    key = m_key_chain.size( ) - 1;
  }
  else
  {
    // Recycle a key
    *key_iter = true;
    key = std::distance( m_key_chain.begin( ), key_iter );
  }
  return key;
}

// key_checkin release the lock held by the key
//
//!param: Dictionary::key_id_type Key - Key to be returned
void Dictionary::
key_checkin( Dictionary::key_id_type Key )
{
  if ( Key >= m_key_chain.size( ) )
  {
    std::ostringstream	e;

    e << "Dictionary::key_checkin attepted to release key out range: "
      << "( " << Key << " >= " << m_key_chain.size( ) << " )";
    throw std::range_error( e.str( ) );
  }
  m_key_chain[ Key ] = false;
}

//-----------------------------------------------------------------------
// Dictionary::Callback
//-----------------------------------------------------------------------
Dictionary::Callback::
~Callback( )
{
}

Dictionary::callback_key_type::
callback_key_type( INT_4U ClassId, key_id_type KeyId )
  : m_class_id( ClassId ),
    m_key_id( KeyId )
{
}
#endif /* 0 */
