#ifndef FrameCPP_VERSION_6_Dictionary_HH
#define FrameCPP_VERSION_6_Dictionary_HH

#include <list>
#include <memory>
#include <set>
#include <string>
#include <vector>

#include "general/hash_map.hh"
#include "general/types.hh"

#include "framecpp/Version6/Reference.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class IFrameStream;
    class FrBase;
    class OFrameStream;

    class Dictionary
    {
    public:
      //: Unique key generated for callback
      //
      // For each callback that is registered with the Dictionary, a unique
      //   key is generated.
      typedef INT_2U	key_id_type;

      struct element_type
      {
	std::string	name;
	std::string	classType;
	std::string	comment;

	element_type( const std::string& Name,
		   const std::string& ClassType,
		   const std::string& Comment );
      }; // struct - element_type

      struct definition_type
      {
	std::string  			name;
	INT_2U				classId;
	std::string			comment;
	std::vector< element_type >	elements;

	definition_type( const std::string& Name,
  		         INT_2U ClassId,
		         const std::string& Comment );
	~definition_type( );

      }; // struct - definition_type

      class Callback
      {
      public:
	virtual ~Callback( );

	virtual void operator( )( FrBase& Base ) = 0;
      };

      typedef FrBase* (*read_func_type)( IFrameStream& Stream,
					 const FrBase& Base );
      typedef void (*write_func_type)( );

      //: Structure to hold io routines and state
      struct io_calls_type
      {
	//: Name of frame structure
	std::string          	name;
	//: Class Id of frame structure
	INT_2U			classId;
	//: Routine to read frame structure
	read_func_type		read;
	//: Routine to write frame structure
	write_func_type		write;

	//: Constructor
	io_calls_type( const std::string& Name,
		       INT_2U ClassId,
		       read_func_type Reader,
		       write_func_type Writer );
      };
      typedef std::pair< key_id_type, Callback* > callback_type;

    public:
      //: Representation of a key
      //
      // This class has all private types so the recipient cannot modify any
      // aspects of the key. The Dictionary class has been made a friend
      // since it is the only object that needs to do any manipulation
      // of this object.
      struct callback_key_type
      {
      private:
	friend class Dictionary;

	//: Constructor
	//
	//!param: INT_4U ClassId - The class id associated with the callback
	//!param: key_id_type KeyId - The unique id associated with the
	//+	callback
	callback_key_type( INT_4U ClassId, key_id_type KeyId );

	INT_4U		m_class_id;
	key_id_type	m_key_id;
      };

      //: Constructor
      Dictionary( const std::vector< io_calls_type >& LibraryDefs,
		  INT_4U NumberOfClassTypes );

      //: Destructor
      ~Dictionary( );

      //: Add a definition to the dictionary
      //
      // This function adds a definition to the dictionary. It is greedy in
      // that it takes ownership of the pointer past and will delete the
       // object when the dictionary 
      void AddDefinition( const definition_type* Definition );

      //: Add a callback method to the dictionary
      const callback_key_type AddCallback( INT_2U Key, Callback& CB );

      void DisableLibraryReaderAll( );

      void EnableLibraryReader( INT_4U ClassId );

      void EnableLibraryReaderAll( );

      //: Get the class id for the library based on text name
      INT_2U GetLibraryClass( const std::string& Name ) const;

      bool IsReadable( IFrameStream& Stream, const FrBase& Base ) const;

      INT_2U MapStreamClassToLibraryClass( INT_2U StreamClass ) const;

      //: Read element off the stream
      FrBase* Read( IFrameStream& Stream, const FrBase& Base );

      //: Remove a callback method from the dictionary
      void RemoveCallback( const callback_key_type Key );

      void TrackObjects( bool Tracking );

      //: Write element the stream
      void Write( OFrameStream& Stream, FrBase& Base );

    private:
      friend class IFrameStream;
      typedef std::vector< bool >	key_chain_type;
      typedef std::set< INT_2U > described_instance_type;
      typedef HASH_MAP_NAMESPACE::hash_map< INT_2U, INT_2U > id_map_type;
      typedef HASH_MAP_NAMESPACE::hash_map< std::string, INT_2U > library_map_type;
      //: Structure to hold information about callbacks
      struct library_io
      {
	//: Default constructor
	library_io( );

	//: True if the callback is enabled for reading
	bool		m_enabled;
	//: Routine to read object
	read_func_type	m_read;
	//: Routine to write object ( Currently not used )
	write_func_type	m_write;
      };
      //: Typedef to simplify programming
      typedef std::vector< struct library_io >	library_defs_type;

      typedef HASH_MAP_NAMESPACE::hash_map< INT_2U, const definition_type* > stream_defs_type;
      typedef std::list< callback_type > callback_list_type;
      typedef HASH_MAP_NAMESPACE::hash_map< INT_2U, callback_list_type> callback_map_type;

      //: key_is functor
      class key_is;

      //: Bookkeeping of keys currently in use
      //
      // If a key is checked out, then a value of true is registered at the
      // the key's index, false otherwise.
      key_chain_type	m_key_chain;
      // Map of structues provied by the library.
      library_defs_type	m_library_defs;
      // Map associating a name with a class id for the library.
      library_map_type	m_library_map;
      // Map of definitions of structures either read in or
      // written to the stream
      stream_defs_type	m_stream_defs;
      // Map of class ids from stream structures to library structures
      id_map_type		m_stream_to_lib;
      // Map of class ids from library structures to stream structures
      mutable id_map_type	m_lib_to_stream;
      // List of callbacks
      callback_map_type	m_callbacks;
      // Mapping of registered callbacks
      // registered_callbacks_type	m_registered_callbacks;
      // Associative array tracking number of instances.
      described_instance_type	m_described_instance;

      Reference		m_reference_info;

      void add_class_mapping( INT_2U StreamClass, INT_2U LibraryClass );

      //: Obtain a key from the pool
      key_id_type key_checkout( );

      //: Release key to the pool
      void key_checkin( key_id_type Key );
    };

    inline void Dictionary::
    TrackObjects( bool Tracking )
    {
      m_reference_info.SetTracking( Tracking );
    }

    inline void Dictionary::
    add_class_mapping( INT_2U StreamClass, INT_2U LibraryClass )
    {
      m_stream_to_lib[ StreamClass ] = LibraryClass;
      m_lib_to_stream[ LibraryClass ] = StreamClass;
    }
  } // namespace - Version_6
} // namespace - FrameCPP
#endif /* FrameCPP_VERSION_6_Dictionary_HH */
