#include "framecpp/config.h"

#include <cmath>
#include <sstream>
#include <stdexcept>

#include "general/util.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrHeader.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/OFrameStream.hh"

using namespace FrameCPP::Version_6;

namespace {
#define	_PI_CONST_ 3.1415926535897932384

  static const char* IGWD_ORIGINATOR = "IGWD";

  static const INT_2U x1234( 0x1234 );
  static const INT_4U x12345678( 0x12345678 );
  static const INT_8U x123456789abcdef( 0x123456789abcdefLL );

  static const REAL_4 pi_REAL_4( _PI_CONST_ );
  static const REAL_8 pi_REAL_8( _PI_CONST_ );

  template< class T >
  void check_magic_number( const unsigned char* const Source,
			   const T Ref,
			   const std::string& Bytes )
  {
    T buf;

    //-------------------------------------------------------------------
    // Load the buffer
    //-------------------------------------------------------------------
    memcpy(&buf, Source, sizeof( buf ) );
    //-------------------------------------------------------------------
    // Check the bytes
    //-------------------------------------------------------------------
    if ( Ref != buf )
    {
      //-----------------------------------------------------------------
      // Reverse the bytes and check again
      //-----------------------------------------------------------------
      reverse< sizeof( buf ) >( &buf, 1 );
      if ( Ref != buf )
      {
	std::ostringstream	err;

	err << "Data corruption detected in FrHeader at byte(s): "
	    << Bytes;
	//---------------------------------------------------------------
	// Pattern does not match either BIGENDIAN or LITTLEENDIAN so
	//   it must not be a frame file.
	//---------------------------------------------------------------
	throw std::runtime_error( err.str( ) );
      }
    }
  }

  template< class T >
  inline void
  check_pi( const T Source,
	    const T Precision,
	    const std::string& Bytes )
  {
    const T pi( _PI_CONST_ );

    //-------------------------------------------------------------------
    // Check the bytes
    //-------------------------------------------------------------------
    if ( std::abs( Source - pi ) > Precision )
    {
      std::ostringstream	err;

      err << "Value of PI out of range in FrHeader at Bytes: "
	  << Bytes;
      throw std::runtime_error( err.str( ) );
    }
  }
#undef _PI_CONST_
} // namespace - anonymous

const char* FrHeader::
GetOriginator( ) const
{
  return m_data.raw.igwd;
}

void FrHeader::
SetOriginator( const std::string& Originator )
{
  const int length( sizeof( m_data.raw.igwd ) - 1);
  strncpy( m_data.raw.igwd, Originator.c_str( ), length );
  m_data.raw.igwd[ length ] = '\0';
}

template< class T > void FrHeader::
Reorder( std::vector< T >& Data ) const
{
  if ( m_reorder_bytes )
  {
    reverse< sizeof( T ) >( &(Data[0]), Data.size( ) );
  }
}

template< class T > void FrHeader::
Reorder( T& Data ) const
{
  if ( m_reorder_bytes )
  {
    reverse< sizeof( T ) >( &Data, 1 );
  }
}

void FrHeader::
read( IFrameStream& fs )
{
  fs.read( &m_data, sizeof( m_data ) );

  if ( 0 ) // :TODO: Query the frame stream for level of validation
  {
    //-------------------------------------------------------------------
    // Check the IGWD field
    //-------------------------------------------------------------------
    if ( strcmp( IGWD_ORIGINATOR, m_data.raw.igwd ) != 0 )
    {
      std::ostringstream	err;
      
      err << "File originator is not " << IGWD_ORIGINATOR
	  << " (Bytes 0-4 of FrHeader).";
      std::runtime_error( err.str( ) );
    }
    //-------------------------------------------------------------------
    // Check the ASCII 'A' && 'Z'
    //-------------------------------------------------------------------
    if ( ( m_data.raw.AZ[0] != 'A' ) || ( m_data.raw.AZ[1] != 'Z' ) )
    {
      std::ostringstream	err;
      
      err << "File does not support IEEE ASCII hardware standard"
	  << " (Bytes 38-39 of FrHeader).";
      std::runtime_error( err.str( ) );
    }
  }
  //---------------------------------------------------------------------
  // Check the "magic number". The "magic number" is:
  //   - the three values specifying the byte order
  //---------------------------------------------------------------------

  check_magic_number( m_data.raw.x1234, x1234, "12-13" );
  check_magic_number( m_data.raw.x12345678, x12345678, "14-17" );
  try
  {
    check_magic_number( m_data.raw.x123456789abcdef, x123456789abcdef,
			"18-25" );
  }
  catch( ... )
  {
    try
    {
      //-----------------------------------------------------------------
      // This is a work around for a bug found in some frames where the
      //   least significant digits were being set to all zeros
      //-----------------------------------------------------------------
      static INT_8U	x123456700000000( 0x123456700000000LL );

      check_magic_number( m_data.raw.x123456789abcdef, x123456700000000,
			  "18-25" );
    }
    catch( ... )
    {
      //-----------------------------------------------------------------
      // This is a work around for a bug found in some frames where the
      //   most significant digits were being set to all zeros
      //-----------------------------------------------------------------
      static INT_8U	x0000000001234567( 0x0000000001234567LL );

      check_magic_number( m_data.raw.x123456789abcdef, x0000000001234567,
			  "18-25" );
    }
  }

  //---------------------------------------------------------------------
  // Figure out how to order the bytes
  //---------------------------------------------------------------------
  {
    INT_2U	x;
    memcpy(&x, m_data.raw.x1234, sizeof( x ) );
    m_reorder_bytes = ( x1234 != x );
  }
  if ( m_reorder_bytes )
  {
    if ( FrameCPP::HostByteOrder == FrameCPP::BYTE_ORDER_BIG_ENDIAN )
    {
      m_byte_order = FrameCPP::BYTE_ORDER_LITTLE_ENDIAN;
    }
    else
    {
      m_byte_order = FrameCPP::BYTE_ORDER_BIG_ENDIAN;
    }
  }
  else
  {
    m_byte_order = FrameCPP::HostByteOrder;
  }
  if ( 0 ) // :TODO: Query the frame stream for level of validation
  {
    //-------------------------------------------------------------------
    // Check the values of pi
    //-------------------------------------------------------------------
    REAL_4	pi_4;
    memcpy( &pi_4, m_data.raw.smallPi, sizeof(pi_4) );
    Reorder( pi_4 );
    check_pi( pi_4, REAL_4( 10e-6 ), "26-29" );
    REAL_8	pi_8;
    memcpy( &pi_8, m_data.raw.smallPi, sizeof(pi_8) );
    Reorder( pi_8 );
    check_pi( pi_8, REAL_8( 10e-14 ), "30-37" );
  }
}        

void FrHeader::
write( OFrameStream& fs )
{
  // Supply the system specific info for the header
  m_data.raw.igwd[ sizeof( m_data.raw.igwd ) - 1 ] = '\0';
  m_data.raw.version = DATA_FORMAT_VERSION;
  m_data.raw.minorVersion = LIBRARY_MINOR_VERSION;
  m_data.raw.sizeInt2 = (unsigned char)sizeof( INT_2U );
  m_data.raw.sizeInt4 = (unsigned char)sizeof( INT_4U );
  m_data.raw.sizeInt8 = (unsigned char)sizeof( INT_8U );
  m_data.raw.sizeReal4 = (unsigned char)sizeof( REAL_4 );
  m_data.raw.sizeReal8 = (unsigned char)sizeof( REAL_8 );
  memcpy( m_data.raw.x1234, &x1234, sizeof( x1234 ) );
  memcpy( m_data.raw.x12345678, &x12345678, sizeof( x12345678 ) );
  memcpy( m_data.raw.x123456789abcdef, &x123456789abcdef,
	  sizeof( x123456789abcdef ) );
  memcpy( m_data.raw.smallPi, &pi_REAL_4, sizeof( pi_REAL_4 ) );
  memcpy( m_data.raw.bigPi, &pi_REAL_8, sizeof( pi_REAL_8 ) );
  m_data.raw.AZ[ 0 ] = 'A';
  m_data.raw.AZ[ 1 ] = 'Z';
  // Write the header to the stream
  fs.write( &m_data.raw, sizeof( m_data.raw ) );
}

#define INSTANTIATE( LM_TYPE ) \
template void FrHeader::Reorder< LM_TYPE >( LM_TYPE& ) const; \
template void FrHeader::Reorder< LM_TYPE >( std::vector< LM_TYPE >& ) const

INSTANTIATE( INT_2U );
INSTANTIATE( INT_2S );
INSTANTIATE( INT_4U );
INSTANTIATE( INT_4S );
INSTANTIATE( INT_8U );
INSTANTIATE( INT_8S );
INSTANTIATE( REAL_4 );
INSTANTIATE( REAL_8 );

#undef INSTANTIATE
