#ifndef FrameCPP_VERSION_6_FrProcData_HH
#define FrameCPP_VERSION_6_FrProcData_HH

#include <vector>

#include "general/types.hh"

#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/FrBase.hh"
#include "framecpp/Version6/FrHistory.hh"
#include "framecpp/Version6/FrTable.hh"
#include "framecpp/Version6/FrVect.hh"
#include "framecpp/Version6/STRING.hh"

namespace FrameCPP
{
  namespace Version_6
  {
    class FrProcData: public FrBase
    {
    public:
      enum type_type {
	UNKNOWN_TYPE = 0,
	TIME_SERIES = 1,
	FREQUENCY_SERIES = 2,
	OTHER_1D_SERIES_DATA = 3,
	TIME_FREQUENCY = 4,
	WAVELETS = 5,
	MULTI_DIMENSIONAL = 6
      };

      enum subType_type {
	UNKNOWN_SUB_TYPE = 0,
	//---------------------------------------------------------------
	// Subtype for fSeries
	//---------------------------------------------------------------
	DFT = 1,
	AMPLITUDE_SPECTRAL_DENSITY = 2,
	POWER_SPECTRAL_DENSITY = 3,
	CROSS_SPECTRAL_DENSITY = 4,
	COHERENCE = 5,
	TRANSFER_FUNCTION = 6
      };

      struct AuxParam_type {
	typedef REAL_8 auxParam_type;
	typedef STRING auxParamName_type;

	auxParam_type	auxParam;
	auxParamName_type	auxParamName;

	inline
	AuxParam_type( )
	{
	};

	inline
	AuxParam_type( auxParam_type Param, auxParamName_type ParamName )
	  : auxParam( Param ),
	    auxParamName( ParamName )
	{
	};

	AuxParam_type( auxParamName_type ParamName, auxParam_type Param )
	  : auxParam( Param ),
	    auxParamName( ParamName )
	{
	};
      };

      typedef std::vector< AuxParam_type >	AuxParamList_type;

      typedef SearchContainer< FrVect, &FrVect::GetName >	data_type;
      typedef SearchContainer< FrVect, &FrVect::GetName >	aux_type;
      typedef SearchContainer< FrTable, &FrTable::GetName >	table_type;
      typedef SearchContainer< FrHistory, &FrHistory::GetName >	history_type;

      static const CHAR* getStaticName();

      typedef data_type::iterator iterator;
      typedef data_type::const_iterator const_iterator;
      typedef aux_type::iterator aux_iterator;
      typedef aux_type::const_iterator const_aux_iterator;
      typedef table_type::iterator table_iterator;
      typedef table_type::const_iterator const_table_iterator;
      typedef history_type::iterator history_iterator;
      typedef history_type::const_iterator const_history_iterator;

      //: Default constructor
      FrProcData( );
      FrProcData( const FrProcData& ProcData );

      FrProcData( const std::string& Name,
		  const std::string& Comment,
		  INT_2U Type,
		  INT_2U SubType,
		  REAL_8 TimeOffset,
		  REAL_8 TRange,
		  REAL_8 FShift,
		  REAL_4 Phase,
		  REAL_8 FRange,
		  REAL_8 BW );

      FrProcData( const std::string& name,
		  REAL_8 sampleRate, const GPSTime& timeOffset,
		  REAL_8 fShift, REAL_4 phase );

      //: Number of bytes needed to write this structure
      INT_8U Bytes( ) const;

      const std::string& GetName() const;
      const std::string& GetComment() const { return m_data.comment; }
      INT_2U GetType() const { return m_data.type; }
      INT_2U GetSubType() const { return m_data.subType; }
      REAL_8 GetTimeOffset() const { return m_data.timeOffset; }
      REAL_8 GetTRange() const;
      REAL_8 GetFShift() const { return m_data.fShift; }
      REAL_4 GetPhase() const { return m_data.phase; }
      REAL_8 GetFRange() const;
      REAL_8 GetBW() const { return m_data.BW; }
      const AuxParamList_type& GetAuxParam( ) const { return m_data.AuxParam; }
      AuxParamList_type& GetAuxParam( ) { return m_data.AuxParam; }

      static const std::string& IDTypeToString( INT_2U Type );
      static const std::string& IDSubTypeToString( INT_2U Type,
						   INT_2U SubType );

      const aux_type& RefAux( ) const { return m_data.aux; }
      aux_type& RefAux( ) { return m_data.aux; }
      const data_type& RefData( ) const { return m_data.data; }
      data_type& RefData( ) { return m_data.data; }
      const history_type& RefHistory( ) const { return m_data.history; }
      history_type& RefHistory( ) { return m_data.history; }
      const table_type& RefTable( ) const { return m_data.table; }
      table_type& RefTable( ) { return m_data.table; }
   
      void  AppendComment( const std::string& comment );

      void SetFRange( REAL_8 FRange )
      {
         m_data.fRange = FRange;
         return;
      }

      void SetSubType( subType_type SubType )
      {
         m_data.subType = SubType;
         return;
      }

      void SetTimeOffset( REAL_8 TimeOffset )
      {
         m_data.timeOffset = TimeOffset;
         return;   
      }

      void SetTRange( REAL_8 TRange )
      {
         m_data.tRange = TRange;
         return;
      }

      void SetType( type_type Type )
      {
         m_data.type = Type;
         return;
      }
   
      //: Merge with another FrAdcData
      FrProcData& Merge( const FrProcData& RHS );

      //: Read the structure from stream
      static FrProcData* ReadDynamic( IFrameStream& Stream,
				  const FrBase& Base );

      //: Retrieve a portion of the current FrProcData
      std::auto_ptr< FrProcData >
      SubFrProcData( REAL_8 Offset, REAL_8 Dt ) const;

      //: Write the structure to the stream
      virtual void Write( OFrameStream& Stream ) const;

      //: Write the structure description to the stream
      virtual void WriteDescription( OFrameStream& Stream ) const;

    private:
      mutable struct data_definition_type {
	STRING	name;
	STRING	comment;
	INT_2U	type;
	INT_2U	subType;
	REAL_8	timeOffset;
	REAL_8	tRange;
	REAL_8	fShift;
	REAL_4	phase;
	REAL_8	fRange;
	REAL_8	BW;
	AuxParamList_type	AuxParam;
	data_type		data;
	aux_type		aux;
	table_type		table;
	history_type	history;


	data_definition_type( );
	data_definition_type( const data_definition_type& Source );

	void copy_core( const data_definition_type& Source );
      } m_data;
      mutable bool	m_synced_with_vector;

      void read( IFrameStream& Stream );

      std::auto_ptr< FrProcData > sub_time_series( REAL_8 Offset,
						   REAL_8 Dt ) const;

      void	sync_with_vector( ) const;

#if ENABLE_DEPRICATED_FRAMECPP
    public:
      typedef data_type DataContainer;
      typedef aux_type AuxContainer;
      typedef table_type TableContainer;

      const std::string& getName() const;

      inline const std::string& getComment() const
      {
	return GetComment( );
      }
      inline REAL_8 getSampleRate() const
      {
	return -1;
      }
      inline const GPSTime& getTimeOffset() const
      {
	offsetVersion5 = GPSTime( ) + m_data.timeOffset;

	return offsetVersion5;
      }
      inline REAL_8 getFShift() const
      {
	return GetFShift( );
      }
      inline REAL_4 getPhase() const
      {
	return GetPhase( );
      }
      inline const aux_type&	refAux( ) const { return RefAux( ); }
      inline aux_type&		refAux( ) { return RefAux( ); }
      inline const data_type&	refData( ) const { return RefData( ); }
      inline data_type&		refData( ) { return RefData( ); }
      inline const table_type&	refTable( ) const { return RefTable( ); }
      inline table_type&		refTable( ) { return RefTable( ); }

      inline void appendComment( const std::string& comment )
      {
	AppendComment( comment );
      }

      inline void setTimeOffset( const INT_4S sec = 0, const INT_4U nan = 0 )
      {
	SetTimeOffset( GPSTime( sec, nan ) - GPSTime( 0, 0 ) );
      }
    private:
      mutable GPSTime	offsetVersion5;
#endif
    };


    const INT_2U FR_PROC_DATA_DEFAULT_PROC_DATA_TYPE =
      FrProcData::UNKNOWN_TYPE;
    const INT_2U FR_PROC_DATA_DEFAULT_SUB_TYPE = FrProcData::UNKNOWN_SUB_TYPE;
    const REAL_8 FR_PROC_DATA_DEFAULT_TRANGE = 0.0;
    const REAL_8 FR_PROC_DATA_DEFAULT_FSHIFT = 0.0;
    const REAL_4 FR_PROC_DATA_DEFAULT_PHASE = 0.0;
    const REAL_8 FR_PROC_DATA_DEFAULT_FRANGE = 0.0;
    const REAL_8 FR_PROC_DATA_DEFAULT_BW = 0.0;

    
    inline REAL_8 FrProcData::
    GetTRange() const
    {
      if ( m_synced_with_vector == false )
      {
	sync_with_vector( );
      }
      return m_data.tRange;
    }

    inline REAL_8 FrProcData::
    GetFRange() const
    {
      if ( m_synced_with_vector == false )
      {
	sync_with_vector( );
      }
      return m_data.fRange;
    }

  } // namespace - Version_6
} // namespace FrameCPP

#endif /* FrameCPP_VERSION_6_FrProcData_HH */
