#ifndef FRAMECPP__IFRAME_STREAM_CC
#define FRAMECPP__IFRAME_STREAM_CC

#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <assert.h>
#include <errno.h>

#include <cstring>

#include <memory>
#include <sstream>
#include <stdexcept>

#include "general/autoarray.hh"
#include "general/gpstime.hh"
#include "general/regex.hh"
#include "general/regexmatch.hh"

#include "framecpp/ContainerBasic.hh"
#include "framecpp/SearchContainer.hh"

#include "framecpp/Version6/Dictionary.hh"
#include "framecpp/Version6/FrameH.hh"
#include "framecpp/Version6/FrAdcData.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/Frame5ReaderBase.hh"
#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/FrHeader.hh"
#include "framecpp/Version6/FrHistory.hh"
#include "framecpp/Version6/FrProcData.hh"
#include "framecpp/Version6/FrRawData.hh"
#include "framecpp/Version6/FrSE.hh"
#include "framecpp/Version6/FrSerData.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/FrSimData.hh"
#include "framecpp/Version6/FrTOC.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"
#include "framecpp/Version6/Reference.icc"
#include "framecpp/Version6/FrEndOfFrame.hh"

#define	LM_DEBUG 0

#if LM_DEBUG
#define MESSAGE( a ) std::cerr << a << __FILE__ << " " << __LINE__ << std::endl;
#define	AT() std::cerr << "INFO: this: " << (void*)this << " " << __FILE__ << " " << __LINE__ << std::endl;
#else
#define MESSAGE( a )
#define	AT()
#endif

using namespace FrameCPP::Version_6;

namespace
{
  static const FrTOC	null_toc;

  template< int FrId, class Map >
  bool
  at_end( const typename Map::const_iterator Data,
	  const FrTOC& TOC );

  template< >
  inline bool
  at_end<CLASS_FR_ADC_DATA, FrTOC::MapADC_type>
  ( const FrTOC::MapADC_type::const_iterator Data,
    const FrTOC& TOC )
  {
    return ( Data == TOC.GetADC( ).end( ) );
  }

  template< >
  inline bool
  at_end<CLASS_FR_PROC_DATA, FrTOC::MapProc_type>
  ( const FrTOC::MapProc_type::const_iterator Data,
    const FrTOC& TOC )
  {
    return ( Data == TOC.GetProc( ).end( ) );
  }

  template< >
  inline bool
  at_end<CLASS_FR_EVENT, FrTOC::MapEventType_type>
  ( const FrTOC::MapEventType_type::const_iterator Data,
    const FrTOC& TOC )
  {
    return ( Data == TOC.GetEventType( ).end( ) );
  }

  template< >
  inline bool
  at_end<CLASS_FR_SER_DATA, FrTOC::MapProc_type>
  ( const FrTOC::MapProc_type::const_iterator Data,
    const FrTOC& TOC )
  {
    return ( Data == TOC.GetSer( ).end( ) );
  }

  template< >
  inline bool
  at_end<CLASS_FR_SIM_DATA, FrTOC::MapProc_type>
  ( const FrTOC::MapProc_type::const_iterator Data,
    const FrTOC& TOC )
  {
    return ( Data == TOC.GetSim( ).end( ) );
  }

  template< >
  inline bool
  at_end<CLASS_FR_SIM_EVENT, FrTOC::MapSimEventType_type>
  ( const FrTOC::MapSimEventType_type::const_iterator Data,
    const FrTOC& TOC )
  {
    return ( Data == TOC.GetSimEventType( ).end( ) );
  }

  template< class Map >
  inline INT_8U
  get_position( const typename Map::const_iterator Data, INT_4U Frame )
  {
    return Data->second[ Frame ];
  }

  template< >
  inline INT_8U
  get_position<FrTOC::MapADC_type>( const FrTOC::MapADC_type::const_iterator Data, INT_4U Frame )
  {
    return Data->second.positionADC[ Frame ];
  }

  template< >
  inline INT_8U
  get_position<FrTOC::MapSimEventType_type>( const FrTOC::MapSimEventType_type::const_iterator Data, INT_4U Frame )
  {
    return Data->second[ Frame ].position;
  }

  class FrameH_FrRawDataCB
    : public Dictionary::Callback
  {
  public:
    FrameH_FrRawDataCB( IFrameStream& Stream );

    virtual void operator( )( FrBase& Base );
  private:
    IFrameStream&	m_stream;
  };

  FrameH_FrRawDataCB::
  FrameH_FrRawDataCB( IFrameStream& Stream )
    : m_stream( Stream )
  {
  }

  void FrameH_FrRawDataCB::
  operator( )( FrBase& Base )
  {
    try
    {
      FrRawData& raw_data( dynamic_cast< FrRawData& >( Base ) );

      m_stream.RemoveContainer( raw_data.RefFirstSer( ) );
      m_stream.RemoveContainer( raw_data.RefFirstAdc( ) );
      m_stream.RemoveContainer( raw_data.RefFirstTable( ) );
      m_stream.RemoveContainer( raw_data.RefLogMsg( ) );
      m_stream.RemoveContainer( raw_data.RefMore( ) );
    }
    catch( ... )
    {
    }
  }
}

//-----------------------------------------------------------------------

const int IFrameStream::MODE_TOC( 0x0001 );
const int IFrameStream::MODE_DAQ( 0x0002 );
const int IFrameStream::MODE_RANDOM_ACCESS( 0x0004 );
const int IFrameStream::MODE_SEQUENTIAL( 0x0000 );

template< class FrData >
std::auto_ptr< FrData > IFrameStream::
read_fr_data( INT_8U Position )
{
  try
  {
    m_dictionary.m_reference_info.Reset( false );
    m_dictionary.m_reference_info.SetRecursiveMode( true );
    MESSAGE( "Data start: " << Position );
    m_stream.seekg( Position );
    
    std::auto_ptr<FrBase>	base( ReadObject( ) );
    if ( base.get( ) )
    {
      m_dictionary.m_reference_info.
	RemoveReference( PTR_STRUCT( base->GetClassId( ),
				     base->GetInstance( ) ) );
    }
    
    FrData*			data( dynamic_cast< FrData* >( base.get( ) ) );
    if ( data )
    {
      read_recursive( );
      m_dictionary.m_reference_info.ResolveLinks( *this );
      base.release( );
      return std::auto_ptr< FrData >( data );
    }
    m_dictionary.m_reference_info.SetRecursiveMode( false );
    MESSAGE( "IFrameStream::read_fr_adc_data: " );
    return std::auto_ptr< FrData >( data );
  }
  catch( ... )
  {
    MESSAGE( "IFrameStream::read_fr_data: " );
    m_dictionary.m_reference_info.SetRecursiveMode( false );
    throw;
  }
}

template< int FrId, class FrData, class Map >
std::auto_ptr< FrData > IFrameStream::
read_fr_data( INT_4U Frame,
	      const typename Map::const_iterator Data )
{
  MESSAGE( "IFrameStream::read_fr_adc_data: " );
  if ( Frame >= GetNumberOfFrames( ) )
  {
    MESSAGE( "IFrameStream::read_fr_adc_data: " );
    throw std::range_error( "Frame request exceeds number of frames in file" );
  }
  MESSAGE( "IFrameStream::read_fr_adc_data: " );
  if ( at_end<FrId, Map>( Data, *m_toc ) )
    
  {
    MESSAGE( "IFrameStream::read_fr_adc_data: " );
    return std::auto_ptr< FrData >( (FrData*)NULL );
  }

  INT_8U	pos( get_position<Map>( Data, Frame ) );
  if ( pos == INT_8U( 0 ) )
    
  {
    MESSAGE( "IFrameStream::read_fr_adc_data: " );
    return std::auto_ptr< FrData >( (FrData*)NULL );
  }

  MESSAGE( "IFrameStream::read_fr_adc_data: " );
  return read_fr_data< FrData >( pos );
}

template< int FrId, class FrData, class Map >
FrData* IFrameStream::
read_fr_struct( INT_4U Frame,
		const typename Map::const_iterator Data )
{
  if ( Frame >= GetNumberOfFrames( ) )
  {
    throw std::range_error( "Frame request exceeds number of frames in file" );
  }
  INT_8U	pos( get_position<Map>( Data, Frame ) );
  if ( ( at_end<FrId, Map>( Data, *m_toc ) ) || ( pos == 0 ) )
    
  {
    return (FrData*)NULL;
  }
  //---------------------------------------------------------------------
  // Go to where the header is located
  //---------------------------------------------------------------------
  m_stream.seekg( pos );
  //---------------------------------------------------------------------
  // Read the object from the file
  //---------------------------------------------------------------------
  std::auto_ptr< FrBase >	base( ReadObject( ) );
  //---------------------------------------------------------------------
  // Ensure that it is of the correct type
  //---------------------------------------------------------------------
  FrData*	data( dynamic_cast< FrData* >( base.get( ) ) );
  if ( data )
  {
    //-------------------------------------------------------------------
    // Take ownership of object and have caller be responsible for
    //   freeing the memory
    //-------------------------------------------------------------------
    m_dictionary.m_reference_info.
      RemoveReference( PTR_STRUCT( data->GetClassId( ),
				   data->GetInstance( ) ) );
    base.release( );
  }
  return data;
}

IFrameStream::
IFrameStream( std::istream& Stream, int Mode )
  : Version_6::FrameStreamBase( GetClassIO( ), CLASS_LAST_ENTRY ),
    m_stream( Stream ),
    m_mode( Mode ),
    m_length_of_last_object( 0 ),
    m_road_block( 0 ),
    m_frame_5_reader( (Frame5ReaderBase*)NULL ),
    m_next_frame( 0 )
{
  AT( );
  init( );
  AT( );
}

IFrameStream::
IFrameStream( std::istream& Stream,
	      const std::vector< Dictionary::io_calls_type >& LibraryDefs,
	      INT_4U NumberOfClassTypes,
	      int Mode )
  : Version_6::FrameStreamBase( LibraryDefs, NumberOfClassTypes ),
    m_stream( Stream ),
    m_mode( Mode ),
    m_length_of_last_object( 0 ),
    m_road_block( 0 ),
    m_frame_5_reader( (Frame5ReaderBase*)NULL ),
    m_next_frame( 0 )
{
  AT( );
  init( );
  AT( );
}

IFrameStream::
~IFrameStream( )
{
  if ( m_frame_5_reader )
  {
    delete m_frame_5_reader;
    m_frame_5_reader = (Frame5ReaderBase*)NULL;
  }
}

namespace FrameCPP
{
  namespace Version_6
  {
    template < >
    IFrameStream& IFrameStream::
    operator>>( STRING& Data )
    {
      m_stream.clear( );
      STRING::length_of_string_type	length;
      *this >> length;
      if ( length == 0 )
      {
	Data = "";
      }
      else
      {
	// Protect ourselves from reading too man bytes
	CheckReadBoundry( length );

	char*	buffer( (char*)NULL );
	try
	{
	  buffer = new char[ length + 1 ];
	  read( buffer, length );
	  if ( buffer[ length - 1 ] != '\0' )
	  {
	    buffer[ length ] = '\0';
	    std::string msg( "String missing null-terminator: \"" );
	    msg += buffer;
	    msg += "\".";
	    throw std::logic_error( msg );
	  }
	  Data = buffer;
	  delete[] buffer;
	  buffer = (char*)NULL;
	}
	catch( ... )
	{
	  delete[] buffer;	// release the resource
	  buffer = (char*)NULL;
	  throw;		// rethrow the exception
	}
#if LM_DEBUG && 0
	std::cerr << "DEBUG: STRING: Length: " << length
		  << " buffer: " << buffer
		  << " data: " << Data
		  << std::endl
	;
#endif /* LM_DEBUG */
      }
    
      return *this;
    }

    template< >
    IFrameStream& IFrameStream::
    operator>>( General::GPSTime& Data )
    {
      INT_4U	seconds;
      INT_4U	nanoseconds;

      *this >> seconds >> nanoseconds;
      Data = General::GPSTime( seconds, nanoseconds );
      return *this;
    }

    template< >
    IFrameStream& IFrameStream::
    operator>>( PTR_STRUCT& Data )
    {
      *this >> Data.data_class >> Data.data_instance;
      return *this;
    }

    template< >
    IFrameStream& IFrameStream::
    operator>>( FrBase** Data )
    {
      PTR_STRUCT	ref;

      *this >> ref;
      m_dictionary.m_reference_info.AddPointer( Data, ref );

      return *this;
    }

    template< class T >
    IFrameStream& IFrameStream::
    operator>>( T& Data )
    {
      // Sanity Check
      CheckReadBoundry( sizeof( T ) );

      read( (char*)( &Data ), sizeof( T ) );
      m_header.Reorder( Data );
      return *this;
    }

    template< class T >
    IFrameStream& IFrameStream::
    operator >>( Container< T >& Data )
    {
      PTR_STRUCT	ref;

      *this >> ref;
      m_dictionary.m_reference_info.AddContainer( Data, ref );
      return *this;
    }

  } // namespace - Version_6
} // namespace - FrameCPP

FrameCPP::Version_3_4_5::FrameReaderBase& IFrameStream::
GetFrame5Reader( )
{
  if (  m_frame_5_reader )
  {
    return *m_frame_5_reader;
  }

  throw std::range_error( "Frame 5 Reader is NULL" );
}

Frame5ReaderBase& IFrameStream::
GetExposedFrame5Reader( )
{
  if (  m_frame_5_reader )
  {
    return *m_frame_5_reader;
  }

  throw std::range_error( "Frame 5 Reader is NULL" );
}

PTR_STRUCT IFrameStream::
GetNextNode5( )
{
  PTR_STRUCT ret( GetExposedFrame5Reader( ).mLinkedListNodes.begin( )->second.
		  getClass( ),
		  GetExposedFrame5Reader( ).mLinkedListNodes.begin( )->second.
		  getInstance( ) );
  GetExposedFrame5Reader( ).mLinkedListNodes.
    erase( GetExposedFrame5Reader( ).mLinkedListNodes.begin( ),
	   GetExposedFrame5Reader( ).mLinkedListNodes.end( ) );
  if ( ret.data_class == 0 ) 
  {
    ret = NULL_PTR_STRUCT;
  }
  return ret;
}

INT_8U IFrameStream::
GetPositionH( INT_4U Frame )
{
  if ( Frame < GetNumberOfFrames( ) )
  {
    return m_toc->GetPositionH( ).operator[]( Frame );
  }
  throw std::range_error( "requested frame exceeds the number of frames in the file" );
}

FrameH* IFrameStream::
ReadNextFrame( )
{
  std::auto_ptr< FrBase >	ret( (FrBase*)NULL );
  std::auto_ptr< FrBase >	base( (FrBase*)NULL );

  if ( ( m_mode& MODE_TOC ) && true )
  {
    if ( m_next_frame < m_toc->GetNFrame( ) )
    {
      m_stream.seekg( m_toc->GetPositionH( )[ m_next_frame ], std::ios::beg );
      base.reset( ReadObjectBase( ) );
    }
  }
  else
  {
    INT_2U id( m_dictionary.GetLibraryClass( FrameH::getStaticName() ) );

    if ( id == FrBase::CLASS_NULL )
    {
      throw std::runtime_error( "Unable to determine the numeric class for FrameH" );
    }
    base.reset( skip_till_class( id ) );
  }
  if ( m_stream.good( ) && base.get( ) )
  {
    //-------------------------------------------------------------------
    // Read in the FrameH structure
    //-------------------------------------------------------------------
    m_dictionary.m_reference_info.Reset( true );
    ret.reset( m_dictionary.Read( *this, *base ) );
    m_dictionary.m_reference_info.
      RemoveReference( PTR_STRUCT( ret->GetClassId( ),
				   ret->GetInstance( ) ) );
    read_objects_through( m_dictionary.GetLibraryClass( FrEndOfFrame::getStaticName() ) );
    m_dictionary.m_reference_info.ResolveLinks( *this );
    m_next_frame++;
  }
  FrameH* frameh( dynamic_cast< FrameH* >( ret.get( ) ) );
  if ( frameh )
  {
    ret.release( );
  }
  return frameh;
}

FrBase* IFrameStream::
ReadObject( )
{
  std::auto_ptr< FrBase >	ret( ReadObjectBase( ) );
  if ( ret.get( ) )
  {
    return ( m_dictionary.Read( *this, *ret ) );
  }
  return (FrBase*)NULL;
}

FrBase* IFrameStream::
ReadObject( FrBase& Base )
{
  return ( m_dictionary.Read( *this, Base ) );
}

FrBase* IFrameStream::
ReadObjectBase( bool SkipUnknownObjects )
{
  FrBase				ret( FrBase::CLASS_NULL );
  std::auto_ptr< Dictionary::definition_type >	def;
  while( m_stream.good( ) )
  {
    ret.read( *this );
    if ( good( ) != true )
    {
      setLastError( "stream is no longer good" );
      return (FrBase*)NULL;
    }
    if ( ret.GetClassId( ) == FrBase::CLASS_FR_SH )
    {
      // Start new definition
      if ( def.get( ) )
      {
	// Save the current definition
	m_dictionary.AddDefinition( def.release( ) );
      }
      // Create an FrSH
      std::auto_ptr< FrBase >	base( m_dictionary.Read( *this, ret ) );
      const FrSH* sh( dynamic_cast< const FrSH* >( base.get( ) ) );

      if ( sh )
      {
	def.reset( new Dictionary::definition_type( sh->GetName(),
   						    sh->GetClass(),
						    sh->GetComment() ) );
      }
    }
    else if ( ret.GetClassId( ) == FrBase::CLASS_FR_SE )
    {
      if ( def.get( ) )
      {
	std::auto_ptr< FrBase >	base( m_dictionary.Read( *this, ret ) );
	const FrSE* se( dynamic_cast< const FrSE* >( base.get( ) ) );
	  
	if ( se )
	{
	  def->elements.push_back( Dictionary::element_type( se->GetName(),
							     se->GetClass( ),
							     se->GetComment( ) ) );
	}
      }
      else
      {
	// There is no active header.
	SkipObject( ret );
	continue;
      }
    }
    else
    {
      if ( def.get( ) )
      {
	// Save the current definition
	m_dictionary.AddDefinition( def.release( ) );
      }
      if ( ( SkipUnknownObjects == false ) ||
	   ( ret.GetClassId( ) != FrBase::CLASS_NULL ) ||
	   ( m_dictionary.IsReadable( *this, ret ) ) )
      {
	// Break out of the look and return the object to the user
	break;
      }
      else
      {
	SkipObject( ret );
	continue;
      }
    }
  }
  if ( ret.GetClassId( ) == FrBase::CLASS_NULL )
  {
    setLastError( "Class Id of object is FrBase::CLASS_NULL" );
    return( (FrBase*)NULL );
  }
  return new FrBase( ret );
}

//-----------------------------------------------------------------------
// ReadVector - specialization
//-----------------------------------------------------------------------
template < >
IFrameStream& IFrameStream::
ReadVector( std::vector< STRING >& Data, INT_4U Size )
{
  //Sanity Check
  CheckReadBoundry( sizeof( STRING::length_of_string_type ) * Size );

  Data.resize( Size );
  for ( std::vector< STRING >::iterator d( Data.begin( ) );
	d != Data.end( );
	d++ )
  {
    *this >> *d;
  }
  return *this;
}

namespace FrameCPP
{
  namespace Version_6
  {
    template<  >
    IFrameStream& IFrameStream::
    ReadVector( FrEvent::ParamList_type& Data, INT_4U Size )
    {
      std::vector< REAL_4 >	parameters;
      std::vector< STRING > parameterNames;

      Data.resize( 0 );

      this->ReadVector( parameters, Size );
      this->ReadVector( parameterNames, Size );

      std::vector< REAL_4 >::const_iterator api( parameters.begin( ) );
      std::vector< STRING >::const_iterator apni( parameterNames.begin( ) );

      while ( api != parameters.end( ) )
      {
	Data.push_back( FrEvent::Param_type( *apni, *api ) );
#if old
	// Data.back( ).second = *api;
	// Data.back( ).first = *apni;
#endif /* old */
	api++;
	apni++;
      }
      return *this;
    }

    //-----------------------------------------------------------------------
    // ReadVector - template specification
    //-----------------------------------------------------------------------
    template< class T >
    IFrameStream& IFrameStream::
    ReadVector( std::vector< T >& Data, INT_4U Size )
    {
      // Sanity Check
      CheckReadBoundry( sizeof( T ) * Size );

      if ( Size )
      {
	Data.resize( Size );
	read( (char*)( &(Data[0]) ), sizeof( T ) * Size);
	m_header.Reorder( Data );
      }

      return *this;
    }
  } // namespace - Version_6
} // namespace - FrameCPP

void IFrameStream::
SetChecksumMethodFile( const CheckSum::kind_type Method )
{
  Common::FrameStreamBase::SetChecksumMethodFile( Method );
  INT_8U			pos( tellg( ) );
  General::AutoArray< char >	buffer( new char[ pos ] );

  m_stream.seekg( 0, std::ios::beg );	// Start from the begining of the file
  read( buffer.get( ), pos );		// Read all bytes up to the present
}

void IFrameStream::
SkipObject( const FrBase& ObjectBase ) const
{
  INT_8U skip( ObjectBase.Skip( *this ) );
  if ( skip > 0 )
  {
    m_stream.seekg( skip, std::ios::cur );
  }
}

void IFrameStream::
init( )
{
  if ( m_stream.good( ) == false )
  {
    std::ostringstream	oss;

    throw std::runtime_error( "Cannot create IFrameStream because input stream is bad" );
  }

  m_toc.reset( (FrTOC*)NULL );
  m_next_frame = 0;

  using FrameCPP::Version_3_4_5::Input;
  using FrameCPP::Version_3_4_5::FileHeader;

  INT_8U	end_of_fr_header;

  AT( );
  // Read in the frame header
  m_header.read( *this );
  end_of_fr_header = m_stream.tellg( );
#if LM_DEBUG
  std::cerr << "INFO: Version: " << (INT_2U)GetVersion( ) << std::endl;
#endif /* LM_DEBUG */
  m_dictionary.m_reference_info.SetFrameVersion( GetVersion( ) );
  m_dictionary.m_reference_info.SetFrameLibraryRevision( GetLibraryRevision( ) );
  //---------------------------------------------------------------------
  // Create the Frame5 reader if needed
  //---------------------------------------------------------------------
  if ( ( GetVersion( ) == 5 ) ||
       ( GetVersion( ) == 4 ) ||
       ( GetVersion( ) == 3 ) )
  {
    AT( );
    // Create a reader that allows for the reading of frames
    //   corresponding to version 3, 4, or 5 of the spec
    m_frame_5_reader =
      new Frame5ReaderBase( m_stream,
			    new FileHeader( (const char*)
					    ( &(m_header.m_data.raw ) ) ) );
  }
  else
  {
    AT( );
    m_frame_5_reader = ( Frame5ReaderBase*)NULL;
  }
  //---------------------------------------------------------------------
  // Check to seek if this stream can be seeked backwards
  //---------------------------------------------------------------------
  std::streampos end_of_header( m_stream.tellg( ) );
  if ( m_mode & MODE_RANDOM_ACCESS )
  {
    AT( );
    m_stream.seekg( 0, std::ios::beg );
    if ( end_of_header != m_stream.tellg( ) )
    {
      AT( );
      // Restore the position
      m_stream.seekg( end_of_header, std::ios::beg );
      m_mode |= MODE_RANDOM_ACCESS;
    }
    else
    {
      AT( );
      m_mode &= ~MODE_RANDOM_ACCESS;
    }
  }
  if ( ( m_mode & MODE_RANDOM_ACCESS ) &&
       ( m_mode & MODE_TOC ) )
  {
    AT( );
    if ( ( m_header.GetVersion( ) == 4 ) ||
	 ( m_header.GetVersion( ) == 6 ) )
    {
      //-----------------------------------------------------------------
      // Since the  stream can be accessed randomly, seek to the end and
      //   look for a table of contents.
      //-----------------------------------------------------------------
      FrBase				eof_base( FrBase::CLASS_NULL );
      std::auto_ptr< FrEndOfFile >	eof;
      std::streamoff			offset( 0 );
      INT_8U				end_of_file( 0 );

      AT( );
      m_stream.seekg( 0, std::ios::end ); // Go to end of file
      end_of_file = m_stream.tellg( );
      offset = FrEndOfFile::Bytes( GetVersion( ) );
      m_stream.seekg( - offset, std::ios::end ); // Go to end of file
      m_stream.clear( );
      
      eof_base.read( *this );

      //-----------------------------------------------------------------
      // Sanity checks on the FrEndOfFile base structure.
      //-----------------------------------------------------------------
      if ( (std::streamoff)eof_base.GetLength( ) != offset )
      {
	AT( );
	throw std::logic_error( "Missing FrEndOfFile" );
      }
      eof.reset( FrEndOfFile::ReadDynamic( *this, eof_base ) );

      //-----------------------------------------------------------------
      // Sanity checks on the FrEndOfFile structure.
      //-----------------------------------------------------------------
      if ( ( eof->GetNBytes( ) != 0 ) && ( eof->GetNBytes( ) != end_of_file ) )
      {
	std::ostringstream	oss;

	oss << "Computed end (" << eof->GetNBytes( )
	    << ") of file differs from true end of file ("
	    << end_of_file << ")";
	throw std::runtime_error(  oss.str( ) );
      }
      if ( eof->GetSeekTOC( ) > ( end_of_file - end_of_fr_header ) )
      {
	throw std::runtime_error( "TOC starts at an invalid position" );
      }

      if ( eof->GetSeekTOC( ) != 0 )
      {
	//---------------------------------------------------------------
	// Go to where the TOC is suppose to start
	//---------------------------------------------------------------
	m_stream.seekg( - eof->GetSeekTOC( ) , std::ios::end );
	//---------------------------------------------------------------
	// Read the base
	//---------------------------------------------------------------
	std::auto_ptr< FrBase > toc_base( ReadObjectBase( false ) );
	if ( toc_base.get( ) )
	{
	  //-------------------------------------------------------------
	  // Provide mapping information for dictionary
	  //-------------------------------------------------------------
	  AT( );
	  m_dictionary.add_class_mapping( toc_base->GetClassId( ),
					  CLASS_FR_TOC );
	  //-------------------------------------------------------------
	  // Read using dictionary to allow for callbacks.
	  //-------------------------------------------------------------
	  AT( );
	  toc_base.reset( m_dictionary.Read( *this, *toc_base ) );
	  
	  AT( );
	  if ( dynamic_cast< FrTOC* >( toc_base.get( ) ) )
	  {
	    //-----------------------------------------------------------
	    // Retain the dictionary for random access.
	    //-----------------------------------------------------------
      
	    AT( );
	    m_toc.reset( dynamic_cast< FrTOC* >( toc_base.release( ) ) );
	    m_dictionary.m_reference_info.
	      RemoveReference( PTR_STRUCT( m_toc->GetClassId( ),
					   m_toc->GetInstance( ) ) );

	    //-----------------------------------------------------------
	    // Add the structure mapping
	    //-----------------------------------------------------------

	    AT( );
	    std::vector< INT_2U >::const_iterator
	      shid( m_toc->GetSHid( ).begin( ) );
	    std::vector< INT_2U >::const_iterator
	      shid_end( m_toc->GetSHid( ).end( ) );
	    std::vector< STRING >::const_iterator
	      shname( m_toc->GetSHname( ).begin( ) );
	    
	    AT( );
	    while( shid != shid_end )
	    {
	      m_dictionary.
		add_class_mapping( *shid,
				   m_dictionary.GetLibraryClass( *shname) );
	      shid++;
	      shname++;
	    }
	  }
	}
      }
    }
    if ( m_toc.get( ) == (FrTOC*)NULL )
    {
      //-----------------------------------------------------------------
      // Scan the file creating a TOC as we go.
      //-----------------------------------------------------------------
      AT( );
      m_stream.clear( );
      m_stream.seekg( end_of_header, std::ios::beg );
      m_toc.reset( new FrTOC( ) );
      m_toc->BuildDynamically( *this );
    }
  }
  else
  {
    AT( );
    m_mode &= ~MODE_TOC;
  }
  AT( );
} // method IFrameStream::init

INT_4U IFrameStream::
read( void* Dest, INT_4U Length, bool CalcChecksum )
{
  errno = 0;
  m_stream.read( (char*)( Dest ), Length );
  if ( ( m_stream.gcount( ) <= 0 ) && ( errno != 0 ) )
  {
    std::ostringstream	msg;
    char		buffer[ 1024];

    strerror_r( errno, buffer, sizeof( buffer ) );
    msg << "Low level read error: " << buffer;
    throw std::runtime_error( msg.str( ) );
  }
  if ( m_stream.good( ) )
  {
    if ( CalcChecksum )
    {
      calcChecksums( (char*)( Dest ), Length );
    }
    return Length;
  }
  else
  {
    return 0;
  }
}

void IFrameStream::
read_recursive( )
{
  PTR_STRUCT	next;

  while ( m_dictionary.m_reference_info.GetRecursiveReference( next ) )
  {
#if LM_DEBUG
    std::cerr << "DEBUG: next: " << next.data_class
	      << "-" << next.data_instance
	      << std::endl;
#endif
    while( next != NULL_PTR_STRUCT )
    {
      m_dictionary.m_reference_info.StartNewObject( );
      read_object( next );
      m_dictionary.m_reference_info.GetRecursiveReference( next );
#if LM_DEBUG
      std::cerr << "DEBUG: next: " << next.data_class
		<< "-" << next.data_instance
		<< std::endl;
#endif
    }
  }
}

FrameH* IFrameStream::
ReadFrameH( INT_4U Frame, INT_4U ContainerSet )
try
{
  //---------------------------------------------------------------------
  // Clear out the previous error
  //---------------------------------------------------------------------
  setLastError( "" );

  m_dictionary.m_reference_info.Reset( true );
  m_dictionary.m_reference_info.SetRecursiveMode( true );
  if ( Frame >= GetNumberOfFrames( ) )
  {
    throw std::range_error( "Frame request exceeds number of frames in file" );
  }
  m_stream.seekg( GetPositionH( Frame ) );
  std::auto_ptr< FrBase >	base( ReadObject( ) );
  if ( base.get( ) )
  {
    m_dictionary.m_reference_info.
      RemoveReference( PTR_STRUCT( base->GetClassId( ),
				   base->GetInstance( ) ) );
  }

  //---------------------------------------------------------------------
  // Add callback so when a FrRawData structure is read, that it
  //   will not be read recursively
  //---------------------------------------------------------------------
  FrameH*		frameh( dynamic_cast< FrameH* >( base.get( ) ) );
  if ( frameh )
  {
    FrameH_FrRawDataCB			cb( *this );
    Dictionary::callback_key_type	cb_key =
      m_dictionary.AddCallback( CLASS_FR_RAW_DATA, cb );

    try
    {
      frameh->ReadSubset( *this, ContainerSet );
      read_recursive( );
      m_dictionary.m_reference_info.ResolveLinks( *this );
      //-----------------------------------------------------------------
      // Remove callback so when a FrRawData structure is read, that it
      //   will not be read recursively
      //-----------------------------------------------------------------
      m_dictionary.RemoveCallback( cb_key );
    }
    catch ( ... )
    {
      //-----------------------------------------------------------------
      // Remove callback so when a FrRawData structure is read, that it
      //   will not be read recursively
      //-----------------------------------------------------------------
      m_dictionary.RemoveCallback( cb_key );
      throw;
    }
  }
  m_dictionary.m_reference_info.SetRecursiveMode( false );
  base.release( );
  return frameh;
}
catch( ... )
{
  m_dictionary.m_reference_info.SetRecursiveMode( false );
  try
  {
    throw;
  }
  catch( const std::exception& Exception )
  {
    setLastError( Exception.what( ) );
    throw;
  }
  catch( ... )
  {
    setLastError( "Unknwon exception" );
    throw;
  }
}

FrAdcData* IFrameStream::
ReadFrAdcData( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrAdcData" << Channel << " " );
  return read_fr_data<CLASS_FR_ADC_DATA, FrAdcData,FrTOC::MapADC_type>
    ( Frame, m_toc->GetADC( Channel ) ).release( );
}

FrAdcData* IFrameStream::
ReadFrAdcData( INT_4U Frame, INT_4U Channel )
{
  MESSAGE( "IFrameStream::ReadFrAdcData" << Channel << " " );
  return read_fr_data<CLASS_FR_ADC_DATA, FrAdcData,FrTOC::MapADC_type>
    ( Frame, m_toc->GetADC( Channel ) ).release( );
}

FrAdcData* IFrameStream::
ReadFrAdcStruct( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrAdcStruct" << Channel << " " );
  return read_fr_struct<CLASS_FR_ADC_DATA, FrAdcData,FrTOC::MapADC_type>
    ( Frame, m_toc->GetADC( Channel ) );
}

FrAdcData* IFrameStream::
ReadFrAdcStruct( INT_4U Frame, INT_4U Channel )
{
  MESSAGE( "IFrameStream::ReadFrAdcStruct" << Channel << " " );
  return read_fr_struct<CLASS_FR_ADC_DATA, FrAdcData,FrTOC::MapADC_type>
    ( Frame, m_toc->GetADC( Channel ) );
}

FrEvent* IFrameStream::
ReadFrEvent( INT_4U Frame, const std::string& Name )
{
  MESSAGE( "IFrameStream::ReadFrEvent" << Name << " " );
  return read_fr_data<CLASS_FR_EVENT, FrEvent,FrTOC::MapEventType_type>
    ( Frame, m_toc->GetEventType( Name ) ).release( );
}

FrProcData* IFrameStream::
ReadFrProcData( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrProcData" << Channel << " " );
  return read_fr_data<CLASS_FR_PROC_DATA, FrProcData,FrTOC::MapProc_type>
    ( Frame, m_toc->GetProc( Channel ) ).release( );
}

FrProcData* IFrameStream::
ReadFrProcData( INT_4U Frame, INT_4U Channel )
{
  MESSAGE( "IFrameStream::ReadFrProcData" << Channel << " " );
  return read_fr_data<CLASS_FR_PROC_DATA, FrProcData,FrTOC::MapProc_type>
    ( Frame, m_toc->GetProc( Channel ) ).release( );
}

FrProcData* IFrameStream::
ReadFrProcStruct( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrProcStruct" << Channel << " " );
  return read_fr_struct<CLASS_FR_PROC_DATA, FrProcData,FrTOC::MapProc_type>
    (Frame, m_toc->GetProc( Channel ) );
}

FrProcData* IFrameStream::
ReadFrProcStruct( INT_4U Frame, INT_4U Channel )
{
  MESSAGE( "IFrameStream::ReadFrProcStruct" << Channel << " " );
  return read_fr_struct<CLASS_FR_PROC_DATA, FrProcData,FrTOC::MapProc_type>
    ( Frame, m_toc->GetProc( Channel ) );
}

FrRawData* IFrameStream::
ReadFrRawData( INT_4U Frame )
{
  if ( Frame >= GetNumberOfFrames( ) )
  {
    throw std::range_error( "Frame request exceeds number of frames in file" );
  }
  m_stream.seekg( GetPositionH( Frame ) );
  std::auto_ptr< FrBase >	base( ReadObject( ) );
  if ( base.get( ) )
  {
    m_dictionary.m_reference_info.
      RemoveReference( PTR_STRUCT( base->GetClassId( ),
				   base->GetInstance( ) ) );
  }

  FrameH*		frameh( dynamic_cast< FrameH* >( base.get( ) ) );
  if ( frameh )
  {
    void* rawData = &(frameh->m_data.rawData );
    PTR_STRUCT	raw_data_ref
      ( m_dictionary.m_reference_info.
	GetReferenceToPointer( (const FrBase**)rawData ) );

    if ( raw_data_ref != NULL_PTR_STRUCT )
    {
      base.reset( read_object( raw_data_ref ) );
      if ( base.get( ) )
      {
	m_dictionary.m_reference_info.
	  RemoveReference( PTR_STRUCT( base->GetClassId( ),
				       base->GetInstance( ) ) );
      }

      FrRawData*
	raw_data( dynamic_cast< FrRawData* >( base.get( ) ) );

      if ( raw_data )
      {
	m_dictionary.m_reference_info.
	  RemoveReference( PTR_STRUCT( raw_data->GetClassId( ),
				       raw_data->GetInstance( ) ) );
	base.release( );
	return raw_data;
      }
    }
  }
  return (FrRawData*)NULL;
}

FrSerData* IFrameStream::
ReadFrSerData( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrSerData" << Channel << " " );
  return read_fr_data<CLASS_FR_SER_DATA, FrSerData,FrTOC::MapSer_type>
    ( Frame, m_toc->GetSer( Channel ) ).release( );
}

FrSerData* IFrameStream::
ReadFrSerStruct( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrSerStruct" << Channel << " " );
  return read_fr_struct<CLASS_FR_SER_DATA, FrSerData,FrTOC::MapSer_type>
    ( Frame, m_toc->GetSer( Channel ) );
}

FrSimData* IFrameStream::
ReadFrSimData( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrSimData" << Channel << " " );
  return read_fr_data<CLASS_FR_SIM_DATA, FrSimData,FrTOC::MapSim_type>
    ( Frame, m_toc->GetSim( Channel ) ).release( );
}

FrSimEvent* IFrameStream::
ReadFrSimEvent( INT_4U Frame, const std::string& Name )
{
  MESSAGE( "IFrameStream::ReadFrSimEvent" << Name << " " );
  return read_fr_data<CLASS_FR_SIM_EVENT, FrSimEvent,FrTOC::MapSimEventType_type>
    ( Frame, m_toc->GetSimEventType( Name ) ).release( );
}

FrSimData* IFrameStream::
ReadFrSimStruct( INT_4U Frame, const std::string& Channel )
{
  MESSAGE( "IFrameStream::ReadFrSimStruct" << Channel << " " );
  return read_fr_struct<CLASS_FR_SIM_DATA, FrSimData,FrTOC::MapSim_type>
    ( Frame, m_toc->GetSim( Channel ) );
}

INT_4U IFrameStream::
ReadFrStatData( const std::string& NamePattern,
		const std::string& DetectorPattern,
		const General::GPSTime& StartTime,
		const General::GPSTime& EndTime,
		const INT_4U Version,
		FrStatData::Query& QueryResults )
{
  QueryResults.Reset( *this );
  QueryResults.Add( NamePattern, DetectorPattern,
		    StartTime, EndTime,
		    Version );
  return QueryResults.size( );
}

std::auto_ptr< FrStatData > IFrameStream::
ReadFrStatData( INT_8U Position )
{
  MESSAGE( "IFrameStream::ReadFrStatData(Position) " << Position << " " );
  return read_fr_data<FrStatData>( Position );
}

FrBase* IFrameStream::
read_object( const PTR_STRUCT& Object )
{
#if LM_DEBUG
  std::cerr << "INFO: Looking for object: "
	    << Object.data_class << "-" << Object.data_instance
	    << std::endl;
#endif
  std::auto_ptr< FrBase >	base;
  PTR_STRUCT 			ps( NULL_PTR_STRUCT );
  FrBase*			request;

  while( m_stream.good( ) )
  {
    base.reset( ReadObjectBase( ) );
    if ( ! base.get( ) )
    {
      break;
    }
#if LM_DEBUG
  std::cerr << "INFO: Read object: "
	    << base->GetClassId( ) << "-" << base->GetInstance( )
	    << std::endl;
#endif
    if ( ( ( Object.data_class == base->GetClassId( ) ) &&
	   ( Object.data_instance == base->GetInstance( ) ) ) ||
	 ( m_dictionary.m_reference_info.IsRecursiveObject( *base ) ) )
    {
      request = m_dictionary.Read( *this, *base );
      if ( request )
      {
	ps = request->GetNext( );
      }
#if LM_DEBUG
      std::cerr << "INFO: Next object: "
		<< ps.data_class << "-" << ps.data_instance
		<< std::endl;
#endif
      m_dictionary.m_reference_info.AddNextRef( ps );
      if ( ( Object.data_class == base->GetClassId( ) ) &&
	   ( Object.data_instance == base->GetInstance( ) ) )
      {
	return request;
      }
    }
    else
    {
      SkipObject( *base );
    }
  }
  return (FrBase*)NULL;
}

FrBase* IFrameStream::
read_objects_through( INT_2U LibraryClass )
{
  FrBase*			ret;

  do
  {
    ret = ReadObject( );

    if ( ( ret != (FrBase*)NULL ) && ( ret->GetClassId( ) == LibraryClass ) )
    {
      return ret;
    }
  } while ( ! m_stream.eof( ) );
  return (FrBase*)NULL;
}

FrBase* IFrameStream::
skip_till_class( INT_2U LibraryClass )
{
  std::auto_ptr< FrBase >	base;

  do
  {
    base.reset( ReadObjectBase( ) );
    if ( base.get( ) == (FrBase*)NULL )
    {
      break;
    }
    if ( base->GetClassId( ) == m_dictionary.m_lib_to_stream[ LibraryClass ] )
    {
      return base.release( );
    }
    SkipObject( *base );
  } while ( ! m_stream.eof( ) );
  return (FrBase*)NULL;
}

#define	INSTANTIATE( LM_TYPE ) \
template IFrameStream& IFrameStream::operator>>( LM_TYPE & ); \
template IFrameStream& IFrameStream::ReadVector( std::vector< LM_TYPE >&, INT_4U )

INSTANTIATE( INT_2U );
INSTANTIATE( INT_2S );
INSTANTIATE( INT_4U );
INSTANTIATE( INT_4S );
INSTANTIATE( INT_8U );
INSTANTIATE( INT_8S );
INSTANTIATE( REAL_4 );
INSTANTIATE( REAL_8 );

#undef INSTANTIATE

#define	INSTANTIATE( LM_TYPE ) \
template IFrameStream& IFrameStream::operator>>( Container< LM_TYPE >& )

#include "ContainerInstantiation.icc"

#undef INSTANTIATE

#endif /* FRAMECPP__IFRAME_STREAM_CC */
