#ifndef OFrameStream_CC
#define OFrameStream_CC

#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <algorithm>
#include <set>
#include <sstream>
#include <stdexcept>

#include "general/gpstime.hh"

#include "framecpp/Version3_4_5/input.hh"
#include "framecpp/Version3_4_5/fileheader.hh"

#include "framecpp/Version6/Container.hh"
#include "framecpp/Version6/FrameH.hh"
#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrDetector.hh"
#include "framecpp/Version6/FrEndOfFile.hh"
#include "framecpp/Version6/FrEndOfFrame.hh"
#include "framecpp/Version6/FrStatData.hh"
#include "framecpp/Version6/FrTOC.hh"
#include "framecpp/Version6/OFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"
#include "framecpp/Version6/SearchContainer.hh"
#include "framecpp/Version6/Util.hh"

#define LM_DEBUG 0

#if LM_DEBUG
#define	DBGMSG( a )  std::cerr << a << __FILE__ << " " << __LINE__ << std::endl
#define METHOD( a ) static const char* method = #a;
#else
#define DBGMSG( a )
#define METHOD( a )
#endif

#define	AT() DBGMSG( "" )
#define	INFO( a )  DBGMSG( a << " " )

using namespace FrameCPP::Version_6;

namespace
{
  class TOC_CB_type: public Dictionary::Callback
  {
  public:
    TOC_CB_type( OFrameStream& Stream, FrTOC& TOC )
      : m_stream( Stream ),
	m_frame_count( 0 ),
	m_toc( TOC )
    {
    }
  
    void operator()( FrBase& Base );
    
  private:
    OFrameStream&	m_stream;
    INT_8U		m_toc_pos;
    INT_4U		m_frame_count;
    FrTOC&		m_toc;
  };

   
  void TOC_CB_type::operator()( FrBase& Base )
  {
      switch( Base.GetClassId( ) )
      {
      case CLASS_FRAME_H:
	m_frame_count++;
	break;
      case CLASS_FR_END_OF_FILE:
	{
	  // The End of File is about to be written.
	  FrEndOfFile& eof( dynamic_cast< FrEndOfFile& >( Base ) );
	  INT_8U	end( m_stream.tellp( ) + eof.Bytes( ) );
	  eof.SetSeekTOC( end - m_toc_pos );
	  eof.SetNBytes( end );
	  eof.SetNFrames( m_frame_count );
	}
	break;
      case CLASS_FR_TOC:
	// The TOC is about to be written.
	// Record were in the stream we are
	m_toc_pos = m_stream.tellp( );
	break;
      }
  }
   
   
   
  class InstanceSetter: public Dictionary::Callback
  {
  private:
    typedef std::set< FrBase* >	instance_set_objects_type;
  public:
    inline InstanceSetter( INT_4U NumberOfClasses )
    {
      m_next_instance.resize( NumberOfClasses );
      std::fill( m_next_instance.begin( ), m_next_instance.end( ), 0 );
    }
    
    inline void Reset( )
    {
      std::fill ( m_next_instance.begin( ), m_next_instance.end( ), 0 );
    }

    void operator()( FrBase& Base );

  private:
    std::vector< INT_4U >	m_next_instance;
  };

   
  inline void
  InstanceSetter::operator()( FrBase& Base )
  {
      if ( Base.IsInstanceSet( ) == false )
      {
	Base.SetInstance( m_next_instance[ Base.GetClassId( ) ] );
#if LM_DEBUG
	if ( Base.GetClassId( ) == CLASS_FR_DETECTOR )
	{
	  INFO( "DEBUG: Setting: class: " << Base.GetClassId( )
		<< " to: " << Base.GetInstance( ) );
	}
#endif
	m_next_instance[ Base.GetClassId( ) ]++;
      }
   
      return;
  }
   
   
  void compress_object( FrBase* Object,
			FrVect::compression_scheme_type Method,
			int Level )
  {
    if ( dynamic_cast< FrVect* >( Object ) )
    {
      FrVect* vect( dynamic_cast< FrVect* >( Object ) );
      vect->Compress( Method, Level );
    }
  }

} // namespace - anonymous

const std::string OFrameStream::DEFAULT_ORIGINATOR( "IGWD" );

OFrameStream::
OFrameStream( std::ostream& Stream,
	      CheckSum::kind_type CheckSumMethod,
	      const std::string& Originator )
  : Version_6::FrameStreamBase( GetClassIO( ), CLASS_LAST_ENTRY ),
    Common::OFrameStream( Stream ),
    m_print_stack(),
    m_print_stack_insertion( m_print_stack.end() )
{
  init( CLASS_LAST_ENTRY, CheckSumMethod, Originator );
}

OFrameStream::
OFrameStream( std::ostream& Stream,
	      const std::vector< Dictionary::io_calls_type >& LibraryDefs,
	      INT_4U NumberOfClassTypes,
	      CheckSum::kind_type CheckSumMethod,
	      const std::string& Originator )
  : Version_6::FrameStreamBase( LibraryDefs, NumberOfClassTypes ),
    Common::OFrameStream( Stream ),
    m_print_stack(),
    m_print_stack_insertion( m_print_stack.end() )
{
  init( NumberOfClassTypes, CheckSumMethod, Originator );
}

OFrameStream::
~OFrameStream( )
{
  close( );
}

//-------------------------------------------------------------------

namespace FrameCPP
{
  namespace Version_6
  {
    template <>
    OFrameStream& OFrameStream::
    operator<<( const STRING& Data )
    {
      STRING::length_of_string_type l( Data.length( ) + 1 );
      *this << l;
      write( Data.c_str( ), l );
      return *this;
    }

    template< >
    OFrameStream& OFrameStream::
    operator<<( const PTR_STRUCT& Data )
    {
      *this << Data.data_class << Data.data_instance;
      return *this;
    }

    template < >
    OFrameStream& OFrameStream::
    operator<<( const std::vector< STRING >& Data )
    {
      for ( std::vector< STRING >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	*this << *d;
      }
      return *this;
    }

    template<>
    OFrameStream& OFrameStream::
    operator<<( const std::vector< std::pair< STRING, REAL_4> >& Data )
    {
      typedef std::vector< std::pair< STRING, REAL_4> >	param_list_type;
      INT_2U	nParam( Data.size( ) );
  
      std::vector< REAL_4 >	parameters;
      std::vector< STRING > parameterNames;
  
      for ( param_list_type::const_iterator i ( Data.begin( ) );
	    i != Data.end( );
	    i++ )
      {
	parameters.push_back( (*i).second );
	parameterNames.push_back( (*i).first );
      }
  
      *this << nParam
	    << parameters
	    << parameterNames;

      return *this;
    }

    template < >
    OFrameStream& OFrameStream::
    operator<<( const std::string& Data )
    {
      STRING::length_of_string_type l( Data.length( ) + 1 );
      *this << l;
      write( Data.c_str( ), l );
      return *this;
    }

    template < >
    OFrameStream& OFrameStream::
    operator<<( const std::list< STRING >& Data )
    {
      for ( std::list< STRING >::const_iterator d( Data.begin( ) );
	    d != Data.end( );
	    d++ )
      {
	*this << *d;
      }
      return *this;
    }

    template< >
    OFrameStream& OFrameStream::
    operator<<( const General::GPSTime& Data )
    {
      *this << Data.GetSeconds( ) << Data.GetNanoseconds( );
      return *this;
    }

    template< >
    OFrameStream& OFrameStream::
    operator<<( const FrBase* Object )
    {
      WriteFrBasePtr( Object );
      return *this;
    }
    
    //-------------------------------------------------------------------

    template< class T >
    OFrameStream& OFrameStream::
    operator<<( Container< T >& Container )
    {
      PTR_STRUCT	ref( NULL_PTR_STRUCT );
  
      if ( Container.size( ) > 0 )
      {
	for( typename FrameCPP::ContainerBasic< T >::iterator
	       c( Container.begin( ) );
	     c != Container.end( );
	     c++ )
	{
	  (*c)->UnsetInstance( );
	  m_instantiate_cb->operator()( *(*c) );	// Establish instance
	  m_print_stack_insertion =
	    m_print_stack.insert( m_print_stack_insertion,
				  *c );			// Put on stack to dump
	}
	for( typename FrameCPP::ContainerBasic< T >::reverse_iterator
	       c( Container.rbegin( ) );
	     c != Container.rend( );
	     c++ )
	{
	  (*c)->SetNext( ref );
	  ref.data_class = (*c)->GetClassId( );
	  ref.data_instance = (*c)->GetInstance( );
	}
      }
      *this << ref;
      return *this;
    }

    //-------------------------------------------------------------------

    template< class T >
    OFrameStream& OFrameStream::
    operator<<( const Container< T >& Container )
    {
      PTR_STRUCT	ref( NULL_PTR_STRUCT );
  
      if ( Container.size( ) > 0 )
      {
	for( typename FrameCPP::ContainerBasic< T >::const_iterator
	       c( Container.begin( ) );
	     c != Container.end( );
	     c++ )
	{
	  (*c)->UnsetInstance( );
	  m_instantiate_cb->operator()( *(*c) );	// Establish instance
	  m_print_stack_insertion =
	    m_print_stack.insert( m_print_stack_insertion,
				  *c );			// Put on stack to dump
	}
	for( typename FrameCPP::ContainerBasic< T >::const_reverse_iterator
	       c( Container.rbegin( ) );
	     c != Container.rend( );
	     c++ )
	{
	  (*c)->SetNext( ref );
	  ref.data_class = (*c)->GetClassId( );
	  ref.data_instance = (*c)->GetInstance( );
	}
      }
      *this << ref;
      return *this;
    }

    //-----------------------------------------------------------------------

    template< class T >
    OFrameStream& OFrameStream::
    operator<<( const T& Data )
    {
      Common::OFrameStream::operator<<( Data );
      return *this;
    }

    //-----------------------------------------------------------------------

    template< class T >
    OFrameStream& OFrameStream::
    operator<<( const std::vector< T >& Data )
    {
      Common::OFrameStream::operator<<( Data );
      return *this;
    }
  } // namespace - Version_6
} // namespace - FrameCPP

//-----------------------------------------------------------------------

void OFrameStream::
AppendObject( FrBase& Object )
{
  seekp( 0, std::ios::end );		// Go to end of stream
  m_dictionary.Write( *this, Object );	// Write the Object
}

void OFrameStream::
WriteDictionary( FrBase& Base )
{
  m_dictionary.Write( *this, Base );
}

void OFrameStream::
WriteDictionary( const FrBase& Base )
{
  m_dictionary.Write( *this, const_cast< FrBase& >( Base ) );
}

void OFrameStream::
WriteFrame( FrameH& Frame, CheckSum::kind_type CheckSumMethod )
{
  INFO( "OFrameStream::WriteFrame - enter" );
  dynamic_cast< InstanceSetter* >( m_instantiate_cb )->Reset( );
  SetChecksumMethodFrame( CheckSumMethod );

  WriteObject( Frame );
  FrEndOfFrame	end_of_frame( Frame.GetRun( ), Frame.GetFrame( ) );
  m_dictionary.Write( *this, end_of_frame );
  // Cleanup the instance counts
  dynamic_cast< InstanceSetter* >( m_instantiate_cb )->Reset( );
  INFO( "OFrameStream::WriteFrame - exit" );
}

void OFrameStream::
WriteFrame( FrameH& Frame,
	    FrVect::compression_scheme_type Method,
	    int Level,
	    CheckSum::kind_type CheckSumMethod )
{
  INFO( "OFrameStream::WriteFrame - enter" );
  dynamic_cast< InstanceSetter* >( m_instantiate_cb )->Reset( );

  SetChecksumMethodFrame( CheckSumMethod );

  WriteObject( Frame, true, Method, Level );
  FrEndOfFrame	end_of_frame( Frame.GetRun( ), Frame.GetFrame( ) );
  //  WriteObject( end_of_frame );
  m_dictionary.Write( *this, end_of_frame );
  // Cleanup the instance counts
  dynamic_cast< InstanceSetter* >( m_instantiate_cb )->Reset( );
  INFO( "OFrameStream::WriteFrame - exit" );
}

void OFrameStream::
WriteFrBasePtr( const FrBase* Object )
{
  INFO( "OFrameStream::WriteFrBasePtr - enter" );
  if ( Object == (const FrBase*)NULL )
  {
    *this << NULL_PTR_STRUCT;
  }
  else
  {
    PTR_STRUCT	ref( Object->GetClassId( ), Object->GetInstance( ) );
    //---------------------------------------------------------------
    // Print PTR_STRUCT to stream
    //---------------------------------------------------------------
    *this << ref;
    //---------------------------------------------------------------
    // Put on stack to dump
    //---------------------------------------------------------------
    m_print_stack_insertion =
      m_print_stack.insert( m_print_stack_insertion,
			    const_cast<FrBase*>( Object ) );
  }
  INFO( "OFrameStream::WriteFrBasePtr - exit" );
}

void OFrameStream::
WriteFrStatData( FrStatData& StatData )
{
  INFO( "OFrameStream::WriteFrStatData - enter" );
  m_fr_stat_map[ StatData.GetDetector( ) ].push_back( &StatData );
  INFO( "OFrameStream::WriteFrStatData - exit" );
}

void OFrameStream::
WriteObject( FrBase& Object,
	     bool Compress,
	     FrVect::compression_scheme_type Method,
	     int Level )
{
  METHOD( OFrameStream::WriteObject );

  INFO( method << " - enter" );
  print_stack_type::size_type	start( m_print_stack.size( ) );

  FrBase* obj;
  
  m_print_stack.push_back( &Object );
  while ( m_print_stack.size( ) > start )
  {
    obj = m_print_stack.back( );
    INFO( method << " Writing obj: " << *obj );
    if ( Compress )
    {
      // Ensure that all compressable objects are compressed
      compress_object( obj, Method, Level );
    }
    if ( ( m_fr_stat_map.size( ) ) &&
	 ( obj->GetClassId( ) == CLASS_FR_DETECTOR ) )
    {
      //-----------------------------------------------------------------
      // Here we put the FrStatData structures onto the print stack
      //   since there is a requirement that the first must appear
      //   after the detector that it references.
      // The spec is silent on what to do about the others so just
      //   put them all after the detector they reference.
      //-----------------------------------------------------------------
      fr_stat_map_type::iterator	fr_stats( m_fr_stat_map.find( obj ) );
      if ( fr_stats != m_fr_stat_map.end( ) )
      {
	fr_stat_list_type::const_reverse_iterator
	  rend( (*fr_stats).second.rend( ) );
	for ( fr_stat_list_type::const_reverse_iterator
		l( (*fr_stats).second.rbegin( ) );
	      l != rend;
	      l++ )
	{
	  m_print_stack.push_back( *l );
	}
	m_fr_stat_map.erase( fr_stats );
	obj = m_print_stack.back( );
      }
    }
    m_print_stack.pop_back( );
    m_print_stack_insertion = m_print_stack.end( );
    m_dictionary.Write( *this, *obj );
    INFO( method << " Wrote obj: " << *obj );
  }
  INFO( method << " - exit" );
}

void OFrameStream::
close( )
{
  METHOD( OFrameStream::close );

  INFO( method << " - enter" );
  if ( is_open( ) )
  {
    INFO( method << ": File is in open state" );
    FrEndOfFile					end_of_file;
    // Write the FrTOC
    m_dictionary.Write( *this, m_toc );
    // Write the FrEndOfFile
    m_dictionary.Write( *this, end_of_file );
    dynamic_cast< InstanceSetter* >( m_instantiate_cb )->Reset( );
    is_open( false );
  }
  INFO( method << ": Unregistering callbacks" );
  m_toc.UnregisterCallbacks( );
  for ( std::vector< Dictionary::callback_key_type >::const_iterator
	  i( m_active_callbacks.begin( ) );
	i != m_active_callbacks.end( );
	i++ )
  {
    RemoveCallback( *i );
  }
  m_active_callbacks.erase( m_active_callbacks.begin( ),
			    m_active_callbacks.end( ) );
  // Go through list of dynamically allocated callbacks and
  //   release the associate resources.
  for ( std::vector< Dictionary::Callback* >::iterator
	  i( m_callback_methods.begin( ) );
	i != m_callback_methods.end( );
	i++ )
  {
    delete (*i);
    (*i) = (Dictionary::Callback*)NULL;
  }
  m_callback_methods.resize( 0 );
  // This has already been freed when deleting objects from m_callback_methods
  m_instantiate_cb = 0;
  INFO( "OFrameStream::close - exit" );
}

void OFrameStream::
init( INT_4U NumberOfClassTypes,
      CheckSum::kind_type CheckSumMethod,
      const std::string& Originator )
{
  INFO( "OFrameStream::init - enter" );
  if ( good( ) == false )
  {
    std::ostringstream	oss;

    throw std::runtime_error( "Cannot create OFrameStream because ouput stream is bad" );
  }

  is_open( true );

  SetChecksumMethodFile( CheckSumMethod );

  TOC_CB_type* toc_cb( new TOC_CB_type( *this, m_toc ) );
  m_instantiate_cb = new InstanceSetter( NumberOfClassTypes );

  m_callback_methods.push_back( toc_cb );
  m_callback_methods.push_back( m_instantiate_cb );

  m_active_callbacks.push_back( AddCallback( CLASS_FRAME_H, *toc_cb ) );
  m_active_callbacks.push_back( AddCallback( CLASS_FR_END_OF_FILE,
					     *toc_cb ) );
  m_active_callbacks.push_back( AddCallback( CLASS_FR_TOC, *toc_cb ) );

  for ( INT_4U x = 0; x < NumberOfClassTypes; x++ )
  {
    m_active_callbacks.push_back( AddCallback( x, *m_instantiate_cb ) );
  }
  // Establish the originator of the frame
  m_header.SetOriginator( Originator );
  // Write the header
  m_header.write( *this );
  m_toc.RegisterCallbacks( m_dictionary, *this );
  INFO( "OFrameStream::init - exit" );
}

void OFrameStream::
write( FrBase& Object )
{
  
}

namespace FrameCPP
{
  namespace Version_6
  {
    //-----------------------------------------------------------------------

#define	INSTANTIATE( LM_TYPE ) \
template OFrameStream& OFrameStream::operator<<( const LM_TYPE & ); \
template OFrameStream& OFrameStream::operator<<( const std::vector< LM_TYPE >& )

    INSTANTIATE( INT_2U );
    INSTANTIATE( INT_2S );
    INSTANTIATE( INT_4U );
    INSTANTIATE( INT_4S );
    INSTANTIATE( INT_8U );
    INSTANTIATE( INT_8S );
    INSTANTIATE( REAL_4 );
    INSTANTIATE( REAL_8 );

#undef INSTANTIATE

//-----------------------------------------------------------------------

#define	INSTANTIATE( LM_TYPE ) \
template OFrameStream& OFrameStream::operator<<( Container< LM_TYPE >& ); \
template OFrameStream& OFrameStream::operator<<( const Container< LM_TYPE >& )

#include "ContainerInstantiation.icc"
#undef INSTANTIATE

  } // namespace - Version_6
} // namespace - FrameCPP

#endif /* OFrameStream_CC */
