//=======================================================================
// This program is useful in dumping out the information from a frame
// file.
//
// Currently supported modes are hex and post processed.
//
// :TODO: Currently the Version 6 frames are supported in the post
// :TODO:   process mode.
//
//=======================================================================
#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <iomanip>
#include <memory>
#include <stdexcept>
#include <string>

#include "general/fstream.hh"

#include "framecpp/Version6/FrCommon.hh"
#include "framecpp/Version6/FrSH.hh"
#include "framecpp/Version6/IFrameStream.hh"
#include "framecpp/Version6/PTR_STRUCT.hh"
#include "framecpp/Version6/Util.hh"
#include "framecpp/Version6/STRING.hh"

using namespace FrameCPP::Version;

#include "DumpObjects6.cc"

namespace
{
  std::map< INT_2U, std::string >	class_names_by_id;
  std::map< std::string, INT_2U >	class_names_by_name;
  std::list< INT_2U >			class_id_index;
  std::list< std::string >		class_name_index;
  Version6::DumpFrStruct		dump_struct_6( false, // Verbose
						       false, // Scientific
						       Version6::DumpFrStruct::OUTPUT_MODE_DUMP_OBJECTS );

  void dump_buffer_hex( IFrameStream& Stream, FrBase& Base );
  void dump_buffer_struct( IFrameStream& Stream, General::ifstream& IStream,
			   FrBase& Base );
  void dump_ptrs_6( IFrameStream& Stream, General::ifstream& IStream,
		    FrBase& Base );

  char* Program;

  enum {
    MODE_HEX,
    MODE_STRUCT
  };

  void usage( )
  {
    std::cerr << "Usage: "
	      << Program
	      << " [-e] [-S] [-s] [-h] [-m hex|struct]"
	      << " frame_file [ frame_file ... ]" << std::endl
	      << "\t-e - Print real numbers in scientific notation"
	      << std::endl
	      << "\t-S - silent with respect to the data" << std::endl
	      << "\t-s - silent mode " << std::endl
	      << "\t-h - display this help message" << std::endl
	      << "\t-m <mode> - specify the mode" << std::endl
	      << std::endl;
  }

  void
  dump_buffer( IFrameStream& Stream, General::ifstream& IStream,
	       FrBase& Base, const int Mode )
  {
    switch( Mode )
    {
    case MODE_HEX:
      dump_buffer_hex( Stream, Base );
      break;
    case MODE_STRUCT:
      dump_buffer_struct( Stream, IStream, Base );
      break;
    }
  }

  void
  dump_buffer_hex( IFrameStream& Stream, FrBase& Base )
  {
    static const int WRAP( 8 );

#if !defined(__GNUG__) || ( __GNUG__ >= 3 )
    std::basic_ios<char>	restore( std::cout.rdbuf( ) );
#endif

    INT_8U	size( Base.Skip( Stream ) );
    CHAR_U*	buffer( new CHAR_U[ size ] );

#if !defined(__GNUG__) || ( __GNUG__ >= 3 )
    restore.copyfmt( std::cout );
#endif

    Stream.read( buffer, size );
    INT_8U	x( 0 );

    std::cout << std::endl;
    while ( x < size )
    {
      if ( ( x % WRAP ) == 0 )
      {
	std::cout << std::setw( 5 )
		  << std::setfill( '0' )
		  << std::dec
		  << x
		  << ": ";
      }
      std::cout << std::setw( 2 )
		<< std::setfill( '0' )
		<< std::hex
		<< (int)buffer[ x ];
      if ( isprint( buffer[x] ) )
      {
	std::cout << "." << buffer[x];
      }
      else
      {
	std::cout << "  ";
      }
      x++;
      if ( ( x % WRAP ) == 0 )
      {
	std::cout << std::endl;
      }
      else if ( x != size )
      {
	std::cout << " ";
      }
    }
    if ( ! ( ( x % WRAP ) == 0 ) )
    {
      std::cout << std::endl;
    }

#if !defined(__GNUG__) || ( __GNUG__ >= 3 )
    std::cout.copyfmt( restore );
#else
    std::cout << std::dec
	      << std::setw( 0 )
	      << std::setfill(' ');
#endif
    delete buffer;
  }

  void
  dump_buffer_struct( IFrameStream& Stream, General::ifstream& IStream,
		      FrBase& Base )
  {
    std::string	name( class_names_by_id[ Base.GetClassId( ) ] );
    std::auto_ptr< FrBase >	obj;

#define READ_DYNAMIC( FrType ) \
    if ( name == #FrType ) \
    { \
      obj.reset( FrType::ReadDynamic( Stream, Base ) ); \
    }

    READ_DYNAMIC( FrSH )
    else READ_DYNAMIC( FrSE )
    else READ_DYNAMIC( FrAdcData )
    else READ_DYNAMIC( FrameH )
    else READ_DYNAMIC( FrDetector )
    else READ_DYNAMIC( FrEndOfFile )
    else READ_DYNAMIC( FrEndOfFrame )
    else READ_DYNAMIC( FrEvent )

    else READ_DYNAMIC( FrHistory )
    else READ_DYNAMIC( FrProcData )
    else READ_DYNAMIC( FrRawData )
    else READ_DYNAMIC( FrSerData )
    else READ_DYNAMIC( FrSimData )
    else READ_DYNAMIC( FrSimEvent )
    else READ_DYNAMIC( FrStatData )
    else READ_DYNAMIC( FrSummary )
    else READ_DYNAMIC( FrTable )
    else READ_DYNAMIC( FrTOC )
    else READ_DYNAMIC( FrVect )
    else
    {
      Stream.SkipObject( Base );
    }

#undef READ_DYNAMIC

    if ( obj.get( ) )
    {
      std::cout << std::endl;
      dump_struct_6.Dump( *(obj.get( ) ) );
      dump_ptrs_6( Stream, IStream, *(obj.get( ) ) );
    }

  }

  void
  dump_ptrs_6( IFrameStream& Stream, General::ifstream& IStream, FrBase& Base )
  {
    const char**	fields( (const char**)NULL );
    INT_4U		field_cnt( 0 );
    INT_4U		width( 10 );

    switch( Base.GetClassId( ) )
    {
    case CLASS_FRAME_H:
      {
	static const char* ptrs[] = {
	  "type",
	  "user",
	  "detectSim",
	  "detectProc",
	  "history",
	  "rawData",
	  "procData",
	  "simData",
	  "event",
	  "simEvent",
	  "summaryData",
	  "auxData",
	  "auxTable",
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
	width = 12;
      }
      break;
    case CLASS_FR_ADC_DATA:
      {
	static const char* ptrs[] = {
	  "data",
	  "aux",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_DETECTOR:
      {
	static const char* ptrs[] = {
	  "aux",
	  "table",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
	width = 15;
      }
      break;
    case CLASS_FR_EVENT:
      {
	static const char* ptrs[] = {
	  "data",
	  "table",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_HISTORY:
      {
	static const char* ptrs[] = {
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_MSG:
      {
	static const char* ptrs[] = {
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_PROC_DATA:
      {
	static const char* ptrs[] = {
	  "data",
	  "aux",
	  "table",
	  "history",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
    case CLASS_FR_RAW_DATA:
      break;
      {
	static const char* ptrs[] = {
	  "firstSer",
	  "firstAdc",
	  "firstTable",
	  "logMsg",
	  "more"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_SER_DATA:
      {
	static const char* ptrs[] = {
	  "serial",
	  "table",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_SIM_DATA:
      {
	static const char* ptrs[] = {
	  "data",
	  "input",
	  "table",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_SIM_EVENT:
      {
	static const char* ptrs[] = {
	  "data",
	  "table",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_STAT_DATA:
      {
	static const char* ptrs[] = {
	  "detector",
	  "data",
	  "table"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_SUMMARY:
      {
	static const char* ptrs[] = {
	  "moments",
	  "table",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_TABLE:
      {
	static const char* ptrs[] = {
	  "column",
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    case CLASS_FR_VECT:
      {
	static const char* ptrs[] = {
	  "next"
	};
	fields = ptrs;
	field_cnt = sizeof(ptrs)/sizeof(*ptrs);
      }
      break;
    default:
      // There is nothing to do. Return to the caller
      return;
      break;
    }
    std::string		pad( "  " );
    PTR_STRUCT		ref;
    Version6::dump_label l( std::cout, pad, width, false );
    IStream.seekg( -( PTR_STRUCT::Bytes( ) * field_cnt ), std::ios::cur );
    for ( INT_4U x = 0; x < field_cnt; ++x )
    {
      Stream >> ref;
      l( fields[ x ], ref );
    }
  }

}

int
main( int ArgC, char**ArgV )
try
{
  Program = ArgV[ 0 ];

  int opt;

  FrSH					frsh;
  bool					verbose( true );
  int					mode( MODE_STRUCT );

  class_names_by_id[ 1 ] = "FrSH";
  class_names_by_id[ 2 ] = "FrSE";

  class_names_by_name[ "FrSH" ] = 1;
  class_names_by_name[ "FrSE" ] = 2;

  while( ( opt = getopt( ArgC, ArgV, "ehm:Ss" ) ) != -1 )
  {
    switch( opt )
    {
    case 'e': /* scientific notation */
      dump_struct_6.SetScientific( true );
      break;
    case 'h':	/* help */
      exit( 1 );
      break;
    case 'S': /* silent with respect to the data */
      dump_struct_6.SetVerbosity( false );
      break;
    case 'm':
      if ( strcmp( optarg, "hex" ) == 0 )
      {
	mode = MODE_HEX;
      }
      else if ( strcmp( optarg, "struct" ) == 0 )
      {
	mode = MODE_STRUCT;
      }
      break;
    case 's':	/* silent */
      verbose = false;
      break;
    default:
      break;
    }
  }

  ArgC -= optind;
  ArgV += optind;

  if ( ArgC < 1 )
  {
    usage( );
    exit( 1 );
  }
  for ( int x = 0; x < ArgC; x++ )
  {
    General::ifstream ifile( ArgV[ x ] );

    if ( ! ifile.good( ) )
    {
      continue;
    }

    IFrameStream  frame_stream( ifile, IFrameStream::MODE_SEQUENTIAL );

    if ( ! frame_stream.good( ) )
    {
      continue;
    }

    int frame_dump_mode = mode;

    std::cout << "Frame Version: " << frame_stream.GetVersion( )
	      << " Library Revision: " << frame_stream.GetLibraryRevision( )
	      << " Originator: " << frame_stream.GetOriginator( )
	      << std::endl;
    switch( frame_stream.GetVersion( ) )
    {
    case 6:
      break;
    default:
      frame_dump_mode = MODE_HEX;
      break;
    }

    INT_8U	pos;

    FrBase	base( FrBase::CLASS_NULL );

    while( frame_stream.good( ) )
    {
      pos = frame_stream.tellg( );
      base.read( frame_stream );
      if ( frame_stream.good( ) )
      {
	std::cout << "Position: " << pos;
	std::cout << " Class: " << base.GetClassId( )
		  << " (" << class_names_by_id[ base.GetClassId( ) ] << ")"
		  << " Instance: " << base.GetInstance( )
		  << " Length: " << base.GetLength( );
	if ( base.GetClassId( ) == FrBase::CLASS_FR_SH )
	{
	  INT_8U pos( frame_stream.tellg( ) );
	  frsh.read( frame_stream, base );
	  class_names_by_id[ frsh.GetClass( ) ] = frsh.GetName( );
	  class_id_index.push_back( frsh.GetClass( ) );
	  class_names_by_name[ frsh.GetName( ) ] = frsh.GetClass( );
	  class_name_index.push_back( frsh.GetName( ) );
	  if ( verbose )
	  {
	    // Rewind so the object can be reread
	    ifile.seekg( pos, std::ios::beg );
	    dump_buffer( frame_stream, ifile, base, frame_dump_mode );
	  }
	}
	else if ( verbose )
	{
	  dump_buffer( frame_stream, ifile, base, frame_dump_mode );
	}
	else
	{
	  frame_stream.SkipObject( base );
	}
      }
      std::cout << std::endl;
    }

    class_id_index.sort( );
    class_name_index.sort( );

    for( std::list< INT_2U >::const_iterator i ( class_id_index.begin( ) );
	 i != class_id_index.end( );
	 i++ )
    {
      std::cout << (*i) << " => " << class_names_by_id[ (*i ) ]
		<< std::endl;
    }
    for( std::list< std::string >::const_iterator i ( class_name_index.begin( ) );
	 i != class_name_index.end( );
	 i++ )
    {
      std::cout << (*i) << " => " << class_names_by_name[ (*i ) ]
		<< std::endl;
    }
  }
  exit( 0 );
}
catch( std::exception& e )
{
  std::cerr << "ABORT: Caught exception: " << e.what( ) << std::endl;
  exit( 1 );
}
