#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>

#include <fstream>
#include <dirent.h>
#include <fnmatch.h>

#include "framedir.hh"
#include "FrameF.hh"

FrameDir::~FrameDir(void) {
    clear();
}

FrameDir::FrameDir(void) 
  : mDebug(0)
{
}

FrameDir::FrameDir(const char* dir) 
  : mDebug(0)
{
    add(dir);
}

const ffData& 
FrameDir::find(const Time& t) const throw(NoData) {
    file_iterator iter = mList.upper_bound(t.getS());
    if (iter == mList.begin()) throw NoData("Specified data not avaiable");
    return (*(--iter)).second;
}

FrameDir::gps_t 
FrameDir::getStart(gps_t time) const {
    file_iterator iter = mList.lower_bound(time);
    if (iter == mList.end()) return 0;
    return iter->second.getStartTime().getS();
}

FrameDir::gps_t 
FrameDir::getLast(gps_t time) const {
    file_iterator iter = mList.upper_bound(time);
    if (iter == mList.begin()) return 0;
    gps_t t = (--iter)->second.getEndGPS();
    while ((iter != mList.end()) && (t == iter->second.getEndGPS())) {
        t = iter->second.getEndGPS();
	iter++;
    }
    return t;
}

void
FrameDir::add(const char* dir) {
    if (!dir || !*dir) return;
    std::string entry(dir);

    //----------------------------------  Add File if no wildcards.
    // std::cout << "Directory: " << entry << std::endl;
    unsigned int wcpos = entry.find("*");
    if (wcpos >= entry.size()) {
        addFile(entry.c_str());
    } else {

        //------------------------------  Find the first name level with a *
        unsigned int last=0, next=0;
	while (next <= wcpos) {
	    last = next;
	    next = entry.substr(last).find("/");
	    if (next > entry.length()-last) next = entry.length()-last;
	    next += last + 1;
	}

	std::string direc(entry.substr(0,last));
	if (direc.empty()) direc = ".";

	//----------------------------------  Open the directory.
	DIR* dd = opendir(direc.c_str());
	if (!dd) {
	    std::cerr << "Directory " << direc << " is unknown" << std::endl;
	    return;
	}

	//------------------------------  Scan for files matching the pattern
	std::string pattern=entry.substr(last, next-1-last);
	// std::cout << "  Test against: " << pattern << std::endl;
	bool atomic;
	if (next >= entry.length()) {
	    atomic = true;
	    wcpos  = next;
	} else {
	    wcpos  = entry.substr(next).find("*") + next;
	    atomic = (wcpos >= entry.size());
	}
	for (dirent* dirt=readdir(dd) ; dirt ; dirt = readdir(dd)) {
	    if (!fnmatch(pattern.c_str(), dirt->d_name, 0)) {
	        std::string path(entry);
		path.replace(last, next-last-1, dirt->d_name);
		// std::cout << "  testing: " << path << std::endl;
	        if (atomic)  addFile(path.c_str());
		else         add(path.c_str());
	    }
	}
	closedir(dd);
	if (wcpos >= entry.size()) checkData();
    }
}

void
FrameDir::addFile(const char* File) throw(BadFile) {
    if (!File || !*File) return;
    if (!strcmp(".", File) || !strcmp("..",File)) return;
    if (getDebug()) std::cout << "Adding file: " << File << std::endl;

    //----------------------------------  Find last number in the file name
    gps_t time=0;
    for (const char* ptr=File ; *ptr ; ) {
        if (*ptr >= '0' && *ptr <= '9') {
	    time = strtol(ptr, (char**)&ptr, 0);
	} else {
	    ptr++;
	}
    }

    //----------------------------------  Use the number as a GPS    
    if (time) {
        mList[time] = ffData(File, Time(time, 0));
	if (getDebug()) std::cout << "Found time: " << time << std::endl;
    } else {
        std::cerr << "no number in file name `" << File << "'" << std::endl;
    }
}

void
FrameDir::remove(const char* dir) {
    for (file_iterator iter=begin() ; iter != end() ; ) {
        const ffData& d = iter->second;
	file_iterator next = iter;
	next++;
	if (!fnmatch(dir, d.getFile(), 0)) {
	    erase(iter->first);
	}
	iter = next;
    }
}

void
FrameDir::clear(void) {
    mList.clear();
}

void
FrameDir::erase(gps_t time) {
    mList.erase(time);
}

void
FrameDir::checkData(CheckLevel lvl) {

    //----------------------------------  Make sure there's data to process.
    if (begin() == end()) return;

    //----------------------------------  Swich on mode.
    ffData::count_t lastNFrame(1);
    Interval        lastDt(1.0);
    switch(lvl) {
    case none:
        break;

    //----------------------------------  Try to fill time gaps;
    case gapsOnly:
        for (file_iterator iter=begin() ; iter != end() ; ) {
	    const ffData& d = iter->second;
	    file_iterator next = iter;
	    next++;
	    if (next == end()) break;
	    if (!Almost(d.getEndTime(), next->second.getStartTime())) {
	        Interval Delta=next->second.getStartTime() - d.getEndTime();
	        if (Delta == lastDt*double(lastNFrame)) {
		    if (getDebug()) {
		        std::cout << "Inferring file: " << d.getFile()
			          << " parameters from previous file." << std::endl;
		    }
		    mList[iter->first]=ffData(d.getFile(), d.getStartTime(),
					      lastNFrame, lastDt, false);
		} else {
		    if (getDebug()) {
		        std::cout << "Getting parameters from file: " 
			          << d.getFile() << "." << std::endl;
		    }
		    try {
		        ffData newFdata = getFileData(d.getFile());
			mList[iter->first] = newFdata;
			lastNFrame = newFdata.getNFrames();
			lastDt     = newFdata.getDt();
		    } catch (BadFile& b) {
		        if (getDebug()) {
			    std::cout << "Exception: " << b.what() 
				 << " caught while processing file " 
				 << d.getFile() << std::endl;
			}
			mList.erase(iter->first);
		    }
		}
	    }
	    iter = next;
	}
	break;

    case allData:
        for (file_iterator iter=begin() ; iter != end() ; ) {
	    const ffData& d = iter->second;
	    file_iterator next = iter;
	    next++;
	    if (!d.isValid()) {
	        try {
	            mList[iter->first] = getFileData(d.getFile());
		} catch (BadFile& b) {
		    if (getDebug()) {
		        std::cout << "Exception: " << b.what() 
			     << " caught while processing file " 
			     << d.getFile() << std::endl;
		    }
		    mList.erase(iter->first);
		}
	    }
	    iter = next;
	}
	break;
    }
}

ffData
FrameDir::getFileData(const char* File) throw (BadFile) {
    ffData::count_t NFrame(1);
    Time Start(0);
    Interval Dt(0.0);

    //----------------------------------  Read in the file header
    std::ifstream s(File, std::ios::in);
    FrameF f(s);
    if (!f.isOK()) throw BadFile("Unable to open File");

    //----------------------------------  Find the FrHeader structure type
    short IDFrHeader(0);
    while (!IDFrHeader && f.NxStruct()) {
        if (f.getID() == 1) {
	    std::string StructName = f.getString();
	    if (StructName == "FrameH") IDFrHeader = f.getShort();
	}
    }
    if (!IDFrHeader) throw BadFile("No FrameH definition");

    //----------------------------------  Look for the first header
    while (f.NxStruct() && f.getID() != IDFrHeader);
    if (f.getID() != IDFrHeader) throw BadFile("Can't find a frame header");

    //----------------------------------  Get the Start time and length
    f.getString();
    if (f.getVersion() == 4) f.Skip(3*sizeof(int));
    else                     f.Skip(2*sizeof(int));
    Start.setS(f.getInt());
    Start.setN(f.getInt());
    f.Skip(sizeof(short)+  sizeof(int));
    Dt = Interval(f.getDouble());

    //----------------------------------  Get the number of frames from the 
    struct FrEOF {
      int nFrames;
      int nBytes;
      int chkFlag;
      int chkSum;
    } eofstruc;
    f.Seek(-(sizeof(FrEOF) + 8), std::ios::end);
    if (f.NxStruct() && f.getLength() == (sizeof(FrEOF) + 8)) {
        eofstruc.nFrames = f.getInt();
        eofstruc.nBytes  = f.getInt();
        eofstruc.chkFlag = f.getInt();
        eofstruc.chkSum  = f.getInt();
	NFrame = eofstruc.nFrames;
	if (!NFrame) NFrame = 1;
    }

    s.close();
    return ffData(File, Start, NFrame, Dt, true);
}

