/*
  Create and write out several frame files. These frames
  are to be used for testing various LDAS functions. Specifically,
  we would like to be able to test currently unused frame data
  structures before they appear in frame builder written frames.
*/
#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <stdlib.h>
#include <unistd.h>

#include <cstdlib>
#include <cstdio>
#include <malloc.h>

#include <fstream>
#include <iostream>
#include <vector>

#include "framecpp/Version3_4_5/frame.hh"
#include "framecpp/Version3_4_5/history.hh"
#include "framecpp/Version3_4_5/rawdata.hh"
#include "framecpp/Version3_4_5/adcdata.hh"
#include "framecpp/Version3_4_5/detector.hh"
#include "framecpp/Version3_4_5/framewritertoc.hh"
#include "framecpp/Version3_4_5/frameappendwritertoc.hh"
#include "framecpp/Version3_4_5/framewriter.hh"
#include "framecpp/Version3_4_5/tocreader.hh"
#include "framecpp/Version3_4_5/framereader.hh"
#include "framecpp/Version3_4_5/dump.hh"
#include "framecpp/Version3_4_5/dimension.hh"
#include "framecpp/Version3_4_5/imageframewriter.hh"

#include "test.hh"

using namespace FrameCPP::Version_3_4_5;
using namespace std;


// The name of the program executable
char* programname = 0;
const char *stdTestFrameName = "std_test_frame";
bool ramp = false;
double amplitude = 1.0;
bool use_image_writer = false;

Frame* makeFrame(unsigned long gps, unsigned int frame_number, unsigned long dt)
{
  Frame* fr = new Frame(
			"LIGO",
			7, /* run */
			frame_number, /* frame number */
			Time( gps , 0 ), /* GPS time */
			13,
			(double)dt);

  // RawData
  RawData rawData( "" );
  fr->setRawData( rawData );

  if (ramp) {
    AdcData b( "ramp", 0, 1, 64, 16.0*1024.0, .00, 1.00, "counts",  .0, Time(0,0), 1 );
    fr->getRawData()->refAdc().append( b );
    {
      REAL_8 *d = new REAL_8[ 16 * 1024 * dt];
      for (unsigned long j = 0; j < dt; j++)
	for ( unsigned long i = 0 ; i < 16 * 1024; i++)
	  d[j*16*1024 + i] = amplitude * (((REAL_8)i)/(16.0*1024.0));
      Dimension ddims[ 1 ] = {
        Dimension( 16*1024*dt, 1.00/(16*1024*dt), "time", 1.00 )
      };
      Vect dvect( "ramp", 1, ddims, d, "counts" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );


      // ProcData
      ProcData p( "ramp", 16.0*1024.0, Time( 0, 0 ), .00, .0 );
      fr->refProcData().append( p );

      // SimData
      SimData sd( "ramp", "", 16.0*1024.0, .0, .0 );
      fr->refSimData().append( sd );
      fr->refSimData()[0]->refData().append( dvect );
      delete [] d;
    }
  } else {

    // Detector (both detectors are the same)
    //	
    // :TODO: put real Hanford values in here
    Detector d("TesT",
			 1.0, 2.0,
			 .0,
			 .0, .0,
			 .0, .0,
			 .0, .0,
			 3600, 0, "");
    fr->refDetectorSim().append( d );
    fr->refDetectorProc().append( d );

    // History 
    fr->refHistory().append( History( "History teaches us lessons", 666666777, "absolutely" ));
    fr->refHistory().append( History( "...", 666666888, "x" ));

    // AdcData
    AdcData a( "TesT", 2, 2, 2, 2.00, 1.00, 1.00, "TesT", 2.00,
			 Time(-2, 2), 1 );
    a.appendComment( "TesT" );
    fr->getRawData()->refAdc().append( a );
    AdcData b( "TesT1", 0, 1, 32, 16.00, 1.00, 1.00, "Volts", .0,
			 Time(0,0), 1 );
    fr->getRawData()->refAdc().append( b );
    
    // append long data vector to the seconds ADC
    {
      REAL_4 d[ 16* 32 ];
      for ( int i = 0 ; i < 16 * 32; i++)
	d[i]= (REAL_4)i;
      Dimension ddims[ 1 ] = {
        Dimension( 16*32, 1.00, "", 1.00 )
      };
      Vect dvect( "a", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[1]->refData().append( dvect );
    }

    // character vector
    {
      CHAR d[ 4 ] = { 'a', 'a', 'a', 'a' };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // unsigned character vector
    {
      CHAR_U d[ 4 ] = { 'b', 'b', 'b', 'b' };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // short vector
    {
      INT_2S d[ 4 ] = { -1, -1, -1, -1 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // unsigned short vector
    {
      INT_2U d[ 4 ] = { 1, 1, 1, 1 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // int vector
    {
      INT_4S d[ 4 ] = { -2, -2, -2, -2 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // unsigned int vector
    {
      INT_2U d[ 4 ] = { 2, 2, 2, 2 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // real vector
    {
      REAL_4 d[ 4 ] = { 1.00, 1.00, 1.00, 1.00 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    // real vector
    {
      REAL_8 d[ 4 ] = { 2.00, 2.00, 2.00, 2.00 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( dvect );
    }

    {
      COMPLEX_8 cdata[ 4 ];
      cdata[0] = COMPLEX_8(1.0, 1.2);
      cdata[1] = COMPLEX_8(1.3, 1.4);
      cdata[2] = COMPLEX_8(1.5, 1.6);
      cdata[3] = COMPLEX_8(1.7, 1.8);
      Dimension cdims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect cvect( "TesT", 1, cdims, cdata, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( cvect );
    }

    {
      COMPLEX_16 cdata[ 4 ];
      cdata[0] = COMPLEX_16(1.0, 1.2);
      cdata[1] = COMPLEX_16(1.3, 1.4);
      cdata[2] = COMPLEX_16(1.5, 1.6);
      cdata[3] = COMPLEX_16(1.7, 1.8);
      Dimension cdims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect cvect( "TesT", 1, cdims, cdata, "TesT" );
      fr->getRawData()->refAdc()[0]->refData().append( cvect );
    }

    // Log Message (a couple of them)
    Msg m( "TesT", "TesT", 2, Time(0, 0) );
    fr->getRawData()->refMsg().append( m );
    fr->getRawData()->refMsg().append( m );

    // ProcData
    ProcData p( "TesT", 2.00, Time( 2, 2 ), 2.00, .0 );
    p.appendComment("TesT");
    fr->refProcData().append( p );
    fr->refProcData().append( p );

    // SimData
    SimData sd( "TesT", "TesT", 1.00, .0, .0 );
    fr->refSimData().append( sd );
    fr->refSimData().append( sd );

    // SerData
    SerData srd( "TesT", Time( 2, 2 ), 1.00 );
    fr->getRawData()->refSerial().append( srd );
    fr->getRawData()->refSerial().append( srd );

    // StatData
    StatData statData("TesT", "TesT", "TesT", 2, 2, 2 );
    // real vector
    {
      REAL_8 d[ 4 ] = { 2.00, 2.00, 2.00, 2.00 };
      Dimension ddims[ 1 ] = {
	Dimension( 2, 2.00, "TesT", 2.00 )
      };
      Vect dvect( "TesT", 1, ddims, d, "TesT" );
      statData.refData().append( dvect );
    }

    statData.setDetector(fr->refDetectorSim()[0]);
    fr->refDetectorSim()[0]->refStatData().append( statData );
    fr->refDetectorSim()[0]->refStatData().append( statData );

    statData.setDetector(fr->refDetectorProc()[0]);
    fr->refDetectorProc()[0]->refStatData().append( statData );
    fr->refDetectorProc()[0]->refStatData().append( statData );

    // Event
    Event event(
        "TesT", "TesT", "TesT", Time( 2, 2 ),
	1.0, 1.0,
	2, 1.00,
        1.00, "TesT",
	std::vector<std::pair<std::string, REAL_4> >());
    fr->refEvent().append( event );
    fr->refEvent().append( event );

    // Summary
    Summary summary( "TesT", "TesT", "TesT", Time(0, 0) );
    fr->refSummary().append( summary );
    fr->refSummary().append( summary );

    // SimEvent
    SimEvent sed( "TesT", "TesT", Time(2, 2), 1.00, 1.00, 1.00 );
    sed.appendComment("TesT");
    fr->refSimEvent().append( sed );


    // Table
    Table table( "TesT", 1 );
    table.appendComment("TesT");
    fr->refTable().append( table );
  }
  return fr;
}

void error_watch (const string& msg)
{
  cerr << msg << endl;
}

// First frame time stamp
unsigned long secini = 666666667;
// Frames per file
unsigned long fpf = 1;
// Frame length
unsigned long dt = 1;
// Sequence length
unsigned long Dt = 10;
void
usage()
{
  cout << endl << "Function: Create a sequence of frame files with meaningfull data in them." << endl;
  cout << "The purpose of this is to allow testing of new frame structures in the LDAS" << endl;
  cout << "before the frame builder is modified to write them out. Thus we are" << endl;
  cout << "trying to achive a smooth LDAS transition to new data in the frames." << endl;
  cout << endl << "Usage: takes numeric argument which is start GPS time" << endl;
  cout << endl << "The following options are supported:" << endl;
  cout << '\t' << "-r\t-generate ramp data" << endl;
  cout << '\t' << "-a amplitude\t-set ramp signal amplitude" << endl;
  cout << '\t' << "-i gps\t-set initial frame timestamp (default values is " << secini << endl;
  cout << '\t' << "-f nframes\t-set the number of frames per file (default is 1)" << endl;
  cout << '\t' << "-l dt\t-set the number of seconds of data per frame (default is 1)" << endl;
  cout << '\t' << "-p Dt\t-set overall length in seconds" << endl;
  cout << '\t' << "-I\t-use image frame writer" << endl;
}


// :TODO: include same data as in ADC into Proc and Sim structures.

int
main(int argc, char* argv[]) try
{
  /* Determine program name (strip filesystem path) */
  for (programname = argv [0] + strlen (argv [0]) - 1;
       programname != argv [0] && *programname != '/';
       programname--)
    ;

  if (*programname == '/') 
    programname++;

  int c;
  extern char *optarg;
  extern int optind;
  bool errflg = false;
  while ((c = getopt(argc, argv, "i:a:f:l:p:?hrI")) != EOF) {
    switch (c) 
      {
      case 'I':
	use_image_writer = true;
	break;
      case 'a':
	{
	  char *endptr;
	  double amp = strtod(optarg, &endptr);
	  if (endptr == optarg) {
	    cerr << "Invalid numeric argument given to -" << c <<" option" << endl;
	    errflg = true;
	  } else
	    amplitude = amp;
	  break;
	}
      case 'r': {
	ramp = true;
	break;
      }
      case '?':
      case 'h':
	errflg = true;
	break;
      case 'f': 
      case 'l': 
      case 'p':
      case 'i':
	{
	  char *endptr;
	  unsigned long f = strtol(optarg, &endptr, 10);
	  if (endptr == optarg || f == 0) {
	    cerr << "Invalid numeric argument given to -" << c <<" option" << endl;
	    errflg = true;
	  } else {
	    switch (c)
	      {
	      case 'f':
		fpf = f;
		break;
	      case 'l': 
		dt = f;
		break;
	      case 'p':
		Dt = f;
		break;
	      case 'i':
		secini = f;
		break;
	      }
	  }
	    break;
	}
      }
  }
  if (errflg || optind < argc) {
    usage();
    exit (2);
  }

  unsigned long nframes = Dt/dt;
  if (nframes < 1)
    {
      cerr << "Invalid number of frames (Dt/dt)" << endl;
      exit(1);
    }

   
#ifdef  LDAS_USE_CGUC_3_2   

     //cout << "Using LDAS_USE_CGUC_3_2"   << endl;
  stringstream ost;
  int image_size = 0;
  ImageFrameWriter fw(ost);
  for (unsigned int i = 0; i < nframes; i += fpf) {
    char buf[1024];
    unsigned long tst = secini+i*dt;
    sprintf(buf,"H-%ld-%ld-%ld.F", tst, fpf, dt);
    ofstream out(buf);
    if (use_image_writer) {
      if (ost.str().size() == 0) {
        for (unsigned int j = 0; j < fpf; j++) {
          Frame* frame = makeFrame(tst + j*dt, j + i, dt);
          fw.writeFrame( *frame );
          delete frame;
        }
        fw.close();
        image_size = ost.str().size();
      } 

      fw.setFrameFileAttributes(7, i, 0, tst, dt, 0, 13, 8);
      out.write(ost.str().c_str(), image_size);
    } else {
      FrameWriterTOC fw( out );
      for (unsigned int j = 0; j < fpf; j++) {
        Frame* frame = makeFrame(tst + j*dt, j + i, dt);
        fw.writeFrame( *frame );
        delete frame;
      }
      fw.close();
    }
    out.close();
  }
   
#endif
   
   
#ifdef  LDAS_USE_CGUC_2_95
   
     //cout << "Using LDAS_USE_CGUC_2_95"   << endl;   
  strstream ost;
  int image_size = 0;
  ImageFrameWriter fw(ost);
  for (unsigned int i = 0; i < nframes; i += fpf) {
    char buf[1024];
    unsigned long tst = secini+i*dt;
    sprintf(buf,"H-%ld-%ld-%ld.F", tst, fpf, dt);
    ofstream out(buf);
    if (use_image_writer) {
      if (ost.pcount() == 0) {
        for (unsigned int j = 0; j < fpf; j++) {
          Frame* frame = makeFrame(tst + j*dt, j + i, dt);
          fw.writeFrame( *frame );
          delete frame;
        }
        fw.close();
        image_size = ost.pcount ();
      } 
/*
  inline void setFrameFileAttributes(INT_4S run, INT_4U frameNumber,
                                     INT_4U dqual, INT_4U gps, INT_2U gpsInc,
                                     INT_4U gpsn,
                                     INT_2U leapS, INT_4S localTime)
*/
      fw.setFrameFileAttributes(7, i, 0, tst, dt, 0, 13, 8);
      out.write(ost.str (), image_size);
    } else {
      FrameWriterTOC fw( out );
      for (unsigned int j = 0; j < fpf; j++) {
        Frame* frame = makeFrame(tst + j*dt, j + i, dt);
        fw.writeFrame( *frame );
        delete frame;
      }
      fw.close();
    }
    out.close();
  }   
   
#endif   
   
   
  return 0;
} catch (...) {
  return 1;
}
