/*
  frtoc -- filter to process a frame file: add the TOC (Table Of Contents)
*/

#include "framecpp/config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <fstream>
#include <iostream>

#include "framecpp/FrameH.hh"
#include "framecpp/FrameWriterTOC.hh"
#include "framecpp/FrameReader.hh"

using namespace FrameCPP;
using namespace std;

// The name of the program executable
char* programname = 0;

void
fatal( string msg ) {
  cerr << "FATAL: " << msg << endl;
  exit (1);
}

void
error_watch (const string& msg)
{
  cerr << "error: " << msg << endl;
}


void
usage( char* f )
{
  if ( ! strncasecmp( f, "-h", 2 ) 
       || ! strncasecmp( f, "--h", 3 ) )
    {
      cout << "Filter to process a frame file: add the Table Of Contents." << endl;
      cout << "Usage: " << (string)programname << " INFILE OUTFILE" << endl;
      cout << "As filter: " << (string)programname << " < INFILE > OUTFILE" << endl;
      exit(0);
    }
}

int
main( int argc, char* argv[] )
{  
  /* Determine program name (strip filesystem path) */
  for (programname = argv [0] + strlen (argv [0]) - 1;
       programname != argv [0] && *programname != '/';
       programname--)
    ;

  if (*programname == '/') 
    programname++;

  if ( argc > 1 )
    usage( argv[1] );
  
  int ifd, ofd;
  if ( argc < 3 )
    {
      if ( argc < 2 )
	{
	  ifd = 0;
	  ofd = 1;
	}
      else
	{
	  ifd = open( argv[ 1 ], O_RDONLY );
	  ofd = 1;
	}
  } else {
    ifd = open( argv[ 1 ], O_RDONLY );
    ofd = open( argv[ 2 ], O_CREAT|O_WRONLY|O_TRUNC, 0644 );
  }

  if ( ifd < 0 )
    {
      fatal( "could not open `" + (string)argv[1] + "' for input" );
    }

  if ( ofd < 0 )
    {
      fatal( "failed to open `" + (string)argv[2] + "' for output" );
    }
  

#if __GNUC__ >= 3
  std::ifstream in;
  {
    FILE *cfile = fdopen(ifd,"r");
    std::ofstream::__filebuf_type* outbuf
        = new std::ofstream::__filebuf_type (cfile, std::ios_base::out);
    std::ofstream::__streambuf_type* oldoutbuf
        = ((std::ofstream::__ios_type* )&in)->rdbuf (outbuf);
    delete oldoutbuf;
  }
  std::ofstream out;
  {
    FILE *cfile = fdopen(ofd,"w");
    std::ofstream::__filebuf_type* outbuf
        = new std::ofstream::__filebuf_type (cfile, std::ios_base::out);
    std::ofstream::__streambuf_type* oldoutbuf
        = ((std::ofstream::__ios_type* )&out)->rdbuf (outbuf);
    delete oldoutbuf;
  }
  
#else
  std::ifstream in( ifd );
  std::ofstream out( ofd );
#endif

  FrameReader fr( in );
  // fr.setErrorWatch( error_watch );
  FrameWriterTOC fw( out );

  // :TODO: why don't we have the error watch in the output class???
  //  fw.setErrorWatch( error_watch );

  try {
    while( fr.isEndOfFile( ) == false )
    {
      FrameH* frame = fr.readFrame();
      if (!frame)
	break;
      fw.writeFrame( *frame );
      delete frame;
    }
  } catch ( std::exception& e ) {
    fatal( e.what() );
  }

  fw.close();   // build TOC
  out.close();
  in.close();
  return 0;
}
