#include "general/config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

#include "general/regex.hh"
#include "general/regexmatch.hh"

#include "general/FileType.hh"
#include "general/System.hh"

namespace
{
  static Regex	BinaryRegex( "[^[:print:][:space:]]", REG_EXTENDED );
  static Regex	BZipRegex( "^BZh91AY\\&SY" );
  static Regex	ElfRegex( "^[\x7F]ELF", REG_EXTENDED );
  static Regex	EPSRegex( "[[:space:]]EPSF[-]" );
  static Regex	FrameRegex( "^IGWD" );
  static Regex	GIFRegex( "^GIF" );
  static Regex	GZipRegex( "^\x1f\x8b" );
  static Regex	HTMLRegex( "([<]html[>]|[<]a[[:space:]]+href=)",
			   REG_EXTENDED | REG_ICASE );
  static Regex	HTML404Regex( "404[[:space:]]+Not[[:space:]]+Found",
			      REG_EXTENDED | REG_ICASE );
  static Regex	HTMLErrorRegex( "Error[[:space:]]+[(][[:digit:]]{3}[)]",
				REG_EXTENDED | REG_ICASE );
  static Regex	ILWDRegex( "[<][?]ilwd[?][>]",
			   REG_EXTENDED | REG_ICASE );
  static Regex	ILWDFrameRegex( "LIGO::Frame" );
  static char	JPegTag[] = { '\xFF','\xD8','\xFF','\xE0','\x00','\x10','J','F','I','F' };
  static char	PDFTag[] = { '%','!','P','D','F','-' };
  static Regex	PGPRegex( "BEGIN PGP MESSAGE" );
  static char	PNGTag[] = { '\x89','P','N','G' };
  static char	PSTag[] = { '%', '!', 'P', 'S', '-' };
  static Regex	ScriptRegex( "^[#][!][[:space:]]*([^[:space:]]+)", REG_EXTENDED );
  static char	TiffTag[] = { 'M','M','\0','*' };
  static Regex	XMLRegex( "[<][?]xml", REG_ICASE );
  static Regex	XMLDocRegex( "[<]!DOCTYPE[[:space:]]+([^[:space:]]+)",
			     REG_EXTENDED | REG_ICASE );
} // namespace - anonymous
namespace General
{
  FileType::
  FileType( const std::string& Filename )
  {
    //-------------------------------------------------------------------
    // :TODO: Check for existance
    //-------------------------------------------------------------------
    bool is_link = false;
    struct stat	file_info;
    int	err = lstat( Filename.c_str( ), &file_info );

    if ( err != 0 )
    {
      std::ostringstream	msg;

      msg << "FileType error when accessing file: "
	  << Filename << " - "
	  << strerror( errno );
      throw std::runtime_error( msg.str( ) );
    }
    if ( file_info.st_mode == S_IFLNK )
    {
      is_link = true;
      err = stat( Filename.c_str( ), &file_info );

      if ( err != 0 )
      {
	std::ostringstream	msg;
	
	msg << "FileType error when accessing symbolic link file: "
	    << Filename << " - "
	    << strerror( errno );
	throw std::runtime_error( msg.str( ) );
      }
    }
    switch( file_info.st_mode & S_IFMT )
    {
    case S_IFIFO:
      m_type_desc += "fifio special";
      break;
    case S_IFCHR:
      m_type_desc += "character special";
      break;
    case S_IFDIR:
      m_type_desc += "directory";
      break;
    case S_IFBLK:
      m_type_desc += "block special";
      break;
    case S_IFREG:
      if ( file_info.st_size == 0 )
      {
	m_type_desc += "empty";
      }
      else
      {
	int fd = open( Filename.c_str( ), O_RDONLY );
	if ( fd < 0 )
	{
	  m_type_desc += "unknown";
	}
	else // if ( fd < 0 )
	{
	  char	buffer[ 1025 ];
	  RegexMatch	compare( 1 );
	  ssize_t	len =
	    read( fd, buffer, sizeof( buffer ) - 1);
          if ( len == (ssize_t)(-1) )
	  {
	    m_type_desc += "unknown (";
	    m_type_desc += General::System::ErrnoMessage( );
	    m_type_desc += " )";
	  }
	  else
	  {
	    buffer[ len ] = '\0';
	    bool binary = false;
	    for ( ssize_t x = 0;
		  ( binary == false ) && ( x < len );
		  ++x )
	    {
	      if ( ! ( isspace( buffer[ x ] ) ||
		       isprint( buffer[ x ] ) ) )
	      {
		binary = true;
	      }
	    } // for

	    //-----------------------------------------------------------
	    // Differentiate between binary and text
	    //-----------------------------------------------------------
	    if ( binary )
	    {
	      m_type_desc = "binary";
	    }
	    else
	    {
	      m_type_desc = "text";
	    }
	    //-----------------------------------------------------------
	    // 
	    //-----------------------------------------------------------
	    if ( compare.match( ScriptRegex, buffer ) )
	    {
	      m_type_desc += " ";
	      m_type_desc += compare.getSubString( 0 );
	    }
	    else if ( binary && compare.match( ElfRegex, buffer ) )
	    {
	      m_type_desc += " elf";
	    }
	    else if ( binary && compare.match( BZipRegex, buffer ) )
	    {
	      m_type_desc += " bzip";
	    }
	    else if ( binary && compare.match( GZipRegex, buffer ) )
	    {
	      m_type_desc += " gzip";
	    }
	    else if ( binary && compare.match( GIFRegex, buffer ) )
	    {
	      m_type_desc += " gif";
	    }
	    else if ( binary &&
		      ( strncmp( PNGTag, buffer, sizeof(PNGTag) ) == 0 ) )
	    {
	      m_type_desc += " png";
	    }
	    else if ( binary &&
		      ( strncmp( JPegTag, buffer, sizeof( JPegTag ) ) == 0 ) )
	    {
	      m_type_desc += " jpeg";
	    }
	    else if ( binary &&
		      ( strncmp( TiffTag, buffer, sizeof( TiffTag ) ) == 0 ) )
	    {
	      m_type_desc += " tiff";
	    }
	    else if ( binary &&	
		      ( strncmp( PDFTag, buffer, sizeof( PDFTag ) ) == 0 ) )
	    {
	      m_type_desc += " pdf";
	    }
	    else if ( !binary && compare.match( HTMLRegex, buffer ) )
	    {
	      m_type_desc += " html";
	      if ( compare.match( HTML404Regex, buffer ) )
	      {
		m_type_desc += " 404";
	      }
	      else if ( compare.match( HTMLErrorRegex, buffer ) )
	      {
		m_type_desc += " ";
		m_type_desc += compare.getSubString( 0 );
	      }
	    }
	    else if ( strncmp( PSTag, buffer, sizeof(PSTag) ) == 0 )
	    {
	      m_type_desc += " ps";
	      if ( compare.match( EPSRegex, buffer ) )
	      {
		m_type_desc += " eps";
	      }
	    }
	    else if ( compare.match( XMLRegex, buffer ) )
	    {
	      m_type_desc += " xml";
	      if ( compare.match( XMLDocRegex, buffer ) )
	      {
		m_type_desc += " ";
		m_type_desc += compare.getSubString( 0 );
	      }
	    }
	    else if ( compare.match( PGPRegex, buffer ) )
	    {
	      m_type_desc += " pgp";
	    }
	    else if ( compare.match( FrameRegex, buffer ) )
	    {
	      m_type_desc += " frame";
	    }
	    else if ( compare.match( ILWDRegex, buffer ) )
	    {
	      m_type_desc += " ilwd";
	      if ( compare.match( ILWDFrameRegex, buffer ) )
	      {
		m_type_desc += " ligo frame";
	      }
	    }
	    else
	    {
	      m_type_desc += " unknown";
	    }
	  } // if ( len < 0 )
	  close( fd );
	}
      }
      break;
    }
    if ( is_link )
    {
      m_type_desc += " link";
    }
  }
}
