#include "general/config.h"

#if HAVE_PROCFS_H
#if !defined(_LP64) && _FILE_OFFSET_BITS == 64
//-----------------------------------------------------------------------
// For 32-bit builds, cannot use LARGE FILE interface
//-----------------------------------------------------------------------
#undef _FILE_OFFSET_BITS
#undef _LARGE_FILES
#endif /* */

#include <sys/time.h>

#include <procfs.h>

#else	/* HAVE_PROCFS_H */

#include <sys/types.h>
#include <sys/sysinfo.h>
#include <sys/stat.h>

#endif /* HAVE_PROCFS_H */

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include <fstream>
#include <iostream>
#include <pwd.h>
#include <sstream>
#include <stdexcept>

#include "general/types.hh"
#include "general/PSInfo.hh"

namespace
{
#if !defined( HAVE_PROCFS_H )
  unsigned long get_total_ram( );

  static const int		PageSize = getpagesize( );
  static const long		Hertz = sysconf( _SC_CLK_TCK );
  static const unsigned long	TotalRam = get_total_ram( );
#endif /* !defined( HAVE_PROCFS_H ) */
}

General::PSInfo::
PSInfo( pid_t PID )
  : m_pid( PID )
{
}

void General::PSInfo::
operator()( General::PSInfo::psinfo_type& Query ) const
{
  char*			pwbuffer;
  struct passwd		pw_entry;
  struct passwd*	pw_entry_result;
  static const int	pwbuffer_size = sysconf( _SC_GETPW_R_SIZE_MAX );

#if HAVE_PROCFS_H
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  // Solaris
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  std::ostringstream	filename;
  psinfo_t		info;
  int			fd;
  struct timeval	now;

  filename << "/proc/" << m_pid << "/psinfo";
  fd = open( filename.str( ).c_str( ), O_RDONLY );
  if ( fd < 0 )
  {
    std::ostringstream	msg;
    msg << "Unable to open: " << filename.str( );
    throw std::runtime_error( msg.str( ) );
  }
  read( fd, &info, sizeof( info ) );
  close( fd );
  //---------------------------------------------------------------------
  // Obtain the current time
  //---------------------------------------------------------------------
  gettimeofday( &now, (void*)NULL );
  //---------------------------------------------------------------------
  // Translate user id to user name
  //---------------------------------------------------------------------
  pwbuffer = (char*)calloc( sizeof( char ), pwbuffer_size );
  getpwuid_r( info.pr_uid, &pw_entry,
	      pwbuffer, pwbuffer_size,
	      &pw_entry_result );
  if ( pw_entry_result )
  {
    Query.s_user = pw_entry_result->pw_name;
  }
  else
  {
    std::ostringstream	uid;

    uid << info.pr_uid;
    Query.s_user = uid.str( );
  }
  free( pwbuffer );
  //---------------------------------------------------------------------
  // Populate the information
  //---------------------------------------------------------------------
  Query.s_pid = m_pid;
  Query.s_state = ( info.pr_wstat == 0 ) ? 'R' : 'Z';
  Query.s_vsz = info.pr_size;
  Query.s_rsz = info.pr_rssize;
  Query.s_pcpu = ( info.pr_pctcpu / float( 0x8000 ) ) * 100.0;
  Query.s_pmem = ( info.pr_pctmem / float( 0x8000 ) ) * 100.0;
  Query.s_etime = ( now.tv_sec - info.pr_start.tv_sec );
  Query.s_fname = info.pr_fname;
  Query.s_args = info.pr_psargs;
#else /* HAVE_PROCFS_H */
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  // Linux
  //---------------------------------------------------------------------
  //---------------------------------------------------------------------
  // Extract information from /proc/<pid>/stat file
  //---------------------------------------------------------------------
  unsigned long	uptime, idle;

  std::ifstream		uptimed( "/proc/uptime" );

  uptimed >> uptime >> idle;
  uptimed.close( );

  std::ostringstream	filename;

  int 		pid, ppid, pgrp, session, tty_nr, tpgid;
  unsigned long	flags, minflt, cminflt, majflt, cmajflt,
    utime, stime, starttime, vsize;
  long		cutime, cstime, priority, nice, ph0, itrealvalue, rss;
  char		comm_buffer[256];
  char*		comm = comm_buffer;
  char		state;

  filename << "/proc/" << m_pid << "/stat";
  std::ifstream		statd( filename.str( ).c_str( ) );

  if ( ! statd.good( ) )
  {
    std::ostringstream	msg;
    msg << "Unable to open: " << filename.str( );
    throw std::runtime_error( msg.str( ) );
  }

  statd >> pid >> comm_buffer >> state >> ppid >> pgrp >> session
	>> tty_nr >> tpgid
	>> flags >> minflt >> cminflt >> majflt >> cmajflt
	>> utime >> stime >> cutime >> cstime >> priority >> nice >> ph0
	>> itrealvalue >> starttime >> vsize >> rss;
  statd.close( );

  if ( ( *comm == '(' ) &&
       ( comm[ strlen( comm ) - 1 ] == ')' ) )
  {
    ++comm;
    comm[ strlen( comm ) - 1 ] = '\0';
  }
  unsigned long seconds = uptime - ( starttime / Hertz );

  //---------------------------------------------------------------------
  // Translate user id to user name
  //---------------------------------------------------------------------
  struct stat	stat_buf;

  ::stat( filename.str( ).c_str( ), &stat_buf );
  pwbuffer = (char*)calloc( sizeof( char ), pwbuffer_size );
  getpwuid_r( stat_buf.st_uid, &pw_entry,
	      pwbuffer, pwbuffer_size,
	      &pw_entry_result );
  if ( pw_entry_result )
  {
    Query.s_user = pw_entry_result->pw_name;
  }
  else
  {
    std::ostringstream	uid;

    uid << stat_buf.st_uid;
    Query.s_user = uid.str( );
  }
  free( pwbuffer );

  //---------------------------------------------------------------------
  // Extract information from /proc/<pid>/cmdline file
  //---------------------------------------------------------------------
  std::ostringstream	cmd_filename;
  int			cmdline_fd;
  char			buffer[ 4096 ];
  int			buffer_read_size;

  cmd_filename << "/proc/" << m_pid << "/cmdline";

  cmdline_fd = open( cmd_filename.str( ).c_str( ), O_RDONLY );
  buffer_read_size = read( cmdline_fd, buffer, sizeof( buffer ) );
  close( cmdline_fd );

  if ( buffer_read_size > 0 )
  {
    int	pos = 0;
    
    Query.s_args = "";
    while( ( pos < buffer_read_size ) &&
	   ( buffer[ pos ] != '\0' ) )
    {
      if ( pos > 0 )
      {
	Query.s_args += " ";
      }
      Query.s_args += &( buffer[ pos ] );
      pos += ::strlen( &( buffer[ pos ] ) ) + 1;
    }
  }

  //---------------------------------------------------------------------

  Query.s_pid = m_pid;
  Query.s_state = state;
  Query.s_vsz = vsize/1024;
  Query.s_rsz = ( rss * PageSize ) / 1024;
  Query.s_pcpu = ( seconds )
    ? ( ( 100.0 * float ( utime + stime + cutime + cstime ) ) / Hertz ) / seconds
    : 0;
  Query.s_pmem = 100.0 * ( Query.s_rsz / TotalRam );
  Query.s_etime = seconds; /* :TODO: */
  Query.s_fname = comm;

#endif /* HAVE_PROCFS_H */
}

#if !defined( HAVE_PROCFS_H )
namespace
{
  unsigned long
  get_total_ram( )
  {
    struct sysinfo	info;

    sysinfo( &info );
    return info.totalram;
  }
}

#endif /* */
