#ifndef GENERAL_AUTO_ARRAY_H
#define GENERAL_AUTO_ARRAY_H

#include <stdexcept>


namespace General
{
  template< typename T>
  struct AutoArrayRef
  {
    T*	m_ptr;
    explicit	AutoArrayRef( T* P )
      : m_ptr( P )
    {
    }
  };

  //---------------------------------------------------------------------
  //
  //: Auto_ptr for arrays.
  //
  template< typename T >
  class AutoArray
  {
  public:

    // Constructor
    explicit AutoArray( T* ptr_to_delete = 0 );

    //-------------------------------------------------------------------
    //
    //: Copy Constructor
    //	
    // Construct a new AutoArray. This object takes ownership from the previous
    // object
    //
    AutoArray( AutoArray<T>& Data )
      : mArray( Data.release( ) )
    {
    }

    // Destructor
    ~AutoArray();

    // auto_ptr compatability
    T& operator[]( const size_t index ) const;
    T* operator->() const;
    AutoArray& operator=( AutoArray& A );
    T* get() const;
    T* release();
    void reset( T* p = 0 );

    //-------------------------------------------------------------------
    // The following methods give reference like symantics to AutoArray
    //-------------------------------------------------------------------
    AutoArray( AutoArrayRef< T > Ref );
  
    AutoArray& operator=( AutoArrayRef< T > Ref )
    {
      reset( Ref.m_ptr );
      return *this;
    }

    template < typename T1 > operator AutoArrayRef< T1 >( )
    {
      return AutoArrayRef< T1 >( this->release( ) );
    }

    template < typename T1 > operator AutoArray< T1 >( )
    {
      return AutoArray< T1 >( this->release( ) );
    }

  private:

    T* mArray;

  };


  //---------------------------------------------------------------------
  //
  //: Constructor
  //
  //!param: T* const ptr_to_delete - A pointer to the dynamically
  //+       allocated array. Default is NULL.
  //
  template< typename T > AutoArray< T >::
  AutoArray( T* ptr_to_delete )
    : mArray( ptr_to_delete )
  {}


  template< typename T > AutoArray< T >::
  AutoArray( AutoArrayRef< T > Ref )
    : mArray( Ref.m_ptr )
  {}
  //---------------------------------------------------------------------
  //
  //: Destructor
  //
  template< typename T > AutoArray< T >::~AutoArray()
  {
    delete[] mArray;
  }


  //---------------------------------------------------------------------
  //
  //: Overloaded = operator.
  //
  // This allows the reassignment of pointer ownership
  //
  //!return: AutoArray< T >& - a reference to this
  //
  template< typename T > AutoArray<T>& AutoArray< T >::
  operator=( AutoArray<T>& A )
  {
    reset( A.release( ) );
    return *this;
  }

  //---------------------------------------------------------------------
  //
  //: Overloaded [] operator.
  //
  //!param: const size_t index - Index into array.
  //
  //!return: T& - A reference to the requested array element.
  //
  template< typename T > T& AutoArray< T >::operator[]( const size_t index ) const
  {
    if( mArray == 0 )
    {
      throw std::runtime_error( "General::AutoArray::operator[]: array is NULL" );
    }   

    return *( mArray + index );
  }


  //---------------------------------------------------------------------
  //
  //: Overloaded -> operator.
  //
  //!return: T& - A reference to the requested array element.
  //
  template< typename T > T* AutoArray< T >::operator->() const
  {
    return mArray;
  }

  //---------------------------------------------------------------------
  //
  //: Get array buffer.
  //
  //!return: T* - A pointer to the array.
  //
  template< typename T > T* AutoArray< T >::get() const
  {
    return mArray;
  }


  //---------------------------------------------------------------------
  //
  //: Release array buffer.
  //
  //!return: T* - A pointer to the array.
  //
  template< typename T > T* AutoArray< T >::release()
  {
    T* data( mArray );
    mArray = 0;

    return data;
  }


  //---------------------------------------------------------------------
  //
  //: Reset array buffer.
  //
  //!param:: T* - A pointer to new array buffer. Default is NULL.
  //
  template< typename T > void AutoArray< T >::reset( T* ptr_to_delete )
  {
    if( mArray != ptr_to_delete )
    {
      delete[] mArray;
      mArray = ptr_to_delete;
    }
   
    return;
  }

#if 0
  template < typename T, T1 > operator
  AutoArray<T>::AutoArrayRef< T1 >( )
  {
    return AutoArrayRef< T1 >( this->release( ) );
  }

  template < typename T, T1 > operator
  AutoArray<T>::
  AutoArray< T1 >( )
  {
    return AutoArray< T1 >( this->release( ) )
  }
#endif /* 0 */

}


#endif
